// Windows Bitmap (*.BMP, *.DIB) loader. Compatible with the generic image loader in Formats ImageFile.hpp.
//
//  developed by Microsoft in 1987 for Windows 2.0
//  fundamental part of graphics systems (icons and images) on Windows and OS/2
//  one of the most widespread formats in the 1990s
//  countless versions and variations
//
// Done:
//   Containers:
//     v1 (Windows 2.0, OS/2 1.x)
//     v2 (OS/2 2.x)
//     v3 (WindowsNT, Windows3.1)
//     WindowsCE 5 flavor
//     Photoshop flavor
//     v4 (WindowsNT 4.0, Windows95)
//     v5 (WindowsNT 2000, Windows98)
//   Compressions:
//     basic: RLE-4, RLE-8
//     OS/2: RLE-24, CCITT T.4 (a.k.a. Huffman1D)
//   Color depths:
//     1-, 2-, 4-, 8-bit indexed
//     24- and 32-bit colors with or without transparency
//     16- and 32-bit colors with all possible bitmasks
//
// To do:
//   PNG & JPEG compression
//   process partial BMPs (already done for RLE & CCITT)
//   expose XOR mask in icons



namespace Formats {


	namespace BMP {


#		pragma pack(push, 1)
		// BITMAPFILEHEADER
		struct FileHeader {
			UInt2B signature;
			UInt4B sizeInBytes;
			UInt4B reserved;
			UInt4B imageOffsetInBytes;
		};
#		pragma pack(pop)


		// BITMAPINFOHEADER
		// The most common BMP header type  not only found in BMP files, but also used for in-memory bitmaps and parameters to the
		//  GDI API.
		//   version 3 on Windows (introduced with WindowsNT and Windows3.x)
		struct Header_v3_NT {
			UInt4B_LE sizeInBytes;
			UInt4B_LE widthInPixels;
			SInt4B_LE heightInPixels; // negative if top-down
			UInt2B_LE one; // was biPlanes once
			UInt2B_LE bitsPerPixel;
			UInt4B_LE compression;
			UInt4B_LE imageSizeInBytes; // may be zero for uncompressed images
			SInt4B_LE pixelsPerMeterX;
			SInt4B_LE pixelsPerMeterY;
			UInt4B_LE colorCount; // only for paletted images
			UInt4B_LE colorsImportant; // optional
		};
		MUST_STATIC(40 == sizeof(Header_v3_NT));


		// A BMP file, loaded into memory.
		struct File {
			UInt4B       widthInPixels;
			//  do not trust  first consider sizeOfPixels (properly handles cropped files)
			//    damaged BMPs are pretty common
			//    very hard to verify for RLE data and such
			//    defer until actual image extraction
			//  negative for top-down BMPs
			SInt4B       heightInPixels;
			UInt1B       bitsPerPixel;
			UInt1B       compression;
			// Number of colors in the palette.
			//   only if bitsPerPixel <= 8
			//   BMP palettes can be truncated, e.g. an 8-bit image might have just 234 colors
			UInt2B       paletteLength;
			// Color masks
			//   for bitmasks compression
			//   maskC is also set if a 32-bit image contains coverage (non-standard)
			UInt4B       maskR;
			UInt4B       maskG;
			UInt4B       maskB;
			UInt4B       maskC;
			// The palette.
			//   only if bitsPerPixel <= 8
			//   always zero-padded to 1 << bitsPerPixel entries (compensating for bad palettes and bad indices)
			//   must be Heap::release()d after use
			GFX::SBGRC_8888 * toPaletteOrNull;
			// The original (compressed?) pixel data in the file.
			Byte const * toPixels;
			USize        sizeOfPixels;
			// If the BMP is embedded in a CUR or ICO file, there is always an implicit 1-bit AND mask.
			Byte const * toMaskOrNull;
			Byte const * toEndOfMask;
		};


		// Checks whether the given 20 B mark the beginning of a Windows Bitmap with some certainty.
		bool isBMP(
			Byte const * toFirst20B,
			USize        size
		);


		// Parses the given BMP.
		//   returns no on failure
		//   otherwise, must be destroy()d after use
		//   keeps references to the given bytes
		bool tryToLoad(
			File &            result,
			Byte const *      toBeginning,
			Byte const *      toEnd,
			Supervisor const *
		);

		// Loads the given raw BMP file content (without file header). Useful for Windows in-memory data structures and for packed
		//  BMPs, e.g. in CUR or ICO files.
		//   returns no on failure
		//   otherwise, must be destroy()d after use
		//   keeps references to the given bytes
		//   requires knowledge of being loaded from a CUR or ICO file
		//   an offset from the base address can be passed (to simulate file header behaviour), or 0 for packed BMPs
		bool tryToLoadNaked(
			File &            ,
			Byte const *      toBeginning,
			Byte const *      toEnd,
			Supervisor const *,
			UInt4B            pixelOffsetOrZero,
			bool              isCURorICO
		);

		// Returns the BMPs properties.
		//   width and height are never zero
		ImageProperties propertiesOf(File const &);

		// Extracts the image from the BMP.
		//   returns no on failure
		//   writes the pixels to the given array, starting with the first line
		//   specify the distance between two consecutive lines (negative for bottom-up)
		bool extractImage(
			File const &      ,
			GFX::SBGRC_8888 * ,
			SSize             linePadding,
			Supervisor const *
		);

		// Extracts the palette from the given BMP.
		//   returns no on failure
		//   palette length can be queried via propertiesOf()
		//   writes the colors in original order to the given array
		bool extractPalette(
			File const &     ,
			GFX::SBGRC_8888 *
		);

		void destroy(File &);


		void tryToSave(
			GFX::Pixels_r8g8b8c8_sRGB const &,
			UTF16Unit const *                path
		);


	} // BMP


} // Formats


#define FORMATS_BMP 1 // use this format in any loader that is #included later