// Generic loader for all kinds of 3D scene files.
//   just the public interface
//   in order to get the implementation:
//    1. choose any combination of the following:
//       #include "Formats STL.hpp"      // stereolithography
//       #include "Formats VRML.hpp"     // VRML 1.0
//       #include "Formats VRML V2.hpp"  // VRML 2.0
//    2. #include "Formats SceneFile.inl"



namespace Formats {


	// A string in a file.
	//   supports all kinds of encodings
	struct FileString {
		// The raw bytes.
		//   may be nullptr or empty
		//   does not bother to convert any data (this is the clients job)
		Byte const *        toBytesOrNull;
		UInt4B              length;
		// The strings codepage.
		//   may be unknown (the client should make an educated guess then)
		//   no guarantee for valid encoding
		OS::Win32::CodePage codePage;

		bool isEmpty() const;

		// Translates the string to UTF-16.
		//   must not be empty
		//   returns nullptr on failure
		//   Heap::release() after use
		//   unknown encoding is analyzed and guessed
		UTF16Unit * asUTF16() const;

		// Translates the string to UTF-16.
		//   must not be empty
		//   returns the new end
		//   unknown encoding is analyzed and guessed
		UTF16Unit * asUTF16(UTF16Unit *) const;

	};


	// All 3D scene file formats which are supported by this framework.
	enum class SceneFileType {
		unknown,
		stl,   // stereolithography (.STL)
		vrml1, // Virtual Reality Modeling Language 1.0 (.WRL)
		vrml2, // Virtual Reality Modeling Language 2.0 (.WRL)
	};


	// Determines the given files format from its first bytes.
	//   you should provide at least 16 B (the more, the better)
	//   returns SceneFileType::unknown on failure
	//   useful to check files for usability without mapping them entirely into memory
	SceneFileType typeOfSceneFile(
		Byte const * toBeginning,
		USize        sizeAvailable,
		USize        fileSize
	);

	
	// Interface to any kind of 3D scene file, loaded into memory.
	//   offers access to metadata via coordinateSpaceOf(), nameOf(), 
	//   offers access to scene graph via numberOfNodesIn(), numberOfChildrenOf(), 
	//   offers access to geometry via trianglesOfMesh(), extractTrianglesOfMesh(), 
	// Supported formats:
	//   stereolithography (.STL)
	//   Virtual Reality Modeling Language 1.0 (.WRL)
	//   Virtual Reality Modeling Language 2.0 (.WRL)
	struct SceneFile;


	// Parses the given file.
	//   returns no on failure
	//   otherwise, must be destroy()d after use
	//   keeps references to the given bytes
	//   SceneFileType::unknown triggers automatic detection
	bool tryToLoad(
		SceneFile &       ,
		SceneFileType     ,
		Byte const *      toBeginning,
		Byte const *      toEnd,
		Supervisor const *
	);


	// Index of a node in a scene.
	using SceneNodeIndex = UIntLeast4B;

	// Index of the root node (present in every scene).
	SceneNodeIndex constexpr sceneRoot = 0;


	// All kinds of coordinate systems used in 3D scene file formats.
	enum class CoordinateSpace3D {
		unknown,
		xRight_yUp_rightHanded,
		xRight_zUp_rightHanded
	};

	struct SceneNode_Coordinates {
		CoordinateSpace3D space;
		// Length of a unit, in meters.
		//   NaN if unknown
		//   compensates for some formats exporting milimeters, feet, etc.
		float             unit;
	};

	// Describes the coordinate space used in the given scene.
	SceneNode_Coordinates coordinateSpaceOf(SceneFile const &);

	Geo::Matrix4x3_float_reg rootTransform__TODO(SceneFile const &);

	// Returns the total number of nodes in the scene.
	SceneNodeIndex numberOfNodesIn(SceneFile const &);

	// Returns the index of the given nodes parent in the scene.
	//   the root is always its own parent
	SceneNodeIndex parentOf(
		SceneFile const &,
		SceneNodeIndex
	);

	// Returns the number of children of the given node in the scene.
	UIntLeast4B numberOfChildrenOf(
		SceneFile const &,
		SceneNodeIndex
	);

	// For the nth child of the given node in the scene, returns its index.
	SceneNodeIndex childIndexOf(
		SceneFile const &,
		SceneNodeIndex   ,
		UIntLeast4B      childsIndex
	);

	// Checks whether the given node is marked as visible in the scene.
	bool isNodeVisible(
		SceneFile const &,
		SceneNodeIndex
	);

	enum class SceneNode_Type {
		//   has no geometry itself
		//   position, deletion, visibility are inherited to children
		group,
		//   should have geometry
		object,
		// Each child node denotes a level of detail.
		//   first child is most detailed
		//   only one child should be active at a time
		lod,
		externalReference
	};

	SceneNode_Type typeOf(
		SceneFile const &,
		SceneNodeIndex
	);

	FileString nameOf(
		SceneFile const &,
		SceneNodeIndex
	);

	// A format-internal name for the node, e.g. DEF names in VRML. Not necessarily human-readable.
	FileString internalNameOf(
		SceneFile const &,
		SceneNodeIndex
	);

	FileString commentOf(
		SceneFile const &,
		SceneNodeIndex
	);


	// Returns the given nodes transformation.
	struct SceneNode_Transformation {
		enum class Type {
			// The transformation is a 43 matrix relative to the parent node.
			//   Examples: scaling, rotation, translation, shearing, mirroring
			affine,
			// The node is transformed to always face the viewer.
			sprite,
			// The transformation depends on the viewer angle. It is a rotation about a fixed axis.
			//   Examples: trees in early video games, VRML 2.0 Billboard node with non-zero axis
			spriteWithAxis
		};
		Type type;
		union {
			struct {
				Geo::Matrix4x3_float localToParent;
			} asAffine;
			struct {
				Geo::XYZ_float normalizedAxis; // in local coordinates
			} asSpriteWithAxis;
		};
	};
	Geo::Matrix4x3_float_reg transformationOf(
		SceneFile const &,
		SceneNodeIndex
	);


	// Index of a mesh in a scene.
	using SceneMeshIndex = UIntLeast4B;

	// Denotes a non-existent mesh.
	auto constexpr noSceneMesh = SceneMeshIndex(-1);

	// Returns the total number of meshes in the scene.
	SceneMeshIndex numberOfMeshesIn(SceneFile const &);

	// If a mesh is assigned to the given node, returns its index. Returns noSceneMesh otherwise.
	SceneMeshIndex meshIndexOf(
		SceneFile const &,
		SceneNodeIndex
	);

	struct SceneMesh_Triangles {
		// The minimal number of vertices used by the nodes triangles.
		//   does not guarantee each vertex is actually used
		//   may exclude vertices of points/lines/curves/etc.
		UCount numberOfVertices;
		// The number of triangles in the node.
		UCount numberOfTriangles;
		// yes if a uniform color is present for the entire object.
		//   does not guarantee meaningful values
		//   can be coupled with other color types
		bool   haveColors_uniform;
		// yes if a color is present for each triangle.
		//   does not guarantee meaningful values
		//   can be coupled with other color types
		bool   haveColors_perTriangle;
		// yes if a color is present for each vertex.
		//   does not guarantee meaningful values
		//   can be coupled with other color types
		bool   haveColors_perVertex;
		// yes if a normal is present for each triangle.
		//   does not guarantee meaningful values
		//   can be coupled with other normal types
		bool   haveNormals_perTriangle;
		// yes if a normal is present for each vertex.
		//   does not guarantee meaningful values
		//   can be coupled with other normal types
		bool   haveNormals_perVertex;
		bool   doubleSided;
	};

	// Describes the triangles of the given mesh in the scene.
	SceneMesh_Triangles trianglesOfMesh(
		SceneFile const &,
		SceneMeshIndex
	);

	struct SceneMesh_TriangleData {
		Geo::XYZ_float *  toVerticesOrNull;
		USize             vertexStride;
		Geo::XYZ_float *  toNormalsOrNull;
		USize             normalsStride;
		GFX::SRGBC_8888 * toAmbientOrNull;
		USize             ambientStride;
		GFX::SRGBC_8888 * toColorsOrNull;
		USize             colorsStride;
		GFX::SRGBC_8888 * toSpecularOrNull;
		USize             specularStride;
		GFX::SRGBC_8888 * toEmissionsOrNull;
		USize             emissionsStride;
		Geo::XY_float *   toUVsOrNull;
		USize             uvStride;
		UInt4B *          toIndicesOrNull;
		UInt4B            baseIndex;
	};

	// Extracts geometry from the given mesh in the scene.
	//   returns no on failure
	//   otherwise:
	//     pre-allocate elements according to object properties
	//     pass the offset from one element to the next
	bool extractTrianglesOfMesh(
		SceneFile const &       ,
		SceneMeshIndex          ,
		SceneMesh_TriangleData &
	);


	// Index of a texture in a scene.
	using SceneTextureIndex = UIntLeast4B;

	// Denotes a non-existent texture.
	auto constexpr noSceneTexture = SceneTextureIndex(-1);

	// Returns the total number of textures in the scene.
	//   logical number of textures, not number of actually different files (its impossible to check all paths, links, etc.)
	SceneTextureIndex numberOfTexturesIn(SceneFile const &);

	// Returns the path of the given texture in the scene.
	//   empty for textures which are embedded in the scene file
	FileString pathOfTexture(
		SceneFile const &,
		SceneTextureIndex
	);

	enum class SceneTexture_Addressing {
		unknown_unknown,
		unknown_relative,
		unknown_absolute,
		repeat_relative,
		repeat_absolute,
		repeatMirrored_relative,
		repeatMirrored_absolute,
		clamp_relative,
		clamp_absolute,
		border_relative,
		border_absolute
	};

	struct SceneTexture_AddressingUVW {
		SceneTexture_Addressing u;
		SceneTexture_Addressing v;
		SceneTexture_Addressing w;
	};

	// Returns the addressing mode of the given texture in the scene.
	SceneTexture_AddressingUVW textureAddressingOf(
		SceneFile const &,
		SceneTextureIndex
	);

	enum class SceneTexture_Filter {
		unknown,
		nearest,
		linear,
		bilinear,
		trilinear,
		anisotropic_x2,
		anisotropic_x4,
		anisotropic_x8,
		anisotropic_x16,
	};

	// Returns the sampling mode of the given texture in the scene.
	SceneTexture_Filter textureFilterOf(
		SceneFile const &,
		SceneTextureIndex
	);

	// Returns the index of the given meshs texture.
	//   noSceneTexture if no texture is assigned
	SceneTextureIndex textureOf(
		SceneFile const &,
		SceneMeshIndex
	);


	// Index of a material in a scene.
	using SceneMaterialIndex = UIntLeast4B;

	// Denotes a non-existent material.
	auto constexpr noSceneMaterial = SceneMaterialIndex(-1);

	// Returns the total number of materials in the scene.
	//   many materials may be identical
	//     however, the author assigned them individually
	//     probably in order to change them later
	SceneMaterialIndex numberOfMaterialsIn(SceneFile const &);

	// Returns the index of the given meshs material.
	//   noSceneMaterial if no material is assigned
	SceneMaterialIndex materialOf(
		SceneFile const &,
		SceneMeshIndex
	);

	enum class LightingModel {
		unknown,
		unlit,   // just forwards to screen (probably precomputed lighting)
		classic, // Phong lighting (emissive + ambient + diffuse + specular ^ exponent)
		fresnel
	};


	// A material with the classic lighting model (OpenGL fixed function pipeline).
	struct Material_Classic {
		GFX::SRGB_FFF emissive;
		GFX::SRGB_FFF ambient;
		GFX::SRGB_FFF diffuse;
		GFX::SRGB_FFF specular;
		float         specularExponent;
		float         coverage;
	};

	struct SceneMaterial {
		LightingModel    model;
		union {
			Material_Classic classic;
		};
	};

	// Returns the properties of the given material.
	SceneMaterial materialPropertiesOf(
		SceneFile const & ,
		SceneMaterialIndex
	);


	void destroy(SceneFile &);


} // Formats