// Generic loader for all kinds of 3D scene files.
//   implementation
//   see Formats SceneFile.hpp for details



namespace Formats {


	// Determines the given files format from its first bytes.
	//   you should provide at least 16 B (the more, the better)
	//   returns SceneFileType::unknown on failure
	//   useful to check files for usability without mapping them entirely into memory
	SceneFileType typeOfSceneFile(
		Byte const * const toBeginning,
		USize const        sizeAvailable,
		USize const        fileSize
	) {
		if(sizeAvailable >= 80) {

#			if FORMATS_STL
				if(STL::isSTL(toBeginning, fileSize, sizeAvailable)) {
					return SceneFileType::stl;
				}
#			endif

		}

		if(sizeAvailable >= 16) {

#			if FORMATS_VRML_1
				if(VRML::V1::isVRML(toBeginning, fileSize, sizeAvailable)) {
					return SceneFileType::vrml1;
				}
#			endif
#			if FORMATS_VRML_2
				if(VRML::V2::isVRML(toBeginning, fileSize, sizeAvailable)) {
					return SceneFileType::vrml2;
				}
#			endif

		}

		return SceneFileType::unknown;
	}


	struct SceneFile {
		SceneFileType myType;
		union {
#			if FORMATS_STL
				STL::File mySTL;
#			endif
#			if FORMATS_VRML_1 || FORMATS_VRML_2
				VRML::SceneGraph myVRML;
#			endif
		};
	};


	// Parses the given file.
	//   returns no on failure
	//   otherwise, must be destroy()d after use
	//   keeps references to the given bytes
	//   SceneFileType::unknown triggers automatic detection
	bool tryToLoad(
		SceneFile &              result,
		SceneFileType            type,
		Byte const * const       toBeginning,
		Byte const * const       toEnd,
		Supervisor const * const supervisor
	) {
		// If no file type is provided, analyze now.
		auto const size = USize(toEnd - toBeginning);
		if(SceneFileType::unknown == type) {
			type = typeOfSceneFile(toBeginning, size, size);
		}

		switch(type) {

#			if FORMATS_STL
				case SceneFileType::stl:
					result.myType = SceneFileType::stl;
					return parse(result.mySTL, toBeginning, toEnd, supervisor);
#			endif

#			if FORMATS_VRML_1
				case SceneFileType::vrml1: {
					result.myType = SceneFileType::vrml1;
					// Load the file:
					VRML::V1::File file;
					if(no == tryToParse(file, toBeginning, toEnd, supervisor)) {
						return no;
					}
					auto const isSuccessful = tryToTraverse(result.myVRML, file, supervisor);
					destroy(file);
					return isSuccessful;
				}
#			endif

#			if FORMATS_VRML_2
				case SceneFileType::vrml2: {
					result.myType = SceneFileType::vrml2;
					// Load the file:
					VRML::V2::File file;
					if(no == tryToParse(file, toBeginning, toEnd, supervisor)) {
						return no;
					}
					// Resolve all prototypes to standard nodes (in order to import them into the scene graph):
					auto const isSuccessful = resolvePrototypes(file, supervisor) && tryToTraverse(result.myVRML, file, supervisor);
					destroy(file);
					return isSuccessful;
				}
#			endif

		}

		return no;
	}


	// Describes the coordinate space used in the given scene.
	SceneNode_Coordinates coordinateSpaceOf(SceneFile const & file) {
		switch(file.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					return coordinateSpaceOf(file.mySTL);
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return { CoordinateSpace3D::xRight_yUp_rightHanded, 1.0f }; // guaranteed by specification
#			endif

			default:
				UNREACHABLE;
		}
	}

	Geo::Matrix4x3_float_reg rootTransform__TODO(SceneFile const & scene) {
		switch(coordinateSpaceOf(scene).space) {

			case Formats::CoordinateSpace3D::xRight_zUp_rightHanded:
				return {
					xyzwFrom(1.0f, 0.0f, 0.0f, 0.0f),
					xyzwFrom(0.0f, 0.0f,-1.0f, 0.0f),
					xyzwFrom(0.0f, 1.0f, 0.0f, 0.0f),
					xyzwFrom(0.0f, 0.0f, 0.0f, 1.0f)
				};

			case Formats::CoordinateSpace3D::xRight_yUp_rightHanded:
				return Geo::Matrix4x3_float::identity(); // no rotation necessary

			default:
				UNREACHABLE;

		}
	}

	// Returns the total number of nodes in the scene.
	SceneNodeIndex numberOfNodesIn(SceneFile const & file) {
		switch(file.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					return 1u + file.mySTL.numberOfObjects; // the root node plus all objects
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return numberOfNodesIn(file.myVRML);
#			endif

			default:
				UNREACHABLE;
		}
	}

	// Returns the index of the given nodes parent in the scene.
	//   the root is always its own parent
	SceneNodeIndex parentOf(
		SceneFile const &    file,
		SceneNodeIndex const nodesIndex
	) {
		switch(file.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					MUST(nodesIndex < 1u + file.mySTL.numberOfObjects);
					//  all STL objects are children to the root
					//  the root is always its own child
					return sceneRoot;
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return parentOf(file.myVRML, nodesIndex);
#			endif

			default:
				UNREACHABLE;
		}
	}

	// Returns the number of children of the given node in the scene.
	UIntLeast4B numberOfChildrenOf(
		SceneFile const &    file,
		SceneNodeIndex const nodesIndex
	) {
		switch(file.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					if(sceneRoot == nodesIndex) {
						return file.mySTL.numberOfObjects;
					} else {
						MUST(nodesIndex < 1u + file.mySTL.numberOfObjects);
						return 0; // there are no hierarchies in STLs
					}
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return numberOfChildrenOf(file.myVRML, nodesIndex);
#			endif

			default:
				UNREACHABLE;
		}
	}

	// For the nth child of the given node in the scene, returns its index.
	SceneNodeIndex childIndexOf(
		SceneFile const &    file,
		SceneNodeIndex const nodesIndex,
		UIntLeast4B const    childsIndex
	) {
		switch(file.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					//  all STL objects are children to the root
					//  the root is always node 0
					MUST(sceneRoot == nodesIndex);
					MUST(childsIndex < file.mySTL.numberOfObjects);
					return 1 + childsIndex;
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return childIndexOf(file.myVRML, nodesIndex, childsIndex);
#			endif

			default:
				UNREACHABLE;
		}
	}

	// Checks whether the given node is marked as visible in the scene.
	bool isNodeVisible(
		SceneFile const &    file,
		SceneNodeIndex const nodesIndex
	) {
		switch(file.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					return yes; // STL does not know hidden objects
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return isNodeVisible(file.myVRML, nodesIndex);
#			endif

			default:
				UNREACHABLE;
		}
	}

	SceneNode_Type typeOf(
		SceneFile const &    file,
		SceneNodeIndex const nodesIndex
	) {
		switch(file.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					if(sceneRoot == nodesIndex) {
						return SceneNode_Type::group;
					} else {
						return SceneNode_Type::object;
					}
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return typeOf(file.myVRML, nodesIndex);
#			endif

			default:
				UNREACHABLE;
		}
	}

	FileString nameOf(
		SceneFile const &    file,
		SceneNodeIndex const nodesIndex
	) {
		switch(file.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					if(sceneRoot == nodesIndex) {
						// STLs have no root node information.
						return { nullptr, 0, OS::Win32::CodePage::unknown };
					} else {
						return nameOf(file.mySTL, nodesIndex - 1);
					}
#			endif

#			if FORMATS_VRML_1
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return nameOf(file.myVRML, nodesIndex);
#			endif

			default:
				UNREACHABLE;
		}
	}

	// A format-internal name for the node, e.g. DEF names in VRML. Not necessarily human-readable.
	FileString internalNameOf(
		SceneFile const &    file,
		SceneNodeIndex const nodesIndex
	) {
		switch(file.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					// STLs have no internal node names.
					return { nullptr, 0, OS::Win32::CodePage::unknown };
#			endif

#			if FORMATS_VRML_1
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return internalNameOf(file.myVRML, nodesIndex);
#			endif

			default:
				UNREACHABLE;
		}
	}

	FileString commentOf(
		SceneFile const &    file,
		SceneNodeIndex const nodesIndex
	) {
		switch(file.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					return { nullptr, 0, OS::Win32::CodePage::unknown }; // no such thing in STLs
#			endif

#			if FORMATS_VRML_1
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return commentOf(file.myVRML, nodesIndex);
#			endif

			default:
				UNREACHABLE;
		}
	}

	// Returns the given nodes transformation.
	Geo::Matrix4x3_float_reg transformationOf(
		SceneFile const &    scene,
		SceneNodeIndex const nodesIndex
	) {
		switch(scene.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					return Geo::Matrix4x3_float::identity(); // STL does not support transformation
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return transformationOf(scene.myVRML, nodesIndex);
#			endif

			default:
				UNREACHABLE;
		}
	}


	// Returns the total number of meshes in the scene.
	SceneMeshIndex numberOfMeshesIn(SceneFile const & scene) {
		switch(scene.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					return scene.mySTL.numberOfObjects;
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return numberOfMeshesIn(scene.myVRML);
#			endif

			default:
				UNREACHABLE;
		}
	}

	// If a mesh is assigned to the given node, returns its index. Returns noSceneMesh otherwise.
	SceneMeshIndex meshIndexOf(
		SceneFile const &    scene,
		SceneNodeIndex const nodesIndex
	) {
		switch(scene.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					MUST(nodesIndex < scene.mySTL.numberOfObjects + 1u);
					if(sceneRoot == nodesIndex) { // Root?
						return noSceneMesh;
					} else {
						return nodesIndex - 1;
					}
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return meshIndexOf(scene.myVRML, nodesIndex);
#			endif

			default:
				UNREACHABLE;
		}
	}

	// Describes the triangles of the given mesh in the scene.
	SceneMesh_Triangles trianglesOfMesh(
		SceneFile const &    scene,
		SceneMeshIndex const meshIndex
	) {
		switch(scene.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					return trianglesOf(scene.mySTL, meshIndex);
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return trianglesOfMesh(scene.myVRML, meshIndex);
#			endif

			default:
				UNREACHABLE;
		}
	}

	// Extracts geometry from the given mesh in the scene.
	//   returns no on failure
	//   otherwise:
	//     pre-allocate elements according to object properties
	//     pass the offset from one element to the next
	bool extractTrianglesOfMesh(
		SceneFile const &        scene,
		SceneMeshIndex const     meshIndex,
		SceneMesh_TriangleData & result
	) {
		switch(scene.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					return extractTrianglesOf(scene.mySTL, meshIndex, result);
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return extractTrianglesOfMesh(scene.myVRML, meshIndex, result);
#			endif

			default:
				UNREACHABLE;
		}
	}


	// Returns the total number of textures in the scene.
	//   logical number of textures, not number of actually different files (its impossible to check all paths, links, etc.)
	SceneTextureIndex numberOfTexturesIn(SceneFile const & file) {
		switch(file.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					return 0; // no texture support
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return numberOfTexturesIn(file.myVRML);
#			endif

			default:
				UNREACHABLE;
		}
	}

	// Returns the path of the given texture in the scene.
	//   empty for textures which are embedded in the scene file
	FileString pathOfTexture(
		SceneFile const &       file,
		SceneTextureIndex const texture
	) {
		switch(file.myType) {

			// STL does not have textures

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return pathOfTexture(file.myVRML, texture);
#			endif

			default:
				UNREACHABLE;
		}
	}

	// Returns the addressing mode of the given texture in the scene.
	SceneTexture_AddressingUVW textureAddressingOf(
		SceneFile const &       file,
		SceneTextureIndex const texture
	) {
		switch(file.myType) {

			// STL does not have textures

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return textureAddressingOf(file.myVRML, texture);
#			endif

			default:
				UNREACHABLE;
		}
	}

	// Returns the sampling mode of the given texture in the scene.
	SceneTexture_Filter textureFilterOf(
		SceneFile const &       file,
		SceneTextureIndex const texture
	) {
		switch(file.myType) {

			// STL does not have textures

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return SceneTexture_Filter::unknown; // VRML does not have any preferences
#			endif

			default:
				UNREACHABLE;
		}
		UNUSED(texture);
	}

	// Returns the index of the given meshs texture.
	//   noSceneTexture if no texture is assigned
	SceneTextureIndex textureOf(
		SceneFile const &    file,
		SceneMeshIndex const meshIndex
	) {
		switch(file.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					return noSceneTexture; // STL does not have textures
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return textureOf(file.myVRML, meshIndex);
#			endif

			default:
				UNREACHABLE;
		}
	}



	// Returns the total number of materials in the scene.
	//   many materials may be identical
	//     however, the author assigned them individually
	//     probably in order to change them later
	SceneMaterialIndex numberOfMaterialsIn(SceneFile const & file) {
		switch(file.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					return 0; // STL does not have materials
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return numberOfMaterialsIn(file.myVRML);
#			endif

			default:
				UNREACHABLE;
		}
	}

	// Returns the index of the given meshs material.
	//   noSceneMaterial if no material is assigned
	SceneMaterialIndex materialOf(
		SceneFile const &    file,
		SceneMeshIndex const meshIndex
	) {
		switch(file.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					return noSceneMaterial; // STL does not have materials
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return materialOf(file.myVRML, meshIndex);
#			endif

			default:
				UNREACHABLE;
		}
	}

	// Returns the properties of the given material.
	SceneMaterial materialPropertiesOf(
		SceneFile const &        file,
		SceneMaterialIndex const materialIndex
	) {
		switch(file.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					UNREACHABLE; // STL does not have materials
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return materialPropertiesOf(file.myVRML, materialIndex);
#			endif

			default:
				UNREACHABLE;
		}
	}




	void destroy(SceneFile & file) {
		switch(file.myType) {

#			if FORMATS_STL
				case SceneFileType::stl:
					return destroy(file.mySTL);
#			endif

#			if FORMATS_VRML_1 || FORMATS_VRML_2
				case SceneFileType::vrml1:
				case SceneFileType::vrml2:
					return destroy(file.myVRML);
#			endif

			default:
				UNREACHABLE;
		}
	}


} // Formats