#include "framework.hpp"
#include "SIMD.hpp"
#include "GFX.hpp"
#include "Formats.hpp"
#include "Memory.hpp" // movsb() & stosb() for run-length encoding

// TODO: just for file string codepages; move to other file
#include "OS Win32.hpp"
#include "Geo.hpp"
#include "Formats SceneFile.hpp"



namespace Formats {


	bool FileString::isEmpty() const {
		MUST(0 == length || nullptr != toBytesOrNull);

		return 0 == length;
	}

	// Translates the string to UTF-16.
	//   must not be empty
	//   returns nullptr on failure
	//   Heap::release() after use
	//   unknown encoding is analyzed and guessed
	UTF16Unit * FileString::asUTF16() const {
		MUST(no == isEmpty());

		// Reserve room.
		//   whatever codepage the source encoding is: no code unit will occupy more than two UTF-16 units
		auto const worstCaseLength = doubled(UCount(length)) + 1;

		auto const result = Heap::tryToReserve<UTF16Unit>(worstCaseLength);
		if(nullptr == result) {
			return nullptr;
		}

		*asUTF16(result) = 0; // zero termination
		return result;
	}

	// Translates the string to UTF-16.
	//   must not be empty
	//   returns the new end
	//   unknown encoding is analyzed and guessed
	UTF16Unit * FileString::asUTF16(UTF16Unit * toEnd) const {
		MUST(no == isEmpty());

		// If the code page is unknown, take an educated guess:
		auto actualCodePage = codePage;
		if(OS::Win32::CodePage::unknown == actualCodePage) {
			actualCodePage = OS::Win32::detectCodePage(toBytesOrNull, length);
		}

		// Convert (use the users code page if the encoding is still unknown):
		toEnd = utf16FromCodePage(
			actualCodePage, toEnd,
			reinterpret_cast<ASCIICharacter const *>(toBytesOrNull), length
		);

		return toEnd;
	}


	namespace PackBits {


		//  source and destination must not be empty
		//  there must be 128 B padding after destination to account for a damaged stream
		//  returns a pointer to the end of the extacted data
		Byte * extract(
			Byte *             toDestination,
			Byte * const       toEndOfDestination,
			Byte const * &     toSource,
			Byte const * const toEndOfSource
		) {
			MUST(toDestination < toEndOfDestination);
			MUST(toSource < toEndOfSource);
			do {

				auto n = UIntLeast1B(*toSource++);
				if(n <= 127) {
					auto size = USize(n + 1); // copy the next n+1 bytes literally
					movsb(toDestination, toSource, size);
					toDestination += size;
					toSource += size;
				} else if(n > 128) {
					auto size = USize(257 - n); // replicate the next byte 257-n times:
					stosb(toDestination, *toSource++, size);
					toDestination += size;
				} else {
					// reserved; ignore
				}

			} while(toSource < toEndOfSource && toDestination < toEndOfDestination);
			return toDestination;
		}
	
	
	} // PackBits


	// 1-D conversions:

	void rgb_to_bgr1(
		GFX::SBGRC_8888 *     d,
		GFX::SRGB_888 const * s,
		UCount const          length
	) {
		MUST(0 < length);

		auto const s_end = s + length;
		do {
			store(*d, bgr1Of(load(*s)));
		} while(++d, ++s < s_end);
	}
	void rgb_to_bgr1(
		GFX::SBGRC_8888 *           d,
		GFX::SXRXGXB_262626 const * s,
		UCount const                length
	) {
		MUST(0 < length);

		auto const s_end = s + length;
		do {
			// Combine an 8-bit and a 16-bit load:
			UInt4B _bgr = UInt4B((s->xb << 16) | *reinterpret_cast<UInt2B const *>(&s->xr));
			// Perform a parallel left-shift on all three components (this cannot run over since the two highest bits are zero). Fill
			//  the two least significant bits with copies of the two most significant bits.
			_bgr = (_bgr << 2) | ((_bgr >> 4) & 0x00030303);
			// Set coverage to 100 % by filling the most significant bits.
			GFX::SRGBC_8888_reg rgbc = { _bgr | 0xFF000000 };
			store(*d, bgrcOf(rgbc));
		} while(++d, ++s < s_end);
	}
	void rgb_to_bgr0(
		GFX::SBGRC_8888 *     d,
		GFX::SRGB_888 const * s,
		UCount const          length
	) {
		MUST(0 < length);

		auto const s_end = s + length;
		do {
			store(*d, bgr0Of(load(*s)));
		} while(++d, ++s < s_end);
	}

	void rgb_to_bgr1(
		GFX::SBGRC_8888_reg * d,
		GFX::SRGB_888 const * s,
		UCount const          length
	) {
		MUST(0 < length);

		auto const s_end = s + length;
		do {
			*d = bgr1Of(load(*s));
		} while(++d, ++s < s_end);
	}

	void rgb_to_bgr1(
		GFX::SBGRC_8888_reg *      d,
		GFX::SXRGB_4444_BE const * s,
		UCount const               length
	) {
		MUST(0 < length);

		auto const s_end = s + length;
		do {
			*d = sBGRC_8888_from(load(*s));
		} while(++d, ++s < s_end);
	}

	void i1_to_bgrc(
		GFX::SBGRC_8888 *             d,
		SSize const                   d_linePad,
		Byte const *                  s,
		SSize const                   s_linePad,
		GFX::SBGRC_8888 const * const palette,
		UIntLeast4B const             width,
		UIntLeast4B const             height
	) {
		MUST(0 < width && 0 < height);

		// Round down the payload: If there are excess pixels (up to seven) in the last byte, handle them individually.
		auto const s_linePayload = width / 8;
		auto const s_lineSize    = (width + 7) / 8 + s_linePad;
		auto const s_excess      = width % 8;

		// Convert all lines in their original order:
		auto const s_end = byteOffset(s, height * s_lineSize);
		do {

			// Convert all pixels in the line in groups of eight:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			while(s < s_lineEnd) { // No do-while with <8 pixels width!
				auto bits = *s++;
				*d++ = palette[(bits >> 7) & 1];
				*d++ = palette[(bits >> 6) & 1];
				*d++ = palette[(bits >> 5) & 1];
				*d++ = palette[(bits >> 4) & 1];
				*d++ = palette[(bits >> 3) & 1];
				*d++ = palette[(bits >> 2) & 1];
				*d++ = palette[(bits >> 1) & 1];
				*d++ = palette[(bits >> 0) & 1];
			}

			// Convert the last group of zero to seven pixels:
			if(s_excess > 0) {
				auto bits = *s++;
				auto i    = s_excess;
				do {
					*d++ = palette[(bits >> 7) & 1];
				} while(bits <<= 1, --i > 0);
			}

			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s < s_end);
	}
	void i1_to_bgrc_flipY(
		GFX::SBGRC_8888 *             d,
		SSize const                   d_linePad,
		Byte const * const            s_begin,
		SSize                         s_linePad,
		GFX::SBGRC_8888 const * const palette,
		UIntLeast4B const             width,
		UIntLeast4B const             height
	) {
		MUST(0 < width && 0 < height);
		
		// Round down the payload: If there are excess pixels (up to seven) in the last byte, handle them individually.
		auto const s_linePayload = width / 8;
		auto const s_lineSize    = (width + 7) / 8 + s_linePad;
		auto const s_excess      = width % 8;

		// Go to the last line:
		auto s = byteOffset(s_begin, (height - 1) * s_lineSize);
		// Use padding to jump backwards:
		s_linePad = -SSize(2 * s_lineSize - s_linePad);
		// Convert all lines in reverse order:
		do {

			// Convert all pixels in the line in groups of eight:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			while(s < s_lineEnd) { // No do-while with <8 pixels width!
				auto bits = *s++;
				*d++ = palette[(bits >> 7) & 1];
				*d++ = palette[(bits >> 6) & 1];
				*d++ = palette[(bits >> 5) & 1];
				*d++ = palette[(bits >> 4) & 1];
				*d++ = palette[(bits >> 3) & 1];
				*d++ = palette[(bits >> 2) & 1];
				*d++ = palette[(bits >> 1) & 1];
				*d++ = palette[(bits >> 0) & 1];
			}

			// Convert the last group of zero to seven pixels:
			if(s_excess > 0) {
				auto bits = *s++;
				auto i    = s_excess;
				do {
					*d++ = palette[(bits >> 7) & 1];
				} while(bits <<= 1, --i > 0);
			}

			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s >= s_begin); // TODO: not C++ standard compliant
	}

	void i2h_to_bgrc(
		GFX::SBGRC_8888 *             d,
		SSize const                   d_linePad,
		Byte const *                  s,
		GFX::SBGRC_8888 const * const palette,
		SSize const                   s_linePad,
		UIntLeast4B const             width,
		UIntLeast4B const             height
	) {
		MUST(0 < width && 0 < height);

		// Round down the payload: If the number of pixels is not a multiple of four, the last pixels are handled individually.
		auto const s_linePayload = width / 4;
		auto const s_lineSize    = (width + 3) / 4 + s_linePad;

		// Convert all lines in their original order:
		auto const s_end = byteOffset(s, height * s_lineSize);
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			while(s < s_lineEnd) { // No do-while with <4 pixels width!
				auto const lr = *s;
				d[0] = palette[ lr >> 6     ];
				d[1] = palette[(lr >> 4) & 3];
				d[2] = palette[(lr >> 2) & 3];
				d[3] = palette[(lr     ) & 3];
				d += 4;
				++s;
			}
			if(width % 4) {
				auto s_ = *s++;
				for(UIntLeast4B i = 0; i < width % 4; ++i) {
					*d++ = palette[s_ >> 6];
					s_ <<= 2;
				}
			}
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s < s_end);
	}
	void i2h_to_bgrc_flipY(
		GFX::SBGRC_8888 *             d,
		SSize const                   d_linePad,
		Byte const * const            s_begin,
		GFX::SBGRC_8888 const * const palette,
		SSize                         s_linePad,
		UIntLeast4B const             width,
		UIntLeast4B const             height
	) {
		MUST(0 < width && 0 < height);

		// Round down the payload: If the number of pixels is not a multiple of four, the last pixels are handled individually.
		auto const s_linePayload = width / 4;
		auto const s_lineSize    = (width + 3) / 4 + s_linePad;

		// Go to the last line:
		auto s = byteOffset(s_begin, (height - 1) * s_lineSize);
		// Use padding to jump backwards:
		s_linePad = -SSize(2 * s_lineSize - s_linePad);
		// Convert all lines in reverse order:
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			while(s < s_lineEnd) { // No do-while with <4 pixels width!
				auto const lr = *s;
				d[0] = palette[ lr >> 6     ];
				d[1] = palette[(lr >> 4) & 3];
				d[2] = palette[(lr >> 2) & 3];
				d[3] = palette[(lr     ) & 3];
				d += 4;
				++s;
			}
			if(width % 4) {
				auto s_ = *s++;
				for(UIntLeast4B i = 0; i < width % 4; ++i) {
					*d++ = palette[s_ >> 6];
					s_ <<= 2;
				}
			}
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s >= s_begin); // TODO: not C++ standard compliant
	}

	void i4h_to_bgrc(
		GFX::SBGRC_8888 *             d,
		SSize const                   d_linePad,
		Byte const *                  s,
		GFX::SBGRC_8888 const * const palette,
		SSize const                   s_linePad,
		UIntLeast4B const             width,
		UIntLeast4B const             height
	) {
		MUST(0 < width && 0 < height);

		// Round down the payload: If the number of pixels is odd, the last pixel is handled individually.
		auto const s_linePayload = width / 2;
		auto const s_lineSize    = (width + 1) / 2 + s_linePad;

		// Convert all lines in their original order:
		auto const s_end = byteOffset(s, height * s_lineSize);
		do {

			// Convert all pixels in the line as groups of two:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			while(s < s_lineEnd) { // No do-while with <4 pixels width!
				auto const lr = *s;
				d[0] = palette[lr >> 4];
				d[1] = palette[lr & 15];
				d += 2;
				++s;
			}
			if(width % 2) {
				*d++ = palette[(*s++) >> 4];
			}
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s < s_end);
	}
	void i4h_to_bgrc_flipY(
		GFX::SBGRC_8888 *             d,
		SSize const                   d_linePad,
		Byte const * const            s_begin,
		GFX::SBGRC_8888 const * const palette,
		SSize                         s_linePad,
		UIntLeast4B const             width,
		UIntLeast4B const             height
	) {
		MUST(0 < width && 0 < height);

		// Round down the payload: If the number of pixels is odd, the last pixel is handled individually.
		auto const s_linePayload = width / 2;
		auto const s_lineSize    = (width + 1) / 2 + s_linePad;

		// Go to the last line:
		auto s = byteOffset(s_begin, (height - 1) * s_lineSize);
		// Use padding to jump backwards:
		s_linePad = -SSize(2 * s_lineSize - s_linePad);
		// Convert all lines in reverse order:
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			while(s < s_lineEnd) { // No do-while with <4 pixels width!
				auto const lr = *s;
				d[0] = palette[lr >> 4];
				d[1] = palette[lr & 15];
				d += 2;
				++s;
			}
			if(width % 2) {
				*d++ = palette[(*s++) >> 4];
			}
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s >= s_begin); // TODO: not C++ standard compliant
	}

	void i4l_to_bgrc(
		GFX::SBGRC_8888 *             d,
		SSize const                   d_linePad,
		Byte const *                  s,
		GFX::SBGRC_8888 const * const palette,
		SSize const                   s_linePad,
		UIntLeast4B const             width,
		UIntLeast4B const             height
	) {
		MUST(0 < width && 0 < height);

		// Round down the payload: If the number of pixels is odd, the last pixel is handled individually.
		auto const s_linePayload = width / 2;
		auto const s_lineSize    = (width + 1) / 2 + s_linePad;

		// Convert all lines in their original order:
		auto const s_end = byteOffset(s, height * s_lineSize);
		do {

			// Convert all pixels in the line as groups of two:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			while(s < s_lineEnd) { // No do-while with <4 pixels width!
				auto const rl = *s;
				d[0] = palette[rl & 15];
				d[1] = palette[rl >> 4];
				d += 2;
				++s;
			}
			if(width % 2) {
				*d++ = palette[(*s++) & 15];
			}
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s < s_end);
	}
	void i4l_to_bgrc_flipY(
		GFX::SBGRC_8888 *             d,
		SSize const                   d_linePad,
		Byte const * const            s_begin,
		GFX::SBGRC_8888 const * const palette,
		SSize                         s_linePad,
		UIntLeast4B const             width,
		UIntLeast4B const             height
	) {
		MUST(0 < width && 0 < height);

		// Round down the payload: If the number of pixels is odd, the last pixel is handled individually.
		auto const s_linePayload = width / 2;
		auto const s_lineSize    = (width + 1) / 2 + s_linePad;

		// Go to the last line:
		auto s = byteOffset(s_begin, (height - 1) * s_lineSize);
		// Use padding to jump backwards:
		s_linePad = -SSize(2 * s_lineSize - s_linePad);
		// Convert all lines in reverse order:
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			while(s < s_lineEnd) { // No do-while with <4 pixels width!
				auto const rl = *s;
				d[0] = palette[rl & 15];
				d[1] = palette[rl >> 4];
				d += 2;
				++s;
			}
			if(width % 2) {
				*d++ = palette[(*s++) & 15];
			}
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s >= s_begin); // TODO: not C++ standard compliant
	}

	void i8_to_bgrc(
		GFX::SBGRC_8888 *             d,
		SSize const                   d_linePad,
		Byte const *                  s,
		GFX::SBGRC_8888 const * const palette,
		SSize const                   s_linePad,
		UIntLeast4B const             width,
		UIntLeast4B const             height
	) {
		MUST(0 < width && 0 < height);

		auto const s_linePayload = width;
		auto const s_lineSize    = s_linePayload + s_linePad;

		// Convert all lines in their original order:
		auto const s_end = byteOffset(s, height * s_lineSize);
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			do {
				*d = palette[*s];
			} while(++d, ++s < s_lineEnd);
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s < s_end);
	}
	void i8_to_bgrc_flipY(
		GFX::SBGRC_8888 *             d,
		SSize const                   d_linePad,
		Byte const * const            s_begin,
		GFX::SBGRC_8888 const * const palette,
		SSize                         s_linePad,
		UIntLeast4B const             width,
		UIntLeast4B const             height
	) {
		MUST(0 < width && 0 < height);
		auto const s_linePayload = width;
		auto const s_lineSize    = s_linePayload + s_linePad;

		// Go to the last line:
		auto s = byteOffset(s_begin, (height - 1) * s_lineSize);
		// Use padding to jump backwards:
		s_linePad = -SSize(2 * s_linePayload + s_linePad);
		// Convert all lines in reverse order:
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			do {
				*d = palette[*s];
			} while(++d, ++s < s_lineEnd);
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s >= s_begin); // TODO: not C++ standard compliant
	}

	void bgr_to_bgrc(
		GFX::SBGRC_8888 *     d,
		SSize const           d_linePad,
		GFX::SBGR_888 const * s,
		SSize const           s_linePad,
		UIntLeast4B const     width,
		UIntLeast4B const     height
	) {
		MUST(0 < width && 0 < height);

		auto const s_linePayload = width * sizeof *s;
		auto const s_lineSize    = s_linePayload + s_linePad;

		// Convert all lines in their original order:
		auto const s_end = byteOffset(s, height * s_lineSize);
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			do {
				d->b = s->b;
				d->g = s->g;
				d->r = s->r;
				d->c = 255;
			} while(++d, ++s < s_lineEnd);
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s < s_end);
	}
	void bgr_to_bgrc_flipY(
		GFX::SBGRC_8888 *     d,
		SSize const           d_linePad,
		GFX::SBGR_888 const * s_begin,
		SSize                 s_linePad,
		UIntLeast4B const     width,
		UIntLeast4B const     height
	) {
		MUST(0 < width && 0 < height);
		auto const s_linePayload = width * sizeof *s_begin;
		auto const s_lineSize    = s_linePayload + s_linePad;

		// Go to the last line:
		auto s = byteOffset(s_begin, (height - 1) * s_lineSize);
		// Use padding to jump backwards:
		s_linePad = -SSize(2 * s_linePayload + s_linePad);
		// Convert all lines in reverse order:
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			do {
				d->b = s->b;
				d->g = s->g;
				d->r = s->r;
				d->c = 255;
			} while(++d, ++s < s_lineEnd);
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s >= s_begin); // TODO: not C++ standard compliant
	}

	void bgr_to_bgrc_flipY(
		GFX::SBGRC_8888 *         d,
		SSize const               d_linePad,
		GFX::SBGRC_565_LE const * s_begin,
		SSize                     s_linePad,
		UIntLeast4B const         width,
		UIntLeast4B const         height
	) {
		MUST(0 < width && 0 < height);
		auto const s_linePayload = width * sizeof *s_begin;
		auto const s_lineSize    = s_linePayload + s_linePad;

		// Go to the last line:
		auto s = byteOffset(s_begin, (height - 1) * s_lineSize);
		// Use padding to jump backwards:
		s_linePad = -SSize(2 * s_linePayload + s_linePad);
		// Convert all lines in reverse order:
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			do {
				d->b = GFX::_5to8( s->data        & 31);
				d->g = GFX::_6to8((s->data >>  5) & 63);
				d->r = GFX::_5to8((s->data >> 11) & 31);
				d->c = 255;
			} while(++d, ++s < s_lineEnd);
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s >= s_begin); // TODO: not C++ standard compliant
	}

	void bgr_to_bgr1(
		GFX::SBGRC_8888 *       d,
		SSize const             d_linePad,
		GFX::SBGRX_8888 const * s,
		SSize const             s_linePad,
		UIntLeast4B const       width,
		UIntLeast4B const       height
	) {
		MUST(0 < width && 0 < height);

		auto const s_linePayload = width * sizeof *s;
		auto const s_lineSize    = s_linePayload + s_linePad;

		// Copy all lines in their original order:
		auto const s_end = byteOffset(s, height * s_lineSize);
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			do {
				d->b = s->b;
				d->g = s->g;
				d->r = s->r;
				d->c = 255;
			} while(++d, ++s < s_lineEnd);
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s < s_end);

	}
	void bgr_to_bgr1_flipY(
		GFX::SBGRC_8888 *             d,
		SSize const                   d_linePad,
		GFX::SBGRX_8888 const * const s_begin,
		SSize                         s_linePad,
		UIntLeast4B const             width,
		UIntLeast4B const             height
	) {
		MUST(0 < width && 0 < height);
		auto const s_linePayload = width * sizeof *s_begin;
		auto const s_lineSize    = s_linePayload + s_linePad;

		// Go to the last line:
		auto s = byteOffset(s_begin, (height - 1) * s_lineSize);
		// Use padding to jump backwards:
		s_linePad = -SSize(2 * s_linePayload + s_linePad);
		// Convert all lines in reverse order:
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			do {
				d->b = s->b;
				d->g = s->g;
				d->r = s->r;
				d->c = 255;
			} while(++d, ++s < s_lineEnd);
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s >= s_begin); // TODO: not C++ standard compliant
	}

	void bgrc_to_bgrc(
		GFX::SBGRC_8888 *       d,
		SSize const             d_linePad,
		GFX::SBGRC_8888 const * s,
		SSize const             s_linePad,
		UIntLeast4B const       width,
		UIntLeast4B const       height
	) {
		MUST(0 < width && 0 < height);

		auto const s_lineSize = s_linePad + width * sizeof *s;
		auto const d_lineSize = d_linePad + width * sizeof *d;

		// Copy all lines in their original order:
		auto const s_end = byteOffset(s, height * s_lineSize);
		do {
			movs(d, s, width);
			s = byteOffset(s, s_lineSize);
			d = byteOffset(d, d_lineSize);
		} while(s < s_end);

	}
	void bgrc_to_bgrc_flipY(
		GFX::SBGRC_8888 *             d,
		SSize const                   d_linePad,
		GFX::SBGRC_8888 const * const s_begin,
		SSize const                   s_linePad,
		UIntLeast4B const             width,
		UIntLeast4B const             height
	) {
		MUST(0 < width && 0 < height);
		auto const linePayload = width * sizeof *s_begin;
		auto const s_lineSize  = linePayload + s_linePad;
		auto const s_lineStep  = -SSize(s_lineSize);
		auto const d_lineStep  = linePayload + d_linePad;

		// Go to the last line and copy all lines in reverse order:
		auto s = byteOffset(s_begin, (height - 1) * s_lineSize);
		do {
			movs(d, s, width);
			s = byteOffset(s, s_lineStep);
			d = byteOffset(d, d_lineStep);
		} while(s >= s_begin); // TODO: not C++ standard compliant
	}

	void bgrc_to_bgrc(
		GFX::SBGRC_8888 *          d,
		SSize const                d_linePad,
		GFX::SBGRC_5551_LE const * s,
		SSize const                s_linePad,
		UIntLeast4B const          width,
		UIntLeast4B const          height
	) {
		MUST(0 < width && 0 < height);

		auto const s_linePayload = width * sizeof *s;
		auto const s_lineSize    = s_linePayload + s_linePad;

		// Convert all lines in their original order:
		auto const s_end = byteOffset(s, height * s_lineSize);
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			do {
				store(*d, sBGRC_8888_from(load(*s)));
			} while(++d, ++s < s_lineEnd);
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s < s_end);
	}
	void bgrc_to_bgrc_flipY(
		GFX::SBGRC_8888 *          d,
		SSize const                d_linePad,
		GFX::SBGRC_5551_LE const * s_begin,
		SSize                      s_linePad,
		UIntLeast4B const          width,
		UIntLeast4B const          height
	) {
		MUST(0 < width && 0 < height);
		auto const s_linePayload = width * sizeof *s_begin;
		auto const s_lineSize    = s_linePayload + s_linePad;

		// Go to the last line:
		auto s = byteOffset(s_begin, (height - 1) * s_lineSize);
		// Use padding to jump backwards:
		s_linePad = -SSize(2 * s_linePayload + s_linePad);
		// Convert all lines in reverse order:
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			do {
				store(*d, sBGRC_8888_from(load(*s)));
			} while(++d, ++s < s_lineEnd);
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s >= s_begin); // TODO: not C++ standard compliant
	}

	void bgrc_to_bgrc(
		GFX::SBGRC_8888 *          d,
		SSize const                d_linePad,
		GFX::SRGBC_5551_LE const * s,
		SSize const                s_linePad,
		UIntLeast4B const          width,
		UIntLeast4B const          height
	) {
		MUST(0 < width && 0 < height);
		auto const s_linePayload = width * sizeof *s;
		auto const s_lineSize    = s_linePayload + s_linePad;

		// Convert all lines in their original order:
		auto const s_end = byteOffset(s, height * s_lineSize);
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			do {
				store(*d, sBGRC_8888_from(load(*s)));
			} while(++d, ++s < s_lineEnd);
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s < s_end);
	}
	void bgrc_to_bgrc_flipY(
		GFX::SBGRC_8888 *          d,
		SSize const                d_linePad,
		GFX::SRGBC_5551_LE const * s_begin,
		SSize                      s_linePad,
		UIntLeast4B const          width,
		UIntLeast4B const          height
	) {
		MUST(0 < width && 0 < height);
		auto const s_linePayload = width * sizeof *s_begin;
		auto const s_lineSize    = s_linePayload + s_linePad;

		// Go to the last line:
		auto s = byteOffset(s_begin, (height - 1) * s_lineSize);
		// Use padding to jump backwards:
		s_linePad = -SSize(2 * s_linePayload + s_linePad);
		// Convert all lines in reverse order:
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			do {
				store(*d, sBGRC_8888_from(load(*s)));
			} while(++d, ++s < s_lineEnd);
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s >= s_begin); // TODO: not C++ standard compliant
	}


	void rgb_to_bgr1(
		GFX::SBGRC_8888 *          d,
		SSize const                d_linePad,
		GFX::SRGBX_5551_LE const * s,
		SSize const                s_linePad,
		UIntLeast4B const          width,
		UIntLeast4B const          height
	) {
		MUST(0 < width && 0 < height);
		auto const s_linePayload = width * sizeof *s;
		auto const s_lineSize    = s_linePayload + s_linePad;

		// Convert all lines in their original order:
		auto const s_end = byteOffset(s, height * s_lineSize);
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			do {
				store(*d, sBGR1_8888_from(load(*s)));
			} while(++d, ++s < s_lineEnd);
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s < s_end);
	}
	void rgb_to_bgr1_flipY(
		GFX::SBGRC_8888 *          d,
		SSize const                d_linePad,
		GFX::SRGBX_5551_LE const * s_begin,
		SSize                      s_linePad,
		UIntLeast4B const          width,
		UIntLeast4B const          height
	) {
		MUST(0 < width && 0 < height);
		auto const s_linePayload = width * sizeof *s_begin;
		auto const s_lineSize    = s_linePayload + s_linePad;

		// Go to the last line:
		auto s = byteOffset(s_begin, (height - 1) * s_lineSize);
		// Use padding to jump backwards:
		s_linePad = -SSize(2 * s_linePayload + s_linePad);
		// Convert all lines in reverse order:
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			do {
				store(*d, sBGR1_8888_from(load(*s)));
			} while(++d, ++s < s_lineEnd);
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s >= s_begin); // TODO: not C++ standard compliant
	}

	// For the given color mask, computes
	//   the number of bits to shift right until the value is normalized
	//   the number of bits occupied by the value
	local void getShiftAndBitCount(
		UInt4B        mask,
		UIntLeast1B & shift,
		UIntLeast1B & bits
	) {
		shift = 0;
		bits  = 0;
		if(0 == mask) {
			return; // cannot handle this
		}

		// Count trailing zeroes (these make the shift count):
		until(mask & 1) {
			mask >>= 1;
			++shift;
		}

		// Count ones:
		do {
			++bits;
			mask >>= 1;
		} while(mask & 1);
	}

	// Extracts a color intensity of arbitrary bit number from the given data.
	//   with bias
	//   useful for Windows BMP bitmask handling
	local UInt1B nTo8(
		UInt4B const data,
		UInt4B const mask,
		UInt4B const shift,
		UInt4B const bits
	) {
		// Place the most significant bit of the mask at the most significant bit of the resulting byte.
		//   8-bit values can be copied with just one shift
		//   fewer bits require bias, i.e. replication of bits
		//   more bits require rounding
		auto result = (data & mask);
		if(bits == 8) {

			result = result >> shift;

		} else if(bits < 8) {

			result = result >> shift << (8 - bits);
			// Replicating one bit to eight places requires three sweeps:
			result |= result >>      bits;
			result |= result >> (2 * bits);
			result |= result >> (4 * bits);

		} else { MUST(bits > 8);

			result = result >> (shift + (bits - 8)); // TODO: rounding

		}
		return UInt1B(result);
	}

	void bitmasked_to_bgrc(
		GFX::SBGRC_8888 * d,
		SSize const       d_linePad,
		Byte const *      s,
		SSize const       s_linePad,
		UIntLeast1B const bytesPerPixel,
		UIntLeast4B const maskR,
		UIntLeast4B const maskG,
		UIntLeast4B const maskB,
		UIntLeast4B const maskC,
		UIntLeast4B const width,
		UIntLeast4B const height
	) {
		MUST(2 == bytesPerPixel || 4 == bytesPerPixel);
		MUST(0 < width && 0 < height);

		auto const s_linePayload = width * bytesPerPixel;
		auto const s_lineSize    = s_linePayload + s_linePad;
		// For each color mask in the bitmap, compute properties like shift count.
		UIntLeast1B shiftR, bitsR;
		UIntLeast1B shiftG, bitsG;
		UIntLeast1B shiftB, bitsB;
		UIntLeast1B shiftC, bitsC;
		getShiftAndBitCount(maskR, shiftR, bitsR);
		getShiftAndBitCount(maskG, shiftG, bitsG);
		getShiftAndBitCount(maskB, shiftB, bitsB);
		getShiftAndBitCount(maskC, shiftC, bitsC);

		// Convert all lines in their original order:
		auto const s_end = byteOffset(s, height * s_lineSize);
		do {

			// Convert all pixels in the line:
			auto const s_lineEnd = byteOffset(s, s_linePayload);
			do {

				// Always process 32-bit data internally. Just pick the correct bytes from the input.
				UInt4B data;
				if(2 == bytesPerPixel) {
					data = littleEndian(*reinterpret_cast<UInt2B const *>(s));
				} else { MUST(4 == bytesPerPixel);
					data = littleEndian(*reinterpret_cast<UInt4B const *>(s));
				}

				d->r = nTo8(data, maskR, shiftR, bitsR);
				d->g = nTo8(data, maskG, shiftG, bitsG);
				d->b = nTo8(data, maskB, shiftB, bitsB);
				if(0 == maskC) { // No alpha channel? Assume one!
					d->c = 255;
				} else {
					d->c = nTo8(data, maskC, shiftC, bitsC);
				}

			} while(++d, (s += bytesPerPixel) < s_lineEnd);
			// Pad:
			s = byteOffset(s, s_linePad);
			d = byteOffset(d, d_linePad);

		} while(s < s_end);
	}


} // Formats