namespace Formats {



	// Microsofts bitmap, a common image container on Windows (used by the ClassicUI and GDI).
	namespace BMP { }

	// A texture format which does not use headers but instead stores pixel format and layout in file size and extension.
	namespace CTX { }

	// Digital Image Design was a British video game developer in the 1990s. DID-specific formats are handled here.
	namespace DID {

		// One of DIDs internal image formats: A very simple bitmap. Used until 1994.
		namespace WAR { }

	}

	// Microsofts icon container on Windows.
	namespace ICO { }

	// Electronic Arts Interchange File Format (IFF), a common container on Amiga.
	namespace IFF { }

	// Standard compression method for Fax machines: CCITT T.4. Used in TIFF & BMP.
	namespace CCITT_T4 { }

	// Namco is a Japanese video game developer. Namco-specific formats are handled here.
	namespace Namco {

		// The compression method used for Ace Combat 3 (1999) and Time Crisis (1995). Headers start with ULZ, hence the name.
		namespace ULZ { }

	}

	// Compression used in ILBM, TGA, and TIFF images; a.k.a. Macintosh PackBits or Macintosh RLE.
	namespace PackBits { }

	// ZSoft Paintbrush, a legacy image format.
	namespace PCX { }

	// Microsofts Resource Interchange File Format (RIFF), a common audio and video container on Windows (.WAV and .AVI files).
	namespace RIFF { }

	// id Software is an American video game developer, best known for Doom and Quake. Their technology has spread widely,
	//  e.g. Valve based their Source engine on Quake.
	namespace id {

		// Image container of Doom, Quake, and Half-Life.
		namespace SPR { }

		// Data container of Doom, Quake, and Half-Life.
		namespace WAD { }

	}

	// The standard interchange format for 3D printing.
	namespace STL { }



	// All kinds of pixel formats which can occur in image files.
	enum class PixelFormat : UInt1B {
		unknown,
		indexed_1bit,   // one-bit indexed (up to two colors)
		indexed_2bit,   // two-bit indexed (up to four colors)
		indexed_3bit,   // three-bit indexed (up to eight colors)
		indexed_4bit,   // four-bit indexed (up to 16 colors)
		indexed_5bit,   // five-bit indexed (up to 32 colors)
		indexed_6bit,   // six-bit indexed (up to 64 colors)
		indexed_7bit,   // seven-bit indexed (up to 128 colors)
		indexed_8bit,   // eight-bit indexed (up to 256 colors)
		indexedHaM_6bit,// six-bit indexed Hold-and-Modify (up to 4096 colors)
		indexedHaM_8bit,// eight-bit indexed Hold-and-Modify (up to 262,144 colors)
		gray_8bit,      // eight-bit grayscale
		bgr_565_le,     // 16-bit High Color
		bgrx_5551_le,   // 15-bit High Color with padding
		bgr_888,        // 24-bit True Color
		bgrc_8888,      // 24-bit True Color with coverage
		bgrc_5551_le,   // 15-bit High Color with coverage
		rgb_888,        // 24-bit True Color
		rgbc_8888,      // 24-bit True Color with coverage
		bgrx_8888,      // 24-bit True Color with padding
		xrxgxb_626262,  // 18-bit High Color with padding
	};


	// All kinds of images appearing in image file formats.
	enum class ImageType : UInt1B {
		unknown,
		// The image is an embedded thumbnail (e.g. the EXIF thumbnail in a JPEG file).
		thumbnail,
		// The image is a page in a multi-page document (e.g. a TIFF containing multiple scanned documents).
		page,
		// The image is a palette variation. Early games used palette swapping extensively to save memory.
		palette,
		// The image is a mip level.
		mip,
		// The image is a frame in an animation.
		frame
	};


	// Properties (width, height, pixel format, ) of a single sub-image in an image file.
	struct ImageProperties {
		UInt4B      widthInPixels;
		UInt4B      heightInPixels;
		ImageType   type;
		PixelFormat pixelFormat;
		UInt2B      paletteLength; // zero for true-color images
	};


	namespace PackBits {


		//  source and destination must not be empty
		//  there must be 128 B padding after destination to account for a damaged stream
		//  returns a pointer to the end of the extacted data
		Byte * extract(
			Byte *         toDestination,
			Byte *         toEndOfDestination,
			Byte const * & toSource,
			Byte const *   toEndOfSource
		);


	} // PackBits


	void rgb_to_bgr1(
		GFX::SBGRC_8888 *    ,
		GFX::SRGB_888 const *,
		UCount
	);
	void rgb_to_bgr1(
		GFX::SBGRC_8888 *          ,
		GFX::SXRXGXB_262626 const *,
		UCount
	);
	void rgb_to_bgr0(
		GFX::SBGRC_8888 *    ,
		GFX::SRGB_888 const *,
		UCount
	);

	void rgb_to_bgr1(
		GFX::SBGRC_8888_reg *,
		GFX::SRGB_888 const *,
		UCount
	);

	void rgb_to_bgr1(
		GFX::SBGRC_8888_reg *     ,
		GFX::SXRGB_4444_BE const *,
		UCount
	);


	void i1_to_bgrc(
		GFX::SBGRC_8888 *       toResult,
		SSize                   resultLinePadding,
		Byte const *            toSource,
		SSize                   sourceLinePadding,
		GFX::SBGRC_8888 const * toSourcePalette,
		UIntLeast4B             widthInPixels,
		UIntLeast4B             heightInPixels
	);
	void i1_to_bgrc_flipY(
		GFX::SBGRC_8888 *       toResult,
		SSize                   resultLinePadding,
		Byte const *            toSource,
		SSize                   sourceLinePadding,
		GFX::SBGRC_8888 const * toSourcePalette,
		UIntLeast4B             widthInPixels,
		UIntLeast4B             heightInPixels
	);

	void i2h_to_bgrc(
		GFX::SBGRC_8888 *       d,
		SSize                   d_linePad,
		Byte const *            s_begin,
		GFX::SBGRC_8888 const * palette,
		SSize                   s_linePad,
		UIntLeast4B             widthInPixels,
		UIntLeast4B             heightInPixels
	);
	void i2h_to_bgrc_flipY(
		GFX::SBGRC_8888 *       d,
		SSize                   d_linePad,
		Byte const *            s_begin,
		GFX::SBGRC_8888 const * palette,
		SSize                   s_linePad,
		UIntLeast4B             widthInPixels,
		UIntLeast4B             heightInPixels
	);

	void i4h_to_bgrc(
		GFX::SBGRC_8888 *       toResult,
		SSize                   resultLinePadding,
		Byte const *            toSource,
		GFX::SBGRC_8888 const * toSourcePalette,
		SSize                   sourceLinePadding,
		UIntLeast4B             widthInPixels,
		UIntLeast4B             heightInPixels
	);
	void i4h_to_bgrc_flipY(
		GFX::SBGRC_8888 *       toResult,
		SSize                   resultLinePadding,
		Byte const *            toBeginningOfSource,
		GFX::SBGRC_8888 const * toSourcePalette,
		SSize                   sourceLinePadding,
		UIntLeast4B             widthInPixels,
		UIntLeast4B             heightInPixels
	);

	void i4l_to_bgrc(
		GFX::SBGRC_8888 *       toResult,
		SSize                   resultLinePadding,
		Byte const *            toBeginningOfSource,
		GFX::SBGRC_8888 const * toSourcePalette,
		SSize                   sourceLinePadding,
		UIntLeast4B             widthInPixels,
		UIntLeast4B             heightInPixels
	);
	void i4l_to_bgrc_flipY(
		GFX::SBGRC_8888 *       toResult,
		SSize                   resultLinePadding,
		Byte const *            toBeginningOfSource,
		GFX::SBGRC_8888 const * toSourcePalette,
		SSize                   sourceLinePadding,
		UIntLeast4B             widthInPixels,
		UIntLeast4B             heightInPixels
	);

	void i8_to_bgrc(
		GFX::SBGRC_8888 *       toResult,
		SSize                   resultLinePadding,
		Byte const *            toSource,
		GFX::SBGRC_8888 const * toSourcePalette,
		SSize                   sourceLinePadding,
		UIntLeast4B             widthInPixels,
		UIntLeast4B             heightInPixels
	);
	void i8_to_bgrc_flipY(
		GFX::SBGRC_8888 *       toResult,
		SSize                   resultLinePadding,
		Byte const *            toBeginningOfSource,
		GFX::SBGRC_8888 const * toSourcePalette,
		SSize                   sourceLinePadding,
		UIntLeast4B             widthInPixels,
		UIntLeast4B             heightInPixels
	);

	void bgr_to_bgrc(
		GFX::SBGRC_8888 *     toResult,
		SSize                 resultLinePadding,
		GFX::SBGR_888 const * toSource,
		SSize                 sourceLinePadding,
		UIntLeast4B           widthInPixels,
		UIntLeast4B           heightInPixels
	);
	void bgr_to_bgrc_flipY(
		GFX::SBGRC_8888 *     toResult,
		SSize                 resultLinePadding,
		GFX::SBGR_888 const * toSource,
		SSize                 sourceLinePadding,
		UIntLeast4B           widthInPixels,
		UIntLeast4B           heightInPixels
	);

	void bgr_to_bgrc_flipY(
		GFX::SBGRC_8888 *         toResult,
		SSize                     resultLinePadding,
		GFX::SBGRC_565_LE const * toSource,
		SSize                     sourceLinePadding,
		UIntLeast4B               widthInPixels,
		UIntLeast4B               heightInPixels
	);

	void bgr_to_bgr1(
		GFX::SBGRC_8888 *       toResult,
		SSize                   resultLinePadding,
		GFX::SBGRX_8888 const * toSource,
		SSize                   sourceLinePadding,
		UIntLeast4B             widthInPixels,
		UIntLeast4B             heightInPixels
	);
	void bgr_to_bgr1_flipY(
		GFX::SBGRC_8888 *       toResult,
		SSize                   resultLinePadding,
		GFX::SBGRX_8888 const * toSource,
		SSize                   sourceLinePadding,
		UIntLeast4B             widthInPixels,
		UIntLeast4B             heightInPixels
	);

	void bgrc_to_bgrc(
		GFX::SBGRC_8888 *       toResult,
		SSize                   resultLinePadding,
		GFX::SBGRC_8888 const * toSource,
		SSize                   sourceLinePadding,
		UIntLeast4B             widthInPixels,
		UIntLeast4B             heightInPixels
	);
	void bgrc_to_bgrc_flipY(
		GFX::SBGRC_8888 *       toResult,
		SSize                   resultLinePadding,
		GFX::SBGRC_8888 const * toSource,
		SSize                   sourceLinePadding,
		UIntLeast4B             widthInPixels,
		UIntLeast4B             heightInPixels
	);

	void bgrc_to_bgrc(
		GFX::SBGRC_8888 *          toResult,
		SSize                      resultLinePadding,
		GFX::SBGRC_5551_LE const * toSource,
		SSize                      sourceLinePadding,
		UIntLeast4B                widthInPixels,
		UIntLeast4B                heightInPixels
	);
	void bgrc_to_bgrc_flipY(
		GFX::SBGRC_8888 *          toResult,
		SSize                      resultLinePadding,
		GFX::SBGRC_5551_LE const * toSource,
		SSize                      sourceLinePadding,
		UIntLeast4B                widthInPixels,
		UIntLeast4B                heightInPixels
	);

	void bgrc_to_bgrc(
		GFX::SBGRC_8888 *          toResult,
		SSize                      resultLinePadding,
		GFX::SRGBC_5551_LE const * toSource,
		SSize                      sourceLinePadding,
		UIntLeast4B                widthInPixels,
		UIntLeast4B                heightInPixels
	);
	void bgrc_to_bgrc_flipY(
		GFX::SBGRC_8888 *          toResult,
		SSize                      resultLinePadding,
		GFX::SRGBC_5551_LE const * toSource,
		SSize                      sourceLinePadding,
		UIntLeast4B                widthInPixels,
		UIntLeast4B                heightInPixels
	);

	void rgb_to_bgr1(
		GFX::SBGRC_8888 *          toResult,
		SSize                      resultLinePadding,
		GFX::SRGBX_5551_LE const * toSource,
		SSize                      sourceLinePadding,
		UIntLeast4B                widthInPixels,
		UIntLeast4B                heightInPixels
	);
	void rgb_to_bgr1_flipY(
		GFX::SBGRC_8888 *          toResult,
		SSize                      resultLinePadding,
		GFX::SRGBX_5551_LE const * toSource,
		SSize                      sourceLinePadding,
		UIntLeast4B                widthInPixels,
		UIntLeast4B                heightInPixels
	);

	void bitmasked_to_bgrc(
		GFX::SBGRC_8888 * toResult,
		SSize             resultLinePadding,
		Byte const *      toSource,
		SSize             sourceLinePadding,
		UIntLeast1B       bytesPerPixel_2or4,
		UIntLeast4B       maskR,
		UIntLeast4B       maskG,
		UIntLeast4B       maskB,
		UIntLeast4B       maskC,
		UIntLeast4B       widthInPixels,
		UIntLeast4B       heightInPixels
	);
	void bitmasked_to_bgrc_flipY(
		GFX::SBGRC_8888 * toResult,
		SSize             resultLinePadding,
		Byte const *      toSource,
		SSize             sourceLinePadding,
		UIntLeast1B       bytesPerPixel_2or4,
		UIntLeast4B       maskR,
		UIntLeast4B       maskG,
		UIntLeast4B       maskB,
		UIntLeast4B       maskC,
		UIntLeast4B       widthInPixels,
		UIntLeast4B       heightInPixels
	);


} // Formats