#include "framework.hpp"
#include "SIMD.hpp"
#include "GFX.hpp"
#include "Memory.hpp"
#include "Math.hpp" // Math::raisedToPower() for gamma correction
#include "Geo.hpp"  // squareRootOf() for gamma correction



namespace GFX {



	LRGBC_reg LRGBC::transparent() {
		return { floatx4Zero() };
	}
	LRGBC_reg LRGBC::opaqueBlack() {
		return { xyzwFrom(0.0f, 0.0f, 0.0f, 1.0f) };
	}
	LRGBC_reg LRGBC::opaqueWhite() {
		return { xyzwFrom(1.0f, 1.0f, 1.0f, 1.0f) };
	}

	LRGBC_reg SIMD_CALL rgbcFrom(
		float const red,
		float const green,
		float const blue,
		float const coverage
	) {
		return { xyzwFrom(red, green, blue, coverage) };
	}

	LRGBC_reg SIMD_CALL load(LRGBC const & memory) {
		return { loadx4(&memory.r) };
	}

	void SIMD_CALL store(
		LRGBC &         memory,
		LRGBC_reg const color
	) {
		xyzwTo(&memory.r, color.rgbc);
	}



	SRGB_888_reg sRGBFrom(
		UInt1B const r,
		UInt1B const g,
		UInt1B const b
	) {
		return { UInt4B((b << 16) | (g << 8) | r) };
	}

	SRGB_888_reg load(SRGB_888 const & rgb) {
		return { UInt4B((rgb.b << 16) | *reinterpret_cast<UInt2B const *>(&rgb.r)) };
	}

	// Converts the given color to linear color space with 100 % coverage.
	LRGBC linearOpaqueFrom(SRGB_888_reg const color) {
		return {
			linearFromSRGB(UInt1B( color.xbgr        & 0xFF)),
			linearFromSRGB(UInt1B((color.xbgr >>  8) & 0xFF)),
			linearFromSRGB(UInt1B((color.xbgr >> 16) & 0xFF)),
			1.0f
		};
	}

	// Interpolates the two colors with a 3:1 ratio. Useful for UI color teints.
	SRGB_888_reg interpolate3to1(SRGB_888_reg const a, SRGB_888_reg const b) {
		auto const a_r = (a.xbgr      ) & 0xFF;
		auto const a_g = (a.xbgr >>  8) & 0xFF;
		auto const a_b = (a.xbgr >> 16) & 0xFF;

		auto const b_r = (b.xbgr      ) & 0xFF;
		auto const b_g = (b.xbgr >>  8) & 0xFF;
		auto const b_b = (b.xbgr >> 16) & 0xFF;

		auto const r_r = (3 * a_r + b_r) / 4;
		auto const r_g = (3 * a_g + b_g) / 4;
		auto const r_b = (3 * a_b + b_b) / 4;

		return { r_r | (r_g << 8) | (r_b << 16) };
	}

	void store(SRGB_888 & memory, SRGB_888_reg const color) {
		*reinterpret_cast<UInt2B *>(&memory) = UInt2B(color.xbgr);
		memory.b = UInt1B(color.xbgr >> 16);
	}


	SRGB_FFF operator *(SRGB_FFF const & l, float const r) {
		return {
			l.r * r,
			l.g * r,
			l.b * r
		};
	}

	SRGB_FFF & operator *=(SRGB_FFF & l, float const r) {
		l.r *= r;
		l.g *= r;
		l.b *= r;
		return l;
	}


	SRGBC_8888_reg SRGBC_8888::solidWhite() {
		return { 0xFFFFFFFF };
	}

	SRGBC_8888_reg SRGBC_8888::transparent() {
		return { 0 };
	}

	// Creates an sRGB value (with linear coverage) from the given values.
	SRGBC_8888_reg sRGBCFrom(
		UInt1B const r,
		UInt1B const g,
		UInt1B const b,
		UInt1B const c
	) {
		return { UInt4B(((c << 24) | (b << 16)) | ((g << 8) | r)) };
	}

	SRGBC_8888_reg load(SRGBC_8888 const & memory) {
		return { *reinterpret_cast<UInt4B const *>(&memory) };
	}

	SRGBC_8888_reg sRGBC_8888_from(SRGB_888_reg const rgb) {
		return { 0xFF000000 | rgb.xbgr };
	}

	SRGBC_8888_reg sRGB1_8888_from(SRGB_FFF const & rgb) {
		// Convert RGB. C is undefined, but it is overwritten anyway.
		return { 0xFF000000 | UNorm1Bx4From(xyzFrom(&rgb.r)) };
	}

	SRGBC_8888_reg sRGBC_8888_from(SRGB_FFF const & rgb, float const c) {
		auto OOOc = wwyxOf(x000From(c));
		auto const rgbc = xyz_wFrom(xyzFrom(&rgb.r), OOOc);
		return { UNorm1Bx4From(rgbc) };
	}

	// Converts the given color to linear color space.
	LRGBC_reg SIMD_CALL linearFrom(SRGBC_8888_reg const color) {
		return { xyzwFrom(
			linearFromSRGB(UInt1B( color.cbgr        & 0xFF)),
			linearFromSRGB(UInt1B((color.cbgr >>  8) & 0xFF)),
			linearFromSRGB(UInt1B((color.cbgr >> 16) & 0xFF)),
			float(color.cbgr >> 24) * (1.0f / 255.0f)
		) };
	}

	bool equalBitwise(SRGBC_8888_reg const a, SRGBC_8888_reg const b) {
		return a.cbgr == b.cbgr;
	}

	void store(
		SRGBC_8888 &         memory,
		SRGBC_8888_reg const color
	) {
		*reinterpret_cast<UInt4B *>(&memory) = color.cbgr;
	}



	SBGRC_8888_reg SBGRC_8888::solidWhite() {
		return { 0xFFFFFFFF };
	}
	SBGRC_8888_reg SBGRC_8888::transparent() {
		return { 0 };
	}

	SBGRC_8888_reg sBGRCFrom(
		UInt1B const b,
		UInt1B const g,
		UInt1B const r,
		UInt1B const c
	) {
		return { UInt4B(((c << 24) | (r << 16)) | ((g << 8) | b)) };
	}

	SBGRC_8888_reg load(SBGRC_8888 const & memory) {
		return { *reinterpret_cast<UInt4B const *>(&memory) };
	}

	// Swizzles the given BGRC color to RGBC.
	SRGBC_8888_reg rgbcOf(SBGRC_8888_reg const color) {

		// Nave approach: 8 instructions on 4 registers.
		// 4-byte endian flip and bit rotation: 4 instructions on 1 register.
		auto const bgrc = flippedEndianness(color.crgb);
		auto const cbgr = rightRotated(bgrc, 8);

		return { cbgr };
	}

	// Swizzles the given RGBC color to BGRC.
	SBGRC_8888_reg bgrcOf(SRGBC_8888_reg const color) {

		// Nave approach: 8 instructions on 4 registers.
		// 4-byte endian flip and bit rotation: 4 instructions on 1 register.
		auto const rgbc = flippedEndianness(color.cbgr);
		auto const crgb = rightRotated(rgbc, 8);

		return { crgb };
	}

	// Swizzles the given RGB color to BGR with 100 % coverage.
	SBGRC_8888_reg bgr1Of(SRGB_888_reg const color) {

		// Nave approach: 9 instructions on 4 registers.
		// 4-byte endian flip and bit rotation: 5 instructions on 1 register.
		auto const rgbx = flippedEndianness(color.xbgr) | 0xFF;
		auto const crgb = rightRotated(rgbx, 8);

		return { crgb };
	}

	// Swizzles the given RGB color to BGR with no coverage.
	SBGRC_8888_reg bgr0Of(SRGB_888_reg const color) {

		// Nave approach: 9 instructions on 4 registers.
		// 4-byte endian flip and bit rotation: 5 instructions on 1 register.
		auto const rgbx = flippedEndianness(color.xbgr) & 0xFFFFFF00;
		auto const crgb = rightRotated(rgbx, 8);

		return { crgb };
	}

	// Converts the given color to linear color space.
	LRGBC_reg SIMD_CALL linearFrom(SBGRC_8888_reg const color) {
		return { xyzwFrom(
			linearFromSRGB(UInt1B((color.crgb >> 16) & 0xFF)),
			linearFromSRGB(UInt1B((color.crgb >>  8) & 0xFF)),
			linearFromSRGB(UInt1B( color.crgb        & 0xFF)),
			float(color.crgb >> 24) * (1.0f / 255.0f)
		) };
	}

	// Computes the given color at half the coverage.
	SBGRC_8888_reg semiTransparent(SBGRC_8888_reg const color) {
		// sRGB => linear =>  2 => sRGB
		local UInt1B const halfs[256] = {
			  0,   0,   1,   2,   2,   3,   3,   4,   4,   5,   5,   6,   6,   7,   7,   8,
			  9,   9,  10,  11,  11,  12,  13,  14,  14,  15,  16,  17,  17,  18,  19,  20,
			 20,  21,  22,  23,  23,  24,  25,  26,  26,  27,  28,  29,  29,  30,  31,  32,
			 32,  33,  34,  35,  35,  36,  37,  38,  38,  39,  40,  41,  41,  42,  43,  44,
			 44,  45,  46,  47,  47,  48,  49,  50,  50,  51,  52,  53,  53,  54,  55,  56,
			 56,  57,  58,  59,  59,  60,  61,  62,  62,  63,  64,  65,  65,  66,  67,  68,
			 68,  69,  70,  71,  71,  72,  73,  74,  74,  75,  76,  77,  77,  78,  79,  80,
			 80,  81,  82,  83,  83,  84,  85,  86,  86,  87,  88,  89,  89,  90,  91,  92,
			 92,  93,  94,  95,  95,  96,  97,  98,  98,  99, 100, 101, 101, 102, 103, 104,
			104, 105, 106, 107, 107, 108, 109, 110, 110, 111, 112, 113, 113, 114, 115, 116,
			116, 117, 118, 119, 119, 120, 121, 122, 122, 123, 124, 125, 125, 126, 127, 128,
			128, 129, 130, 131, 131, 132, 133, 134, 134, 135, 136, 137, 137, 138, 139, 140,
			140, 141, 142, 143, 143, 144, 145, 146, 146, 147, 148, 149, 149, 150, 151, 152,
			152, 153, 154, 155, 155, 156, 157, 158, 158, 159, 160, 161, 161, 162, 163, 164,
			164, 165, 166, 167, 167, 168, 169, 170, 170, 171, 172, 173, 173, 174, 175, 176,
			176, 177, 178, 179, 179, 180, 181, 182, 182, 183, 184, 185, 185, 186, 187, 188
		};

		return sBGRCFrom(
			halfs[ color.crgb        & 0xFF],
			halfs[(color.crgb >>  8) & 0xFF],
			halfs[(color.crgb >> 16) & 0xFF],
			UInt1B(color.crgb / 2)
		);
	}

	bool equalBitwise(SBGRC_8888_reg const a, SBGRC_8888_reg const b) {
		return a.crgb == b.crgb;
	}

	void store(SBGRC_8888 & memory, SBGRC_8888_reg const color) {
		*reinterpret_cast<UInt4B *>(&memory) = color.crgb;
	}

	// Performs 8-bit alpha blending.
	//   not premultiplied
	//   the foregrounds lane [3] specifies the coverage
	//   expects 16-bit colors because its the fast path, but uses only the lowest 8 bits
	local UNorm1Bx4 SIMD_CALL alphaBlended_8bit(
		UNorm1Bx4 const ffff,
		UNorm1Bx4 const bbbb
	) {
		// Coverage (multiplicand for foreground):
		auto const cccc = dddd_of(ffff); 
		// Transmittance (multiplicand for background):
		auto const tttt = loadUNorm1Bx4(255) - cccc;

		return dotProductOf(ffff, cccc, bbbb, tttt);
	}

	// Performs 8-bit alpha blending.
	//   premultiplied
	//   the foregrounds lane [3] specifies the coverage
	//   expects 16-bit colors because its the fast path, but uses only the lowest 8 bits
	local UNorm1Bx4 SIMD_CALL alphaBlended_8bit_premultiplied(
		UNorm1Bx4 const ffff,
		UNorm1Bx4 const bbbb
	) {
		// White (multiplicand for foreground):
		auto const white = loadUNorm1Bx4(255);
		// Transmittance (multiplicand for background):
		auto const tttt = white - dddd_of(ffff);

		return dotProductOf(ffff, white, bbbb, tttt);
	}

	// Blends the given pixels in-place with the given background color.
	void alphaBlend_inPlace(
		GFX::SBGRC_8888_reg const background_raw,
		GFX::SBGRC_8888           pixels[],
		GFX::SBGRC_8888 * const   toEndOfPixels,
		bool const                isPremultiplied
	) {
		// TODO: Avoid this store!
		GFX::SBGRC_8888 background_mem; store(background_mem, background_raw);
		auto const background = loadUNormx4(&background_mem.b);

		auto toPixel = pixels;
		do {
			auto const foreground = loadUNormx4(&toPixel->b);
			UNorm1Bx4  result;
			if(isPremultiplied) {
				result = alphaBlended_8bit_premultiplied(foreground, background);
			} else {
				result = alphaBlended_8bit(foreground, background);
			}
			store(&toPixel->b, result);
		} while(++toPixel < toEndOfPixels);

	}


	QBGRC_8888_reg QBGRC_8888::opaqueBlack() {
		return { 0xFF000000 };
	}
	QBGRC_8888_reg QBGRC_8888::solidWhite() {
		return { 0xFFFFFFFF };
	}

	QBGRC_8888_reg load(QBGRC_8888 const & memory) {
		return { *reinterpret_cast<UInt4B const *>(&memory) };
	}

	QBGRC_8888_reg qBGRCFrom(
		UInt1B const b,
		UInt1B const g,
		UInt1B const r,
		UInt1B const c
	) {
		return { UInt4B(((c << 24) | (r << 16)) | ((g << 8) | b)) };
	}

	QBGRC_8888_reg qBGRCFromLinear(
		float const linearRed,
		float const linearGreen,
		float const linearBlue,
		float const coverage
	) {
		MUST(0.0f <= coverage && 1.0f >= coverage);

		return qBGRCFrom(
			squareRootFromLinear(linearBlue),
			squareRootFromLinear(linearGreen),
			squareRootFromLinear(linearRed),
			UInt1B(rounded(255.0f * coverage))
		);
	}

	// Converts the given color to linear color space.
	LRGBC_reg SIMD_CALL linearFrom(QBGRC_8888_reg const color) {
		auto const rgbc = loadx4(1.0f / 255.0f) * zyxwOf(asFloats(xyzwFrom(color.crgb)));
		return { xyz_wFrom(rgbc * rgbc, rgbc) };
	}

	// Converts the given color to linear color space and ignores coverage (slightly faster).
	LRGBC_reg SIMD_CALL linearRGBFrom(QBGRC_8888_reg const color) {
		auto const rgbc = loadx4(1.0f / 255.0f) * zyxwOf(asFloats(xyzwFrom(color.crgb)));
		return { rgbc * rgbc };
	}

	// Converts the given color from square root to sRGB color space.
	SBGRC_8888_reg sBGRCFrom(QBGRC_8888_reg const color) {
		// TODO: do this straighforward
		auto linear = linearFrom(color);
		return sBGRCFrom(
			sRGBFromLinear(xOf(linear.rgbc)),
			sRGBFromLinear(yOf(linear.rgbc)),
			sRGBFromLinear(zOf(linear.rgbc)),
			UInt1B(wOf(linear.rgbc) * 255.0f)
		);
	}

	// Converts the given color from sRGB to square root color space.
	QBGRC_8888_reg qBGRCFrom(SBGRC_8888 const sRGBC) {
		return qBGRCFrom(
			qBGRCFrom(sRGBC.b),
			qBGRCFrom(sRGBC.g),
			qBGRCFrom(sRGBC.r),
			sRGBC.c
		);
	}

	// Computes the given color at half the coverage.
	QBGRC_8888_reg semiTransparent(QBGRC_8888_reg const color) {
		// square root => linear =>  2 => square root
		local UInt1B const halfs[256] = {
			  0,   3,   4,   5,   6,   7,   8,   8,   9,   9,  10,  10,  11,  11,  12,  12,
			 13,  13,  14,  15,  15,  16,  16,  17,  17,  18,  18,  19,  19,  20,  21,  21,
			 22,  22,  23,  23,  24,  25,  25,  26,  26,  27,  27,  28,  29,  29,  30,  30,
			 31,  32,  32,  33,  33,  34,  35,  35,  36,  36,  37,  38,  38,  39,  40,  40,
			 41,  41,  42,  43,  43,  44,  45,  45,  46,  47,  47,  48,  48,  49,  50,  50,
			 51,  52,  52,  53,  54,  54,  55,  56,  56,  57,  58,  58,  59,  60,  60,  61,
			 62,  62,  63,  64,  64,  65,  66,  66,  67,  68,  68,  69,  70,  71,  71,  72,
			 73,  73,  74,  75,  75,  76,  77,  77,  78,  79,  80,  80,  81,  82,  82,  83,
			 84,  84,  85,  86,  87,  87,  88,  89,  89,  90,  91,  92,  92,  93,  94,  95,
			 95,  96,  97,  97,  98,  99, 100, 100, 101, 102, 102, 103, 104, 105, 105, 106,
			107, 108, 108, 109, 110, 111, 111, 112, 113, 114, 114, 115, 116, 117, 117, 118,
			119, 120, 120, 121, 122, 123, 123, 124, 125, 126, 126, 127, 128, 129, 129, 130,
			131, 132, 132, 133, 134, 135, 135, 136, 137, 138, 139, 139, 140, 141, 142, 142,
			143, 144, 145, 146, 146, 147, 148, 149, 149, 150, 151, 152, 153, 153, 154, 155,
			156, 156, 157, 158, 159, 160, 160, 161, 162, 163, 164, 164, 165, 166, 167, 168,
			168, 169, 170, 171, 172, 172, 173, 174, 175, 175, 176, 177, 178, 179, 180, 180
		};

		return qBGRCFrom(
			halfs[ color.crgb        & 0xFF],
			halfs[(color.crgb >>  8) & 0xFF],
			halfs[(color.crgb >> 16) & 0xFF],
			UInt1B((color.crgb >> 24) / 2)
		);
	}

	bool equalBitwise(QBGRC_8888_reg const a, QBGRC_8888_reg const b) {
		return a.crgb == b.crgb;
	}

	void store(
		QBGRC_8888 &         memory,
		QBGRC_8888_reg const color
	) {
		*reinterpret_cast<UInt4B *>(&memory) = color.crgb;
	}


	QRGBC_8888_reg QRGBC_8888::solidWhite() {
		return { 0xFFFFFFFF };
	}

	QRGBC_8888_reg load(QRGBC_8888 const & memory) {
		return { *reinterpret_cast<UInt4B const *>(&memory) };
	}

	QRGBC_8888_reg qRGBCFrom(
		UInt1B const r,
		UInt1B const g,
		UInt1B const b,
		UInt1B const c
	) {
		return { UInt4B(((c << 24) | (b << 16)) | ((g << 8) | r)) };
	}

	// Converts the given color from sRGB to square root color space.
	QRGBC_8888_reg qRGBCFrom(SRGBC_8888 const sRGBC) {
		return qRGBCFrom(
			qBGRCFrom(sRGBC.r),
			qBGRCFrom(sRGBC.g),
			qBGRCFrom(sRGBC.b),
			sRGBC.c
		);
	}

	void store(QRGBC_8888 & memory, QRGBC_8888_reg const color) {
		*reinterpret_cast<UInt4B *>(&memory) = color.cbgr;
	}



	SBGRC_5551_reg load(SBGRC_5551_LE const & memory) {
		return { memory.data };
	}

	// Expands the given color to 32-bit color depth.
	//   with bias
	SBGRC_8888_reg sBGRC_8888_from(SBGRC_5551_reg const color) {

		// The transparency control bit causes some weirdness here:
		//  color T R G B . . with alpha . . without alpha
		//        0 0 0 0     transparent    transparent
		//        0 X X X     opaque         opaque
		//        1 X X X     additive       opaque
		//        1 0 0 0     opaque black   opaque black
		if(0 == color.crgb) {
			return { 0 };
		}
#		if 0
			auto const r = _5to8(UInt1B(0xFF & (color.crgb & 0x7c00) >> 10))      ;
			auto const g = _5to8(UInt1B(0xFF & (color.crgb & 0x03e0) >>  5)) <<  8;
			auto const b = _5to8(UInt1B(0xFF & (color.crgb & 0x001f)      )) << 16;
			if(color.crgb & 0x8000) {
				return { UInt4B(r | g | b) };
			} else {
				return { UInt4B(0xFF000000 | r | g | b) };
			}
#		else
			auto const r = _5to8(UInt1B(0xFF & (color.crgb & 0x7c00) >> 10))      ;
			auto const g = _5to8(UInt1B(0xFF & (color.crgb & 0x03e0) >>  5)) <<  8;
			auto const b = _5to8(UInt1B(0xFF & (color.crgb & 0x001f)      )) << 16;
			return { UInt4B(0xFF000000 | r | g | b) };
#		endif
	}

	// Expands the given color to 24-bit color depth with full coverage.
	//   with bias
	SRGBC_8888_reg sRGB1_8888_from(SBGRC_5551_reg const color) {
		auto const r = _5to8(UInt1B(0xFF & (color.crgb & 0x7c00) >> 10))      ;
		auto const g = _5to8(UInt1B(0xFF & (color.crgb & 0x03e0) >>  5)) <<  8;
		auto const b = _5to8(UInt1B(0xFF & (color.crgb & 0x001f)      )) << 16;
		return { UInt4B(0xFF000000 | r | g | b) };
	}



	SRGBC_5551_reg load(SRGBC_5551_LE const & memory) {
		return { memory.data };
	}

	void store(SRGBC_5551_LE & memory, SRGBC_5551_reg const color) {
		memory.data = color.cbgr;
	}

	// Expands the given color to 24-bit color depth with full coverage.
	//   with bias
	SRGBC_8888_reg sRGB1_8888_from(SRGBC_5551_reg const color) {
		auto const r = _5to8(UInt1B(0xFF & (color.cbgr & 0x001f)      ))      ;
		auto const g = _5to8(UInt1B(0xFF & (color.cbgr & 0x03e0) >>  5)) <<  8;
		auto const b = _5to8(UInt1B(0xFF & (color.cbgr & 0x7c00) >> 10)) << 16;
		return { UInt4B(0xFF000000 | r | g | b) };
	}

	// Expands the given color to 32-bit color depth.
	//   with bias
	SBGRC_8888_reg sBGRC_8888_from(SRGBC_5551_reg const color) {
		auto const b = _5to8(UInt1B(0xFF & (color.cbgr & 0x7c00) >> 10)) << 16;
		auto const g = _5to8(UInt1B(0xFF & (color.cbgr & 0x03e0) >>  5)) <<  8;
		auto const r = _5to8(UInt1B(0xFF & (color.cbgr & 0x001f)      ))      ;
		if(color.cbgr & 0x8000) {
			return { UInt4B(0xFF000000 | r | g | b) };
		} else {
			return { UInt4B(r | g | b) };
		}
	}

	SRGBC_5551_reg sRGBC_5551_from(SRGBC_8888_reg const color) {
		auto const r = (((color.cbgr      ) & 0xFF) /   8)      ;
		auto const g = (((color.cbgr >>  8) & 0xFF) /   8) <<  5;
		auto const b = (((color.cbgr >> 16) & 0xFF) /   8) << 10;
		auto const c = (((color.cbgr >> 24) & 0xFF) / 128) << 15;
		return { UInt2B(c | b | g | r) };
	}

	bool hasCoverage(SRGBC_5551_reg const color) {
		return 0 != (color.cbgr & 0x8000);
	}
	


	SRGBX_5551_reg load(SRGBX_5551_LE const & memory) {
		return { memory.data };
	}

	void store(SRGBX_5551_LE & memory, SRGBX_5551_reg const color) {
		memory.data = color.cbgr;
	}

	// Expands the given color to 24-bit color depth with full coverage.
	//   with bias
	SRGBC_8888_reg sRGB1_8888_from(SRGBX_5551_reg const color) {
		auto const r = _5to8(UInt1B(0xFF & (color.cbgr & 0x001f)      ))      ;
		auto const g = _5to8(UInt1B(0xFF & (color.cbgr & 0x03e0) >>  5)) <<  8;
		auto const b = _5to8(UInt1B(0xFF & (color.cbgr & 0x7c00) >> 10)) << 16;
		return { UInt4B(0xFF000000 | r | g | b) };
	}

	// Expands the given color to 24-bit color depth with full coverage.
	//   with bias
	SBGRC_8888_reg sBGR1_8888_from(SRGBX_5551_reg const color) {
		auto const b = _5to8(UInt1B(0xFF & (color.cbgr & 0x7c00) >> 10)) << 16;
		auto const g = _5to8(UInt1B(0xFF & (color.cbgr & 0x03e0) >>  5)) <<  8;
		auto const r = _5to8(UInt1B(0xFF & (color.cbgr & 0x001f)      ))      ;
		return { UInt4B(0xFF000000 | r | g | b) };
	}





	SXRGB_4444_reg load(SXRGB_4444_BE const & memory) {
		return { bigEndian(memory.data) };
	}

	// Expands the given color to 32-bit color depth.
	//   with bias
	SBGRC_8888_reg sBGRC_8888_from(SXRGB_4444_reg const color) {
		MUST(0 == color.xrgb >> 12);

		return sBGRCFrom(
			_4to8((color.xrgb >> 0) & 0xF),
			_4to8((color.xrgb >> 4) & 0xF),
			_4to8((color.xrgb >> 8) & 0xF),
			0xFF
		);
	}






	// Converts the given 4-bit unsigned normalized value to an 8-bit unsigned normalized value with bias.
	// Useful to unpack 12-bpp texture formats.
	UInt1B _4to8(UInt1B value) {
		// Bias: If we just multiplied by 8, the old maximum (15) would not be mapped to the new maximum (240 instead of 255).
		//  Therefore, the new least significant bits must be padded with a copy of the most significant bits.
		return UInt1B(value << 4) | value;
	}

	// Converts the given 5-bit unsigned normalized value to an 8-bit unsigned normalized value with bias.
	// Useful to unpack 16-bpp texture formats.
	UInt1B _5to8(UInt1B value) {
		// Bias: If we just multiplied by 8, the old maximum (31) would not be mapped to the new maximum (248 instead of 255).
		//  Therefore, the new least significant bits must be padded with a copy of the most significant bits.
		return UInt1B((value << 3) | (value >> 2));
	}

	// Converts the given 6-bit unsigned normalized value to an 8-bit unsigned normalized value with bias.
	// Useful to unpack 16-bpp texture formats.
	UInt1B _6to8(UInt1B value) {
		// Bias: If we just multiplied by 4, the old maximum (63) would not be mapped to the new maximum (252 instead of 255).
		//  Therefore, the new least significant bits must be padded with a copy of the most significant bits.
		return UInt1B((value << 2) | (value >> 4));
	}

	// Converts the given 8-bit unsigned normalized luminance in sRGB color space to a linear color space luminance.
	float linearFromSRGB(UInt1B sRGBValue) {
		// See Jason Summers (for ImageWorsener), A close look at the sRGB formula.
		//auto const sRGBValue = rawSRGBValue / 255.0f;
		//if(0.0404482362771082f >= sRGBValue) {
		//	return sRGBValue / 12.92f;
		//} else {
		//	return Math::raisedToPower((sRGBValue + 0.055f) / 1.055f, 2.4f);
		//}

		local float const sRGBToLinearTable[] = {
			0.0f,          0.00030352699f, 0.00060705398f, 0.00091058097f, 0.0012141080f, 0.0015176350f, 0.0018211619f, 0.0021246888f,
			0.0024282159f, 0.0027317430f,  0.0030352699f,  0.0033465356f,  0.0036765069f, 0.0040247170f, 0.0043914421f, 0.0047769533f,
			0.0051815170f, 0.0056053917f,  0.0060488326f,  0.0065120910f,  0.0069954102f, 0.0074990317f, 0.0080231922f, 0.0085681248f,
			0.0091340570f, 0.0097212177f,  0.010329823f,   0.010960094f,   0.011612245f,  0.012286487f,  0.012983031f,  0.013702081f,
			0.014443844f,  0.015208514f,   0.015996292f,   0.016807375f,   0.017641952f,  0.018500218f,  0.019382361f,  0.020288562f,
			0.021219010f,  0.022173883f,   0.023153365f,   0.024157630f,   0.025186857f,  0.026241222f,  0.027320892f,  0.028426038f,
			0.029556833f,  0.030713441f,   0.031896029f,   0.033104762f,   0.034339808f,  0.035601314f,  0.036889445f,  0.038204364f,
			0.039546236f,  0.040915199f,   0.042311411f,   0.043735027f,   0.045186203f,  0.046665084f,  0.048171822f,  0.049706563f,
			0.051269468f,  0.052860655f,   0.054480281f,   0.056128494f,   0.057805434f,  0.059511241f,  0.061246071f,  0.063010029f,
			0.064803280f,  0.066625953f,   0.068478182f,   0.070360109f,   0.072271861f,  0.074213579f,  0.076185390f,  0.078187428f,
			0.080219828f,  0.082282715f,   0.084376216f,   0.086500466f,   0.088655606f,  0.090841733f,  0.093058981f,  0.095307484f,
			0.097587362f,  0.099898741f,   0.10224175f,    0.10461649f,    0.10702311f,   0.10946172f,   0.11193243f,   0.11443538f,
			0.11697067f,   0.11953843f,    0.12213880f,    0.12477184f,    0.12743770f,   0.13013649f,   0.13286833f,   0.13563335f,
			0.13843162f,   0.14126331f,    0.14412849f,    0.14702728f,    0.14995980f,   0.15292616f,   0.15592647f,   0.15896086f,
			0.16202940f,   0.16513222f,    0.16826940f,    0.17144109f,    0.17464739f,   0.17788841f,   0.18116423f,   0.18447499f,
			0.18782076f,   0.19120167f,    0.19461781f,    0.19806930f,    0.20155624f,   0.20507871f,   0.20863685f,   0.21223073f,
			0.21586053f,   0.21952623f,    0.22322798f,    0.22696589f,    0.23074007f,   0.23455065f,   0.23839766f,   0.24228120f,
			0.24620140f,   0.25015837f,    0.25415218f,    0.25818291f,    0.26225072f,   0.26635566f,   0.27049786f,   0.27467737f,
			0.27889434f,   0.28314880f,    0.28744090f,    0.29177070f,    0.29613832f,   0.30054384f,   0.30498737f,   0.30946895f,
			0.31398875f,   0.31854683f,    0.32314324f,    0.32777813f,    0.33245158f,   0.33716366f,   0.34191445f,   0.34670410f,
			0.35153270f,   0.35640025f,    0.36130688f,    0.36625269f,    0.37123778f,   0.37626222f,   0.38132611f,   0.38642952f,
			0.39157256f,   0.39675531f,    0.40197787f,    0.40724030f,    0.41254270f,   0.41788515f,   0.42326775f,   0.42869055f,
			0.43415371f,   0.43965724f,    0.44520125f,    0.45078585f,    0.45641106f,   0.46207705f,   0.46778384f,   0.47353154f,
			0.47932023f,   0.48514998f,    0.49102089f,    0.49693304f,    0.50288659f,   0.50888145f,   0.51491779f,   0.52099568f,
			0.52711523f,   0.53327650f,    0.53947961f,    0.54572457f,    0.55201149f,   0.55834049f,   0.56471163f,   0.57112491f,
			0.57758051f,   0.58407849f,    0.59061891f,    0.59720188f,    0.60382742f,   0.61049563f,   0.61720663f,   0.62396044f,
			0.63075721f,   0.63759696f,    0.64447975f,    0.65140569f,    0.65837485f,   0.66538733f,   0.67244321f,   0.67954254f,
			0.68668544f,   0.69387192f,    0.70110202f,    0.70837593f,    0.71569365f,   0.72305524f,   0.73046088f,   0.73791057f,
			0.74540436f,   0.75294232f,    0.76052463f,    0.76815128f,    0.77582234f,   0.78353792f,   0.79129803f,   0.79910284f,
			0.80695236f,   0.81484669f,    0.82278585f,    0.83076996f,    0.83879912f,   0.84687328f,   0.85499269f,   0.86315727f,
			0.87136722f,   0.87962234f,    0.88792318f,    0.89626938f,    0.90466136f,   0.91309869f,   0.92158204f,   0.93011087f,
			0.93868589f,   0.94730657f,    0.95597351f,    0.96468627f,    0.97344548f,   0.98225057f,   0.99110222f,   1.0f
		};
		MUST_STATIC(256 == CAPACITY_OF(sRGBToLinearTable));

		return sRGBToLinearTable[sRGBValue];
	}

	// Converts the given linear color space luminance to an 8-bit unsigned normalized luminance in sRGB color space.
	UInt1B sRGBFromLinear(float linearValue) {
		MUST(isWithin(0.0f, linearValue, 1.0f));

		// See Jason Summers (for ImageWorsener), A close look at the sRGB formula.
		if(0.00313066844250063f >= linearValue) {
			return UInt1B(rounded(float(255.0 * 12.92) * linearValue));
		} else {
			// Modify the constants of the sRGB formula to automatically scale to [0, 255] instead of [0, 1].
			auto sRGBValue = float(255.0 * 1.055) * Math::raisedToPower(linearValue, float(1.0 / 2.4)) - float(255.0 * 0.055);
			return UInt1B(rounded(sRGBValue));
		}
	}

	// Converts the given linear color space luminance to an 8-bit unsigned luminance in the square root space.
	// Useful to deliver parameters to high-performance shaders.
	UInt1B squareRootFromLinear(float linearValue) {
		MUST(isWithin(0.0f, linearValue, 1.0f));

		return UInt1B(rounded(255.0f * squareRootOf(linearValue)));
	}

	UInt1B qBGRCFrom(UInt1B srgb) {
		static UInt1B const lookup[] = {
			  0,   4,   6,   8,   9,  10,  11,  12,  13,  13,  14,  15,  15,  16,  17,  18,
			 18,  19,  20,  21,  21,  22,  23,  24,  24,  25,  26,  27,  27,  28,  29,  30,
			 31,  31,  32,  33,  34,  35,  36,  36,  37,  38,  39,  40,  40,  41,  42,  43,
			 44,  45,  46,  46,  47,  48,  49,  50,  51,  52,  52,  53,  54,  55,  56,  57,
			 58,  59,  60,  60,  61,  62,  63,  64,  65,  66,  67,  68,  69,  69,  70,  71,
			 72,  73,  74,  75,  76,  77,  78,  79,  80,  81,  82,  82,  83,  84,  85,  86,
			 87,  88,  89,  90,  91,  92,  93,  94,  95,  96,  97,  98,  99, 100, 101, 102,
			103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 112, 113, 114, 115, 116, 117,
			118, 119, 120, 121, 122, 123, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134,
			135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150,
			151, 152, 153, 154, 155, 156, 157, 159, 160, 161, 162, 163, 164, 165, 166, 167,
			168, 169, 170, 171, 172, 173, 174, 175, 177, 178, 179, 180, 181, 182, 183, 184,
			185, 186, 187, 188, 189, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201,
			203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 214, 215, 216, 217, 218, 219,
			220, 221, 222, 223, 225, 226, 227, 228, 229, 230, 231, 232, 234, 235, 236, 237,
			238, 239, 240, 241, 243, 244, 245, 246, 247, 248, 249, 250, 252, 253, 254, 255
		};
		return lookup[srgb];
	}

	// Tries to reserve space for an image of the given dimensions.
	//   destroy() after use
	//   returns nullptr on failure
	//   content undefined
	Pixels_r8g8b8c8_sRGB * tryToReserve(
		UIntLeast2B width,
		UIntLeast2B height
	) {
		auto const result = reinterpret_cast<Pixels_r8g8b8c8_sRGB *>(Heap::tryToReserve(
			sizeof(Pixels_r8g8b8c8_sRGB) + width * height * sizeof(GFX::SRGBC_8888)
		));
		if(nullptr != result) {
			result->width  = UInt2B(width);
			result->height = UInt2B(height);
		}
		return result;
	}

	void destroy(Pixels_r8g8b8c8_sRGB & pixels) {
		Heap::release(&pixels);
	}


	void sRGBFromLinear(
		UCount const        width,
		SRGBC_8888 *        toDst,
		USize const         dstStride,
		LRGBC const *       toSrc,
		USize const         srcStride,
		LRGBC const * const toEndOfSrc
	) {
		MUST(toSrc < toEndOfSrc);
		MUST(0 < width);
		MUST(dstStride >= width * sizeof *toDst);
		MUST(srcStride >= width * sizeof *toSrc);

		auto const dstPadding = dstStride - width * sizeof *toDst;
		auto const srcPadding = srcStride - width * sizeof *toSrc;
		do {
			UCount x = 0;
			do {

				toDst->r = sRGBFromLinear(toSrc->r);
				toDst->g = sRGBFromLinear(toSrc->g);
				toDst->b = sRGBFromLinear(toSrc->b);
				toDst->c = UInt1B(rounded(255.0f * toSrc->c)); // coverage is always linear

			} while(++toDst, ++toSrc, ++x < width);
			toSrc = byteOffset(toSrc, srcPadding);
			toDst = byteOffset(toDst, dstPadding);
		} while(toSrc < toEndOfSrc);
	}


	LRGBC const & access(
		LRGBC const * const toBase,
		USize const         stride,
		UCount const        x,
		UCount const        y
	) {
		return byteOffset(toBase, y * stride)[x];
	}

	void sRGBFromLinear_dithered(
		UCount        width,
		SRGBC_8888 *  toDst,
		USize         dstStride,
		LRGBC const * toSrc,
		USize         srcStride,
		LRGBC const * toEndOfSrc
	) {
		MUST(toSrc < toEndOfSrc);
		MUST(0 < width);
		MUST(dstStride >= width * sizeof *toDst);
		MUST(srcStride >= width * sizeof *toSrc);

		UCount const height = (reinterpret_cast<Byte const *>(toEndOfSrc) - reinterpret_cast<Byte const *>(toSrc)) / srcStride;

		auto const buffer = Heap::tryToReserve<LRGBC>(width * height);
		for(auto y = 0u; y < height; ++y) {
			for(auto x = 0u; x < width; ++x) {
				buffer[y * width + x] = access(toSrc, srcStride, x, y);
			}
		}

		auto const offset_right = 1;
		auto const offset_bleft = width - 1;
		auto const offset_bottom = width;
		auto const offset_bright = width + 1;
		for(auto y = 0u; y < height; ++y) {
			for(auto x = 0u; x < width; ++x) {

				// The value with the previous quantization errors accumulated, in linear color space.
				auto const & accumulated_linear = buffer[y * width + x];

				// The value with the previous quantization errors accumulated, in sRGB color space.
				SRGBC_8888 const accumulated_sRGB = {
					sRGBFromLinear(clamped(0.0f, accumulated_linear.r, 1.0f)),
					sRGBFromLinear(clamped(0.0f, accumulated_linear.g, 1.0f)),
					sRGBFromLinear(clamped(0.0f, accumulated_linear.b, 1.0f))
				};

				// The value that is written to the file.
				toDst[y * width + x] = accumulated_sRGB;

				// The value that has been restored from the bits written to the file, in linear color space. I.e. this is what the GPU
				//  samples.
				LRGBC const reconstructed_linear = {
					linearFromSRGB(accumulated_sRGB.r),
					linearFromSRGB(accumulated_sRGB.g),
					linearFromSRGB(accumulated_sRGB.b)
				};

				// Spread the error over the neighbouring pixels.
				LRGBC const errorBy16_linear = {
					 (accumulated_linear.r - reconstructed_linear.r) / 16.0f,
					 (accumulated_linear.g - reconstructed_linear.g) / 16.0f,
					 (accumulated_linear.b - reconstructed_linear.b) / 16.0f
				};

				if((width * height) > y * width + x + offset_right) {
					auto & right = buffer[y * width + x + offset_right];
					right.r += errorBy16_linear.r * 7.0f;
					right.g += errorBy16_linear.g * 7.0f;
					right.b += errorBy16_linear.b * 7.0f;
				}

				if((width * height) > y * width + x + offset_bleft) {
					auto & bleft = buffer[y * width + x + offset_bleft];
					bleft.r += errorBy16_linear.r * 3.0f;
					bleft.g += errorBy16_linear.g * 3.0f;
					bleft.b += errorBy16_linear.b * 3.0f;
				}

				if((width * height) > y * width + x + offset_bottom) {
					auto & bottom = buffer[y * width + x + offset_bottom];
					bottom.r += errorBy16_linear.r * 5.0f;
					bottom.g += errorBy16_linear.g * 5.0f;
					bottom.b += errorBy16_linear.b * 5.0f;
				}

				if((width * height) > y * width + x + offset_bright) {
					auto & bright = buffer[y * width + x + offset_bright];
					bright.r += errorBy16_linear.r;
					bright.g += errorBy16_linear.g;
					bright.b += errorBy16_linear.b;
				}

			}
		}

		Heap::release(buffer);
	}


	// Interpolates the color in a triangle according to barycentric coordinates.
	//   linear interpolation
	SRGBC_8888 SIMD_CALL interpolatedTriangleColor(
		SRGBC_8888 const & colorAtU,
		SRGBC_8888 const & colorAtV,
		SRGBC_8888 const & colorAtW,
		floatx4 const      uvw
	) {
		GFX::SRGBC_8888 result;

		auto const colorU_rgbc = asFloats(asSInt4B(loadx4(&colorAtU.r)));
		auto const colorV_rgbc = asFloats(asSInt4B(loadx4(&colorAtV.r)));
		auto const colorW_rgbc = asFloats(asSInt4B(loadx4(&colorAtW.r)));

		auto const uuu_ = xxxxOf(uvw);
		auto const vvv_ = yyyyOf(uvw);
		auto const www_ = zzzzOf(uvw);

		auto const color_rgbc = asUNorm1B_0_255(
			  uuu_ * colorU_rgbc
			+ vvv_ * colorV_rgbc
			+ www_ * colorW_rgbc
		);

		store(&result.r, color_rgbc);
		return result;
	}


} // GFX



// The VGA palette.
//   first 16 entries are the standard CGA/EGA palette
static GFX::SBGRC_8888 const vgaPalette_bgrc[] = {
	// The standard EGA palette:
	{ 0x00, 0x00, 0x00, 0xFF }, // 0: black
	{ 0xAA, 0x00, 0x00, 0xFF }, // 1: blue
	{ 0x00, 0xAA, 0x00, 0xFF }, // 2: green
	{ 0xAA, 0xAA, 0x00, 0xFF }, // 3: cyan
	{ 0x00, 0x00, 0xAA, 0xFF }, // 4: red
	{ 0xAA, 0x00, 0xAA, 0xFF }, // 5: magenta
	{ 0x00, 0x55, 0xAA, 0xFF }, // 6: brown
	{ 0xAA, 0xAA, 0xAA, 0xFF }, // 7: light gray
	{ 0x55, 0x55, 0x55, 0xFF }, // 8: gray
	{ 0xFF, 0x55, 0x55, 0xFF }, // 9: light blue
	{ 0x55, 0xFF, 0x55, 0xFF }, // A: light green
	{ 0xFF, 0xFF, 0x55, 0xFF }, // B: light cyan
	{ 0x55, 0x55, 0xFF, 0xFF }, // C: light red
	{ 0xFF, 0x55, 0xFF, 0xFF }, // D: light magenta
	{ 0x55, 0xFF, 0xFF, 0xFF }, // E: yellow
	{ 0xFF, 0xFF, 0xFF, 0xFF }, // F: white
	// Linear, non-gammacorrected grayscale:
	{ 0x00, 0x00, 0x00, 0xFF }, { 0x14, 0x14, 0x14, 0xFF }, { 0x20, 0x20, 0x20, 0xFF }, { 0x2C, 0x2C, 0x2C, 0xFF },
	{ 0x38, 0x38, 0x38, 0xFF }, { 0x45, 0x45, 0x45, 0xFF }, { 0x51, 0x51, 0x51, 0xFF }, { 0x61, 0x61, 0x61, 0xFF },
	{ 0x71, 0x71, 0x71, 0xFF }, { 0x82, 0x82, 0x82, 0xFF }, { 0x92, 0x92, 0x92, 0xFF }, { 0xA2, 0xA2, 0xA2, 0xFF },
	{ 0xB6, 0xB6, 0xB6, 0xFF }, { 0xCB, 0xCB, 0xCB, 0xFF }, { 0xE3, 0xE3, 0xE3, 0xFF }, { 0xFF, 0xFF, 0xFF, 0xFF },
	// Hue mix (no white):
	{ 0xFF, 0x00, 0x00, 0xFF }, { 0xFF, 0x00, 0x41, 0xFF }, { 0xFF, 0x00, 0x7D, 0xFF }, { 0xFF, 0x00, 0xBE, 0xFF },
	{ 0xFF, 0x00, 0xFF, 0xFF }, { 0xBE, 0x00, 0xFF, 0xFF }, { 0x7D, 0x00, 0xFF, 0xFF }, { 0x41, 0x00, 0xFF, 0xFF },
	{ 0x00, 0x00, 0xFF, 0xFF }, { 0x00, 0x41, 0xFF, 0xFF }, { 0x00, 0x7D, 0xFF, 0xFF }, { 0x00, 0xBE, 0xFF, 0xFF },
	{ 0x00, 0xFF, 0xFF, 0xFF }, { 0x00, 0xFF, 0xBE, 0xFF }, { 0x00, 0xFF, 0x7D, 0xFF }, { 0x00, 0xFF, 0x41, 0xFF },
	{ 0x00, 0xFF, 0x00, 0xFF }, { 0x41, 0xFF, 0x00, 0xFF }, { 0x7D, 0xFF, 0x00, 0xFF }, { 0xBE, 0xFF, 0x00, 0xFF },
	{ 0xFF, 0xFF, 0x00, 0xFF }, { 0xFF, 0xBE, 0x00, 0xFF }, { 0xFF, 0x7D, 0x00, 0xFF }, { 0xFF, 0x41, 0x00, 0xFF },
	{ 0xFF, 0x7D, 0x7D, 0xFF }, { 0xFF, 0x7D, 0x9E, 0xFF }, { 0xFF, 0x7D, 0xBE, 0xFF }, { 0xFF, 0x7D, 0xDF, 0xFF },
	{ 0xFF, 0x7D, 0xFF, 0xFF }, { 0xDF, 0x7D, 0xFF, 0xFF }, { 0xBE, 0x7D, 0xFF, 0xFF }, { 0x9E, 0x7D, 0xFF, 0xFF },
	{ 0x7D, 0x7D, 0xFF, 0xFF }, { 0x7D, 0x9E, 0xFF, 0xFF }, { 0x7D, 0xBE, 0xFF, 0xFF }, { 0x7D, 0xDF, 0xFF, 0xFF },
	{ 0x7D, 0xFF, 0xFF, 0xFF }, { 0x7D, 0xFF, 0xDF, 0xFF }, { 0x7D, 0xFF, 0xBE, 0xFF }, { 0x7D, 0xFF, 0x9E, 0xFF },
	{ 0x7D, 0xFF, 0x7D, 0xFF }, { 0x9E, 0xFF, 0x7D, 0xFF }, { 0xBE, 0xFF, 0x7D, 0xFF }, { 0xDF, 0xFF, 0x7D, 0xFF },
	{ 0xFF, 0xFF, 0x7D, 0xFF }, { 0xFF, 0xDF, 0x7D, 0xFF }, { 0xFF, 0xBE, 0x7D, 0xFF }, { 0xFF, 0x9E, 0x7D, 0xFF },
	{ 0xFF, 0xB6, 0xB6, 0xFF }, { 0xFF, 0xB6, 0xC7, 0xFF }, { 0xFF, 0xB6, 0xDB, 0xFF }, { 0xFF, 0xB6, 0xEB, 0xFF },
	{ 0xFF, 0xB6, 0xFF, 0xFF }, { 0xEB, 0xB6, 0xFF, 0xFF }, { 0xDB, 0xB6, 0xFF, 0xFF }, { 0xC7, 0xB6, 0xFF, 0xFF },
	{ 0xB6, 0xB6, 0xFF, 0xFF }, { 0xB6, 0xC7, 0xFF, 0xFF }, { 0xB6, 0xDB, 0xFF, 0xFF }, { 0xB6, 0xEB, 0xFF, 0xFF },
	{ 0xB6, 0xFF, 0xFF, 0xFF }, { 0xB6, 0xFF, 0xEB, 0xFF }, { 0xB6, 0xFF, 0xDB, 0xFF }, { 0xB6, 0xFF, 0xC7, 0xFF },
	{ 0xB6, 0xFF, 0xB6, 0xFF }, { 0xC7, 0xFF, 0xB6, 0xFF }, { 0xDB, 0xFF, 0xB6, 0xFF }, { 0xEB, 0xFF, 0xB6, 0xFF },
	{ 0xFF, 0xFF, 0xB6, 0xFF }, { 0xFF, 0xEB, 0xB6, 0xFF }, { 0xFF, 0xDB, 0xB6, 0xFF }, { 0xFF, 0xC7, 0xB6, 0xFF },
	{ 0x71, 0x00, 0x00, 0xFF }, { 0x71, 0x00, 0x1C, 0xFF }, { 0x71, 0x00, 0x38, 0xFF }, { 0x71, 0x00, 0x55, 0xFF },
	{ 0x71, 0x00, 0x71, 0xFF }, { 0x55, 0x00, 0x71, 0xFF }, { 0x38, 0x00, 0x71, 0xFF }, { 0x1C, 0x00, 0x71, 0xFF },
	{ 0x00, 0x00, 0x71, 0xFF }, { 0x00, 0x1C, 0x71, 0xFF }, { 0x00, 0x38, 0x71, 0xFF }, { 0x00, 0x55, 0x71, 0xFF },
	{ 0x00, 0x71, 0x71, 0xFF }, { 0x00, 0x71, 0x55, 0xFF }, { 0x00, 0x71, 0x38, 0xFF }, { 0x00, 0x71, 0x1C, 0xFF },
	{ 0x00, 0x71, 0x00, 0xFF }, { 0x1C, 0x71, 0x00, 0xFF }, { 0x38, 0x71, 0x00, 0xFF }, { 0x55, 0x71, 0x00, 0xFF },
	{ 0x71, 0x71, 0x00, 0xFF }, { 0x71, 0x55, 0x00, 0xFF }, { 0x71, 0x38, 0x00, 0xFF }, { 0x71, 0x1C, 0x00, 0xFF },
	{ 0x71, 0x38, 0x38, 0xFF }, { 0x71, 0x38, 0x45, 0xFF }, { 0x71, 0x38, 0x55, 0xFF }, { 0x71, 0x38, 0x61, 0xFF },
	{ 0x71, 0x38, 0x71, 0xFF }, { 0x61, 0x38, 0x71, 0xFF }, { 0x55, 0x38, 0x71, 0xFF }, { 0x45, 0x38, 0x71, 0xFF },
	{ 0x38, 0x38, 0x71, 0xFF }, { 0x38, 0x45, 0x71, 0xFF }, { 0x38, 0x55, 0x71, 0xFF }, { 0x38, 0x61, 0x71, 0xFF },
	{ 0x38, 0x71, 0x71, 0xFF }, { 0x38, 0x71, 0x61, 0xFF }, { 0x38, 0x71, 0x55, 0xFF }, { 0x38, 0x71, 0x45, 0xFF },
	{ 0x38, 0x71, 0x38, 0xFF }, { 0x45, 0x71, 0x38, 0xFF }, { 0x55, 0x71, 0x38, 0xFF }, { 0x61, 0x71, 0x38, 0xFF },
	{ 0x71, 0x71, 0x38, 0xFF }, { 0x71, 0x61, 0x38, 0xFF }, { 0x71, 0x55, 0x38, 0xFF }, { 0x71, 0x45, 0x38, 0xFF },
	{ 0x71, 0x51, 0x51, 0xFF }, { 0x71, 0x51, 0x59, 0xFF }, { 0x71, 0x51, 0x61, 0xFF }, { 0x71, 0x51, 0x69, 0xFF },
	{ 0x71, 0x51, 0x71, 0xFF }, { 0x69, 0x51, 0x71, 0xFF }, { 0x61, 0x51, 0x71, 0xFF }, { 0x59, 0x51, 0x71, 0xFF },
	{ 0x51, 0x51, 0x71, 0xFF }, { 0x51, 0x59, 0x71, 0xFF }, { 0x51, 0x61, 0x71, 0xFF }, { 0x51, 0x69, 0x71, 0xFF },
	{ 0x51, 0x71, 0x71, 0xFF }, { 0x51, 0x71, 0x69, 0xFF }, { 0x51, 0x71, 0x61, 0xFF }, { 0x51, 0x71, 0x59, 0xFF },
	{ 0x51, 0x71, 0x51, 0xFF }, { 0x59, 0x71, 0x51, 0xFF }, { 0x61, 0x71, 0x51, 0xFF }, { 0x69, 0x71, 0x51, 0xFF },
	{ 0x71, 0x71, 0x51, 0xFF }, { 0x71, 0x69, 0x51, 0xFF }, { 0x71, 0x61, 0x51, 0xFF }, { 0x71, 0x59, 0x51, 0xFF },
	{ 0x41, 0x00, 0x00, 0xFF }, { 0x41, 0x00, 0x10, 0xFF }, { 0x41, 0x00, 0x20, 0xFF }, { 0x41, 0x00, 0x30, 0xFF },
	{ 0x41, 0x00, 0x41, 0xFF }, { 0x30, 0x00, 0x41, 0xFF }, { 0x20, 0x00, 0x41, 0xFF }, { 0x10, 0x00, 0x41, 0xFF },
	{ 0x00, 0x00, 0x41, 0xFF }, { 0x00, 0x10, 0x41, 0xFF }, { 0x00, 0x20, 0x41, 0xFF }, { 0x00, 0x30, 0x41, 0xFF },
	{ 0x00, 0x41, 0x41, 0xFF }, { 0x00, 0x41, 0x30, 0xFF }, { 0x00, 0x41, 0x20, 0xFF }, { 0x00, 0x41, 0x10, 0xFF },
	{ 0x00, 0x41, 0x00, 0xFF }, { 0x10, 0x41, 0x00, 0xFF }, { 0x20, 0x41, 0x00, 0xFF }, { 0x30, 0x41, 0x00, 0xFF },
	{ 0x41, 0x41, 0x00, 0xFF }, { 0x41, 0x30, 0x00, 0xFF }, { 0x41, 0x20, 0x00, 0xFF }, { 0x41, 0x10, 0x00, 0xFF },
	{ 0x41, 0x20, 0x20, 0xFF }, { 0x41, 0x20, 0x28, 0xFF }, { 0x41, 0x20, 0x30, 0xFF }, { 0x41, 0x20, 0x38, 0xFF },
	{ 0x41, 0x20, 0x41, 0xFF }, { 0x38, 0x20, 0x41, 0xFF }, { 0x30, 0x20, 0x41, 0xFF }, { 0x28, 0x20, 0x41, 0xFF },
	{ 0x20, 0x20, 0x41, 0xFF }, { 0x20, 0x28, 0x41, 0xFF }, { 0x20, 0x30, 0x41, 0xFF }, { 0x20, 0x38, 0x41, 0xFF },
	{ 0x20, 0x41, 0x41, 0xFF }, { 0x20, 0x41, 0x38, 0xFF }, { 0x20, 0x41, 0x30, 0xFF }, { 0x20, 0x41, 0x28, 0xFF },
	{ 0x20, 0x41, 0x20, 0xFF }, { 0x28, 0x41, 0x20, 0xFF }, { 0x30, 0x41, 0x20, 0xFF }, { 0x38, 0x41, 0x20, 0xFF },
	{ 0x41, 0x41, 0x20, 0xFF }, { 0x41, 0x38, 0x20, 0xFF }, { 0x41, 0x30, 0x20, 0xFF }, { 0x41, 0x28, 0x20, 0xFF },
	{ 0x41, 0x2C, 0x2C, 0xFF }, { 0x41, 0x2C, 0x30, 0xFF }, { 0x41, 0x2C, 0x34, 0xFF }, { 0x41, 0x2C, 0x3C, 0xFF },
	{ 0x41, 0x2C, 0x41, 0xFF }, { 0x3C, 0x2C, 0x41, 0xFF }, { 0x34, 0x2C, 0x41, 0xFF }, { 0x30, 0x2C, 0x41, 0xFF },
	{ 0x2C, 0x2C, 0x41, 0xFF }, { 0x2C, 0x30, 0x41, 0xFF }, { 0x2C, 0x34, 0x41, 0xFF }, { 0x2C, 0x3C, 0x41, 0xFF },
	{ 0x2C, 0x41, 0x41, 0xFF }, { 0x2C, 0x41, 0x3C, 0xFF }, { 0x2C, 0x41, 0x34, 0xFF }, { 0x2C, 0x41, 0x30, 0xFF },
	{ 0x2C, 0x41, 0x2C, 0xFF }, { 0x30, 0x41, 0x2C, 0xFF }, { 0x34, 0x41, 0x2C, 0xFF }, { 0x3C, 0x41, 0x2C, 0xFF },
	{ 0x41, 0x41, 0x2C, 0xFF }, { 0x41, 0x3C, 0x2C, 0xFF }, { 0x41, 0x34, 0x2C, 0xFF }, { 0x41, 0x30, 0x2C, 0xFF },
	// Black:
	{ 0x00, 0x00, 0x00, 0xFF }, { 0x00, 0x00, 0x00, 0xFF }, { 0x00, 0x00, 0x00, 0xFF }, { 0x00, 0x00, 0x00, 0xFF },
	{ 0x00, 0x00, 0x00, 0xFF }, { 0x00, 0x00, 0x00, 0xFF }, { 0x00, 0x00, 0x00, 0xFF }, { 0x00, 0x00, 0x00, 0xFF },
};

namespace IBM {


	namespace CGA {


		// Returns the standard 16-color palette (for 4-bit modes).
		GFX::SBGRC_8888 const * palette_4bit_asBGRC() {
			return vgaPalette_bgrc; // fully compatible
		}


	} // CGA


	namespace EGA {


		// Returns the standard 16-color palette (for 4-bit modes).
		//   identical to the CGA standard palette
		GFX::SBGRC_8888 const * palette_4bit_asBGRC() {
			return vgaPalette_bgrc; // fully compatible
		}


	} // EGA


	namespace VGA {


		// Returns the standard 256-color palette (for 8-bit modes).
		//   first 16 colors are identical to the CGA/EGA standard palette
		GFX::SBGRC_8888 const * palette_8bit_asBGRC() {
			return vgaPalette_bgrc;
		}


	} // VGA


} // IBM