namespace GFX {
	struct SRGB_888_reg;
	struct SRGBC_8888_reg;
	struct SBGRC_8888_reg;
	struct QRGBC_8888_reg;
	struct QBGRC_8888_reg;
	struct LRGBC_reg;



	// Four 32-bit floating-point numbers in RGBC order.
	//   linear color space
	//   D3DFMT_A32B32G32R32F in Direct3D9
	//   DXGI_FORMAT_R32G32B32A32_FLOAT in Direct3D1012
	struct LRGBC {
		float r, g, b, c;

		static LRGBC_reg SIMD_CALL transparent();
		static LRGBC_reg SIMD_CALL opaqueBlack();
		static LRGBC_reg SIMD_CALL opaqueWhite();
	};



	// Three 32-bit floating-point numbers in RGB order.
	//   linear color space
	//   D3DFMT_B32G32R32F in Direct3D9
	//   DXGI_FORMAT_R32G32B32_FLOAT in Direct3D1012
	struct LRGB {
		float r, g, b;
	};



	// Three 8-bit unsigned integers in RGB order.
	//   sRGB color space
	//   native to Win32, e.g. GetSysColor() and all GDI functions
	struct SRGB_888 {
		UInt1B r, g, b;
	};



	// Three 32-bit floating-point numbers in RGB order.
	//   sRGB color space
	//   common format for legacy code without sRGB knowledge, e.g. OpenGL 1.x, Direct3D 8, VRML
	struct SRGB_FFF {
		float r, g, b;
	};



	// Three 6-bit unsigned integers in RGB order.
	//   sRGB color space
	//   two most significant bits of each component unused
	//   native to Amiga
	struct SXRXGXB_262626 {
		UInt1B xr, xg, xb;
	};



	// Three 8-bit unsigned integers in BGR order.
	//   sRGB color space
	//   native to Win32, e.g. BMP format
	struct SBGR_888 {
		UInt1B b, g, r;
	};



	// Four 8-bit unsigned integers in RGBC order.
	//   sRGB color space with linear coverage
	//   D3DFMT_A8B8G8R8 in Direct3D9
	//   DXGI_FORMAT_R8G8B8A8_UINT_SRGB in Direct3D1012
	struct SRGBC_8888 {
		UInt1B r, g, b, c;

		static SRGBC_8888_reg solidWhite();
		static SRGBC_8888_reg transparent();
	};



	// Four 8-bit unsigned integers in BGRC order.
	//   sRGB color space with linear coverage
	//   D3DFMT_A8R8G8B8 in Direct3D9
	//   DXGI_FORMAT_B8G8R8A8_UINT_SRGB in Direct3D1012
	//   PixelFormat32bppPARGB in GDI+
	struct SBGRC_8888 {
		UInt1B b, g, r, c;

		static SBGRC_8888_reg solidWhite();
		static SBGRC_8888_reg transparent();
	};


	// Four 8-bit unsigned integers in BGR order with the last byte unused.
	//   sRGB color space
	//   native to Win32, e.g. 32-bit BMP format
	struct SBGRX_8888 {
		UInt1B b, g, r, x;
	};



	// Four 8-bit unsigned integers in RGBC order.
	//   square root color space with linear coverage (very handy for shaders  instead of doing a full sRGB-to-linear
	//    conversion, the linear value is simply the square)
	struct QRGBC_8888 {
		UInt1B r, g, b, c;

		static QRGBC_8888_reg solidWhite();
	};




	// Four 8-bit unsigned integers in BGRC order.
	//   square root color space with linear coverage (very handy for shaders  instead of doing a full sRGB-to-linear
	//    conversion, the linear value is simply the square)
	struct QBGRC_8888 {
		UInt1B b, g, r, c;

		static QBGRC_8888_reg opaqueBlack();
		static QBGRC_8888_reg solidWhite();
	};



	// Three 5-bit unsigned integers and 1-bit coverage in BGRC order, Little Endian.
	//   sRGB color space
	//   common texture format on PlayStation 1
	//   usage of coverage varies:
	//     if all 16 bits are zero, the PSX treats it fully transparent
	//     anything else depends on blending modes
	struct SBGRC_5551_LE {
		UInt2B_LE data;
	};



	// Three 5-bit unsigned integers and 1-bit coverage in RGBC order, Little Endian.
	//   sRGB color space
	//   can be found in Windows Bitmaps
	struct SRGBC_5551_LE {
		UInt2B_LE data;
	};



	// Three 5-bit unsigned integers and an unused bit in RGB order, Little Endian.
	//   sRGB color space
	//   can be found in binary STL files and Windows Bitmaps
	struct SRGBX_5551_LE {
		UInt2B_LE data;
	};



	// Three unsigned integers with 5, 6, and 5 bits in BGR order, Little Endian.
	//   sRGB color space
	//   common texture format on Direct3D 8-era PCs
	struct SBGRC_565_LE {
		UInt2B_LE data;
	};



	// Three 4-bit unsigned integers and four bits padding in XBGR order, Big Endian.
	//   sRGB color space
	//   can be found in color extensions for Amiga ILBMs
	struct SXRGB_4444_BE {
		UInt2B_BE data;
	};



	// Four 32-bit floating-point numbers in a register.
	struct LRGBC_reg {
		floatx4 rgbc;
	};

	LRGBC_reg SIMD_CALL rgbcFrom(
		float red,
		float green,
		float blue,
		float coverage
	);

	LRGBC_reg SIMD_CALL load(LRGBC const &);

	void SIMD_CALL store(LRGBC &, LRGBC_reg);



	// Three 8-bit unsigned integers in a register.
	struct SRGB_888_reg {
		UInt4B xbgr; // from most to least significant bit
	};

	SRGB_888_reg sRGBFrom(
		UInt1B red,
		UInt1B green,
		UInt1B blue
	);

	SRGB_888_reg load(SRGB_888 const &);

	// Converts the given color to linear color space with 100 % coverage.
	LRGBC linearOpaqueFrom(SRGB_888_reg);

	// Interpolates the two colors with a 3:1 ratio. Useful for UI color teints.
	SRGB_888_reg interpolate3to1(SRGB_888_reg, SRGB_888_reg);

	void store(SRGB_888 &, SRGB_888_reg);



	SRGB_FFF operator *(SRGB_FFF const &, float);

	SRGB_FFF & operator *=(SRGB_FFF &, float);



	// Four 8-bit unsigned integers in a register.
	//   sRGB color space with linear coverage
	struct SRGBC_8888_reg {
		UInt4B cbgr; // from most to least significant bit
	};

	SRGBC_8888_reg sRGBCFrom(UInt1B, UInt1B, UInt1B, UInt1B);

	SRGBC_8888_reg load(SRGBC_8888 const &);

	SRGBC_8888_reg sRGBC_8888_from(SRGB_888_reg);

	SRGBC_8888_reg sRGB1_8888_from(SRGB_FFF const &);

	SRGBC_8888_reg sRGBC_8888_from(SRGB_FFF const & rgb, float c);

	// Converts the given color to linear color space.
	LRGBC_reg SIMD_CALL linearFrom(SRGBC_8888_reg);

	bool equalBitwise(SRGBC_8888_reg, SRGBC_8888_reg);

	void store(SRGBC_8888 &, SRGBC_8888_reg);



	// Four 8-bit unsigned integers in a register.
	//   sRGB color space with linear coverage
	struct SBGRC_8888_reg {
		UInt4B crgb; // from most to least significant bit
	};

	SBGRC_8888_reg sBGRCFrom(
		UInt1B blue,
		UInt1B green,
		UInt1B red,
		UInt1B coverage
	);

	SBGRC_8888_reg load(SBGRC_8888 const &);

	// Swizzles the given BGRC color to RGBC.
	SRGBC_8888_reg rgbcOf(SBGRC_8888_reg);

	// Swizzles the given RGBC color to BGRC.
	SBGRC_8888_reg bgrcOf(SRGBC_8888_reg);

	// Swizzles the given RGB color to BGR with 100 % coverage.
	SBGRC_8888_reg bgr1Of(SRGB_888_reg);

	// Swizzles the given RGB color to BGR with no coverage.
	SBGRC_8888_reg bgr0Of(SRGB_888_reg);

	// Converts the given color to linear color space.
	LRGBC_reg SIMD_CALL linearFrom(SBGRC_8888_reg);

	// Computes the given color at half the coverage.
	SBGRC_8888_reg semiTransparent(SBGRC_8888_reg);

	bool equalBitwise(SBGRC_8888_reg, SBGRC_8888_reg);

	void store(SBGRC_8888 &, SBGRC_8888_reg);

	// Blends the given pixels in-place with the given background color.
	void alphaBlend_inPlace(
		GFX::SBGRC_8888_reg background,
		GFX::SBGRC_8888     pixels[],
		GFX::SBGRC_8888 *   toEndOfPixels,
		bool                isPremultiplied
	);



	// Four 8-bit unsigned integers in a register.
	//   square root color space with linear coverage
	struct QRGBC_8888_reg {
		UInt4B cbgr; // from most to least significant bit
	};

	QRGBC_8888_reg load(QRGBC_8888 const &);

	QRGBC_8888_reg qRGBCFrom(
		UInt1B r,
		UInt1B g,
		UInt1B b,
		UInt1B c
	);

	// Converts the given color from sRGB to square root color space.
	QRGBC_8888_reg qRGBCFrom(SRGBC_8888);

	void store(QRGBC_8888 &, QRGBC_8888_reg);



	// Four 8-bit unsigned integers in a register.
	//   square root color space with linear coverage
	struct QBGRC_8888_reg {
		UInt4B crgb; // from most to least significant bit
	};

	QBGRC_8888_reg load(QBGRC_8888 const &);

	QBGRC_8888_reg qBGRCFrom(
		UInt1B blue,
		UInt1B green,
		UInt1B red,
		UInt1B coverage
	);

	QBGRC_8888_reg qBGRCFromLinear(
		float red,
		float green,
		float blue,
		float coverage
	);

	// Converts the given color to linear color space.
	LRGBC_reg SIMD_CALL linearFrom(QBGRC_8888_reg);

	// Converts the given color to linear color space and ignores coverage (slightly faster).
	LRGBC_reg SIMD_CALL linearRGBFrom(QBGRC_8888_reg);

	// Converts the given color from square root to sRGB color space.
	SBGRC_8888_reg sBGRCFrom(QBGRC_8888_reg);

	// Converts the given color from sRGB to square root color space.
	QBGRC_8888_reg qBGRCFrom(SBGRC_8888);

	// Computes the given color at half the coverage.
	QBGRC_8888_reg semiTransparent(QBGRC_8888_reg);

	bool equalBitwise(QBGRC_8888_reg, QBGRC_8888_reg);

	void store(QBGRC_8888 &, QBGRC_8888_reg);





	// Three 8-bit unsigned integers and a coverage bit in a register.
	//   sRGB color space
	struct SBGRC_5551_reg {
		UInt2B crgb; // from most to least significant bit
	};

	SBGRC_5551_reg load(SBGRC_5551_LE const &);

	// Expands the given color to 32-bit color depth.
	//   with bias
	SBGRC_8888_reg sBGRC_8888_from(SBGRC_5551_reg);

	// Expands the given color to 24-bit color depth with full coverage.
	//   with bias
	SRGBC_8888_reg sRGB1_8888_from(SBGRC_5551_reg);



	// Three 5-bit unsigned integers and a coverage bit in a register.
	//   sRGB color space
	struct SRGBC_5551_reg {
		UInt2B cbgr; // from most to least significant bit
	};

	SRGBC_5551_reg load(SRGBC_5551_LE const &);

	void store(SRGBC_5551_LE &, SRGBC_5551_reg);

	// Expands the given color to 24-bit color depth with full coverage.
	//   with bias
	SRGBC_8888_reg sRGB1_8888_from(SRGBC_5551_reg);

	// Expands the given color to 32-bit color depth.
	//   with bias
	SBGRC_8888_reg sBGRC_8888_from(SRGBC_5551_reg);

	SRGBC_5551_reg sRGBC_5551_from(SRGBC_8888_reg);

	bool hasCoverage(SRGBC_5551_reg);



	// Three 5-bit unsigned integers in a register.
	//   sRGB color space
	struct SRGBX_5551_reg {
		UInt2B cbgr; // from most to least significant bit
	};

	SRGBX_5551_reg load(SRGBX_5551_LE const &);

	void store(SRGBX_5551_LE &, SRGBX_5551_reg);

	// Expands the given color to 24-bit color depth with full coverage.
	//   with bias
	SRGBC_8888_reg sRGB1_8888_from(SRGBX_5551_reg);

	// Expands the given color to 24-bit color depth with full coverage.
	//   with bias
	SBGRC_8888_reg sBGR1_8888_from(SRGBX_5551_reg);



	struct SXRGB_4444_reg {
		UInt2B xrgb; // from most to least significant bit
	};

	SXRGB_4444_reg load(SXRGB_4444_BE const &);

	// Expands the given color to 32-bit color depth.
	//   with bias
	SBGRC_8888_reg sBGRC_8888_from(SXRGB_4444_reg);



	// Converts the given 4-bit unsigned normalized value to an 8-bit unsigned normalized value with bias.
	// Useful to unpack 12-bpp texture formats.
	UInt1B _4to8(UInt1B);

	// Converts the given 5-bit unsigned normalized value to an 8-bit unsigned normalized value with bias.
	// Useful to unpack 16-bpp texture formats.
	UInt1B _5to8(UInt1B);

	// Converts the given 6-bit unsigned normalized value to an 8-bit unsigned normalized value with bias.
	// Useful to unpack 16-bpp texture formats.
	UInt1B _6to8(UInt1B);

	// Converts the given 8-bit unsigned normalized luminance in sRGB color space to a linear color space luminance.
	float linearFromSRGB(UInt1B);

	// Converts the given linear color space luminance to an 8-bit unsigned normalized luminance in sRGB color space.
	UInt1B sRGBFromLinear(float);

	// Converts the given linear color space luminance to an 8-bit unsigned luminance in the square root space.
	// Useful to deliver parameters to high-performance shaders.
	UInt1B squareRootFromLinear(float);

	UInt1B qBGRCFrom(UInt1B);


	// A collection of pixels:
	//   8-bpc RGBC in sRGB color space
	//   no padding
	struct Pixels_r8g8b8c8_sRGB {
		UInt2B     width;
		UInt2B     height;
		SRGBC_8888 pixels[];
	};

	// Tries to reserve space for an image of the given dimensions.
	//   destroy() after use
	//   returns nullptr on failure
	//   content undefined
	Pixels_r8g8b8c8_sRGB * tryToReserve(
		UIntLeast2B width,
		UIntLeast2B height
	);

	void destroy(Pixels_r8g8b8c8_sRGB & pixels);


	void sRGBFromLinear(
		UCount        widthInPixels,
		SRGBC_8888 *  toDestination,
		USize         destinationLineSize,
		LRGBC const * toSource,
		USize         sourceLineSize,
		LRGBC const * toEndOfSource
	);
	void sRGBFromLinear_dithered(
		UCount        widthInPixels,
		SRGBC_8888 *  toDestination,
		USize         destinationLineSize,
		LRGBC const * toSource,
		USize         sourceLineSize,
		LRGBC const * toEndOfSource
	);


	// Interpolates the color in a triangle according to barycentric coordinates.
	//   linear interpolation
	SRGBC_8888 SIMD_CALL interpolatedTriangleColor(
		SRGBC_8888 const & colorAtU,
		SRGBC_8888 const & colorAtV,
		SRGBC_8888 const & colorAtW,
		floatx4            uvw
	);


} // GFX


namespace IBM {


	// Color Graphics Adapter, the 1981 PC graphics standard. Most commercial software was compatible until the early 1990s.
	namespace CGA {


		// Returns the standard 16-color palette (for 4-bit modes).
		GFX::SBGRC_8888 const * palette_4bit_asBGRC();


	} // CGA


	// Enhanced Graphics Adapter, a 1984 PC graphics standard. Widespread among software from that era.
	namespace EGA {


		// Returns the standard 16-color palette (for 4-bit modes).
		//   identical to the CGA standard palette
		GFX::SBGRC_8888 const * palette_4bit_asBGRC();


	} // EGA


	// Video Graphics Array, a 1987 PC graphics standard. Color palettes and display modes are widespread among software from that
	//  era, and the physical display connection is still supported today.
	namespace VGA {


		// Returns the standard 256-color palette (for 8-bit modes).
		//   first 16 colors are identical to the CGA/EGA standard palette
		GFX::SBGRC_8888 const * palette_8bit_asBGRC();


	} // VGA


} // IBM