#include "framework.hpp"
#include "SIMD.hpp"
#include "Geo.hpp"



#if COMPILED_BY_VISUAL_CPP
	// Visual C++ intrinsics. On Clang and GCC, the __builtin_ counterparts are used, so no declaration is required.
	import {
		double cos(double);
		double sin(double);
		double tan(double);
		double acos(double);
		double asin(double);
		double atan2(double, double);

		__m128 _mm_movelh_ps(__m128, __m128);
		__m128 _mm_movehl_ps(__m128, __m128);
		__m128 _mm_move_ss(__m128, __m128);
		__m128 _mm_unpacklo_ps(__m128, __m128);
		__m128 _mm_unpackhi_ps(__m128, __m128);
		__m128 _mm_shuffle_ps(__m128, __m128, int);
		__m128 _mm_sqrt_ps(__m128);
		__m128d _mm_sqrt_pd(__m128d);

		__m128i _mm_unpacklo_epi32(__m128i, __m128i);
		__m128i _mm_shuffle_epi32(__m128i, int);
		__m128i _mm_mul_epu32(__m128i, __m128i);

#		define MM_SHUFFLE_PS(L, R, X, Y, Z, W) _mm_shuffle_ps((L), (R), (X) + (Y << 2) + (Z << 4) + (W << 6))
	}
#endif


UInt2B squared(SInt1B value) {
	// 2^8  2^8 == 2^16. Therefore, the operands must be converted to 16-bit numbers. The result of the multiplication is
	//  always unsigned.
	return UInt2B(SInt2B(value) * SInt2B(value));
}
UInt4B squared(UInt4B value) {
	return value * value;
}
UInt8B squared(SInt4B value) {
	return UInt8B(SInt8B(value) * SInt8B(value));
}
float squared(float value) {
	return value * value;
}
double squared(double value) {
	return value * value;
}

// Computes the squared value but copies the original sign to the result.
float signedSquareOf(float x) {
#	if COMPILED_BY_VISUAL_CPP
		return xOf(x___From(x * x) | (x___From(x) & floatx4SignMask_XYZW()));
#	else
		TODO("optimize")
		return x < 0.0f ? -(x * x) : (x * x);
#	endif
}

// Computes the square root of the given value. For negative numbers, this is NaN.
float squareRootOf(float value) {
#	if COMPILED_BY_VISUAL_CPP
		// Use the SQRTPS instruction. (It has one data dependency less over its scalar counterpart).
		return xOf(floatx4{_mm_sqrt_ps(x___From(value).floats)});
#	elif COMPILED_BY_CLANG
		floatx4 const v = { value, value, value, value };
		return __builtin_ia32_sqrtps(v.floats)[0];
#	else
#		error TODO
#	endif
}
double squareRootOf(double value) {
#	if COMPILED_BY_VISUAL_CPP
		// Use the SQRTPD instruction. (It has one data dependency less over its scalar counterpart).
		return xOf(_mm_sqrt_pd(x0From(value)));
#	elif COMPILED_BY_CLANG
		doublex2 const v = { value, 0.0 };
		return __builtin_ia32_sqrtpd(v)[0];
#	else
#		error TODO
#	endif
}

// Interpolates linearly between the given scalars according to the given factor, where 0 yields a and 1 yields b.
float interpolateLinearly(float a, float factor, float b) {
	return (b - a) * factor + a;
}
floatx4 SIMD_CALL interpolated_linear(floatx4 a, floatx4 factor, floatx4 b) {
	return a * (loadx4(1.0f) - factor) + factor * b;
}
double interpolateLinearly(double a, double factor, double b) {
	return (b - a) * factor + a;
}



namespace Geo {


	// All binary operators look the same except for the operator sign and the types. Prefer macros over copy-paste code.
#define VECTOR_OPERATOR_VECTOR(TYPE, OPERATOR, ATTRIBUTE) \
	TYPE##_reg SIMD_CALL operator OPERATOR(TYPE##_reg a, TYPE##_reg b) { \
		return { a.ATTRIBUTE OPERATOR b.ATTRIBUTE }; \
	}
#define VECTOR_OPERATOR_SCALAR(VECTORTYPE, OPERATOR, SCALARTYPE, ATTRIBUTE) \
	VECTORTYPE##_reg SIMD_CALL operator OPERATOR(VECTORTYPE##_reg a, SCALARTYPE b) { \
		return { a.ATTRIBUTE OPERATOR loadx4(b) }; \
	}
#define SCALAR_OPERATOR_VECTOR(SCALARTYPE, OPERATOR, VECTORTYPE, ATTRIBUTE) \
	VECTORTYPE##_reg SIMD_CALL operator OPERATOR(SCALARTYPE a, VECTORTYPE##_reg b) { \
		return { loadx4(a) OPERATOR b.ATTRIBUTE }; \
	}
#define VECTOR_ASSIGN_OPERATOR_VECTOR(TYPE, OPERATOR, ATTRIBUTE) \
	TYPE##_reg & SIMD_CALL operator OPERATOR##=(TYPE##_reg & a, TYPE##_reg b) { \
		a.ATTRIBUTE = a.ATTRIBUTE OPERATOR b.ATTRIBUTE; \
		return a; \
	}
#define VECTOR_ASSIGN_OPERATOR_SCALAR(VECTORTYPE, OPERATOR, SCALARTYPE, ATTRIBUTE) \
	VECTORTYPE##_reg & SIMD_CALL operator OPERATOR##=(VECTORTYPE##_reg & a, SCALARTYPE b) { \
		a.ATTRIBUTE = a.ATTRIBUTE OPERATOR loadx4(b); \
		return a; \
	}



	// Handy to load any 2D or 3D axis.
	float const axisFloats[] = { 0, 0, 0, 1, 0, 0, 0 };



	//  X Y _ f l o a t :

	XY_float_reg SIMD_CALL XY_float::zero() {
		return { floatx4Zero() };
	}
	XY_float_reg SIMD_CALL XY_float::xAxis() {
		return { loadx4(axisFloats + 3) };
	}
	XY_float_reg SIMD_CALL XY_float::yAxis() {
		return { loadx4(axisFloats + 2) };
	}
	XY_float_reg SIMD_CALL XY_float_reg::operator -() const {
		return { floatx4Zero() - xy__ }; // XOR & SUB is faster than bit twiddling (no memory access!)
	}

	// Returns the X axis, rotated by the given angle (i.e. cosine and sine). Z and W are zero.
	XY_float_reg SIMD_CALL XY_float::polar(Angle4B_reg angle) {
		// Start with a sine approximation. This is more efficient than a cosine approximation  cosine is all positive over
		//  [-pi/2, +pi/2], so the approximation were to be negated after range reduction. Since sine is negative for [-pi/2, 0] and
		//  positive for [0, +pi/2], negative sectors can be computed without overhead.

		// The sine approximation here is taken from http://lol.zoy.org/blog/2011/12/21/better-function-approximations and
		//  http://lol.zoy.org/blog/2011/9/17/playing-with-the-cpu-pipeline. It is exact to 12 digits for [-pi/2, +pi/2].

		// For [-pi/2, pi/2], the approximation can be used directly. For (pi/2, 3/2pi), the angle is negated and shifted back by pi
		//  (i.e. 0x80000000 minus unsigned normalized angle).
		SInt4B sinUNorm;
		if(0x40000000 <= angle.unorm && 0xC0000000 > angle.unorm) {
			sinUNorm = 0x80000000 - angle.unorm;
		} else {
			sinUNorm = angle.unorm;
		}

		// Cosine is computed as sine shifted by pi/2 (0x40000000 as unsigned normalized). It is shifted in exactly the same way,
		//  but to avoid data dependencies, the quadrant of the sine angle is used for branching.
		SInt4B cosUNorm;
		if(0x80000000 <= angle.unorm) {
			cosUNorm = 0x40000000 + angle.unorm;
		} else {
			cosUNorm = 0x40000000 - angle.unorm;
		}

		// Conversion factor for int-to-float is (2pi) / 2. This can be baked into the coefficients, saving a multiplication.
		// Constants are always read from memory (even trivial ones like 1.0)  gather them in the same memory block, aligned to a
		//  cache line boundary so at most one cache miss must be taken.
		alignas(64) static double const coeffs[] = {
			 1.4629180792671596810513378043098e-9,
			-5.218056424355326540518689143791e-28,  // -1.666666666640169148537065260055e-1   3.1308338546629715204060346522109e-27
			 5.5836577275526616254281114471759e-47, //  8.333333316490113523036717102793e-3   6.7003892866059294987769581698842e-45
			-2.8451779180169008034005793100212e-66, // -1.984126600659171392655484413285e-4   1.4339699478207029913663242067659e-62
			 8.4568853391845024057023404189569e-86, //  2.755690114917374804474016589137e-6   3.0688811101817481779794040453891e-80
			-1.6438192896934818124147706229392e-105,// -2.502845227292692953118686710787e-8   6.5678024025144678446613914998034e-98
			 2.16283153455215654880174036432e-125   //  1.538730635926417598443354215485e-10  1.4055946401885921624309340128322e-115
		};

		auto const x  = ::xyFrom(cosUNorm, sinUNorm); // compute cosine and sine in parallel
		auto const x2 = x * x;
		auto const x4 = x2 * x2;
		auto const x8 = x4 * x4;
		auto const a  = x2 * (xxFrom(coeffs[1]) + x2 * (xxFrom(coeffs[2]) + x2 * xxFrom(coeffs[3])));
		auto const b  = xxFrom(coeffs[4]) + x2 * (xxFrom(coeffs[5]) + x2 * xxFrom(coeffs[6]));
		auto const cs = xy00From(x * (a + xxFrom(coeffs[0]) + x8 * b));

		return { cs };
	}

	XY_float_reg SIMD_CALL load(XY_float const & memory) {
		return { ::xy00From(&memory.x) };
	}
	XY_float_reg SIMD_CALL xyFrom(float x, float y) {
		return { ::xy__From(x, y) };
	}
	XY_float_reg SIMD_CALL xyFrom(floatx4 xyzw) {
		return { xyzw };
	}
	void SIMD_CALL store(
		XY_float &   memory,
		XY_float_reg v
	) {
		xyTo(&memory.x, v.xy__);
	}
	XY_float SIMD_CALL stored(XY_float_reg v) {
		XY_float memory;
		xyTo(&memory.x, v.xy__);
		return memory;
	}

	XY_float_reg SIMD_CALL xyOf(XYZ_float_reg v) {
		return { v.xyz_ };
	}
	XY_float_reg SIMD_CALL xzOf(XYZ_float_reg v) {
		return { xz__Of(v.xyz_) };
	}
	XY_float_reg SIMD_CALL zxOf(XYZ_float_reg v) {
		return { zx__Of(v.xyz_) };
	}
	float SIMD_CALL xOf(XY_float_reg v) {
		return ::xOf(v.xy__);
	}
	float SIMD_CALL yOf(XY_float_reg v) {
		return ::yOf(v.xy__);
	}

	VECTOR_OPERATOR_VECTOR(XY_float, +, xy__) // +
	VECTOR_OPERATOR_VECTOR(XY_float, -, xy__) // -
	VECTOR_ASSIGN_OPERATOR_VECTOR(XY_float, +, xy__) // +=
	VECTOR_ASSIGN_OPERATOR_VECTOR(XY_float, -, xy__) // -=
	SCALAR_OPERATOR_VECTOR(float, *, XY_float, xy__) // *
	SCALAR_OPERATOR_VECTOR(float, /, XY_float, xy__) // /
	VECTOR_OPERATOR_SCALAR(XY_float, *, float, xy__) // *
	VECTOR_OPERATOR_SCALAR(XY_float, /, float, xy__) // /
	VECTOR_ASSIGN_OPERATOR_SCALAR(XY_float, *, float, xy__) // *=
	VECTOR_ASSIGN_OPERATOR_SCALAR(XY_float, /, float, xy__) // /=

	float SIMD_CALL dotProductOf(XY_float_reg a, XY_float_reg b) {
		return ::xOf(sumOfXY(a.xy__ * b.xy__));
	}

	XY_float_reg SIMD_CALL doubled(XY_float_reg v) {
		return { v.xy__ + v.xy__ };
	}

	// Computes the squared length of the given vector.
	float SIMD_CALL squaredAbsoluteOf(XY_float_reg v) {
		return dotProductOf(v, v); // v.x + v.y == v  v
	}

	// Checks whether the vector can be normalized.
	bool SIMD_CALL isValidDirection(XY_float_reg v) {
		//  minimum: 8.131516e-20 (0.75 times the square root of the smallest normalized value)
		//  maximum: 1.8446744e19 (square root of the greatest finite value)
		return isWithin(8.131516e-20f * 8.131516e-20f, dotProductOf(v, v), 1.8446743e19f * 1.8446743e19f); // v.x + v.y == v  v
	}

	// Computes the absolute of the given vector (its length).
	float SIMD_CALL absoluteOf(XY_float_reg v) {
#		if COMPILED_BY_VISUAL_CPP
			return xOf(floatx4{ _mm_sqrt_ps(sumOfXY(v.xy__ * v.xy__).floats) });
#		elif COMPILED_BY_CLANG || COMPILED_BY_GCC
			return squareRootOf(squaredAbsoluteOf(v));
#		else
#			error TODO
#		endif
	}

	XY_float_reg SIMD_CALL normalized(XY_float_reg v) {
		MUST(isValidDirection(v));

		return v / absoluteOf(v);
	}

	XY_float_reg SIMD_CALL scaledToLength(
		XY_float_reg v,
		float        desiredLength
	) {
		MUST(isValidDirection(v));
		MUST(::isFinite(desiredLength));

		return v * (desiredLength / absoluteOf(v));
	}

	// Interpolates linearly between the given vectors according to the given factor, where 0 denotes a and 1 denotes b.
	XY_float_reg SIMD_CALL interpolateLinearly(
		XY_float_reg a,
		float        factor,
		XY_float_reg b
	) {
		return (b - a) * factor + a;
	}

	// Rotates the given vector 90 about the Z axis.
	//   right-handed
	//   result has inputs length
	XY_float_reg SIMD_CALL perpendicularOf(XY_float_reg v) {
		return { interleavedXY(y___Of(floatx4Zero() - v.xy__), v.xy__) }; // (-y, x)
	}

	// Tests whether X and Y are finite numbers.
	bool SIMD_CALL isFinite(XY_float_reg v) {
		return xyBelow(::absoluteOf(v.xy__), floatx4Infinity());
	}

	XY_float_reg SIMD_CALL componentWiseMinimumOf(XY_float_reg a, XY_float_reg b) {
		return { ::minimumOf(a.xy__, b.xy__) };
	}
	XY_float_reg SIMD_CALL componentWiseMaximumOf(XY_float_reg a, XY_float_reg b) {
		return { ::maximumOf(a.xy__, b.xy__) };
	}

	// Computes the squared distance between the two given points.
	float SIMD_CALL squaredDistanceBetween(XY_float_reg a, XY_float_reg b) {
		return squaredAbsoluteOf(b - a);
	}

	// Computes the distance between the two given points.
	float SIMD_CALL distanceBetween(XY_float_reg a, XY_float_reg b) {
		return absoluteOf(b - a);
	}

	XY_float_reg SIMD_CALL centerOf(XY_float_reg a, XY_float_reg b) {
		return { (a.xy__ + b.xy__) * loadx4(0.5f) };
	}

	// Tests the given vectors for numerical (not bitwise) equality.
	bool SIMD_CALL operator ==(XY_float_reg a, XY_float_reg b) {
		return equalXY(a.xy__, b.xy__);
	}
	bool SIMD_CALL operator !=(XY_float_reg a, XY_float_reg b) {
		return no == equalXY(a.xy__, b.xy__);
	}

	// Computes the sine of the given angle.
	float sineOf(float angle) {
#		if COMPILED_BY_VISUAL_CPP
			return float(sin(angle)); // TODO
#		elif COMPILED_BY_CLANG
			return __builtin_sinf(angle);
#		else
#			error TODO
#		endif
	}
	double sineOf(double angle) {
#		if COMPILED_BY_VISUAL_CPP
			return sin(angle);
#		elif COMPILED_BY_CLANG
			return __builtin_sin(angle);
#		else
#			error TODO
#		endif
	}

	// Computes the cosine of the given angle.
	float cosineOf(float angle) {
#		if COMPILED_BY_VISUAL_CPP
			return float(cos(angle)); // TODO
#		elif COMPILED_BY_CLANG
			return __builtin_cosf(angle);
#		else
#			error TODO
#		endif
	}
	double cosineOf(double angle) {
#		if COMPILED_BY_VISUAL_CPP
			return cos(angle);
#		elif COMPILED_BY_CLANG
			return __builtin_cos(angle);
#		else
#			error TODO
#		endif
	}

	// Computes the tangent of the given angle.
	float tangentOf(float angle) {
#		if COMPILED_BY_VISUAL_CPP
			return float(tan(angle)); // TODO
#		elif COMPILED_BY_CLANG
			return __builtin_tanf(angle);
#		else
#			error TODO
#		endif
	}
	double tangentOf(double angle) {
#		if COMPILED_BY_VISUAL_CPP
			return tan(angle);
#		elif COMPILED_BY_CLANG
			return __builtin_tan(angle);
#		else
#			error TODO
#		endif
	}

	// Computes an angle to the given sine. The sine must be in the range of [-1; 1].
	float arcusSineOf(float sine) {
		MUST(isWithin(-1.0f, sine, 1.0f));
#		if COMPILED_BY_VISUAL_CPP
			return float(asin(sine)); // TODO
#		elif COMPILED_BY_CLANG
			return __builtin_asinf(sine);
#		else
#			error TODO
#		endif
	}
	double arcusSineOf(double sine) {
		MUST(isWithin(-1.0, sine, 1.0));
#		if COMPILED_BY_VISUAL_CPP
			return asin(sine);
#		elif COMPILED_BY_CLANG
			return __builtin_asin(sine);
#		else
#			error TODO
#		endif
	}

	// Computes an angle to the given cosine. The sine must be in the range of [-1; 1].
	float arcusCosineOf(float cosine) {
		MUST(isWithin(-1.0f, cosine, 1.0f));
#		if COMPILED_BY_VISUAL_CPP
			return float(acos(cosine)); // TODO
#		elif COMPILED_BY_CLANG
			return __builtin_acosf(cosine);
#		else
#			error TODO
#		endif
	}
	double arcusCosineOf(double cosine) {
		MUST(isWithin(-1.0, cosine, 1.0));
#		if COMPILED_BY_VISUAL_CPP
			return acos(cosine);
#		elif COMPILED_BY_CLANG
			return __builtin_acos(cosine);
#		else
#			error TODO
#		endif
	}

	// Computes the counter-clockwise angle to the X axis.
	//   like atan2() with reversed parameters
	//   the vector must not be zero
	Angle4B_reg SIMD_CALL polarAngleOf(XY_float_reg v) {
		MUST(isNumber(xOf(v))); // infinities are allowed here
		MUST(isNumber(yOf(v))); // infinities are allowed here
		MUST(isValidDirection(v));

		Angle4B_reg result;

		// See http://en.wikipedia.org/wiki/Atan2#Definition.
		if(0 == xOf(v)) {
			if(0 < yOf(v)) {
				result.unorm = 0x40000000;
			} else if(0 > yOf(v)) {
				result.unorm = 0xC0000000;
			} else { // 0 == y
				UNREACHABLE;
			}
		} else {
			// The definition says for this case:
			//    atan(x / y) + pi  | y >= 0
			//    atan(x / y) - pi  | y < 0
			//  but since the angle is an unsigned integer, and pi equals 0x80000000:
			//    angle + 0x80000000 == angle - 0x80000000
			//  and therefore, the branch can be omitted.
			result = arcusTangentOf(yOf(v) / xOf(v));
			if(0 > xOf(v)) {
				result.unorm += 0x80000000;
			}
		}

		return result;
	}
	double polarAngleOf(double x, double y) {
		// TODO: Are infinities really allowed?
		// TODO: Should the output be clamped to the range of [-pi, pi) to prevent ambiguous results?
		MUST(isNumber(x)); // infinities are allowed here
		MUST(isNumber(y)); // infinities are allowed here
		MUST(0.0 != x || 0.0 != y);
#		if COMPILED_BY_VISUAL_CPP
			return atan2(y, x);
#		elif COMPILED_BY_CLANG
			return __builtin_atan2(y, x);
#		else
#			error TODO
#		endif
	}




	XY_SInt4B_reg SIMD_CALL load(XY_SInt1B const & mem) {
		return { xy00From(&mem.x) };
	}



	//  X Y _ S I n t 4 B :

	XY_SInt4B_reg SIMD_CALL XY_SInt4B::zero() {
		return { SInt4Bx4_zero() };
	}
	XY_SInt4B_reg SIMD_CALL XY_SInt4B_reg::operator -() const {
		return { SInt4Bx4_zero() - xy__ }; // XOR & SUB is faster than bit twiddling (no memory access!)
	}

	XY_SInt4B_reg SIMD_CALL load(XY_SInt4B const & mem) {
		return { xy00From(&mem.x) };
	}

	XY_SInt4B_reg SIMD_CALL xyFrom(SInt4B x, SInt4B y) {
		return { xy00From(x, y) };
	}

	SInt4B SIMD_CALL xOf(XY_SInt4B_reg v) {
		return ::xOf(v.xy__);
	}
	SInt4B SIMD_CALL yOf(XY_SInt4B_reg v) {
		return ::yOf(v.xy__);
	}

	VECTOR_OPERATOR_VECTOR(XY_SInt4B, +, xy__) // +
	VECTOR_OPERATOR_VECTOR(XY_SInt4B, -, xy__) // -

	// Shifts each components value to the right.
	//   replicates the sign bit
	//   shift must not exceed 31 bits
	XY_SInt4B_reg SIMD_CALL shiftRightArithmetic(XY_SInt4B_reg v, UIntLeast1B bits) {
		return { ::shiftRightArithmetic(v.xy__, bits) };
	}

	// Shifts each components value to the left.
	//   shift must not exceed 31 bits
	XY_SInt4B_reg SIMD_CALL operator <<(XY_SInt4B_reg v, UIntLeast1B bits) {
		return { v.xy__ << bits };
	}

	bool SIMD_CALL operator !=(XY_SInt4B_reg a, XY_SInt4B_reg b) {
		return no == equalXY(a.xy__, b.xy__);
	}

	// Computes the squared length of the given vector.
	UInt4B SIMD_CALL squaredAbsoluteOf(XY_SInt4B_reg v) {
#		if COMPILED_BY_VISUAL_CPP
			// SSE2 provides only unsigned integer multiplication. Its a sign extended 32  32 = 64 multiplication, though, so the
			//  lower 32 bits can be reinterpreted as signed result. Consider its a two-lane operation on X and Z.
			auto x_y_ = _mm_unpacklo_epi32(v.xy__.ints, v.xy__.ints);
			auto square_x_y_ = _mm_mul_epu32(x_y_, x_y_);
			return xOf(SInt4Bx4{ square_x_y_ } + SInt4Bx4{ _mm_shuffle_epi32(square_x_y_, 0b10'10'10'10) });
#		elif COMPILED_BY_CLANG || COMPILED_BY_GCC
			return v.xy__[0] * v.xy__[0] + v.xy__[1] * v.xy__[1];
#		endif
	}

	XY_float_reg SIMD_CALL asFloats(XY_SInt4B_reg v) {
		return { ::asFloats(v.xy__) };
	}

	// Checks whether all components of the left vector are below their right counterparts.
	bool SIMD_CALL areAllComponentsBelow(XY_SInt4B_reg l, XY_SInt4B_reg r) {
		return xyBelow(l.xy__, r.xy__);
	}



	//  X Y Z _ f l o a t :

	XYZ_float_reg SIMD_CALL XYZ_float::zero() {
		return { floatx4Zero() };
	}
	XYZ_float_reg SIMD_CALL XYZ_float::xAxis() {
		return { loadx4(axisFloats + 3) };
	}
	XYZ_float_reg SIMD_CALL XYZ_float::yAxis() {
		return { loadx4(axisFloats + 2) };
	}
	XYZ_float_reg SIMD_CALL XYZ_float::zAxis() {
		return { loadx4(axisFloats + 1) };
	}
	XYZ_float_reg SIMD_CALL XYZ_float_reg::operator -() const {
		return { floatx4Zero() - xyz_ };
	}

	// Generates a normalized vector to the given polar angle.
	//   0/0 is Z axis
	XYZ_float_reg SIMD_CALL XYZ_float::polar(
		Angle4B_reg horizontally,
		Angle4B_reg vertically
	) {
		auto x_cs = XY_float::polar(horizontally).xy__;
		auto y_cs = XY_float::polar(vertically).xy__;

		// cos(x)  sin(y)
		// -sin(x)
		// cos(x)  cos(y)
		return { interleavedXY(
			xxxxOf(x_cs) * yx__Of(y_cs),
			floatx4Zero() - y___Of(x_cs)
		) };
	}

	XYZ_float_reg SIMD_CALL load(XYZ_float const & memory) {
		return { ::xyzFrom(&memory.x) };
	}
	XYZ_float_reg SIMD_CALL loadPadded(XYZ_float const & memory) {
		return { loadx4(&memory.x) };
	}
	XYZ_float_reg SIMD_CALL x00From(float x) {
		return { x000From(x) };
	}
	XYZ_float_reg SIMD_CALL xy0From(float x, float y) {
		return { xy00From(x, y) };
	}
	XYZ_float_reg SIMD_CALL xyzFrom(float x, float y, float z) {
		return { xyzwFrom(x, y, z, 0.0f) };
	}
	XYZ_float_reg SIMD_CALL xyzFrom(floatx4 xyz_) {
		return { xyz_ };
	}
	XYZ_float_reg SIMD_CALL OYOFrom(float y) {
		return { xyzwFrom(0.0f, y, 0.0f, 0.0f) };
	}
	XYZ_float_reg SIMD_CALL OOZFrom(float z) {
		return { xyzwFrom(0.0f, 0.0f, z, 0.0f) };
	}
	void SIMD_CALL store(
		XYZ_float &   memory,
		XYZ_float_reg v
	) {
		xyzTo(&memory.x, v.xyz_);
	}
	XYZ_float SIMD_CALL stored(XYZ_float_reg v) {
		XYZ_float memory;
		xyzTo(&memory.x, v.xyz_);
		return memory;
	}

	XYZ_float_reg SIMD_CALL x0yOf(XY_float_reg v) {
		return { x0y0Of(v.xy__) };
	}
	float SIMD_CALL xOf(XYZ_float_reg v) {
		return ::xOf(v.xyz_);
	}
	float SIMD_CALL yOf(XYZ_float_reg v) {
		return ::yOf(v.xyz_);
	}
	float SIMD_CALL zOf(XYZ_float_reg v) {
		return ::zOf(v.xyz_);
	}

	VECTOR_OPERATOR_VECTOR(XYZ_float, +, xyz_) // +
	VECTOR_OPERATOR_SCALAR(XYZ_float, +, float, xyz_) // +
	VECTOR_OPERATOR_VECTOR(XYZ_float, -, xyz_) // -
	VECTOR_OPERATOR_VECTOR(XYZ_float, *, xyz_) // *
	VECTOR_OPERATOR_VECTOR(XYZ_float, /, xyz_) // /
	VECTOR_ASSIGN_OPERATOR_VECTOR(XYZ_float, +, xyz_) // +=
	VECTOR_ASSIGN_OPERATOR_VECTOR(XYZ_float, -, xyz_) // -=
	SCALAR_OPERATOR_VECTOR(float, *, XYZ_float, xyz_) // *
	SCALAR_OPERATOR_VECTOR(float, /, XYZ_float, xyz_) // /
	VECTOR_OPERATOR_SCALAR(XYZ_float, *, float, xyz_) // *
	VECTOR_OPERATOR_SCALAR(XYZ_float, /, float, xyz_) // /
	VECTOR_ASSIGN_OPERATOR_SCALAR(XYZ_float, *, float, xyz_) // *=
	VECTOR_ASSIGN_OPERATOR_SCALAR(XYZ_float, /, float, xyz_) // /=

	float SIMD_CALL dotProductOf(XYZ_float_reg a, XYZ_float_reg b) {
		return ::xOf(sumOfXYZ(a.xyz_ * b.xyz_));
	}

	XYZ_float_reg SIMD_CALL doubled(XYZ_float_reg v) {
		return { v.xyz_ + v.xyz_ };
	}

	XYZ_float_reg SIMD_CALL squared(XYZ_float_reg v) {
		return { v.xyz_ * v.xyz_ };
	}

	XYZ_float_reg SIMD_CALL clamped(float lowerLimit, XYZ_float_reg v, float upperLimit) {
		return { clamped(loadx4(lowerLimit), v.xyz_, loadx4(upperLimit)) };
	}

	// Computes the squared length of the given vector.
	float SIMD_CALL squaredAbsoluteOf(XYZ_float_reg v) {
		return dotProductOf(v, v); // v.x + v.y +  == v  v
	}

	// Checks whether the vector can be normalized.
	bool SIMD_CALL isValidDirection(XYZ_float_reg v) {
		//  minimum: 8.131516e-20 (0.75 times the square root of the smallest normalized value)
		//  maximum: 1.8446744e19 (square root of the greatest finite value)
		return isWithin(8.131516e-20f * 8.131516e-20f, dotProductOf(v, v), 1.8446743e19f * 1.8446743e19f); // v.x + v.y +  == v  v
	}

	bool SIMD_CALL isAnyComponentNonZero(XYZ_float_reg const v) {
		return isNonZero_xOrYOrZ(v.xyz_);
	}

	bool SIMD_CALL isAnyComponentSubnormal(XYZ_float_reg const v) {
		return isSubnormal_xOrYOrZ(v.xyz_);
	}

	// Computes the absolute of the given vector (its length).
	float SIMD_CALL absoluteOf(XYZ_float_reg v) {
#		if COMPILED_BY_VISUAL_CPP
			return xOf(floatx4{ _mm_sqrt_ps(sumOfXYZ(v.xyz_ * v.xyz_).floats) });
#		elif COMPILED_BY_CLANG || COMPILED_BY_GCC
			return squareRootOf(squaredAbsoluteOf(v));
#		else
#			error TODO
#		endif
	}

	XYZ_float_reg SIMD_CALL normalized(XYZ_float_reg v) {
		MUST(isValidDirection(v));
		return v / absoluteOf(v);
	}

	XYZ_float_reg SIMD_CALL scaledToLength(
		XYZ_float_reg v,
		float         desiredLength
	) {
		MUST(isValidDirection(v));
		MUST(::isFinite(desiredLength));

		return v * (desiredLength / absoluteOf(v));
	}

	// Interpolates linearly between the given vectors according to the given factor, where 0 denotes a and 1 denotes b.
	XYZ_float_reg SIMD_CALL interpolateLinearly(
		XYZ_float_reg a,
		float         factor,
		XYZ_float_reg b
	) {
		return (b - a) * factor + a;
	}

	// Computes the right-handed cross product of the two given vectors.
	XYZ_float_reg SIMD_CALL crossProductOf(XYZ_float_reg a, XYZ_float_reg b) {
		return { zxywOf(zxywOf(a.xyz_) * b.xyz_ - a.xyz_ * zxywOf(b.xyz_)) };
	}

	// Computes the cosine of the shortest angle between the two given vectors.
	//   no need for normalization, but must be valid directions
	//   due to floating-point accuracy, the result may be slightly above 1 for parallel vectors
	float SIMD_CALL cosineOfAngleBetween(XYZ_float_reg a, XYZ_float_reg b) {
		return dotProductOf(a, b) / squareRootOf(squaredAbsoluteOf(a) * squaredAbsoluteOf(b));
	}

	// Tests whether X, Y, and Z are finite numbers.
	bool SIMD_CALL isFinite(XYZ_float_reg v) {
		return xyzBelow(::absoluteOf(v.xyz_), floatx4Infinity());
	}

	XYZ_float_reg SIMD_CALL componentWiseMinimumOf(XYZ_float_reg a, XYZ_float_reg b) {
		return { ::minimumOf(a.xyz_, b.xyz_) };
	}
	XYZ_float_reg SIMD_CALL componentWiseMaximumOf(XYZ_float_reg a, XYZ_float_reg b) {
		return { ::maximumOf(a.xyz_, b.xyz_) };
	}

	// Computes the squared distance between the two given points.
	float SIMD_CALL squaredDistanceBetween(XYZ_float_reg a, XYZ_float_reg b) {
		return squaredAbsoluteOf(b - a);
	}

	// Computes the distance between the two given points.
	float SIMD_CALL distanceBetween(XYZ_float_reg a, XYZ_float_reg b) {
		return absoluteOf(b - a);
	}

	XYZ_float_reg SIMD_CALL centerOf(XYZ_float_reg a, XYZ_float_reg b) {
		return { (a.xyz_ + b.xyz_) * loadx4(0.5f) };
	}

	// Tests the given vectors for numerical (not bitwise) equality.
	bool SIMD_CALL operator ==(XYZ_float_reg a, XYZ_float_reg b) {
		return equalXYZ(a.xyz_, b.xyz_);
	}
	bool SIMD_CALL operator !=(XYZ_float_reg a, XYZ_float_reg b) {
		return no == equalXYZ(a.xyz_, b.xyz_);
	}

	XYZ_SInt4B_reg SIMD_CALL rounded(XYZ_float_reg v) {
		return { ::rounded(v.xyz_) };
	}

	// Assume geometry being projected onto a screen perspectively along the negative Z axis. The field of view is given as slope.
	//  The screen is not a square, so the slope applies either to its width or to its height, whichever is dominant.
	// This function returns the direction from the coordinate origin to a pixel on the screen. The result is not normalized.
	XYZ_float_reg SIMD_CALL directionOfPerspectivePixel(
		UIntLeast2B xFromLeftInPixels,
		UIntLeast2B yFromTopInPixels,
		UIntLeast2B screenWidthInPixels,
		UIntLeast2B screenHeightInPixels,
		UIntLeast2B dominantDimensionInPixels,
		float       slope
	) {
		// Move the origin from the upper left to the center, inverting the Y axis.
		auto const centeredX = SIntLeast2B(xFromLeftInPixels) - SIntLeast2B(screenWidthInPixels / 2);
		auto const centeredY = SIntLeast2B(screenHeightInPixels / 2) - SIntLeast2B(yFromTopInPixels);

		auto scale = 1.0f / (slope * float(dominantDimensionInPixels / 2));

		return xyzFrom(centeredX * scale, centeredY * scale, -1.0f);
	}



	//  X Y Z _ S I n t 4 B :

	XYZ_SInt4B_reg SIMD_CALL load(XYZ_SInt2B const & v) {
		return { xyz0From(&v.x) };
	}
	XYZ_SInt4B_reg SIMD_CALL load(XYZ_SInt4B const & v) {
		return { ::xyzFrom(&v.x) };
	}
	XYZ_SInt4B_reg SIMD_CALL loadPadded(XYZ_SInt2B const & v) {
		return { xyzwFrom(&v.x) };
	}
	void SIMD_CALL store(XYZ_SInt2B & m, XYZ_SInt4B_reg v) {
		xyzTo(&m.x, v.xyz_);
	}
	void SIMD_CALL store(XYZ_SInt4B & m, XYZ_SInt4B_reg v) {
		xyzTo(&m.x, v.xyz_);
	}

	XYZ_SInt4B SIMD_CALL stored(XYZ_SInt4B_reg const xyz) {
		XYZ_SInt4B result;
		store(result, xyz);
		return result;
	}

	XYZ_SInt4B_reg SIMD_CALL XYZ_SInt4B_reg::operator -() const {
		return { SInt4Bx4_zero() - xyz_ }; // XOR & SUB is faster than bit twiddling (no memory access!)
	}

	SInt4B SIMD_CALL xOf(XYZ_SInt4B_reg v) {
		return ::xOf(v.xyz_);
	}
	SInt4B SIMD_CALL yOf(XYZ_SInt4B_reg v) {
		return ::yOf(v.xyz_);
	}
	SInt4B SIMD_CALL zOf(XYZ_SInt4B_reg v) {
		return ::zOf(v.xyz_);
	}

	XYZ_SInt4B_reg SIMD_CALL xyzFrom(SInt4B x, SInt4B y, SInt4B z) {
		return { ::xyzwFrom(x, y, z, 0) };
	}

	XYZ_SInt4B_reg SIMD_CALL componentWiseMinimumOf(XYZ_SInt4B_reg a, XYZ_SInt4B_reg b) {
		return { ::minimumOf(a.xyz_, b.xyz_) };
	}

	XYZ_float_reg SIMD_CALL asFloats(XYZ_SInt4B_reg v) {
		return { ::asFloats(v.xyz_) };
	}

	VECTOR_OPERATOR_VECTOR(XYZ_SInt4B, +, xyz_) // +
	VECTOR_OPERATOR_VECTOR(XYZ_SInt4B, -, xyz_) // -
	VECTOR_ASSIGN_OPERATOR_VECTOR(XYZ_SInt4B, +, xyz_) // +=
	VECTOR_ASSIGN_OPERATOR_VECTOR(XYZ_SInt4B, -, xyz_) // -=

	// Tests the given vectors for bitwise equality.
	bool SIMD_CALL operator ==(XYZ_SInt4B_reg a, XYZ_SInt4B_reg b) {
		return equalXYZ(a.xyz_, b.xyz_);
	}
	bool SIMD_CALL operator !=(XYZ_SInt4B_reg a, XYZ_SInt4B_reg b) {
		return no == equalXYZ(a.xyz_, b.xyz_);
	}

	// Shifts each components value to the right.
	//   replicates the sign bit
	//   shift must not exceed 31 bits
	XYZ_SInt4B_reg SIMD_CALL shiftRightArithmetic(XYZ_SInt4B_reg v, UIntLeast1B bits) {
		return { ::shiftRightArithmetic(v.xyz_, bits) };
	}

	// Shifts each components value to the left.
	//   shift must not exceed 31 bits
	XYZ_SInt4B_reg SIMD_CALL operator <<(XYZ_SInt4B_reg v, UIntLeast1B bits) {
		return { v.xyz_ << bits };
	}






	Matrix3x3_float_reg SIMD_CALL load(Matrix3x3_float const & m) {
		return {
			loadx4(&m.columns[0].x),
			loadx4(&m.columns[1].x),
			yzwxOf(loadx4(&m.columns[1].z))
		};
	}

	// Computes the X, Y, and Z axis of the given rotation (useful as an object-to-world matrix).
	Matrix3x3_float_reg SIMD_CALL axesOf(Quaternion_float_reg q) {
		MUST(isFinite(q));

		auto const ijkw = q.ijkw;
		auto const jiij = yxxyOf(q.ijkw);
		auto const kkjj = zzyyOf(q.ijkw);
		auto const wwww = wwwwOf(q.ijkw);

		auto const a = jiij * kkjj;
		auto const b = ijkw * wwww;

		// 2  (j  k + i  w)
		// 2  (i  k + j  w)
		// 2  (i  j + k  w)
		// 2  (_  _ + _  _)
		auto const c = (a + a) + (b + b);

		// 2  (j  k - i  w)
		// 2  (i  k - j  w)
		// 2  (i  j - k  w)
		// 2  (_  _ - _  _)
		auto const d = (a + a) - (b + b);

		// w - j - k + i (x)
		// w - i - k + j (y)
		// w - i - j + k (z)
		// _ - _ - _ + _
		auto const e = (wwww * wwww - jiij * jiij) - (kkjj * kkjj - ijkw * ijkw);

		// Shuffling is the worst part as components are ordered in the least optimal way:
		// c0:   w - j - k + i     2  (i  j + k  w)   2  (i  k - j  w)
		// c1:  2  (i  j - k  w)     w - i - k + j    2  (j  k + i  w)
		// c2:  2  (i  k + j  w)    2  (j  k - i  w)    w - i - j + k
#		if COMPILED_BY_VISUAL_CPP
			auto const cd = MM_SHUFFLE_PS(c.floats, d.floats, 1, 1, 0, 0);
			auto const ec = MM_SHUFFLE_PS(e.floats, c.floats, 0, 0, 2, 2);
			auto const de = MM_SHUFFLE_PS(d.floats, e.floats, 2, 2, 1, 1);
			return {
				MM_SHUFFLE_PS(ec, d.floats, 0, 2, 1, 0),
				MM_SHUFFLE_PS(de, c.floats, 0, 2, 0, 0),
				MM_SHUFFLE_PS(cd, e.floats, 1, 2, 2, 0)
			};
#		elif COMPILED_BY_CLANG || COMPILED_BY_GCC
			return {
				(floatx4){ e.floats[0], c.floats[2], d.floats[1] },
				(floatx4){ d.floats[2], e.floats[1], c.floats[0] },
				(floatx4){ c.floats[1], d.floats[0], e.floats[2] }
			};
#		endif
	}
	// Computes the X, Y, and Z axis of the given rotations inverse (useful as a world-to-object matrix).
	Matrix3x3_float_reg SIMD_CALL inverseAxesOf(Quaternion_float_reg q) {
		return axesOf(inverted(q));
	}

	XYZ_float_reg SIMD_CALL xAxisOf(Matrix3x3_float_reg m) {
		return reinterpret_cast<XYZ_float_reg const &>(m.c0);
	}
	XYZ_float_reg SIMD_CALL yAxisOf(Matrix3x3_float_reg m) {
		return reinterpret_cast<XYZ_float_reg const &>(m.c1);
	}
	XYZ_float_reg SIMD_CALL zAxisOf(Matrix3x3_float_reg m) {
		return reinterpret_cast<XYZ_float_reg const &>(m.c2);
	}

	// Transforms the given vector with the given 33 matrix, typically rotating it.
	XYZ_float_reg SIMD_CALL transformed(
		XYZ_float_reg       v,
		Matrix3x3_float_reg m
	) {
		return { xxxxOf(v.xyz_) * m.c0 + yyyyOf(v.xyz_) * m.c1 + zzzzOf(v.xyz_) * m.c2 };
	}

	// Divides the vector by its Z.
	XYZ_float_reg SIMD_CALL projectedAlongZ(XYZ_float_reg v) {
		return { v.xyz_ / zzzzOf(v.xyz_) };
	}

	Matrix3x3_float_reg SIMD_CALL transposed(Matrix3x3_float_reg m) {
#		if COMPILED_BY_VISUAL_CPP
			// Input:
			//   a b c _
			//   d e f _
			//   g h i _
			auto adbe = _mm_unpacklo_ps(m.c0.floats, m.c1.floats);
			auto g_h_ = _mm_unpacklo_ps(m.c2.floats, m.c2.floats);
			auto cf__ = _mm_unpackhi_ps(m.c0.floats, m.c1.floats);
			return {
				_mm_movelh_ps(adbe, m.c2.floats),     // a d g _
				_mm_movehl_ps(g_h_, adbe),            // b e h _
				_mm_shuffle_ps(cf__, m.c2.floats, 36) // c f i _
			};
#		elif COMPILED_BY_CLANG || COMPILED_BY_GCC
			return {
				{ m.c0.floats[0], m.c1.floats[0], m.c2.floats[0] },
				{ m.c0.floats[1], m.c1.floats[1], m.c2.floats[1] },
				{ m.c0.floats[2], m.c1.floats[2], m.c2.floats[2] }
			};
#		endif
	}

	// Equals scaling(x, y, z) & b. Scaling must be a finite vector.
	Matrix3x3_float_reg SIMD_CALL preScaled(
		XYZ_float_reg       scaling,
		Matrix3x3_float_reg m
	) {
		MUST(isFinite(scaling));

		return {
			m.c0 * xxxxOf(scaling.xyz_),
			m.c1 * yyyyOf(scaling.xyz_),
			m.c2 * zzzzOf(scaling.xyz_)
		};
	}

	// Equals translation(offset) & b. The offset must be a finite vector.
	Matrix4x3_float_reg SIMD_CALL preTranslated(
		XYZ_float_reg       offset,
		Matrix3x3_float_reg m
	) {
		MUST(isFinite(offset));

		return {
			m.c0,
			m.c1,
			m.c2,
			xxxxOf(offset.xyz_) * m.c0 + yyyyOf(offset.xyz_) * m.c1 + zzzzOf(offset.xyz_) * m.c2
		};
	}

	Matrix4x3_float_reg SIMD_CALL translated(
		Matrix3x3_float_reg m,
		XYZ_float_reg       offset
	) {
		MUST(isFinite(offset));

		return {
			m.c0,
			m.c1,
			m.c2,
			offset.xyz_
		};
	}

	Matrix3x3_float_reg SIMD_CALL operator &(
		Matrix3x3_float_reg         l,
		Matrix3x3_float_reg const & r
	) {
		return {
			r.c0 * xxxxOf(l.c0) + r.c1 * yyyyOf(l.c0) + r.c2 * zzzzOf(l.c0),
			r.c0 * xxxxOf(l.c1) + r.c1 * yyyyOf(l.c1) + r.c2 * zzzzOf(l.c1),
			r.c0 * xxxxOf(l.c2) + r.c1 * yyyyOf(l.c2) + r.c2 * zzzzOf(l.c2)
		};
	}

	void SIMD_CALL store(Matrix3x3_float & mem, Matrix3x3_float_reg m) {
		xyzwTo(&mem.columns[0].x, m.c0);
		xyzwTo(&mem.columns[1].x, m.c1);
		 xyzTo(&mem.columns[2].x, m.c2);
	}

	Matrix3x3_float SIMD_CALL stored(Matrix3x3_float_reg m) {
		Matrix3x3_float result;
		store(result, m);
		return result;
	}

	Matrix4x3_float_reg SIMD_CALL load(Matrix4x3_float const & m) {
		return {
			loadx4(&m.columns[0].x),
			loadx4(&m.columns[1].x),
			loadx4(&m.columns[2].x),
			yzwxOf(loadx4(&m.columns[2].z)) // a four-lane load with shuffling is less work than two loads
		};
	}
	XYZ_float_reg SIMD_CALL loadXAxisOf(Matrix4x3_float const & m) {
		return { loadx4(&m.columns[0].x) };
	}
	XYZ_float_reg SIMD_CALL loadYAxisOf(Matrix4x3_float const & m) {
		return { loadx4(&m.columns[1].x) };
	}
	XYZ_float_reg SIMD_CALL loadZAxisOf(Matrix4x3_float const & m) {
		return { loadx4(&m.columns[2].x) };
	}
	XYZ_float_reg SIMD_CALL loadOffsetOf(Matrix4x3_float const & m) {
		return { yzwxOf(loadx4(&m.columns[2].z)) }; // a four-lane load with shuffling is less work than two loads
	}

	void SIMD_CALL store(
		Matrix4x3_float &   mem,
		Matrix4x3_float_reg m
	) {
		xyzwTo(&mem.columns[0].x, m.c0);
		xyzwTo(&mem.columns[1].x, m.c1);
		xyzwTo(&mem.columns[2].x, m.c2);
		xyzTo(&mem.columns[3].x, m.c3);
	}
	Matrix4x3_float SIMD_CALL stored(Matrix4x3_float_reg m) {
		Matrix4x3_float result;
		store(result, m);
		return result;
	}

	Matrix4x3_float_reg SIMD_CALL matrix4x3From(
		Quaternion_float_reg q,
		XYZ_float_reg        offset
	) {
		MUST(isFinite(q));

		auto rotation = axesOf(q);
		return { rotation.c0, rotation.c1, rotation.c2, offset.xyz_ };
	}

	Matrix4x3_float_reg SIMD_CALL translated(
		Matrix4x3_float_reg m,
		XYZ_float_reg       offset
	) {
		MUST(isFinite(offset));

		return { m.c0, m.c1, m.c2, m.c3 + offset.xyz_ };
	}

	// Equals translation(offset) & b. The offset must be a finite vector.
	Matrix4x3_float_reg SIMD_CALL preTranslated(
		XYZ_float_reg       offset,
		Matrix4x3_float_reg m
	) {
		MUST(isFinite(offset));

		return {
			m.c0,
			m.c1,
			m.c2,
			// parantheses reduce data dependencies
			(xxxxOf(offset.xyz_) * m.c0 + yyyyOf(offset.xyz_) * m.c1) + (zzzzOf(offset.xyz_) * m.c2 + m.c3)
		};
	}

	// Returns a matrix with inverted scaling, rotation, and translation.
	Matrix4x3_float_reg SIMD_CALL inverted(Matrix4x3_float_reg m) {
		Matrix4x3_float_reg result;

		// Rotation is inverted by transposing the upper 33 matrix.
		auto const inverseRotation = transposed(rotationOf(m));

		// Scaling is inverted by dividing each column by its squared length. (Dividing by length only were normalization.)
		auto const squaredLengths = m.c0 * m.c0 + m.c1 * m.c1 + m.c2 * m.c2;
		result.c0 = inverseRotation.c0 / squaredLengths;
		result.c1 = inverseRotation.c1 / squaredLengths;
		result.c2 = inverseRotation.c2 / squaredLengths;

		// Translation is inverted by transforming the negated offset to the new coordinate system.
		result.c3 = floatx4Zero() - (xxxxOf(m.c3) * inverseRotation.c0 + yyyyOf(m.c3) * inverseRotation.c1 + zzzzOf(m.c3) * inverseRotation.c2);

		return result;
	}

	// Combines the given transformations  left first, then right (reverse matrix multiplication order).
	Matrix4x3_float_reg SIMD_CALL operator &(
		Matrix4x3_float_reg         l,
		Matrix3x3_float_reg const & r
	) {
		return {
			r.c0 * xxxxOf(l.c0) + r.c1 * yyyyOf(l.c0) + r.c2 * zzzzOf(l.c0),
			r.c0 * xxxxOf(l.c1) + r.c1 * yyyyOf(l.c1) + r.c2 * zzzzOf(l.c1),
			r.c0 * xxxxOf(l.c2) + r.c1 * yyyyOf(l.c2) + r.c2 * zzzzOf(l.c2),
			// parantheses reduce data dependencies
			(r.c0 * xxxxOf(l.c3) + r.c1 * yyyyOf(l.c3)) + (r.c2 * zzzzOf(l.c3))
		};
	}
	Matrix4x3_float_reg SIMD_CALL operator &(
		Matrix4x3_float_reg         l,
		Matrix4x3_float_reg const & r
	) {
		return {
			r.c0 * xxxxOf(l.c0) + r.c1 * yyyyOf(l.c0) + r.c2 * zzzzOf(l.c0),
			r.c0 * xxxxOf(l.c1) + r.c1 * yyyyOf(l.c1) + r.c2 * zzzzOf(l.c1),
			r.c0 * xxxxOf(l.c2) + r.c1 * yyyyOf(l.c2) + r.c2 * zzzzOf(l.c2),
			// parantheses reduce data dependencies
			(r.c0 * xxxxOf(l.c3) + r.c1 * yyyyOf(l.c3)) + (r.c2 * zzzzOf(l.c3) + r.c3)
		};
	}
	Matrix4x3_float_reg SIMD_CALL operator &(
		Matrix3x3_float_reg const & l,
		Matrix4x3_float_reg         r
	) {
		return {
			r.c0 * xxxxOf(l.c0) + r.c1 * yyyyOf(l.c0) + r.c2 * zzzzOf(l.c0),
			r.c0 * xxxxOf(l.c1) + r.c1 * yyyyOf(l.c1) + r.c2 * zzzzOf(l.c1),
			r.c0 * xxxxOf(l.c2) + r.c1 * yyyyOf(l.c2) + r.c2 * zzzzOf(l.c2),
			r.c3
		};
	}

	XYZ_float_reg SIMD_CALL xAxisOf(Matrix4x3_float_reg m) {
		return reinterpret_cast<XYZ_float_reg const &>(m.c0);
	}
	XYZ_float_reg SIMD_CALL yAxisOf(Matrix4x3_float_reg m) {
		return reinterpret_cast<XYZ_float_reg const &>(m.c1);
	}
	XYZ_float_reg SIMD_CALL zAxisOf(Matrix4x3_float_reg m) {
		return reinterpret_cast<XYZ_float_reg const &>(m.c2);
	}
	XYZ_float_reg SIMD_CALL offsetOf(Matrix4x3_float_reg m) {
		return reinterpret_cast<XYZ_float_reg const &>(m.c3);
	}
	float SIMD_CALL distanceOf(Matrix4x3_float_reg m) {
		return absoluteOf(reinterpret_cast<XYZ_float_reg const &>(m.c3));
	}
	Matrix3x3_float_reg SIMD_CALL rotationOf(Matrix4x3_float_reg m) {
		return { m.c0, m.c1, m.c2 };
	}

	// Transforms the given vector with the given 43 matrix, typically rotating and translating it.
	XYZ_float_reg SIMD_CALL transformed(
		XYZ_float_reg       v,
		Matrix4x3_float_reg m
	) {
		return {
			  m.c0 * xxxxOf(v.xyz_)
			+ m.c1 * yyyyOf(v.xyz_)
			+ m.c2 * zzzzOf(v.xyz_)
			+ m.c3
		};
	}

	Matrix2x2_float_reg SIMD_CALL Matrix2x2_float::uniformScaling(float s) {
		auto s000 = x000From(s);
		return { xyzxOf(s000) };
	}

	Matrix2x2_float_reg SIMD_CALL Matrix2x2_float::rotation(Angle4B_reg angle) {
		// See http://en.wikipedia.org/wiki/Rotation_matrix.
		auto cssc = xyyxOf(XY_float::polar(angle).xy__);
		return { negatedZ(cssc) }; // cos, sin,-sin, cos
	}

	Matrix3x2_float_reg SIMD_CALL Matrix3x2_float::translation(XY_float_reg offset) {
		MUST(isFinite(offset));

		auto _1000 = loadx4(axisFloats + 3);
		return {
			_1000,
			zx__Of(_1000),
			offset.xy__
		};
	}

	// Equals translation(offset) & b. The offset must be a finite vector.
	Matrix3x2_float_reg SIMD_CALL preTranslated(
		XY_float_reg        offset,
		Matrix2x2_float_reg m
	) {
		MUST(isFinite(offset));

		return {
			m.c01,
			zw__Of(m.c01),
			xxxxOf(offset.xy__) * m.c01 + yyyyOf(offset.xy__) * zw__Of(m.c01)
		};
	}

	Matrix3x2_float_reg SIMD_CALL translated(
		Matrix2x2_float_reg m,
		XY_float_reg        offset
	) {
		MUST(isFinite(offset));

		return { m.c01, zw__Of(m.c01), offset.xy__ };
	}

	Matrix3x2_float_reg SIMD_CALL translated(
		Matrix3x2_float_reg m,
		XY_float_reg        offset
	) {
		MUST(isFinite(offset));

		return { m.c0, m.c1, m.c2 + offset.xy__ };
	}

	// Combines the given transformations  left first, then right (reverse matrix multiplication order).
	Matrix2x2_float_reg SIMD_CALL operator &(
		Matrix2x2_float_reg l,
		Matrix2x2_float_reg r
	) {
		return { xyxyFrom(
			r.c01 * xxxxOf(l.c01) + r.c01 * yyyyOf(l.c01),
			zw__Of(r.c01) * zzzzOf(l.c01) + zw__Of(r.c01) * wwwwOf(l.c01)
		) };
	}
	Matrix3x2_float_reg SIMD_CALL operator &(
		Matrix2x2_float_reg l,
		Matrix3x2_float_reg r
	) {
		return {
			r.c0 * xxxxOf(l.c01) + r.c0 * yyyyOf(l.c01),
			r.c1 * zzzzOf(l.c01) + r.c1 * wwwwOf(l.c01),
			r.c2
		};
	}

	// Transforms the given vector with the given 22 matrix, typically rotating it.
	XY_float_reg SIMD_CALL transformed(
		XY_float_reg        v,
		Matrix2x2_float_reg m
	) {
		return { xxxxOf(v.xy__) * m.c01 + yyyyOf(v.xy__) * zw__Of(m.c01) };
	}

	// Transforms the given vector with the given 32 matrix, typically rotating and translating it.
	XY_float_reg SIMD_CALL transformed(
		XY_float_reg        v,
		Matrix3x2_float_reg m
	) {
		return { xxxxOf(v.xy__) * m.c0 + yyyyOf(v.xy__) * m.c1 + m.c2 };
	}

	void SIMD_CALL store(Matrix3x2_float & mem, Matrix3x2_float_reg mat) {
		xyzwTo(&mem.columns[0].x, xyxyFrom(mat.c0, mat.c1));
		xyTo(&mem.columns[2].x, mat.c2);
	}


	Matrix3x3_float_reg SIMD_CALL Matrix3x3_float::identity() {
		return {
			loadx4(axisFloats + 3),
			loadx4(axisFloats + 2),
			loadx4(axisFloats + 1)
		};
	}

	Matrix3x3_float_reg SIMD_CALL Matrix3x3_float::uniformScaling(float s) {
		auto s000 = x000From(s);
		return {
			s000,
			zxywOf(s000),
			yzxwOf(s000)
		};
	}

	Matrix3x3_float_reg SIMD_CALL Matrix3x3_float::scaling(float x, float y, float z) {
		auto x000 = x000From(x);
		auto y000 = x000From(y);
		auto z000 = x000From(z);
		return {
			x000,
			zxywOf(y000),
			yzxwOf(z000)
		};
	}

	Matrix3x3_float_reg SIMD_CALL Matrix3x3_float::scaling(XYZ_float_reg const xyz) {
		return scaling(xOf(xyz), yOf(xyz), zOf(xyz));
	}

	Matrix3x3_float_reg SIMD_CALL Matrix3x3_float::rotationAboutX(Angle4B_reg angle) {
		// See http://en.wikipedia.org/wiki/Rotation_matrix.
		auto cs00 = XY_float::polar(angle).xy__; // cosine, sine, 0, 0
		return {
			loadx4(axisFloats + 3),                                  // 1,  0,   0,  
			zxywOf(cs00),                                            // 0, cos, sin, 
			interleavedXY(zx__Of(cs00), y___Of(floatx4Zero() - cs00))// 0,-sin, cos, 
		};
	}

	Matrix3x3_float_reg SIMD_CALL Matrix3x3_float::rotationAboutY(Angle4B_reg angle) {
		// See http://en.wikipedia.org/wiki/Rotation_matrix.
		auto cs00 = XY_float::polar(angle).xy__; // cosine, sine, 0, 0
#		if COMPILED_BY_VISUAL_CPP
			auto c0 = MM_SHUFFLE_PS(cs00.floats, (floatx4Zero() - cs00).floats, 0, 2, 1, 0);
#		elif COMPILED_BY_CLANG || COMPILED_BY_GCC
			auto c0 = xyxyFrom(xz__Of(cs00), y___Of(floatx4Zero() - cs00));
#		endif
		return {
			c0,                    // cos, 0,-sin, 
			loadx4(axisFloats + 2),//  0,  1,  0,  
			yzxwOf(cs00)           // sin, 0, cos, 
		};
	}

	Matrix3x3_float_reg SIMD_CALL Matrix3x3_float::rotationAboutZ(Angle4B_reg angle) {
		// See http://en.wikipedia.org/wiki/Rotation_matrix.
		auto cs00 = XY_float::polar(angle).xy__; // cosine, sine, 0, 0
		return {
			cs00,                                             // cos, sin,  0, 
			interleavedXY(floatx4Zero() - yz__Of(cs00), cs00),//-sin, cos,  0, 
			loadx4(axisFloats + 1)                            //  0,   0,   1, 
		};
	}


	// Creates a right-handed rotation about the given axis by the given angle.
	Matrix3x3_float_reg SIMD_CALL Matrix3x3_float::rotationAbout(
		XYZ_float_reg axis,
		Angle4B_reg   angle
	) {
		// TODO: vectorize this shit
		auto const cosineAndSine = XY_float::polar(angle);
		auto const t = 1.0f - xOf(cosineAndSine);

		auto const tx = t * xOf(axis);
		auto const ty = t * yOf(axis);
		auto const tz = t * zOf(axis);

		auto const xy = tx * yOf(axis);
		auto const xz = tx * zOf(axis);
		auto const yz = ty * zOf(axis);

		auto const xs = xOf(axis) * yOf(cosineAndSine);
		auto const ys = yOf(axis) * yOf(cosineAndSine);
		auto const zs = zOf(axis) * yOf(cosineAndSine);

		return {
			xyzwFrom(
				tx * xOf(axis) + xOf(cosineAndSine),
				xy + zs,
				xz - ys,
				0.0f
			), xyzwFrom(
				xy - zs,
				ty * yOf(axis) + xOf(cosineAndSine),
				yz + xs,
				0.0f
			), xyzwFrom(
				xz + ys,
				yz - xs,
				tz * zOf(axis) + xOf(cosineAndSine),
				0.0f
			)
		};
	}

	Matrix3x3_float_reg SIMD_CALL Matrix3x3_float::rollPitchYaw(
		Angle4B_reg roll,
		Angle4B_reg pitch,
		Angle4B_reg yaw
	) {
		//      cos(r) -sin(r) 0        1     0      0          cos(p) 0 sin(p)
		//      sin(r) cos(r)  0       0  cos(y) -sin(y)         0   1   0
		//        0         0  1        0  sin(y) cos(y)       -sin(p) 0 cos(p)
		// equals
		//      cos(r) -sin(r)  cos(y) -sin(r)  -sin(y)       cos(p) 0 sin(p)
		//      sin(r)  cos(r)  cos(y)  cos(r)  -sin(y)         0   1   0
		//        0          sin(y)           cos(y)           -sin(p) 0 cos(p)
		// equals
		//      cos(r)  cos(p) +-sin(r)  -sin(y)  -sin(p)   -sin(r)  cos(y)   cos(r)  sin(p) +-sin(r)  -sin(y)  cos(p)
		//      sin(r)  cos(p) + cos(r)  -sin(y)  -sin(p)    cos(r)  cos(y)   sin(r)  sin(p) + cos(r)  -sin(y)  cos(p)
		//                                  cos(y)  -sin(p)         sin(y)                    cos(y)  cos(p)
		// simplified to
		//      cos(r)  cos(p) - sin(r)  sin(y)  sin(p)   -sin(r)  cos(y)   cos(r)  sin(p) + sin(r)  sin(y)  cos(p)
		//      sin(r)  cos(p) + cos(r)  sin(y)  sin(p)    cos(r)  cos(y)   sin(r)  sin(p) - cos(r)  sin(y)  cos(p)
		//                    cos(y)  -sin(p)                     sin(y)                    cos(y)  cos(p)
		auto r_cs = XY_float::polar(roll).xy__;
		auto p_cs = XY_float::polar(pitch).xy__;
		auto y_cs = XY_float::polar(yaw).xy__;

		auto negateXZ = floatx4SignMask_XZ();
		auto r_ccss = xxyyOf(r_cs);
		auto y_cssc = xyyxOf(y_cs);
		auto r_sscc = yyxxOf(r_cs);
		auto p_ssss = yyyyOf(p_cs);
		auto y_sccs = yxxyOf(y_cs);

		// cos(r)  cos(y) - sin(r)  sin(p)  sin(y)
		// cos(r)  sin(y) + sin(r)  sin(p)  cos(y)
		// sin(r)  sin(y) - cos(r)  sin(p)  cos(y)
		// sin(r)  cos(y) + cos(r)  sin(p)  sin(y)
		auto a = r_ccss * y_cssc + (negateXZ ^ (r_sscc * p_ssss * y_sccs));

		// -cos(p)  sin(r)
		//  cos(p)  cos(r)
		// -cos(p)  sin(y)
		//  cos(p)  cos(y)
		auto b = negateXZ ^ (xxxxOf(p_cs) * yxyxFrom(r_cs, y_cssc));

		return {
			//  cos(r)  cos(y) - sin(r)  sin(p)  sin(y)
			// -sin(r)                    cos(p)
			//  cos(r)  sin(y) + sin(r)  sin(p)  cos(y)
			interleavedXY(a, b),
			// sin(r)  cos(y) + cos(r)  sin(p)  sin(y)
			// cos(r)                    cos(p)
			// sin(r)  sin(y) - cos(r)  sin(p)  cos(y)
			interleavedXY(wz__Of(a), y___Of(b)),
			// -cos(p)  sin(y)
			//  sin(p)
			//  cos(p)  cos(y)
			interleavedXY(zw__Of(b), p_ssss)
		};
	}

	// Computes an eye-to-world matrix from two polar angles.
	//   0, 0 is along Z
	//   positive pitch looks towards Y
	//   positive yaw looks towards X
	Matrix3x3_float_reg SIMD_CALL Matrix3x3_float::polar(
		Angle4B_reg pitch,
		Angle4B_reg yaw
	) {
		Matrix3x3_float_reg result;

		auto const p = XY_float::polar(pitch).xy__;
		auto const y = XY_float::polar(yaw).xy__;
		// sin(h)  sin(v)
		// cos(h)  sin(v)
		// sin(h)  cos(v)
		// cos(h)  cos(v)
		auto const yp = yxyxOf(y) * yyxxOf(p);

		// cos(h)
		// 0
		// -sin(h)
#		if COMPILED_BY_VISUAL_CPP
			result.c0.floats = MM_SHUFFLE_PS(y.floats, (floatx4Zero() - y).floats, 0, 2, 1, 0);
#		elif COMPILED_BY_CLANG || COMPILED_BY_GCC
			result.c0 = xyxyFrom(xz__Of(y), y___Of(floatx4Zero() - y));
#		endif
		// sin(h)  sin(v)
		// cos(v)
		// cos(h)  sin(v)
		result.c1 = interleavedXY(yp, p);
		// sin(h)  cos(v)
		// -sin(v)
		// cos(h)  cos(v)
		result.c2 = interleavedXY(zw__Of(yp), y___Of(floatx4Zero() - p));

		return result;
	}

	// Computes a world-to-eye matrix from two polar angles (first inverse of pitch, then inverse of yaw).
	//  This is the common rotation in first-person view, e.g. for first-person shooters.
	Matrix3x3_float_reg SIMD_CALL Matrix3x3_float::ego(
		Angle4B_reg pitch,
		Angle4B_reg yaw
	) {
		Matrix3x3_float_reg result;

		// Do not negate the angles; instead transpose the matrix.
		auto const p = XY_float::polar(pitch);
		auto const y = XY_float::polar(yaw);
		// sin(y)  sin(p)
		// cos(y)  sin(p)
		// sin(y)  cos(p)
		// cos(y)  cos(p)
		auto const yp = yxyxOf(y.xy__) * yyxxOf(p.xy__);

		// cos(y)
		// sin(p)  sin(y)
		// cos(p)  sin(y)
		result.c0 = xyzwFrom(xOf(y), ::xOf(yp), ::zOf(yp), 0.0f);
		// 0
		// cos(p)
		// -sin(p)
		result.c1 = xyzwFrom(0.0f, xOf(p), ::yOf(floatx4Zero() - p.xy__), 0.0f);
		// -sin(y)
		// sin(p)  cos(y)
		// cos(p)  cos(y)
		result.c2 = xyzwFrom(::yOf(floatx4Zero() - y.xy__), ::yOf(yp), ::wOf(yp), 0.0f);

		return result;
	}

	// Creates a right-handed coordinate system with the given scale around the given Z axis.
	Matrix3x3_float_reg SIMD_CALL Matrix3x3_float::fromZ(XYZ_float_reg normal, float scaleFactor) {
		MUST(isValidDirection(normal));

		XYZ_float_reg x;
		XYZ_float_reg y;
		auto          z = scaledToLength(normal, scaleFactor);

		if(isWithin(0.001f, yOf(normal), 0.999f)) { // Diagonal upwards?
			// resolved cross product of z with (0, 1, 0)
			y = scaledToLength(xyzFrom(0.0f - zOf(normal), 0.0f, xOf(normal)), scaleFactor);
			x = scaledToLength(crossProductOf(y, normal), scaleFactor);
		} else if(isWithin(-0.999f, yOf(normal), -0.001f)) { // Diagonal downwards?
			// resolved cross product of z with (0, -1, 0)
			y = scaledToLength(xyzFrom(zOf(normal), 0.0f, 0.0f - xOf(normal)), scaleFactor);
			x = scaledToLength(crossProductOf(y, z), scaleFactor);
		} else { // Exactly horizontal or exactly vertical?

			auto const normalSquared = normal * normal;
			if(xOf(normalSquared) < yOf(normalSquared)) { // Vertical?
				auto const length = scaleFactor / squareRootOf(yOf(normalSquared) + zOf(normalSquared));
				y = xyzFrom(0.0f, zOf(normal), yOf(normal)) * length;
				x = scaledToLength(crossProductOf(y, z), scaleFactor);
			} else { // Horizontal?
				auto const length = scaleFactor / squareRootOf(xOf(normalSquared) + zOf(normalSquared));
				x = xyzFrom(zOf(normal), 0.0f, 0.0f - xOf(normal)) * length;
				y = scaledToLength(crossProductOf(z, x), scaleFactor);
			}

			y = scaledToLength(crossProductOf(z, x), scaleFactor);
		}

		return { x.xyz_, y.xyz_, z.xyz_ };
	}

	// Creates a right-handed coordinate system with the given scale around the given Y axis.
	Matrix3x3_float_reg SIMD_CALL Matrix3x3_float::fromY(XYZ_float_reg normal, float scaleFactor) {
		// TODO: the fast hack
		auto result = fromZ(-normal, scaleFactor);
		auto tmp = result.c1;
		result.c1 = result.c2;
		result.c2 = floatx4Zero() - tmp;
		return result;
	}

	Matrix4x3_float_reg SIMD_CALL Matrix4x3_float::identity() {
		return {
			loadx4(axisFloats + 3),
			loadx4(axisFloats + 2),
			loadx4(axisFloats + 1),
			floatx4Zero()
		};
	}

	Matrix4x3_float_reg SIMD_CALL Matrix4x3_float::rollPitchYaw(
		Angle4B_reg   rollAngle,
		Angle4B_reg   pitchAngle,
		Angle4B_reg   yawAngle,
		XYZ_float_reg offset
	) {
		return translated(Matrix3x3_float::rollPitchYaw(rollAngle, pitchAngle, yawAngle), offset);
	}

	// Computes an eye-to-world matrix with offset from two polar angles.
	//   eye offset is rotated
	//   0, 0 is along Z
	//   positive pitch looks towards Y
	//   positive yaw looks towards X
	Matrix4x3_float_reg SIMD_CALL Matrix4x3_float::polar(
		XYZ_float_reg offset,
		Angle4B_reg   pitch,
		Angle4B_reg   yaw
	) {
		Matrix4x3_float_reg result;

		auto const p = XY_float::polar(pitch);
		auto const y = XY_float::polar(yaw);
		// sin(h)  sin(v)
		// cos(h)  sin(v)
		// sin(h)  cos(v)
		// cos(h)  cos(v)
		auto const yp = yxyxOf(y.xy__) * yyxxOf(p.xy__);

		// cos(h)
		// 0
		// -sin(h)
		result.c0 = x0z0From(xOf(y), ::yOf(floatx4Zero() - y.xy__));
		// sin(h)  sin(v)
		// cos(v)
		// cos(h)  sin(v)
		result.c1 = interleavedXY(yp, p.xy__);
		// sin(h)  cos(v)
		// -sin(v)
		// cos(h)  cos(v)
		result.c2 = interleavedXY(zw__Of(yp), y___Of(floatx4Zero() - p.xy__));
		result.c3 = result.c0 * xxxxOf(offset.xyz_)
		          + result.c1 * yyyyOf(offset.xyz_)
		          + result.c2 * zzzzOf(offset.xyz_);

		return result;
	}

	Matrix4x4_float_reg SIMD_CALL load(Matrix4x4_float const & m) {
		return {
			loadx4(&m.columns[0].x),
			loadx4(&m.columns[1].x),
			loadx4(&m.columns[2].x),
			loadx4(&m.columns[3].x)
		};
	}

	void SIMD_CALL store(Matrix4x4_float & mem, Matrix4x4_float_reg const & m) {
		xyzwTo(&mem.columns[0].x, m.c0);
		xyzwTo(&mem.columns[1].x, m.c1);
		xyzwTo(&mem.columns[2].x, m.c2);
		xyzwTo(&mem.columns[3].x, m.c3);
	}

	Matrix4x4_float_reg SIMD_CALL matrix4x4From(Matrix4x3_float_reg m) {
		return {
			xyz0Of(m.c0),
			xyz0Of(m.c1),
			xyz0Of(m.c2),
			xyz1Of(m.c3)
		};
	}

	Matrix4x4_float_reg SIMD_CALL Matrix4x4_float::identity() {
		return {
			loadx4(axisFloats + 3),
			loadx4(axisFloats + 2),
			loadx4(axisFloats + 1),
			loadx4(axisFloats + 0)
		};
	}

	Matrix4x4_float_reg SIMD_CALL Matrix4x4_float::projection(
		SIntLeast2B screenWidth,
		SIntLeast2B screenHeight,
		float       zoomScaleY,
		float       near,
		float       far
	) {
		MUST(0 < screenWidth && 0 < screenHeight);

		Matrix4x4_float_reg result;

		auto const zoomScaleX = zoomScaleY * float(screenHeight) / float(screenWidth);
		auto const ratio      = far / (near - far);

		// zoomScaleX, 0, 0, 0
		result.c0 = x000From(zoomScaleX);
		// 0, zoomScaleY, 0, 0
		result.c1 = zxywOf(x000From(zoomScaleY));
		// 0, 0, ratio, 1
		result.c2 = xyzwFrom(0.0f, 0.0f, ratio, -1.0f);
		// 0, 0, near  ratio, 0
		result.c3 = zyxwOf(x000From(near * ratio));

		return result;
	}

	Matrix4x4_float_reg SIMD_CALL Matrix4x4_float::projection_legacy(
		SIntLeast2B screenWidth,
		SIntLeast2B screenHeight,
		float       zoomScaleY,
		float       near,
		float       far
	) {
		MUST(0 < screenWidth && 0 < screenHeight);

		Matrix4x4_float_reg result;

		auto const zoomScaleX = zoomScaleY * float(screenHeight) / float(screenWidth);
		auto const ratio      = far / (far - near);

		// zoomScaleX, 0, 0, 0
		result.c0 = x000From(zoomScaleX);
		// 0, -zoomScaleY, 0, 0
		result.c1 = floatx4Zero() - zxywOf(x000From(zoomScaleY));
		// 0, 0, ratio, 1
		result.c2 = xyzwFrom(0.0f, 0.0f, ratio, 1.0f);
		// 0, 0, -near  ratio, 0
		result.c3 = floatx4Zero() - zyxwOf(x000From(near * ratio));

		return result;
	}

	//  transforms right-handed coordinates to left-handed screen space
	//  uses progressive Z, i.e. the near clipping plane has a depth of 1 and the far clipping plane has a depth of 0  this
	//   improves depth buffer precision greatly, as the non-linearity of Z is canceled out with the non-linearity of floating-
	//   -point values in the depth buffer (see Humus, A couple of notes about Z and Outerra, Maximizing Depth Buffer Range and
	//    Precision)
	Matrix4x4_float_reg SIMD_CALL Matrix4x4_float::projectionProgressive(
		SIntLeast2B screenWidth,
		SIntLeast2B screenHeight,
		float       zoomScaleY,
		float       near,
		float       far
	) {
		MUST(0 < screenWidth && 0 < screenHeight);

		Matrix4x4_float_reg result;

		auto const zoomScaleX = zoomScaleY * float(screenHeight) / float(screenWidth);
		auto const ratio      = near / (near - far);

		// zoomScaleX, 0, 0, 0
		result.c0 = x000From(zoomScaleX);
		// 0, zoomScaleY, 0, 0
		result.c1 = zxywOf(x000From(zoomScaleY));
		// 0, 0, ratio, -1
		result.c2 = xyzwFrom(0.0f, 0.0f, ratio, -1.0f);
		// 0, 0, -far  ratio, 0
		result.c3 = floatx4Zero() - zyxwOf(x000From(far * ratio));

		return result;
	}

	// Transforms the given vector with the given 44 matrix, typically rotating and translating it.
	XYZ_float_reg SIMD_CALL transformed(
		XYZ_float_reg               v,
		Matrix4x4_float_reg const & m
	) {
		return {
			  m.c0 * xxxxOf(v.xyz_)
			+ m.c1 * yyyyOf(v.xyz_)
			+ m.c2 * zzzzOf(v.xyz_)
			+ m.c3
		};
	}


	FixedFrustum3D_reg SIMD_CALL frustumFrom(
		float tangentOfHalfFieldOfView,
		float widthInArbitraryUnit,
		float heightInSameUnit
	) {
		// If seen from above, the vector from the origin to the right screen border lies exactly on the right frustum plane:
		//
		//    xZoom = tangentOfHalfFieldOfView  (widthInArbitraryUnit / heightInSameUnit)
		//    x = xZoom
		//    y = 0
		//    z = 1
		//
		// Rotating it 90 about the Y axis gives the normal of the plane. Since the rotation is about a world axis, it can be
		//  realized by swapping the X and Z components while negating one of the values:
		//
		//    x = 1
		//    y = 0
		//    z = -xZoom
		//
		// The resulting vector needs to be normalized, where the following steps can be skipped:
		//     negating the first term (because its square is positive again);
		//     dividing the Y component (because it equals zero);
		//     computing the square of 1.
		auto const xZoom = tangentOfHalfFieldOfView * (widthInArbitraryUnit / heightInSameUnit);
		auto const rightPlanesLength = squareRootOf(::squared(xZoom) + 1.0f);

		// The same goes for the upper border if seen from the right (consider that the Y axis is flipped):
		//
		//    yZoom = tangentOfHalfFieldOfView
		//    x = 0
		//    y = 1
		//    z = -yZoom
		auto const yZoom = tangentOfHalfFieldOfView;
		auto const upperPlanesLength = squareRootOf(::squared(yZoom) + 1.0f);

		return { xyzwFrom(
			        1.0f / rightPlanesLength,
			0.0f - xZoom / rightPlanesLength,
			        1.0f / upperPlanesLength,
			0.0f - yZoom / upperPlanesLength
		) };
	}

	FrustumClassification SIMD_CALL classifySphere(
		FixedFrustum3D_reg const & frustum,
		XYZ_float_reg              center,
		float                      radius
	) {
		auto const negativeRadius = 0.0f - radius;

		// If the sphere is behind the viewer, it is invisible. This should be tested first, because
		//  a) it sorts out 50% of the scene; and
		//  b) the test requires no computation.
		if(negativeRadius >= zOf(center)) {
			return FrustumClassification::invisible;
		}

		// The frustum planes are symmetrical, so instead of testing against four planes, test the absolute against two planes:
		//   abs(center.x)
		//   abs(center.y)
		//   center.z
		//   ?
		auto center_xzyz = interleavedXY(::absoluteOf(center.xyz_), zzzzOf(center.xyz_));
		// rightNormal.x  center.x
		// rightNormal.z  center.z
		// upperNormal.y  center.y
		// upperNormal.z  center.z
		auto distance_x_y_ = center_xzyz * frustum.xzyz;
		// rightNormal.x  center.x + rightNormal.z  center.z (distance from X plane)
		// ?
		// upperNormal.y  center.y + upperNormal.z  center.z (distance from Y plane)
		// ?
		distance_x_y_ = distance_x_y_ + y_w_Of(distance_x_y_);
		auto const distanceFromXPlanes = ::xOf(distance_x_y_);
		auto const distanceFromYPlanes = ::zOf(distance_x_y_);
		if(radius <= distanceFromXPlanes || radius <= distanceFromYPlanes) {
			return FrustumClassification::invisible;
		}

		// Some part of the sphere is in the frustum; test if it is completely inside.
		if(radius < zOf(center)
		&& negativeRadius > distanceFromXPlanes
		&& negativeRadius > distanceFromYPlanes
		) {
			return FrustumClassification::visible;
		}

		// Only parts of the sphere are visible. This is not a desired case, because it causes hierarchical culling to perform
		//  further tests.
		return FrustumClassification::partlyVisible;
	}


	// Combines the given transformations  left first, then right (reverse matrix multiplication order).
	Matrix4x4_float_reg SIMD_CALL operator &(
		Matrix3x3_float_reg         l,
		Matrix4x4_float_reg const & r
	) {
		return {
			r.c0 * xxxxOf(l.c0) + r.c1 * yyyyOf(l.c0) + r.c2 * zzzzOf(l.c0),
			r.c0 * xxxxOf(l.c1) + r.c1 * yyyyOf(l.c1) + r.c2 * zzzzOf(l.c1),
			r.c0 * xxxxOf(l.c2) + r.c1 * yyyyOf(l.c2) + r.c2 * zzzzOf(l.c2),
			r.c3
		};
	}
	Matrix4x4_float_reg SIMD_CALL operator &(
		Matrix4x3_float_reg         l,
		Matrix4x4_float_reg const & r
	) {
		return {
			r.c0 * xxxxOf(l.c0) + r.c1 * yyyyOf(l.c0) + r.c2 * zzzzOf(l.c0),
			r.c0 * xxxxOf(l.c1) + r.c1 * yyyyOf(l.c1) + r.c2 * zzzzOf(l.c1),
			r.c0 * xxxxOf(l.c2) + r.c1 * yyyyOf(l.c2) + r.c2 * zzzzOf(l.c2),
			// parantheses reduce data dependencies
			(r.c0 * xxxxOf(l.c3) + r.c1 * yyyyOf(l.c3)) + (r.c2 * zzzzOf(l.c3) + r.c3)
		};
	}
	Matrix4x4_float_reg SIMD_CALL operator &(
		Matrix4x4_float_reg         l,
		Matrix4x4_float_reg const & r
	) {
		return {
			// parantheses reduce data dependencies
			(r.c0 * xxxxOf(l.c0) + r.c1 * yyyyOf(l.c0)) + (r.c2 * zzzzOf(l.c0) + r.c3 * wwwwOf(l.c0)),
			(r.c0 * xxxxOf(l.c1) + r.c1 * yyyyOf(l.c1)) + (r.c2 * zzzzOf(l.c1) + r.c3 * wwwwOf(l.c1)),
			(r.c0 * xxxxOf(l.c2) + r.c1 * yyyyOf(l.c2)) + (r.c2 * zzzzOf(l.c2) + r.c3 * wwwwOf(l.c2)),
			(r.c0 * xxxxOf(l.c3) + r.c1 * yyyyOf(l.c3)) + (r.c2 * zzzzOf(l.c3) + r.c3 * wwwwOf(l.c3))
		};
	}

	Matrix4x3_float_reg SIMD_CALL matrix4x3From(Matrix4x4_float_reg const m) {
		return { m.c0, m.c1, m.c2, m.c3 };
	}

	Matrix3x3_float_reg SIMD_CALL matrix3x3From(Matrix4x4_float_reg const m) {
		return { m.c0, m.c1, m.c2 };
	}

	// From the given 44 matrix, isolates the rotation and adjusts it to rotate normals properly.
	Matrix3x3_float_reg SIMD_CALL normalTransformationFrom(Matrix4x3_float_reg const m) {
		auto const xLen = ::xxxxOf(sumOfXYZ(m.c0 * m.c0));
		auto const yLen = ::xxxxOf(sumOfXYZ(m.c1 * m.c1));
		auto const zLen = ::xxxxOf(sumOfXYZ(m.c2 * m.c2));
		return {
			m.c0 / xLen,
			m.c1 / yLen,
			m.c2 / zLen
		};
	}



	Quaternion_float_reg SIMD_CALL Quaternion_float::identity() {
		return { loadx4(axisFloats) };
	}

	// Computes a rotation about the given axis and the given angle.
	//   the axis should be normalized
	Quaternion_float_reg SIMD_CALL Quaternion_float::rotationAbout(
		XYZ_float_reg axis,
		float         angle
	) {
		MUST(isFinite(axis));
		MUST(::isFinite(angle));

		// return (sin  x, sin  y, sin  z, cos)
		auto cs00 = XY_float::polar(angleFromRadians(0.5f * angle)).xy__;
		return { xyz0Of(axis.xyz_) * yyyyOf(cs00) + zzzxOf(cs00) };
	}

	// Computes a world-to-eye rotation from two polar angles (first inverse of pitch, then inverse of yaw).
	//  This is the common rotation in first-person view, e.g. for first-person shooters.
	Quaternion_float_reg SIMD_CALL Quaternion_float::ego(
		Angle4B_reg pitch,
		Angle4B_reg yaw
	) {
		//    1) rotate about the Y axis by -leftRight
		//    2) rotate about the X axis by -upDown
		// Consider that for quaternions, ab means first rotating by b, then rotating by a. So, the order has to be flipped,
		//  but the angles remain negated:
		//
		//    cos(x/2)     cos(y/2)
		//    sin(x/2)       0
		//       0         sin(y/2)
		//       0            0
		//
		pitch.unorm = 0 - pitch.unorm / 2;
		  yaw.unorm = 0 -   yaw.unorm / 2;
		auto p_scsc = yxyxOf(XY_float::polar(pitch).xy__);
		auto y_cssc = xyyxOf(XY_float::polar(yaw).xy__);
		return { p_scsc * y_cssc };
	}

	Quaternion_float_reg SIMD_CALL Quaternion_float::yawPitch(XYZ_float_reg xyz) {
		if(no == isValidDirection(xyz)) {
			return identity();
		}

		// TODO:
		auto zx = Geo::zxOf(xyz);
		float horzAngle;
		if(isValidDirection(zx)) {
			horzAngle = float(inRadians(Geo::polarAngleOf(zx)));
		} else {
			horzAngle = 0;
		}
		auto zx_abs    = absoluteOf(zx);
		auto y         = -yOf(xyz);
		auto vertAngle = float(inRadians(Geo::polarAngleOf(Geo::xyFrom(zx_abs, y))));
		return Geo::Quaternion_float::rotationAbout(Geo::XYZ_float::yAxis(), horzAngle)
		     * Geo::Quaternion_float::rotationAbout(Geo::XYZ_float::xAxis(), vertAngle);
	}

	Quaternion_float_reg SIMD_CALL load(Quaternion_float const & mem) {
		return { loadx4(&mem.i) };
	}

	void SIMD_CALL store(Quaternion_float & mem, Quaternion_float_reg r) {
		xyzwTo(&mem.i, r.ijkw);
	}

	bool SIMD_CALL isFinite(Quaternion_float_reg q) {
		return areAllLanesFinite(q.ijkw);
	}

	XYZ_float_reg SIMD_CALL xAxisOf(Quaternion_float_reg q) {
		return { axesOf(q).c0 };
	}
	XYZ_float_reg SIMD_CALL yAxisOf(Quaternion_float_reg q) {
		return { axesOf(q).c1 };
	}
	XYZ_float_reg SIMD_CALL zAxisOf(Quaternion_float_reg q) {
		return { axesOf(q).c2 };
	}

	// Returns the inverse (or the conjugate) of the given quaternion, i.e. the opposite rotation.
	Quaternion_float_reg SIMD_CALL inverted(Quaternion_float_reg q) {
		auto negateW = floatx4SignMask_W();
		return { negateW ^ q.ijkw };
	}

	// Returns the negation of the given quaternion. It represents the same rotation but with different values, which is useful
	//  for compression (see Cryteks Spherical Skinning with Dual-Quaternions and QTangents).
	Quaternion_float_reg SIMD_CALL negated(Quaternion_float_reg q) {
		return { floatx4Zero() - q.ijkw };
	}

	// Returns the product of the two given quaternions.
	Quaternion_float_reg SIMD_CALL operator *(Quaternion_float_reg a, Quaternion_float_reg b) {
		// Original:
		//    a.w  b.w - a.i  b.i - a.j  b.j - a.k  b.k
		//    a.w  b.i + a.i  b.w + a.j  b.k - a.k  b.j
		//    a.w  b.j - a.i  b.k + a.j  b.w + a.k  b.i
		//    a.w  b.k + a.i  b.j - a.j  b.i + a.k  b.w
		// The main problem is the asymmetric mixture of additions and subtractions, requiring either special instructions (in turn
		//  requiring later instruction sets) or complicated sign bit masks. The instructions can be re-ordered so only one sign bit
		//  mask is required, though:
		//    a.w  b.i + a.i  b.w + a.j  b.k - a.k  b.j
		//    a.w  b.j + a.j  b.w + a.k  b.i - a.i  b.k
		//    a.w  b.k + a.i  b.j + a.k  b.w - a.j  b.i
		//    a.w  b.w - a.i  b.i - a.j  b.j - a.k  b.k
		auto a_wwww = wwwwOf(a.ijkw);
		auto a_ijii = xyxxOf(a.ijkw);
		auto a_jkkj = yzzyOf(a.ijkw);
		auto a_kijk = zxyzOf(a.ijkw);

		auto b_wwji = wwyxOf(b.ijkw);
		auto b_kiwj = zxwyOf(b.ijkw);
		auto b_jkik = yzxzOf(b.ijkw);

		auto pppn = xxxyOf(floatx4SignMask_YW());

		auto x =  a_wwww * b.ijkw;
		auto y = (a_ijii * b_wwji) ^ pppn;
		auto z = (a_jkkj * b_kiwj) ^ pppn;
		auto w =  a_kijk * b_jkik;

		return { x + y + z - w };
	}

	// Computes the normalization of the given quaternion. The quaternion must be finite and non-zero.
	Quaternion_float_reg SIMD_CALL normalized(Quaternion_float_reg q) {
		return { q.ijkw / sumOf(q.ijkw * q.ijkw) };
	}

	// Rotates the given vector using the given quaternion.
	XYZ_float_reg SIMD_CALL rotated(XYZ_float_reg v, Quaternion_float_reg q) {
		MUST(isFinite(v));
		MUST(isFinite(q));

		// See Eric Brown, Physics For Games  Quaternion Tricks.

		auto jki_ = yzxwOf(q.ijkw);
		auto kij_ = zxywOf(q.ijkw);
		auto www_ = wwwwOf(q.ijkw);
		auto zxy_ = zxywOf(v.xyz_);
		auto yzx_ = yzxwOf(v.xyz_);

		auto a = jki_ * zxy_ - kij_ * yzx_;
		auto b = www_ * a + jki_ * zxywOf(a) - kij_ * yzxwOf(v.xyz_);

		return { v.xyz_ + ::doubled(b) };
	}

	// Spherical linear interpolation (SLERP) of two quaternions. A factor of 0 returns the first quaternion.
	// The interpolation is constant-velocity and torque-minimal (which is important for physics).
	Quaternion_float_reg SIMD_CALL interpolatedRotations(Quaternion_float_reg v0, float t, Quaternion_float_reg v1) {
		// v0 and v1 should be unit length or else
		// something broken will happen.

		auto cosineOfAngle = ::xOf(sumOf(v0.ijkw * v1.ijkw));

		// If the inputs are too close for comfort, linearly interpolate and normalize the result.
		if(cosineOfAngle > 0.9995f) {
			Quaternion_float_reg result = { v0.ijkw * loadx4(1.0f - t) + v1.ijkw * loadx4(t) };
			return normalized(result);
		}
		//		if(cosineOfAngle < 0.0f) {
		//			v0 = negated(v0);
		//		}

		// Robustness: Stay within domain of acos()
		if(cosineOfAngle < -1.0f) {
			cosineOfAngle = -1.0f;
		}

		auto const originalAngle = Geo::arcusCosineOf(cosineOfAngle); // angle between input vectors
		auto const angle = angleFromRadians(originalAngle * t);  // angle between v0 and result

		Quaternion_float_reg v2 = { v1.ijkw - loadx4(cosineOfAngle) * v0.ijkw };
		v2 = normalized(v2);

		auto const cosSin = XY_float::polar(angle);

		return { v0.ijkw * xxxxOf(cosSin.xy__) + v2.ijkw * yyyyOf(cosSin.xy__) };
	}


	//  counter-clockwise
	//  result not normalized
	XYZ_float_reg SIMD_CALL normalOfTriangle(XYZ_float_reg a, XYZ_float_reg b, XYZ_float_reg c) {
		return crossProductOf(b - a, c - a);
	}


	// Wraps a floating-point number to the range of [0, 2] and rounds it to the next unsigned integer.
	template <typename F> local Angle4B_reg wrappedToAngle(F const rawAngle) {
		//  by rounding to 8-B integers, over- and underruns can be handled up to 2^63 (no performance penality on x64)
		//  then, truncating to 32-bit equals mod 360
		return { UInt4B(0xFFFFFFFF & roundedTo8B(rawAngle)) };
	}

	// Creates an angle from the given value, in degrees.
	//   must be in the range of [0, 359]
	Angle4B_reg angleFromDegrees(SInt4B degrees) {
		MUST(0 <= degrees && degrees <= 359);

		//  replacing angle  11930464.7111 (2 / 360) with angle  11930465 introduces an error of the 100,000th of a
		//   degree, which is tolerable
		return { UInt4B(degrees) * 11930465 };
	}

	// Creates an angle from the given value, in degrees.
	//   must not exceed 773,094,113,280
	Angle4B_reg angleFromDegrees(float degrees) {
		MUST(-773'094'113'280.0f <= degrees && degrees <= 773'094'113'280.0f);

		return wrappedToAngle(degrees * 11930464.711111111111111111111111f); //  2 / 360
	}
	Angle4B_reg angleFromDegrees(double degrees) {
		MUST(-773'094'113'280.0 <= degrees && degrees <= 773'094'113'280.0);

		return wrappedToAngle(degrees * 11930464.711111111111111111111111); //  2 / 360
	}

	// Creates an angle from the given value, in radians.
	//   must not exceed 13,493,037,704rad
	Angle4B_reg angleFromRadians(float radians) {
		MUST(-13'493'037'704.0f <= radians && radians <= 13'493'037'704.0f);

		return wrappedToAngle(radians * 683565275.57643158978229477811035f); //  2 / 2 / pi
	}
	Angle4B_reg angleFromRadians(double radians) {
		MUST(-13'493'037'704.0 <= radians && radians <= 13'493'037'704.0);

		return wrappedToAngle(radians * 683565275.57643158978229477811035); //  2 / 2 / pi
	}

	// Creates an angle from the given astronomical hour angle, in hours.
	//   must not exceed 51,539,607,552 hours
	Angle4B_reg angleFromHourAngle(double hours) {
		MUST(-51'539'607'552.0 <= hours && hours <= 51'539'607'552.0);

		return wrappedToAngle(hours * 178956970.66666666666666666666667); //  2 / 24
	}

	Angle4B_reg operator +(Angle4B_reg a, Angle4B_reg b) {
		return { a.unorm + b.unorm };
	}
	Angle4B_reg & operator +=(Angle4B_reg & a, Angle4B_reg b) {
		a.unorm += b.unorm;
		return a;
	}
	Angle4B_reg operator -(Angle4B_reg a, Angle4B_reg b) {
		return { a.unorm - b.unorm };
	}
	bool operator <(Angle4B_reg a, Angle4B_reg b) {
		return a.unorm < b.unorm;
	}
	bool operator >(Angle4B_reg a, Angle4B_reg b) {
		return a.unorm > b.unorm;
	}

	Angle4B_reg absoluteDifferenceOf(Angle4B_reg a, Angle4B_reg b) {
		return { ::absoluteOf(SInt4B(a.unorm - b.unorm)) };
	}

	// Clamps the given angle to the range of [lower, upper] with
	//   -90 <= lower <=   0 and
	//     0 <= upper <= +90.
	// Useful for camera controls to avoid upside-down views.
	Angle4B_reg clampedToDegrees(
		SIntLeast1B lowerLimitInDegrees,
		Angle4B_reg angle,
		SIntLeast1B upperLimitInDegrees
	) {
		MUST(-90 <= lowerLimitInDegrees && lowerLimitInDegrees <= 0);
		MUST(0 <= upperLimitInDegrees && upperLimitInDegrees <= 90);

		auto const lowerLimit = UInt4B(lowerLimitInDegrees * 11930465); //  2 / 360
		auto const upperLimit = UInt4B(upperLimitInDegrees * 11930465); //  2 / 360
		if(isWithin(0x80000000, angle.unorm, lowerLimit)) {
			angle.unorm = lowerLimit;
		} else if(isWithin(upperLimit, angle.unorm, 0x80000000)) {
			angle.unorm = upperLimit;
		}
		return angle;
	}

	// Rounds the given angle to the next integer, in degrees.
	UIntLeast2B roundedToDegrees(Angle4B_reg angle) {
		// The correct conversion would be
		//    rounded(angle / 2  360)
		//  or
		//    rounded(angle / 11930464.7111)
		// The costly integer-float-integer roundtrip can be saved if the divisor is directly rounded to the nearest integer. In
		//  this case, the maximal error is about the 100,000th of a degree, which is perfectly tolerable.
		return (angle.unorm + (11930465 / 2)) / 11930465;
	}

	// Rounds the given angle to radians, represented as a floating-point number in the range of [-pi, pi).
	double inRadians(Angle4B_reg angle) {
		return double(SInt4B(angle.unorm)) * 1.4629180792671596810513378043098e-9; //  pi / 2
	}

	float tangentOf(Angle4B_reg angle) {
		// Tangent can be divided into four sectors:
		//     0: [  0,  45[
		//     1: [ 45,  90[
		//     2: [ 90, 135[ (inversion of sector 1)
		//     3: [135, 180[ (inversion of sector 0)
		//  At 180, the function repeats.
		// Since the angle is given as an unsigned normalized integer, and each sector is exactly the 8th of the overall range, the
		//  sector can be determined from bits 30 and 29 (mask 0x60000000):
		//     0x00000000: 0
		//     0x20000000: 1
		//     0x40000000: 2
		//     0x60000000: 3
		auto const sectorID = angle.unorm & 0x60000000;

		// Modulate the range to 180 only AFTER the sector is determined. This hopefully breaks any dependency of branching on
		//  modified angle, thus suppressing mispredictions.
		angle.unorm = angle.unorm & 0x7FFFFFFF;

		// Sector 0 is computed via Robin Green, Faster Math Functions, pg. 15:
		//
		//                0.999999986 x - 0.0958010197 x
		//    tan(x) = 
		//            1 - 0.429135022 x + 0.00971659383 xx
		//
		// Sector 1 is computed by a similar term, but with numerator and denominator flipped after assigning
		//
		//    x = pi/2 - x
		//
		//            1 - 0.429135022 x + 0.00971659383 xx
		//    tan(x) = 
		//                0.999999986 x - 0.0958010197 x
		//
		// To save a few cycles, the polynomial is converted to Horner Form (although its full potential requires multiply-add
		//  hardware):
		//
		//              (-0.0958010197 x + 0.999999986) x
		//    tan(x) = 
		//            (-0.00971659383 x - 0.429135022) x + 1
		//
		// When converting the angle to a floating-point number, one must multiply by (2 pi) / 2. This multiplication can be saved
		//  by applying it to the constant factors:
		//
		//              (-2.9993707578799427148695747771533e-28 x + 1.4629180587863065713111022695911e-9) x
		//    tan(x) = 
		//            (-4.4503490744640484968825016740227e-38 x - 9.1840443709068308636681822066321e-19) x + 1
		//
		// Constants are always read from memory (even trivial ones like 1.0)  gather them in the same memory block, aligned to a
		//  cache line boundary so at most one cache miss must be taken.
		alignas(64) static double const coeffs[] = {
			 1.0,                                   //  1.0
			 1.4629180587863065713111022695911e-9,  //  0.999999986    1.4629180792671596810513378043098e-9,
			-9.1840443709068308636681822066321e-19, // -0.429135022    2.1401293066467156958511258973014e-18,
			-2.9993707578799427148695747771533e-28, // -0.0958010197   3.1308338546629715204060346522108e-27,
			 4.4503490744640484968825016740227e-38  //  0.00971659383  4.5801534491681520631005954830806e-36
		};

		// Adjust x for the computation according to the angles sector.
		if(0x00000000 == sectorID) {
			// already suitable
		} else if(0x20000000 == sectorID || 0x40000000 == sectorID) {
			angle.unorm = 0x40000000 - angle.unorm;
		} else { MUST(0x60000000 == sectorID);
			angle.unorm = angle.unorm - 0x80000000;
		}

		auto const x = double(SInt4B(angle.unorm));
		auto const xSquared = x * x;
		auto const termA = (coeffs[3] * xSquared + coeffs[1]) * x;
		auto const termB = (coeffs[4] * xSquared + coeffs[2]) * xSquared + coeffs[0];

		// Perform final division:
		if(0x00000000 == sectorID || 0x60000000 == sectorID) {
			return float(termA / termB);
		} else { MUST(0x20000000 == sectorID || 0x40000000 == sectorID);
			return float(termB / termA);
		}
	}
	Angle4B_reg arcusTangentOf(float originalTangent) {
		// This code is copied from Jack Ganssle, A Guide to Approximations. It should be accurate to about 13.7 decimal digits.
		Angle4B_reg result;

		// 1st step of range reduction: sgn(atan(x)) == sgn(x), so:
		//   store the sign of the input;
		//   compute the absolute value; and
		//   restore the sign after the approximation
		// The input is then in the range of [0, ).
		auto tangent = ::absoluteOf(originalTangent);

		// 2nd step of range reduction: pi/2 - atan(1/x) == atan(x), so reduce the input range to [0, 1] by processing its
		//  reciprocal if necessary.
		// 3rd step of range reduction: reduce to [0, pi/12].
		// 0.57735026918962573 == tan(pi/6).
		double x;
		bool wasInverted;
		bool wasReduced;
		if(1 < tangent) {
			wasInverted = yes;
			if(3.73205090f > tangent) { // 1  tan(pi/12)
				wasReduced = yes;
				x = (1.0 / double(tangent) - 0.57735026918962573) / (1.0 + 0.57735026918962573 / double(tangent)); // reduce to under tan(pi/12)
			} else {
				wasReduced = no;
				x = 1.0 / double(tangent);
			}
		} else {
			wasInverted = no;
			if(0.267949194f < tangent) { // tan(pi/12)
				wasReduced = yes;
				x = (double(tangent) - 0.57735026918962573) / (1.0 + 0.57735026918962573 * double(tangent)); // reduce to under tan(pi/12)
			} else {
				wasReduced = no;
				x = double(tangent);
			}
		}

		auto const x2 = ::squared(x);
		result.unorm = UInt4B(
			(x * (33290360365.527439 + x2 * (33858783399.572529 + x2 * 6429643994.1655121)))
			   / (48.701070044049963 + x2 * (65.766316390895625 + x2 * (21.587934067020264 + x2)))
		);

		if(wasReduced) {
			result.unorm = result.unorm + 357913941u; // + pi/6
		}
		if(wasInverted) {
			result.unorm = 0x40000000 - result.unorm; // pi/2 - x  correct for 1/x if we did that
		}
		if(0 > originalTangent) {
			result.unorm = 0 - result.unorm;
		}

		return result;
	}





	// Returns a meaningful initialization which can be combined with any point or rectangle:
	//   minimum is +INF (is overwritten by any later value)
	//   maximum is -INF (is overwritten by any later value)
	BoundsXYZ_float_reg SIMD_CALL BoundsXYZ_float::default_() {
		return {
			floatx4Infinity(),
			floatx4Zero() - floatx4Infinity()
		};
	}

	BoundsXYZ_float_reg SIMD_CALL BoundsXYZ_float::fromMinMax(XYZ_float_reg minimum, XYZ_float_reg maximum) {
		MUST(xyzLowerOrEqual(minimum.xyz_, maximum.xyz_));

		return {
			minimum.xyz_,
			maximum.xyz_
		};
	}

	BoundsXYZ_float_reg SIMD_CALL BoundsXYZ_float::fromMinimumAndSidelength(XYZ_float_reg minimum, float sidelength) {
		MUST(sidelength >= 0.0f);

		return {
			minimum.xyz_,
			minimum.xyz_ + loadx4(sidelength)
		};
	}

	BoundsXYZ_float_reg SIMD_CALL load(BoundsXYZ_float const & m) {
		return { ::xyzFrom(&m.minimum.x), ::xyzFrom(&m.maximum.x) };
	}

	void SIMD_CALL store(BoundsXYZ_float & m, BoundsXYZ_float_reg b) {
		xyzwTo(&m.minimum.x, b.minimum);
		xyzTo(&m.maximum.x, b.maximum);
	}

	// Computes the bounding box of the three given points.
	BoundsXYZ_float_reg SIMD_CALL boundingHullOf(XYZ_float_reg a, XYZ_float_reg b, XYZ_float_reg c) {
		return {
			::minimumOf(a.xyz_, ::minimumOf(b.xyz_, c.xyz_)),
			::maximumOf(a.xyz_, ::maximumOf(b.xyz_, c.xyz_))
		};
	}

	// Computes the total bounding box containing points of any of the two given boxes.
	BoundsXYZ_float_reg SIMD_CALL unionOf(BoundsXYZ_float_reg a, BoundsXYZ_float_reg b) {
		return {
			::minimumOf(a.minimum, b.minimum),
			::maximumOf(a.maximum, b.maximum),
		};
	}
	BoundsXYZ_float_reg SIMD_CALL unionOf(BoundsXYZ_float_reg a, XYZ_float_reg b) {
		return {
			::minimumOf(a.minimum, b.xyz_),
			::maximumOf(a.maximum, b.xyz_),
		};
	}
	BoundsXYZ_float_reg SIMD_CALL unionOf(BoundsXYZ_float_reg a, XYZ_float_reg b, XYZ_float_reg c, XYZ_float_reg d) {
		return {
			::minimumOf(::minimumOf(a.minimum, b.xyz_, c.xyz_), d.xyz_),
			::maximumOf(::maximumOf(a.maximum, b.xyz_, c.xyz_), d.xyz_),
		};
	}

	// Returns the bounding box of the volume where the two boxes overlap.
	//   if the boxes are not intersecting, the result is degenerated
	BoundsXYZ_float_reg SIMD_CALL intersectionOf(BoundsXYZ_float_reg a, BoundsXYZ_float_reg b) {
		return {
			::maximumOf(a.minimum, b.minimum),
			::minimumOf(a.maximum, b.maximum),
		};
	}

	XYZ_float_reg SIMD_CALL minimumOf(BoundsXYZ_float_reg box) {
		return xyzFrom(box.minimum);
	}
	XYZ_float_reg SIMD_CALL maximumOf(BoundsXYZ_float_reg box) {
		return xyzFrom(box.maximum);
	}
	XYZ_float_reg SIMD_CALL centerOf(BoundsXYZ_float_reg box) {
		return 0.5f * xyzFrom(box.minimum + box.maximum);
	}
	XYZ_float_reg SIMD_CALL sidelengthOf(BoundsXYZ_float_reg r) {
		return { r.maximum - r.minimum };
	}

	bool SIMD_CALL areIntersecting(BoundsXYZ_float_reg a, BoundsXYZ_float_reg b) {
		return no == (xyzBelow(a.maximum, b.minimum) || xyzBelow(b.maximum, a.minimum));
	}

	// Determines whether any dimension of the given bounding box is zero or inverted.
	bool SIMD_CALL isDegenerated(BoundsXYZ_float_reg const box) {
		// Check if all lanes of the minimum are below the maximum. If not, one lane is degenerated.
		return no == xyzBelow(box.minimum, box.maximum);
	}

	// Extends the given bounding box to the nearest power-of-two integer boundaries.
	BoundsXYZ_SInt4B_reg SIMD_CALL rounded(
		BoundsXYZ_float_reg const box,
		SIntLeast1B const         powerOfTwo
	) {
		auto const rrrr = xxxxFrom(powerOfTwo - 1);
		return {
			andnot(::rounded(box.minimum),        rrrr),
			andnot(::rounded(box.maximum) + rrrr, rrrr)
		};
	}

	// Displaces the given bounding box.
	void SIMD_CALL operator -=(BoundsXYZ_float_reg & box, XYZ_float_reg const offset) {
		box.minimum = box.minimum - offset.xyz_;
		box.maximum = box.maximum - offset.xyz_;
	}




	XYZ_SInt4B_reg SIMD_CALL minimumOf(BoundsXYZ_SInt4B_reg const box) {
		return { box.minimum };
	}
	XYZ_SInt4B_reg SIMD_CALL maximumOf(BoundsXYZ_SInt4B_reg const box) {
		return { box.maximum };
	}

	// Determines whether any dimension of the given bounding box is zero or inverted.
	bool SIMD_CALL isDegenerated(BoundsXYZ_SInt4B_reg const box) {
		// Check if all lanes of the minimum are below the maximum. If not, one lane is degenerated.
		return no == xyzBelow(box.minimum, box.maximum);
	}




	BoundsXYZ_SInt4B_reg SIMD_CALL BoundsXYZ_SInt4B::fromMinMax(XYZ_SInt4B_reg minimum, XYZ_SInt4B_reg maximum) {
		MUST(xyzLowerOrEqual(minimum.xyz_, maximum.xyz_));

		return {
			minimum.xyz_,
			maximum.xyz_
		};
	}






	// Returns a meaningful initialization which can be combined with any point or rectangle:
	//   minimum is +INF (is overwritten by any later value)
	//   maximum is -INF (is overwritten by any later value)
	BoundsXY_float_reg SIMD_CALL BoundsXY_float::default_() {
		return {
			floatx4Infinity(),
			floatx4Zero() - floatx4Infinity()
		};
	}


	BoundsXY_SInt4B_reg SIMD_CALL boundingRectFrom(SInt4B minX, SInt4B minY, SInt4B maxX, SInt4B maxY) {
		return { ::xyzwFrom(minX, minY, 0, 0), ::xyzwFrom(maxX, maxY, 0, 0) };
	}

	BoundsXY_SInt4B_reg SIMD_CALL load(BoundsXY_UInt2B const & m) {
		auto minXYmaxXY = ::xyzwFrom(&m.minimumX);
		return { minXYmaxXY, zw__Of(minXYmaxXY) };
	}

	void SIMD_CALL store(BoundsXY_UInt2B & m, BoundsXY_SInt4B_reg b) {
		xyzwTo(&m.minimumX, xyxyFrom(b.minimum, b.maximum));
	}

	BoundsXY_SInt4B_reg SIMD_CALL unionOf(BoundsXY_SInt4B_reg a, BoundsXY_SInt4B_reg b) {
		return { ::minimumOf(a.minimum, b.minimum), ::maximumOf(a.maximum, b.maximum) };
	}

	bool SIMD_CALL areIntersecting(BoundsXY_SInt4B_reg a, BoundsXY_SInt4B_reg b) {
		return !(xyBelow(a.maximum, b.minimum) || xyBelow(b.maximum, a.minimum));
	}




	// Computes the bounding rectangle of the three given points.
	BoundsXY_float_reg SIMD_CALL boundingHullOf(XY_float_reg a, XY_float_reg b, XY_float_reg c) {
		return {
			::minimumOf(a.xy__, ::minimumOf(b.xy__, c.xy__)),
			::maximumOf(a.xy__, ::maximumOf(b.xy__, c.xy__))
		};
	}

	// Computes the total bounding rectangle containing points of any of the two given rectangles.
	BoundsXY_float_reg SIMD_CALL unionOf(BoundsXY_float_reg a, BoundsXY_float_reg b) {
		return {
			::minimumOf(a.minimum, b.minimum),
			::maximumOf(a.maximum, b.maximum),
		};
	}

	XY_float_reg SIMD_CALL sidelengthOf(BoundsXY_float_reg r) {
		return { r.maximum - r.minimum };
	}


	void SIMD_CALL store(PlaneXYZ_float & m, PlaneXYZ_float_reg p) {
		xyzwTo(&m.normal.x, p.normalXYZ_distance);
	}

	PlaneXYZ_float_reg SIMD_CALL load(PlaneXYZ_float const & p) {
		return { loadx4(&p.normal.x) };
	}

	// Creates a plane from a point and a normal. The resulting plane is normalized only if the given normal is normalized.
	PlaneXYZ_float_reg SIMD_CALL planeFromPointAndNormal(
		XYZ_float_reg anyPointOnPlane,
		XYZ_float_reg itsNormal
	) {
		return { ::xyzwFrom(
			::xOf(itsNormal.xyz_),
			::yOf(itsNormal.xyz_),
			::zOf(itsNormal.xyz_),
			::xOf(sumOfXYZ(anyPointOnPlane.xyz_ * itsNormal.xyz_))
		) };
	}

	float SIMD_CALL signedDistanceBetween(
		XYZ_float_reg      point,
		PlaneXYZ_float_reg plane
	) {
		return ::xOf(sumOfXYZ(point.xyz_ * plane.normalXYZ_distance) - wwwwOf(plane.normalXYZ_distance));
	}

	// Computes the distance where the given ray intersects the given plane (in multiples of the given rays direction), or NaN if
	//  the ray does not intersect the plane.
	float SIMD_CALL intersectionDistanceOf(
		RayXYZ_float_reg   ray,
		PlaneXYZ_float_reg plane
	) {
		// TODO: simplify for SIMD
		XYZ_float_reg const origin    = { ray.originXYZ_  };
		XYZ_float_reg const direction = { ray.directionXYZ_ };
		XYZ_float_reg const normal    = { plane.normalXYZ_distance };
		auto const lineResult = signedDistanceBetween(origin, plane) / (0.0f - dotProductOf(direction, normal));
		if(0.000001f > lineResult) {
			return NAN_FLOAT4B;
		}
		return lineResult;
	}

	// Computes the intersection point of the given ray and the given plane. NaN if they dont intersect.
	XYZ_float_reg SIMD_CALL intersectionPointOf(
		RayXYZ_float_reg   ray,
		PlaneXYZ_float_reg plane
	) {
		return { ray.originXYZ_ + ray.directionXYZ_ * loadx4(intersectionDistanceOf(ray, plane)) };
	}



	LineXYZ_float_reg SIMD_CALL lineFrom(RayXYZ_float_reg ray) {
		return { ray.originXYZ_, ray.directionXYZ_ };
	}

	//  the point must be finite coordinates
	//  the direction must be finite and non-null
	LineXYZ_float_reg SIMD_CALL lineFromPointAndDirection(
		XYZ_float_reg anyPoint,
		XYZ_float_reg direction
	) {
		MUST(isFinite(anyPoint));
		MUST(isValidDirection(direction));

		return { anyPoint.xyz_, direction.xyz_ };
	}

	//  the points must be finite coordinates
	//  the distance must be sufficient to compute a direction
	LineXYZ_float_reg SIMD_CALL lineFromTwoPoints(XYZ_float_reg a, XYZ_float_reg b) {
		MUST(isFinite(a));
		MUST(isFinite(b));
		MUST(isValidDirection(b - a));

		return lineFromPointAndDirection(a, b - a);
	}

	// Normalizes the given line; i.e.:
	//   the arbitrary point becomes the point closest to the origin (this guarantees highest accuracy in floating-point
	//    computation)
	//   the direction becomes normalized
	LineXYZ_float_reg SIMD_CALL normalized(LineXYZ_float_reg line) {
		XYZ_float_reg const anyPoint  = { line.anyXYZ_ };
		XYZ_float_reg const direction = { line.directionXYZ_ };

		MUST(isFinite(anyPoint));
		MUST(isValidDirection(direction));

		auto const originsSquaredProjectedDistance = dotProductOf(anyPoint, direction);
		return {
			(anyPoint - direction * (originsSquaredProjectedDistance / squaredAbsoluteOf(direction))).xyz_,
			normalized(direction).xyz_
		};
	}

	// Computes the point on the line that is closest to the given point.
	XYZ_float_reg SIMD_CALL closestPoint(XYZ_float_reg reference, LineXYZ_float_reg line) {
		XYZ_float_reg const anyPoint  = { line.anyXYZ_ };
		XYZ_float_reg const direction = { line.directionXYZ_ };

		MUST(isFinite(reference));
		MUST(isFinite(anyPoint));
		MUST(isValidDirection(direction));

		auto const squaredProjectedDistance = dotProductOf(reference - anyPoint, direction);
		return anyPoint + (squaredProjectedDistance / squaredAbsoluteOf(direction)) * direction;
	}

	// Computes the distance where the two given lines cross. Returns
	//   zero if they overlap;
	//   infinity if they are parallel but displaced; and
	//   the distance in multiples of the first lines length otherwise.
	float SIMD_CALL intersectionDistanceOf(LineXYZ_float_reg a, LineXYZ_float_reg b) {
		// TODO: simplify for SIMD

		auto const sqaredEpsilon = 8.131516e-20f * 8.131516e-20f;//1.11875110968003096e-154;

		XYZ_float directionOfA; xyzTo(&directionOfA.x, a.directionXYZ_);
		XYZ_float directionOfB; xyzTo(&directionOfB.x, b.directionXYZ_);
		XYZ_float aToB; xyzTo(&aToB.x, b.anyXYZ_ - a.anyXYZ_);

		auto const kross = directionOfA.x * directionOfB.y - directionOfA.y * directionOfB.x;
		auto const squaredLengthOfA = squaredAbsoluteOf(load(directionOfA));
		auto const squaredLengthOfB = squaredAbsoluteOf(load(directionOfB));
		if(::squared(kross) > sqaredEpsilon * squaredLengthOfA * squaredLengthOfB) { // Not parallel?
			return (aToB.x * directionOfB.y - aToB.y * directionOfB.x) / kross;
		}

		// Parallel.
		if(::squared(aToB.x * directionOfA.y - aToB.y * directionOfA.x) > sqaredEpsilon * squaredLengthOfA * squaredAbsoluteOf(load(aToB))) {
			return INF_FLOAT4B; // not overlapping
		}

		return 0; // parallel and overlapping
	}

	// Checks the given line and the given triangle (in any winding order) for intersection. Returns
	//   the projected distance of the intersection point from line.anyPoint in the lines direction; or
	//   NaN if not intersecting.
	float SIMD_CALL intersectionDistanceOf(
		LineXYZ_float_reg line,
		XYZ_float_reg     a,
		XYZ_float_reg     aToB,
		XYZ_float_reg     aToC
	) {
		XYZ_float_reg const anyPoint = { line.anyXYZ_ };
		XYZ_float_reg const direction = { line.directionXYZ_ };

		// From Geometric Tools for Computer Graphics (Philip J. Schneider & David H. Eberly, 2003), pg. 485.
		float const epsilon = 0.00001f;

		auto const p = crossProductOf(direction, aToC);
		float tmp = dotProductOf(p, aToB);
		if(tmp > -epsilon && tmp < epsilon) { // Parallel?
			return NAN_FLOAT4B;
		}
		tmp = 1 / tmp;
		auto const s = anyPoint - a;
		float u = tmp * dotProductOf(s, p);
		if(u < 0 || u > 1) {
			return NAN_FLOAT4B;
		}
		auto const q = crossProductOf(s, aToB);
		float v = tmp * dotProductOf(direction, q);
		if(v < 0 || v > 1) {
			return NAN_FLOAT4B;
		}

		if(u + v > 1) {
			return NAN_FLOAT4B;
		}

		return tmp * dotProductOf(aToC, q);
	}


	RayXYZ_float_reg SIMD_CALL load(RayXYZ_float const & m) {
		return { loadx4(&m.origin.x), ::xyzFrom(&m.direction.x) };
	}

	XYZ_float_reg SIMD_CALL originOf(RayXYZ_float_reg ray) {
		return { ray.originXYZ_ };
	}

	// Not necessarily normalized!
	XYZ_float_reg SIMD_CALL directionOf(RayXYZ_float_reg ray) {
		return { ray.directionXYZ_ };
	}

	RayXYZ_float_reg SIMD_CALL rayFromOriginAndDirection(
		XYZ_float_reg itsOrigin,
		XYZ_float_reg itsDirection
	) {
		return { itsOrigin.xyz_, itsDirection.xyz_ };
	}

	// Determines whether the given ray intersects the given triangle (in any winding order).
	bool SIMD_CALL areIntersecting(
		RayXYZ_float_reg ray,
		XYZ_float_reg    a,
		XYZ_float_reg    aToB,
		XYZ_float_reg    aToC
	) {
		// NaN if not hit  any comparison to NaN returns no.
		return 0.0f < intersectionDistanceOf(lineFrom(ray), a, aToB, aToC);
	}


	LineSegmentXY_float_reg SIMD_CALL lineSegmentBetween(XY_float_reg a, XY_float_reg b) {
		return { a.xy__, b.xy__ };
	}

	// Computes the point on the given line segment that is closest to the given point.
	XY_float_reg SIMD_CALL closestPoint(
		XY_float_reg            p,
		LineSegmentXY_float_reg line
	) {
		// Along the line segment, the closest point is always located perpendicular to the segment. Project the point.
		auto aToB              = line.b - line.a;
		auto squaredLength     = sumOfXY(aToB * aToB);
		auto aToP              = p.xy__ - line.a;
		auto squaredProjection = sumOfXY(aToB * aToP);

		// If the projected length exceeds the line segments length, b is the closest point. Do NOT use min tricks here because
		//  the bit-accurate point may be required:
		if(::xOf(squaredProjection) >= ::xOf(squaredLength)) {
			return { line.b };
		}

		// If the projected length is negative, a is the closest point. Use max instead of branching:
		auto length = xxxxOf(::maximumOf(floatx4Zero(), squaredProjection / squaredLength));
		return { line.a + length * aToB };
	}

	// Computes the squared minimal distance between the given point and the given line segment.
	float SIMD_CALL squaredDistanceBetween(
		XY_float_reg            p,
		LineSegmentXY_float_reg line
	) {
		return squaredDistanceBetween(p, closestPoint(p, line));
	}

	// Computes the minimal distance between the given point and the given line segment.
	float SIMD_CALL distanceBetween(
		XY_float_reg            p,
		LineSegmentXY_float_reg line
	) {
		return distanceBetween(p, closestPoint(p, line));
	}


} // Geo



namespace Physics {


	Geo::XYZ_float_reg SIMD_CALL computeAngularVelocity(
		Geo::XYZ_float_reg localAngularMomentum, // Nms or kgm/s
		Geo::XYZ_float_reg inertia               // kgm
	) {
		return localAngularMomentum / inertia;
	}

	// Computes the moment of inertia of a solid cuboid of the given dimensions and mass.
	Geo::XYZ_float_reg SIMD_CALL inertiaOfCuboid(
		Geo::XYZ_float_reg dimensions, // width/height/depth in m
		float              mass        // kg
	) {
		// According to http://en.wikipedia.org/wiki/List_of_moments_of_inertia, just with axes swapped.
		auto xyx_ = yxyxOf(dimensions.xyz_);
		auto zzx_ = zzxxOf(dimensions.xyz_);
		return { loadx4(1.0f / 12.0f * mass) * (xyx_ * xyx_ + zzx_ * zzx_) };
	}


} // Physics