UInt2B squared(SInt1B);
UInt4B squared(UInt4B);
UInt8B squared(SInt4B);
float squared(float);
double squared(double);

// Computes the squared value but copies the original sign to the result.
float signedSquareOf(float);

// Computes the square root of the given value. For negative numbers, this is NaN.
float squareRootOf(float);
double squareRootOf(double);

// Interpolates linearly between the given scalars according to the given factor, where 0 yields a and 1 yields b.
float interpolateLinearly(
	float a,
	float factor,
	float b
);
floatx4 SIMD_CALL interpolated_linear(
	floatx4 a,
	floatx4 factor,
	floatx4 b
);
double interpolateLinearly(
	double a,
	double factor,
	double b
);


namespace Geo {
	// SIMD-optimized vector and matrix functions. There are two data layouts for each object:
	//    memory layout, optimized for size-efficient storage
	//    register layout (_reg postfix), optimized for SIMD processing
	// load() and store() convert one layout to another. There is no point in using only one for both storage and processing:
	//    memory layout requires lane swizzling on each operation, up to the point where it eats up any gain in ALU power
	//    register layout requires too much space, worsening the Von Neumann bottleneck


	// A vector of two IEEE754 floating-point numbers in a register.
	struct XY_float_reg {
		floatx4 xy__; // 3rd & 4th lane undefined

		XY_float_reg SIMD_CALL operator -() const;
	};


	// A vector of two signed 4-B integers in a register.
	struct XY_SInt4B_reg {
		SInt4Bx4 xy__; // 3rd & 4th lane undefined

		XY_SInt4B_reg SIMD_CALL operator -() const;
	};


	// A vector of three IEEE754 floating-point numbers in a register.
	struct XYZ_float_reg {
		floatx4 xyz_; // 4th lane undefined

		XYZ_float_reg SIMD_CALL operator -() const;
	};


	// A vector of three signed 4-B integers in a register.
	struct XYZ_SInt4B_reg {
		SInt4Bx4 xyz_; // 4th lane undefined

		XYZ_SInt4B_reg SIMD_CALL operator -() const;
	};


	// Two columns and two rows of IEEE floating-point numbers in a register. Handy for 2D rotation.
	struct Matrix2x2_float_reg {
		floatx4 c01;
	};


	// Three columns and two rows of IEEE floating-point numbers in three registers. Handy for 2D rotation and translation.
	struct Matrix3x2_float_reg {
		floatx4 c0, c1, c2; // 3rd & 4th lane undefined
	};


	// Three columns and three rows of IEEE floating-point numbers in three registers. Handy for 3D rotation.
	struct Matrix3x3_float_reg {
		floatx4 c0, c1, c2; // 4th lane undefined
	};


	// Four columns and three rows of IEEE floating-point numbers in four registers. Handy for 3D rotation and translation.
	struct Matrix4x3_float_reg {
		floatx4 c0, c1, c2, c3; // 4th lane undefined
	};


	// Four columns and four rows of IEEE floating-point numbers in four registers. Handy for 3D projection.
	struct Matrix4x4_float_reg {
		floatx4 c0, c1, c2, c3;
	};


	// A quaternion of IEEE floating-point numbers in a register. Handy for 3D rotation.
	struct Quaternion_float_reg {
		floatx4 ijkw;
	};


	typedef struct Angle4B {
		// Floating-point numbers for angles have disadvantages:
		//   non-uniform accuracy: for binary32, the accuracy at 0 is about one 10^70th of a degree, but at 180 degrees its just
		//    a 100,000th  adding up angles eventually leads to absorption
		//   excess range: most transcendental functions spend lots of time wrapping angles to a range of [0, 2pi)
		// Therefore, angles are stored as four-byte unsigned integers, where 0 represents 0 and 2 - 1 represents 360 - 1  2.
		//  The advantages are:
		//   uniform accuracy (4+4B address any square centimeter on Earth via latitude and longitude)
		//   qutomatic wraparounds on addition (C++ guarantees 2s complement wraparounds for unsigned integers)
		//   quick detection of the quadrant/octant/etc. by checking the most significant bits (speeds up transcendental functions)
		UInt4B unorm;
	} Angle4B_reg;


	// A bounding rectangle of IEEE floating-point numbers in two registers.
	struct BoundsXY_float_reg {
		// No sense in packing one register because min/max are hard to mix!
		floatx4 minimum;
		floatx4 maximum;
	};


	// A bounding rectangle of signed 4-B integers in two registers.
	struct BoundsXY_SInt4B_reg {
		// No sense in packing one register because min/max are hard to mix!
		SInt4Bx4 minimum;
		SInt4Bx4 maximum;
	};


	// A bounding box of IEEE floating-point numbers in two registers.
	struct BoundsXYZ_float_reg {
		floatx4 minimum;
		floatx4 maximum;
	};


	// A bounding box of signed 4-B integers in two registers.
	struct BoundsXYZ_SInt4B_reg {
		SInt4Bx4 minimum;
		SInt4Bx4 maximum;
	};


	// A plane (normal-distance representation) of IEEE floating-point numbers in a register.
	struct PlaneXYZ_float_reg {
		floatx4 normalXYZ_distance;
	};


	// A 3D line (infinite) of IEEE floating-point numbers in two registers.
	struct LineXYZ_float_reg {
		floatx4 anyXYZ_;
		floatx4 directionXYZ_; // Not necessarily normalized!
	};


	// A 3D ray of IEEE floating-point numbers in two registers.
	struct RayXYZ_float_reg {
		floatx4 originXYZ_;
		floatx4 directionXYZ_; // Not necessarily normalized!
	};


	// A 2D line segment of IEEE floating-point numbers in two registers.
	struct LineSegmentXY_float_reg {
		floatx4 a;
		floatx4 b;
	};


	// A rectangular 3D frustum, fixed to the Z axis, in one register. Handy for culling in eye space.
	struct FixedFrustum3D_reg {
		//  The planes are symmetrically, so only the right and upper plane need to be stored.
		//  Eache planes distance from the origin is zero, so it can be omitted.
		//  The upper planes X is always zero, so it can be omitted.
		//  The right planes Y is always zero, so it can be omitted.
		//  Store in XZYZ order so SIMD can be used effectively.
		floatx4 xzyz;
	};


	// A vector of two IEEE floating-point numbers in memory.
	struct XY_float {
		float x, y;

		static XY_float_reg SIMD_CALL zero();
		static XY_float_reg SIMD_CALL xAxis();
		static XY_float_reg SIMD_CALL yAxis();

		// Returns the X axis, rotated by the given angle (i.e. cosine and sine). Z and W are zero.
		static XY_float_reg SIMD_CALL polar(Angle4B_reg);
	};


	// A vector of two signed 1-B integers in memory.
	struct XY_SInt1B {
		SInt1B x, y;
	};


	// A vector of two signed 2-B integers in memory.
	struct XY_SInt2B {
		SInt2B x, y;
	};


	// A vector of two signed 4-B integers in memory.
	struct XY_SInt4B {
		SInt4B x, y;

		static XY_SInt4B_reg SIMD_CALL zero();
	};


	// A vector of three IEEE floating-point numbers in memory.
	struct XYZ_float {
		float x, y, z;

		static XYZ_float_reg SIMD_CALL zero();
		static XYZ_float_reg SIMD_CALL xAxis();
		static XYZ_float_reg SIMD_CALL yAxis();
		static XYZ_float_reg SIMD_CALL zAxis();

		// Generates a normalized vector to the given polar angle.
		//   0/0 is Z axis
		static XYZ_float_reg SIMD_CALL polar(
			Angle4B_reg horizontally,
			Angle4B_reg vertically
		);

	};


	// A vector of three IEEE floating-point numbers in memory.
	struct XYZ_double {
		double x, y, z;
	};


	// A vector of three signed 2-B integers in memory.
	struct XYZ_SInt2B {
		SInt2B x, y, z;
	};


	// A vector of three signed 4-B integers in memory.
	struct XYZ_SInt4B {
		SInt4B x, y, z;
	};


	// A vector of four IEEE floating-point numbers in memory.
	struct XYZW_float {
		float x, y, z, w;
	};


	// Two columns and two rows of IEEE floating-point numbers, in memory. Handy for 2D rotation.
	struct Matrix2x2_float {
		XY_float columns[2];

		static Matrix2x2_float_reg SIMD_CALL uniformScaling(float);
		static Matrix2x2_float_reg SIMD_CALL rotation(Angle4B_reg);
	};


	// Three columns and two rows of IEEE floating-point numbers, in memory. Handy for 2D rotation and translation.
	struct Matrix3x2_float {
		XY_float columns[3];

		static Matrix3x2_float_reg SIMD_CALL identity();

		static Matrix3x2_float_reg SIMD_CALL translation(XY_float_reg);
	};


	// Three columns and three rows of IEEE floating-point numbers, in memory. Handy for 3D rotation.
	struct Matrix3x3_float {
		XYZ_float columns[3];

		static Matrix3x3_float_reg SIMD_CALL identity();

		static Matrix3x3_float_reg SIMD_CALL uniformScaling(float);
		static Matrix3x3_float_reg SIMD_CALL scaling(float, float, float);
		static Matrix3x3_float_reg SIMD_CALL scaling(XYZ_float_reg);

		static Matrix3x3_float_reg SIMD_CALL rotationAboutX(Angle4B_reg);
		static Matrix3x3_float_reg SIMD_CALL rotationAboutY(Angle4B_reg);
		static Matrix3x3_float_reg SIMD_CALL rotationAboutZ(Angle4B_reg);
		static Matrix3x3_float_reg SIMD_CALL rotationAbout(XYZ_float_reg, Angle4B_reg);

		static Matrix3x3_float_reg SIMD_CALL rollPitchYaw(
			Angle4B_reg roll,
			Angle4B_reg pitch,
			Angle4B_reg yaw
		);

		// Computes an eye-to-world matrix from two polar angles.
		//   0, 0 is along Z
		//   positive pitch looks towards Y
		//   positive yaw looks towards X
		static Matrix3x3_float_reg SIMD_CALL polar(
			Angle4B_reg pitch,
			Angle4B_reg yaw
		);

		// Computes a world-to-eye matrix from two polar angles (first inverse of pitch, then inverse of yaw).
		//  This is the common rotation in first-person view, e.g. for first-person shooters.
		static Matrix3x3_float_reg SIMD_CALL ego(
			Angle4B_reg pitch,
			Angle4B_reg yaw
		);

		// Creates a right-handed coordinate system with the given scale around the given Z axis.
		static Matrix3x3_float_reg SIMD_CALL fromZ(XYZ_float_reg z, float scaleFactor);

		// Creates a right-handed coordinate system with the given scale around the given Y axis.
		static Matrix3x3_float_reg SIMD_CALL fromY(XYZ_float_reg y, float scaleFactor);

	};


	// A matrix of four columns with three IEEE754 floating-point numbers in memory. Handy for 3D rotation, scaling, shearing,
	//  and translation.
	struct Matrix4x3_float {
		XYZ_float columns[4];

		static Matrix4x3_float_reg SIMD_CALL identity();

		static Matrix4x3_float_reg SIMD_CALL rollPitchYaw(
			Angle4B_reg   roll,
			Angle4B_reg   pitch,
			Angle4B_reg   yaw,
			XYZ_float_reg offset
		);

		// Computes an eye-to-world matrix with offset from two polar angles.
		//   eye offset is rotated
		//   0, 0 is along Z
		//   positive pitch looks towards Y
		//   positive yaw looks towards X
		static Matrix4x3_float_reg SIMD_CALL polar(
			XYZ_float_reg offset,
			Angle4B_reg   vertical,
			Angle4B_reg   horizontal
		);

	};


	// A matrix of four columns with four IEEE754 floating-point numbers in memory. Handy for 3D projection.
	struct Matrix4x4_float {
		XYZW_float columns[4];

		static Matrix4x4_float_reg SIMD_CALL identity();

		static Matrix4x4_float_reg SIMD_CALL projection(
			SIntLeast2B screenWidthInAnyUnit,
			SIntLeast2B screenHeightInSameUnit,
			float       zoomScale,
			float       near,
			float       far
		);

		static Matrix4x4_float_reg SIMD_CALL projection_legacy(
			SIntLeast2B screenWidthInAnyUnit,
			SIntLeast2B screenHeightInSameUnit,
			float       zoomScale,
			float       near,
			float       far
		);

		//  transforms right-handed coordinates to left-handed screen space
		//  uses progressive Z, i.e. the near clipping plane has a depth of 1 and the far clipping plane has a depth of 0  this
		//   improves depth buffer precision greatly, as the non-linearity of Z is canceled out with the non-linearity of floating-
		//   -point values in the depth buffer (see Humus, A couple of notes about Z and Outerra, Maximizing Depth Buffer Range
		//    and Precision)
		static Matrix4x4_float_reg SIMD_CALL projectionProgressive(
			SIntLeast2B screenWidthInAnyUnit,
			SIntLeast2B screenHeightInSameUnit,
			float       zoomScale,
			float       near,
			float       far
		);

	};


	// A quaternion of IEEE floating-point numbers, in memory. Handy for 3D rotation.
	struct Quaternion_float {
		float i, j, k, w;

		static Quaternion_float_reg SIMD_CALL identity();

		// Computes a rotation about the given axis and the given angle.
		//   the axis should be normalized
		static Quaternion_float_reg SIMD_CALL rotationAbout(
			XYZ_float_reg axis,
			float         angle
		);

		// Computes a world-to-eye rotation from two polar angles (first inverse of pitch, then inverse of yaw).
		//  This is the common rotation in first-person view, e.g. for first-person shooters.
		static Quaternion_float_reg SIMD_CALL ego(
			Angle4B_reg pitch,
			Angle4B_reg yaw
		);

		static Quaternion_float_reg SIMD_CALL yawPitch(XYZ_float_reg);

	};


	// A bounding rectangle.
	struct BoundsXY_float {
		XY_float minimum;
		XY_float maximum;

		// Returns a meaningful initialization which can be combined with any point or rectangle:
		//   minimum is +INF (is overwritten by any later value)
		//   maximum is -INF (is overwritten by any later value)
		static BoundsXY_float_reg SIMD_CALL default_();

	};


	struct BoundsXY_UInt2B {
		UInt2B minimumX;
		UInt2B minimumY;
		UInt2B maximumX;
		UInt2B maximumY;
	};


	// An axis-aligned bounding box.
	struct BoundsXYZ_float {
		XYZ_float minimum;
		XYZ_float maximum;

		// Returns a meaningful initialization which can be combined with any point or rectangle:
		//   minimum is +INF (is overwritten by any later value)
		//   maximum is -INF (is overwritten by any later value)
		static BoundsXYZ_float_reg SIMD_CALL default_();

		static BoundsXYZ_float_reg SIMD_CALL fromMinMax(XYZ_float_reg, XYZ_float_reg);

		static BoundsXYZ_float_reg SIMD_CALL fromMinimumAndSidelength(XYZ_float_reg, float);

	};


	// An axis-aligned bounding box.
	struct BoundsXYZ_SInt4B {
		XYZ_SInt4B minimum;
		XYZ_SInt4B maximum;

		static BoundsXYZ_SInt4B_reg SIMD_CALL fromMinMax(XYZ_SInt4B_reg, XYZ_SInt4B_reg);

	};


	struct PlaneXYZ_float {
		XYZ_float normal;
		float     distance;
	};


	struct LineXY_SInt4B {
		XY_SInt4B anyPoint;
		XY_SInt4B direction;
	};


	struct RayXYZ_float {
		XYZ_float origin;
		XYZ_float direction;
	};


	struct LineSegmentXY_float {
		Geo::XY_float a;
		Geo::XY_float b;
	};


	enum FrustumClassification {
		invisible     = -1,
		partlyVisible =  0,
		visible       =  1
	};

	FixedFrustum3D_reg SIMD_CALL frustumFrom(
		float tangentOfHalfFieldOfView,
		float widthInArbitraryUnit,
		float heightInSameUnit
	);

	FrustumClassification SIMD_CALL classifySphere(
		FixedFrustum3D_reg const & ,
		XYZ_float_reg              center,
		float                      radius
	);


	XY_float_reg SIMD_CALL load(XY_float const &);
	XY_float_reg SIMD_CALL xyFrom(floatx4);
	XY_float_reg SIMD_CALL xyFrom(float, float);
	void SIMD_CALL store(XY_float &, XY_float_reg);
	XY_float SIMD_CALL stored(XY_float_reg);

	XY_float_reg SIMD_CALL xyOf(XYZ_float_reg);
	XY_float_reg SIMD_CALL xzOf(XYZ_float_reg);
	XY_float_reg SIMD_CALL zxOf(XYZ_float_reg);
	float SIMD_CALL xOf(XY_float_reg);
	float SIMD_CALL yOf(XY_float_reg);

	XY_float_reg   SIMD_CALL operator + (XY_float_reg, XY_float_reg);
	XY_float_reg   SIMD_CALL operator - (XY_float_reg, XY_float_reg);
	XY_float_reg & SIMD_CALL operator +=(XY_float_reg &, XY_float_reg);
	XY_float_reg & SIMD_CALL operator -=(XY_float_reg &, XY_float_reg);
	XY_float_reg   SIMD_CALL operator * (float, XY_float_reg);
	XY_float_reg   SIMD_CALL operator / (float, XY_float_reg);
	XY_float_reg   SIMD_CALL operator * (XY_float_reg, float);
	XY_float_reg   SIMD_CALL operator / (XY_float_reg, float);
	XY_float_reg & SIMD_CALL operator *=(XY_float_reg &, float);
	XY_float_reg & SIMD_CALL operator /=(XY_float_reg &, float);

	float SIMD_CALL dotProductOf(XY_float_reg, XY_float_reg);

	XY_float_reg SIMD_CALL doubled(XY_float_reg);

	// Computes the squared length of the given vector.
	float SIMD_CALL squaredAbsoluteOf(XY_float_reg);

	// Checks whether the vector can be normalized.
	bool SIMD_CALL isValidDirection(XY_float_reg);

	// Computes the absolute of the given vector (its length).
	float SIMD_CALL absoluteOf(XY_float_reg);

	XY_float_reg SIMD_CALL normalized(XY_float_reg);

	XY_float_reg SIMD_CALL scaledToLength(XY_float_reg, float desiredLength);

	// Interpolates linearly between the given vectors according to the given factor, where 0 denotes a and 1 denotes b.
	XY_float_reg SIMD_CALL interpolateLinearly(
		XY_float_reg a,
		float        factor,
		XY_float_reg b
	);

	// Rotates the given vector 90 about the Z axis.
	//   right-handed
	//   result has inputs length
	XY_float_reg SIMD_CALL perpendicularOf(XY_float_reg);

	// Tests whether X and Y are finite numbers.
	bool SIMD_CALL isFinite(XY_float_reg);

	XY_float_reg SIMD_CALL componentWiseMinimumOf(XY_float_reg, XY_float_reg);
	XY_float_reg SIMD_CALL componentWiseMaximumOf(XY_float_reg, XY_float_reg);

	// Computes the squared distance between the two given points.
	float SIMD_CALL squaredDistanceBetween(XY_float_reg, XY_float_reg);

	// Computes the distance between the two given points.
	float SIMD_CALL distanceBetween(XY_float_reg, XY_float_reg);

	XY_float_reg SIMD_CALL centerOf(XY_float_reg, XY_float_reg);

	// Tests the given vectors for numerical (not bitwise) equality.
	bool SIMD_CALL operator ==(XY_float_reg, XY_float_reg);
	bool SIMD_CALL operator !=(XY_float_reg, XY_float_reg);

	// Computes the sine of the given angle.
	float sineOf(float);
	double sineOf(double);

	// Computes the cosine of the given angle.
	float cosineOf(float);
	double cosineOf(double);

	// Computes the tangent of the given angle.
	float tangentOf(float);
	double tangentOf(double);

	// Computes an angle to the given sine. The sine must be in the range of [-1; 1].
	float arcusSineOf(float);
	double arcusSineOf(double);

	// Computes an angle to the given cosine. The sine must be in the range of [-1; 1].
	float arcusCosineOf(float);
	double arcusCosineOf(double);

	// Computes the counter-clockwise angle to the X axis.
	//   like atan2() with reversed parameters
	//   the vector must not be zero
	Angle4B_reg SIMD_CALL polarAngleOf(XY_float_reg);
	double polarAngleOf(double x, double y);






	XY_SInt4B_reg SIMD_CALL load(XY_SInt1B const &);




	XY_SInt4B_reg SIMD_CALL load(XY_SInt4B const &);
	XY_SInt4B_reg SIMD_CALL xyFrom(SInt4B x, SInt4B y);
	SInt4B SIMD_CALL xOf(XY_SInt4B_reg);
	SInt4B SIMD_CALL yOf(XY_SInt4B_reg);

	XY_SInt4B_reg SIMD_CALL operator +(XY_SInt4B_reg, XY_SInt4B_reg);
	XY_SInt4B_reg SIMD_CALL operator -(XY_SInt4B_reg, XY_SInt4B_reg);

	// Shifts each components value to the right.
	//   replicates the sign bit
	//   shift must not exceed 31 bits
	XY_SInt4B_reg SIMD_CALL shiftRightArithmetic(XY_SInt4B_reg, UIntLeast1B);

	// Shifts each components value to the left.
	//   shift must not exceed 31 bits
	XY_SInt4B_reg SIMD_CALL operator <<(XY_SInt4B_reg, UIntLeast1B);

	bool SIMD_CALL operator !=(XY_SInt4B_reg, XY_SInt4B_reg);

	// Computes the squared length of the given vector.
	UInt4B SIMD_CALL squaredAbsoluteOf(XY_SInt4B_reg);

	XY_float_reg SIMD_CALL asFloats(XY_SInt4B_reg);

	// Checks whether all components of the left vector are below their right counterparts.
	bool SIMD_CALL areAllComponentsBelow(XY_SInt4B_reg, XY_SInt4B_reg);






	XYZ_float_reg SIMD_CALL load(XYZ_float const &);
	XYZ_float_reg SIMD_CALL loadPadded(XYZ_float const & v);
	XYZ_float_reg SIMD_CALL OYOFrom(float y);
	XYZ_float_reg SIMD_CALL OOZFrom(float z);
	XYZ_float_reg SIMD_CALL x00From(float x);
	XYZ_float_reg SIMD_CALL xy0From(float x, float y);
	XYZ_float_reg SIMD_CALL xyzFrom(float x, float y, float z);
	XYZ_float_reg SIMD_CALL xyzFrom(floatx4);
	void SIMD_CALL store(XYZ_float &, XYZ_float_reg);
	XYZ_float SIMD_CALL stored(XYZ_float_reg);

	XYZ_float_reg SIMD_CALL x0yOf(XY_float_reg);
	float SIMD_CALL xOf(XYZ_float_reg);
	float SIMD_CALL yOf(XYZ_float_reg);
	float SIMD_CALL zOf(XYZ_float_reg);

	XYZ_float_reg   SIMD_CALL operator + (XYZ_float_reg, XYZ_float_reg);
	XYZ_float_reg   SIMD_CALL operator + (XYZ_float_reg, float);
	XYZ_float_reg   SIMD_CALL operator - (XYZ_float_reg, XYZ_float_reg);
	XYZ_float_reg   SIMD_CALL operator * (XYZ_float_reg, XYZ_float_reg);
	XYZ_float_reg   SIMD_CALL operator / (XYZ_float_reg, XYZ_float_reg);
	XYZ_float_reg & SIMD_CALL operator +=(XYZ_float_reg &, XYZ_float_reg);
	XYZ_float_reg & SIMD_CALL operator -=(XYZ_float_reg &, XYZ_float_reg);
	XYZ_float_reg   SIMD_CALL operator * (float, XYZ_float_reg);
	XYZ_float_reg   SIMD_CALL operator / (float, XYZ_float_reg);
	XYZ_float_reg   SIMD_CALL operator * (XYZ_float_reg, float);
	XYZ_float_reg   SIMD_CALL operator / (XYZ_float_reg, float);
	XYZ_float_reg & SIMD_CALL operator *=(XYZ_float_reg &, float);
	XYZ_float_reg & SIMD_CALL operator /=(XYZ_float_reg &, float);

	float SIMD_CALL dotProductOf(XYZ_float_reg, XYZ_float_reg);

	XYZ_float_reg SIMD_CALL doubled(XYZ_float_reg);

	XYZ_float_reg SIMD_CALL squared(XYZ_float_reg);

	XYZ_float_reg SIMD_CALL clamped(float, XYZ_float_reg, float);

	// Computes the squared length of the given vector.
	float SIMD_CALL squaredAbsoluteOf(XYZ_float_reg);

	// Checks whether the vector can be normalized.
	bool SIMD_CALL isValidDirection(XYZ_float_reg);

	bool SIMD_CALL isAnyComponentNonZero(XYZ_float_reg);

	bool SIMD_CALL isAnyComponentSubnormal(XYZ_float_reg);

	// Computes the absolute of the given vector (its length).
	float SIMD_CALL absoluteOf(XYZ_float_reg);

	XYZ_float_reg SIMD_CALL normalized(XYZ_float_reg);

	XYZ_float_reg SIMD_CALL scaledToLength(XYZ_float_reg v, float desiredLength);

	// Interpolates linearly between the given vectors according to the given factor, where 0 denotes a and 1 denotes b.
	XYZ_float_reg SIMD_CALL interpolateLinearly(
		XYZ_float_reg a,
		float         factor,
		XYZ_float_reg b
	);

	// Computes the right-handed cross product of the two given vectors.
	XYZ_float_reg SIMD_CALL crossProductOf(XYZ_float_reg, XYZ_float_reg);

	// Computes the cosine of the shortest angle between the two given vectors.
	//   no need for normalization, but must be valid directions
	//   due to floating-point accuracy, the result may be slightly above 1 for parallel vectors
	float SIMD_CALL cosineOfAngleBetween(XYZ_float_reg, XYZ_float_reg);

	// Tests whether X, Y, and Z are finite numbers.
	bool SIMD_CALL isFinite(XYZ_float_reg);

	XYZ_float_reg SIMD_CALL componentWiseMinimumOf(XYZ_float_reg, XYZ_float_reg);
	XYZ_float_reg SIMD_CALL componentWiseMaximumOf(XYZ_float_reg, XYZ_float_reg);

	// Computes the squared distance between the two given points.
	float SIMD_CALL squaredDistanceBetween(XYZ_float_reg, XYZ_float_reg);

	// Computes the distance between the two given points.
	float SIMD_CALL distanceBetween(XYZ_float_reg, XYZ_float_reg);

	XYZ_float_reg SIMD_CALL centerOf(XYZ_float_reg, XYZ_float_reg);

	// Tests the given vectors for numerical (not bitwise) equality.
	bool SIMD_CALL operator ==(XYZ_float_reg, XYZ_float_reg);
	bool SIMD_CALL operator !=(XYZ_float_reg, XYZ_float_reg);

	XYZ_SInt4B_reg SIMD_CALL rounded(XYZ_float_reg);

	// Assume geometry being projected onto a screen perspectively along the negative Z axis. The field of view is given as slope.
	//  The screen is not a square, so the slope applies either to its width or to its height, whichever is dominant.
	// This function returns the direction from the coordinate origin to a pixel on the screen. The result is not normalized.
	XYZ_float_reg SIMD_CALL directionOfPerspectivePixel(
		UIntLeast2B xFromLeftInPixels,
		UIntLeast2B yFromTopInPixels,
		UIntLeast2B screenWidthInPixels,
		UIntLeast2B screenHeightInPixels,
		UIntLeast2B dominantDimensionInPixels,
		float       slope
	);




	XYZ_SInt4B_reg SIMD_CALL load(XYZ_SInt2B const &);
	XYZ_SInt4B_reg SIMD_CALL load(XYZ_SInt4B const &);
	XYZ_SInt4B_reg SIMD_CALL loadPadded(XYZ_SInt2B const &);
	void SIMD_CALL store(XYZ_SInt2B &, XYZ_SInt4B_reg);
	void SIMD_CALL store(XYZ_SInt4B &, XYZ_SInt4B_reg);
	XYZ_SInt4B SIMD_CALL stored(XYZ_SInt4B_reg);

	SInt4B SIMD_CALL xOf(XYZ_SInt4B_reg);
	SInt4B SIMD_CALL yOf(XYZ_SInt4B_reg);
	SInt4B SIMD_CALL zOf(XYZ_SInt4B_reg);

	XYZ_SInt4B_reg SIMD_CALL xyzFrom(SInt4B x, SInt4B y, SInt4B z);

	XYZ_SInt4B_reg SIMD_CALL componentWiseMinimumOf(XYZ_SInt4B_reg, XYZ_SInt4B_reg);

	XYZ_float_reg SIMD_CALL asFloats(XYZ_SInt4B_reg);

	XYZ_SInt4B_reg   SIMD_CALL operator + (XYZ_SInt4B_reg, XYZ_SInt4B_reg);
	XYZ_SInt4B_reg   SIMD_CALL operator - (XYZ_SInt4B_reg, XYZ_SInt4B_reg);
	XYZ_SInt4B_reg & SIMD_CALL operator +=(XYZ_SInt4B_reg &, XYZ_SInt4B_reg);
	XYZ_SInt4B_reg & SIMD_CALL operator -=(XYZ_SInt4B_reg &, XYZ_SInt4B_reg);

	// Tests the given vectors for bitwise equality.
	bool SIMD_CALL operator ==(XYZ_SInt4B_reg, XYZ_SInt4B_reg);
	bool SIMD_CALL operator !=(XYZ_SInt4B_reg, XYZ_SInt4B_reg);

	// Shifts each components value to the right.
	//   replicates the sign bit
	//   shift must not exceed 31 bits
	XYZ_SInt4B_reg SIMD_CALL shiftRightArithmetic(XYZ_SInt4B_reg, UIntLeast1B);

	// Shifts each components value to the left.
	//   shift must not exceed 31 bits
	XYZ_SInt4B_reg SIMD_CALL operator <<(XYZ_SInt4B_reg, UIntLeast1B);



	// Equals translation(offset) & b. The offset must be a finite vector.
	Matrix3x2_float_reg SIMD_CALL preTranslated(XY_float_reg, Matrix2x2_float_reg);

	Matrix3x2_float_reg SIMD_CALL translated(Matrix2x2_float_reg, XY_float_reg);

	Matrix3x2_float_reg SIMD_CALL translated(Matrix3x2_float_reg, XY_float_reg);

	// Combines the given transformations  left first, then right (reverse matrix multiplication order).
	Matrix2x2_float_reg SIMD_CALL operator &(Matrix2x2_float_reg, Matrix2x2_float_reg);
	Matrix3x2_float_reg SIMD_CALL operator &(Matrix2x2_float_reg, Matrix3x2_float_reg);

	// Transforms the given vector with the given 22 matrix, typically rotating it.
	XY_float_reg SIMD_CALL transformed(
		XY_float_reg       ,
		Matrix2x2_float_reg
	);

	// Transforms the given vector with the given 32 matrix, typically rotating and translating it.
	XY_float_reg SIMD_CALL transformed(
		XY_float_reg        v,
		Matrix3x2_float_reg m
	);

	void SIMD_CALL store(Matrix3x2_float &, Matrix3x2_float_reg);


	Matrix3x3_float_reg SIMD_CALL load(Matrix3x3_float const &);

	// Computes the X, Y, and Z axis of the given rotation (useful as an object-to-world matrix).
	Matrix3x3_float_reg SIMD_CALL axesOf(Quaternion_float_reg);
	// Computes the X, Y, and Z axis of the given rotations inverse (useful as a world-to-object matrix).
	Matrix3x3_float_reg SIMD_CALL inverseAxesOf(Quaternion_float_reg);

	XYZ_float_reg SIMD_CALL xAxisOf(Matrix3x3_float_reg);
	XYZ_float_reg SIMD_CALL yAxisOf(Matrix3x3_float_reg);
	XYZ_float_reg SIMD_CALL zAxisOf(Matrix3x3_float_reg);

	// Transforms the given vector with the given 33 matrix, typically rotating it.
	XYZ_float_reg SIMD_CALL transformed(
		XYZ_float_reg      ,
		Matrix3x3_float_reg
	);

	// Divides the vector by its Z.
	XYZ_float_reg SIMD_CALL projectedAlongZ(XYZ_float_reg);

	Matrix3x3_float_reg SIMD_CALL transposed(Matrix3x3_float_reg);

	// Equals scaling(x, y, z) & b. Scaling must be a finite vector.
	Matrix3x3_float_reg SIMD_CALL preScaled(
		XYZ_float_reg      ,
		Matrix3x3_float_reg
	);

	// Equals translation(offset) & b. The offset must be a finite vector.
	Matrix4x3_float_reg SIMD_CALL preTranslated(
		XYZ_float_reg      ,
		Matrix3x3_float_reg
	);

	Matrix4x3_float_reg SIMD_CALL translated(
		Matrix3x3_float_reg,
		XYZ_float_reg
	);

	Matrix3x3_float_reg SIMD_CALL operator &(Matrix3x3_float_reg, Matrix3x3_float_reg const &);

	void SIMD_CALL store(Matrix3x3_float &, Matrix3x3_float_reg);
	Matrix3x3_float SIMD_CALL stored(Matrix3x3_float_reg);

	Matrix4x3_float_reg SIMD_CALL load(Matrix4x3_float const &);
	XYZ_float_reg SIMD_CALL loadXAxisOf(Matrix4x3_float const &);
	XYZ_float_reg SIMD_CALL loadYAxisOf(Matrix4x3_float const &);
	XYZ_float_reg SIMD_CALL loadZAxisOf(Matrix4x3_float const &);
	XYZ_float_reg SIMD_CALL loadOffsetOf(Matrix4x3_float const &);

	void SIMD_CALL store(Matrix4x3_float &, Matrix4x3_float_reg);
	Matrix4x3_float SIMD_CALL stored(Matrix4x3_float_reg);

	Matrix4x3_float_reg SIMD_CALL matrix4x3From(
		Quaternion_float_reg,
		XYZ_float_reg
	);

	Matrix4x3_float_reg SIMD_CALL translated(
		Matrix4x3_float_reg,
		XYZ_float_reg
	);

	Matrix4x3_float_reg SIMD_CALL preTranslated(
		XYZ_float_reg      ,
		Matrix4x3_float_reg
	);

	// Returns a matrix with inverted scaling, rotation, and translation.
	Matrix4x3_float_reg SIMD_CALL inverted(Matrix4x3_float_reg);

	// Combines the given transformations  left first, then right (reverse matrix multiplication order).
	Matrix4x3_float_reg SIMD_CALL operator &(Matrix4x3_float_reg, Matrix4x3_float_reg const &);
	Matrix4x3_float_reg SIMD_CALL operator &(Matrix4x3_float_reg, Matrix3x3_float_reg const &);
	Matrix4x3_float_reg SIMD_CALL operator &(Matrix3x3_float_reg const &, Matrix4x3_float_reg);

	XYZ_float_reg SIMD_CALL xAxisOf(Matrix4x3_float_reg);
	XYZ_float_reg SIMD_CALL yAxisOf(Matrix4x3_float_reg);
	XYZ_float_reg SIMD_CALL zAxisOf(Matrix4x3_float_reg);
	XYZ_float_reg SIMD_CALL offsetOf(Matrix4x3_float_reg);
	float SIMD_CALL distanceOf(Matrix4x3_float_reg);
	Matrix3x3_float_reg SIMD_CALL rotationOf(Matrix4x3_float_reg);

	// Transforms the given vector with the given 43 matrix, typically rotating and translating it.
	XYZ_float_reg SIMD_CALL transformed(
		XYZ_float_reg      ,
		Matrix4x3_float_reg
	);


	Matrix4x4_float_reg SIMD_CALL load(Matrix4x4_float const &);

	void SIMD_CALL store(Matrix4x4_float &, Matrix4x4_float_reg const &);

	Matrix4x4_float_reg SIMD_CALL matrix4x4From(Matrix4x3_float_reg);

	// Transforms the given vector with the given 44 matrix, typically rotating and translating it.
	XYZ_float_reg SIMD_CALL transformed(
		XYZ_float_reg              ,
		Matrix4x4_float_reg const &
	);

	// Combines the given transformations  left first, then right (reverse matrix multiplication order).
	Matrix4x4_float_reg SIMD_CALL operator &(Matrix3x3_float_reg, Matrix4x4_float_reg const &);
	Matrix4x4_float_reg SIMD_CALL operator &(Matrix4x3_float_reg, Matrix4x4_float_reg const &);
	Matrix4x4_float_reg SIMD_CALL operator &(Matrix4x4_float_reg, Matrix4x4_float_reg const &);

	Matrix4x3_float_reg SIMD_CALL matrix4x3From(Matrix4x4_float_reg);

	Matrix3x3_float_reg SIMD_CALL matrix3x3From(Matrix4x4_float_reg);

	// From the given 44 matrix, isolates the rotation and adjusts it to rotate normals properly.
	Matrix3x3_float_reg SIMD_CALL normalTransformationFrom(Matrix4x3_float_reg);



	Quaternion_float_reg SIMD_CALL load(Quaternion_float const &);

	void SIMD_CALL store(Quaternion_float &, Quaternion_float_reg);

	bool SIMD_CALL isFinite(Quaternion_float_reg);

	XYZ_float_reg SIMD_CALL xAxisOf(Quaternion_float_reg);
	XYZ_float_reg SIMD_CALL yAxisOf(Quaternion_float_reg);
	XYZ_float_reg SIMD_CALL zAxisOf(Quaternion_float_reg);

	// Returns the inverse (or the conjugate) of the given quaternion, i.e. the opposite rotation.
	Quaternion_float_reg SIMD_CALL inverted(Quaternion_float_reg);

	// Returns the negation of the given quaternion. It represents the same rotation but with different values, which is useful
	//  for compression (see Cryteks Spherical Skinning with Dual-Quaternions and QTangents).
	Quaternion_float_reg SIMD_CALL negated(Quaternion_float_reg);

	// Returns the product of the two given quaternions.
	Quaternion_float_reg SIMD_CALL operator *(Quaternion_float_reg, Quaternion_float_reg);

	// Computes the normalization of the given quaternion. The quaternion must be finite and non-zero.
	Quaternion_float_reg SIMD_CALL normalized(Quaternion_float_reg);

	// Rotates the given vector using the given quaternion.
	XYZ_float_reg SIMD_CALL rotated(XYZ_float_reg, Quaternion_float_reg);

	// Spherical linear interpolation (SLERP) of two quaternions. A factor of 0 returns the first quaternion.
	// The interpolation is constant-velocity and torque-minimal (which is important for physics).
	Quaternion_float_reg SIMD_CALL interpolatedRotations(Quaternion_float_reg, float, Quaternion_float_reg);


	//  counter-clockwise
	//  result not normalized
	XYZ_float_reg SIMD_CALL normalOfTriangle(XYZ_float_reg, XYZ_float_reg, XYZ_float_reg);


	// Creates an angle from the given value, in degrees.
	//   must be in the range of [0, 359]
	Angle4B_reg angleFromDegrees(SInt4B);

	// Creates an angle from the given value, in degrees.
	//   must not exceed 773,094,113,280
	Angle4B_reg angleFromDegrees(float);
	Angle4B_reg angleFromDegrees(double);

	// Creates an angle from the given value, in radians.
	//   must not exceed 13,493,037,704rad
	Angle4B_reg angleFromRadians(float);
	Angle4B_reg angleFromRadians(double);

	// Creates an angle from the given astronomical hour angle, in hours.
	//   must not exceed 51,539,607,552 hours
	Angle4B_reg angleFromHourAngle(double);

	Angle4B_reg operator +(Angle4B_reg, Angle4B_reg);
	Angle4B_reg & operator +=(Angle4B_reg &, Angle4B_reg);
	Angle4B_reg operator -(Angle4B_reg, Angle4B_reg);
	bool operator <(Angle4B_reg, Angle4B_reg);
	bool operator >(Angle4B_reg, Angle4B_reg);

	// Clamps the given angle to the range of [lower, upper] with
	//   -90 <= lower <=   0 and
	//     0 <= upper <= +90.
	// Useful for camera controls to avoid upside-down views.
	Angle4B_reg clampedToDegrees(
		SIntLeast1B lowerLimitInDegrees,
		Angle4B_reg angle,
		SIntLeast1B upperLimitInDegrees
	);

	Angle4B_reg absoluteDifferenceOf(Angle4B_reg, Angle4B_reg);

	template <
		unsigned fractionBits
	> void adaptGradually(
		Angle4B_reg &       angle,
		UInt4B              factor,
		Angle4B_reg const & reference
	) {
		auto const diff = SInt4B(reference.unorm - angle.unorm);
		angle.unorm += SInt4B((SInt8B(factor) * diff) >> fractionBits);
	}

	// Rounds the given angle to the next integer, in degrees.
	UIntLeast2B roundedToDegrees(Angle4B_reg);

	// Rounds the given angle to radians, represented as a floating-point number.
	double inRadians(Angle4B_reg);

	float tangentOf(Angle4B_reg);

	Angle4B_reg arcusTangentOf(float);




	BoundsXYZ_float_reg SIMD_CALL load(BoundsXYZ_float const &);
	void SIMD_CALL store(BoundsXYZ_float &, BoundsXYZ_float_reg);

	// Computes the bounding box of the three given points.
	BoundsXYZ_float_reg SIMD_CALL boundingHullOf(XYZ_float_reg, XYZ_float_reg, XYZ_float_reg);

	// Computes the total bounding box containing points of any of the two given boxes.
	BoundsXYZ_float_reg SIMD_CALL unionOf(BoundsXYZ_float_reg, BoundsXYZ_float_reg);
	BoundsXYZ_float_reg SIMD_CALL unionOf(BoundsXYZ_float_reg, XYZ_float_reg);
	BoundsXYZ_float_reg SIMD_CALL unionOf(BoundsXYZ_float_reg, XYZ_float_reg, XYZ_float_reg, XYZ_float_reg);

	// Returns the bounding box of the volume where the two boxes overlap.
	//   if the boxes are not intersecting, the result is degenerated
	BoundsXYZ_float_reg SIMD_CALL intersectionOf(BoundsXYZ_float_reg, BoundsXYZ_float_reg);

	XYZ_float_reg SIMD_CALL minimumOf(BoundsXYZ_float_reg);
	XYZ_float_reg SIMD_CALL maximumOf(BoundsXYZ_float_reg);
	XYZ_float_reg SIMD_CALL centerOf(BoundsXYZ_float_reg);
	XYZ_float_reg SIMD_CALL sidelengthOf(BoundsXYZ_float_reg);

	bool SIMD_CALL areIntersecting(BoundsXYZ_float_reg, BoundsXYZ_float_reg);

	// Determines whether any dimension of the given bounding box is zero or inverted.
	bool SIMD_CALL isDegenerated(BoundsXYZ_float_reg);

	// Extends the given bounding box to the nearest power-of-two integer boundaries.
	BoundsXYZ_SInt4B_reg SIMD_CALL rounded(BoundsXYZ_float_reg, SIntLeast1B powerOfTwo);

	// Displaces the given bounding box.
	void SIMD_CALL operator -=(BoundsXYZ_float_reg &, XYZ_float_reg);




	XYZ_SInt4B_reg SIMD_CALL minimumOf(BoundsXYZ_SInt4B_reg);
	XYZ_SInt4B_reg SIMD_CALL maximumOf(BoundsXYZ_SInt4B_reg);

	// Determines whether any dimension of the given bounding box is zero or inverted.
	bool SIMD_CALL isDegenerated(BoundsXYZ_SInt4B_reg);




	BoundsXY_SInt4B_reg SIMD_CALL boundingRectFrom(SInt4B minX, SInt4B minY, SInt4B maxX, SInt4B maxY);

	BoundsXY_SInt4B_reg SIMD_CALL load(BoundsXY_UInt2B const &);

	void SIMD_CALL store(BoundsXY_UInt2B &, BoundsXY_SInt4B_reg);

	BoundsXY_SInt4B_reg SIMD_CALL unionOf(BoundsXY_SInt4B_reg, BoundsXY_SInt4B_reg);

	bool SIMD_CALL areIntersecting(BoundsXY_SInt4B_reg, BoundsXY_SInt4B_reg);



	// Computes the bounding rectangle of the three given points.
	BoundsXY_float_reg SIMD_CALL boundingHullOf(XY_float_reg, XY_float_reg, XY_float_reg);

	// Computes the total bounding rectangle containing points of any of the two given rectangles.
	BoundsXY_float_reg SIMD_CALL unionOf(BoundsXY_float_reg, BoundsXY_float_reg);

	XY_float_reg SIMD_CALL sidelengthOf(BoundsXY_float_reg);





	void SIMD_CALL store(PlaneXYZ_float &, PlaneXYZ_float_reg);
	PlaneXYZ_float_reg SIMD_CALL load(PlaneXYZ_float const &);

	// Creates a plane from a point and a normal. The resulting plane is normalized only if the given normal is normalized.
	PlaneXYZ_float_reg SIMD_CALL planeFromPointAndNormal(
		XYZ_float_reg anyPointOnPlane,
		XYZ_float_reg itsNormal
	);

	float SIMD_CALL signedDistanceBetween(XYZ_float_reg, PlaneXYZ_float_reg);

	// Computes the distance where the given ray intersects the given plane (in multiples of the given rays direction), or NaN if
	//  the ray does not intersect the plane.
	float SIMD_CALL intersectionDistanceOf(RayXYZ_float_reg, PlaneXYZ_float_reg);

	// Computes the intersection point of the given ray and the given plane. NaN if they dont intersect.
	XYZ_float_reg SIMD_CALL intersectionPointOf(RayXYZ_float_reg, PlaneXYZ_float_reg);







	LineXYZ_float_reg SIMD_CALL lineFrom(RayXYZ_float_reg);

	//  the point must be finite coordinates
	//  the direction must be finite and non-null
	LineXYZ_float_reg SIMD_CALL lineFromPointAndDirection(
		XYZ_float_reg anyPoint,
		XYZ_float_reg direction
	);

	//  the points must be finite coordinates
	//  the distance must be sufficient to compute a direction
	LineXYZ_float_reg SIMD_CALL lineFromTwoPoints(XYZ_float_reg, XYZ_float_reg);

	// Normalizes the given line; i.e.:
	//   the arbitrary point becomes the point closest to the origin (this guarantees highest accuracy in floating-point
	//    computation)
	//   the direction becomes normalized
	LineXYZ_float_reg SIMD_CALL normalized(LineXYZ_float_reg);

	// Computes the point on the line that is closest to the given point.
	XYZ_float_reg SIMD_CALL closestPoint(XYZ_float_reg, LineXYZ_float_reg);

	// Computes the distance where the two given lines cross. Returns
	//   zero if they overlap;
	//   infinity if they are parallel but displaced; and
	//   the distance in multiples of the first lines length otherwise.
	float SIMD_CALL intersectionDistanceOf(LineXYZ_float_reg, LineXYZ_float_reg);

	// Checks the given line and the given triangle (in any winding order) for intersection. Returns
	//   the projected distance of the intersection point from line.anyPoint in the lines direction; or
	//   NaN if not intersecting.
	float SIMD_CALL intersectionDistanceOf(
		LineXYZ_float_reg,
		XYZ_float_reg    a,
		XYZ_float_reg    aToB,
		XYZ_float_reg    aToC
	);



	RayXYZ_float_reg SIMD_CALL load(RayXYZ_float const &);

	XYZ_float_reg SIMD_CALL originOf(RayXYZ_float_reg);

	// Not necessarily normalized!
	XYZ_float_reg SIMD_CALL directionOf(RayXYZ_float_reg);

	RayXYZ_float_reg SIMD_CALL rayFromOriginAndDirection(
		XYZ_float_reg itsOrigin,
		XYZ_float_reg itsDirection
	);

	// Determines whether the given ray intersects the given triangle (in any winding order).
	bool SIMD_CALL areIntersecting(
		RayXYZ_float_reg ,
		XYZ_float_reg    a,
		XYZ_float_reg    aToB,
		XYZ_float_reg    aToC
	);



	LineSegmentXY_float_reg SIMD_CALL lineSegmentBetween(XY_float_reg, XY_float_reg);

	// Computes the point on the given line segment that is closest to the given point.
	XY_float_reg SIMD_CALL closestPoint(XY_float_reg, LineSegmentXY_float_reg);

	// Computes the squared minimal distance between the given point and the given line segment.
	float SIMD_CALL squaredDistanceBetween(XY_float_reg, LineSegmentXY_float_reg);

	// Computes the minimal distance between the given point and the given line segment.
	float SIMD_CALL distanceBetween(XY_float_reg, LineSegmentXY_float_reg);


} // namespace Geo



namespace Physics {


	Geo::XYZ_float_reg SIMD_CALL computeAngularVelocity(
		Geo::XYZ_float_reg localAngularMomentum, // Nms or kgm/s
		Geo::XYZ_float_reg inertia               // kgm
	);

	// Computes the moment of inertia of a solid cuboid of the given dimensions and mass.
	Geo::XYZ_float_reg SIMD_CALL inertiaOfCuboid(
		Geo::XYZ_float_reg dimensions, // width/height/depth in m
		float              mass        // kg
	);


} // Physics