#include "framework.hpp"
#include "SIMD.hpp"
#include "Math.hpp"
#include "Geo.hpp"



#if COMPILED_BY_VISUAL_CPP
	// Visual C++ intrinsics. On Clang and GCC, the __builtin_ counterparts are used, so no declaration is required.
	import {
		double exp(double);
		double log(double);
		double pow(double, double);
#		if _M_X64
			float expf(float);
			float logf(float);
			float powf(float, float);
#		endif
	}
#endif



namespace Math {

	// Computes the natural logarithm of the given value. The value must be greater than zero.
	float naturalLogarithmOf(float x) {
#		if COMPILED_BY_VISUAL_CPP
#			if _M_X64
				return logf(x);
#			else
				return float(log(x));
#			endif
#		elif COMPILED_BY_CLANG
			return __builtin_logf(x);
#		endif
	}
	double naturalLogarithmOf(double value) {
#		if COMPILED_BY_VISUAL_CPP
			return log(value);
#		elif COMPILED_BY_CLANG
			return __builtin_log(value);
#		endif
	}

	// Computes the base 2 logarithm of the given value. The value must be greater than zero.
	float base2LogarithmOf(float x) {
		return float(base2LogarithmOf(double(x))); // TODO
	}
	double base2LogarithmOf(double x) {
#		if COMPILED_BY_VISUAL_CPP
			return log(x) / 0.69314718055994530941723212145818; // log2(x) = logN(x)  logN(2)
#		elif COMPILED_BY_CLANG
			return __builtin_log(x) / 0.69314718055994530941723212145818; // log2(x) = logN(x)  logN(2)
#		endif
	}

	// Computes the given base raised to the given power.
	float raisedToPower(float base, float exponent) {
#		if COMPILED_BY_VISUAL_CPP
#			if _M_X64
				return powf(base, exponent);
#			else
				return float(pow(base, exponent));
#			endif
#		elif COMPILED_BY_CLANG
			return __builtin_powf(base, exponent);
#		endif
	}
	double raisedToPower(double base, double exponent) {
#		if COMPILED_BY_VISUAL_CPP
			return pow(base, exponent);
#		elif COMPILED_BY_CLANG
			return __builtin_pow(base, exponent);
#		endif
	}

	// Computes e raised to the given power.
	float eRaisedToPower(float exponent) {
#		if COMPILED_BY_VISUAL_CPP
#			if _M_X64
				return expf(exponent);
#			else
				return float(exp(exponent));
#			endif
#		elif COMPILED_BY_CLANG
			return __builtin_expf(exponent);
#		endif
	}
	double eRaisedToPower(double exponent) {
#		if COMPILED_BY_VISUAL_CPP
			return exp(exponent);
#		elif COMPILED_BY_CLANG
			return __builtin_exp(exponent);
#		endif
	}

	// Computes the base raised to the given power. Optimized for integer powers.
	// (complexity: logarithmic in magnitude of exponent)
	float raisedToPower(float currentBase, SIntLeast2B exponent) {
		auto result = 1.0f;
		auto currentBinaryExponent = ::absoluteOf(exponent);

		raise:

			if(currentBinaryExponent & 1) {
				result = result * currentBase;
			}
			currentBinaryExponent = currentBinaryExponent >> 1;
			if(0 != currentBinaryExponent) {
				currentBase = squared(currentBase);
				goto raise;
			}

		return (0 <= exponent) ? result : (1.0f / result);
	}
	double raisedToPower(double currentBase, SIntLeast2B exponent) {
		auto result = 1.0;
		auto currentBinaryExponent = ::absoluteOf(exponent);

		raise:

			if(currentBinaryExponent & 1) {
				result = result * currentBase;
			}
			currentBinaryExponent = currentBinaryExponent >> 1;
			if(0 != currentBinaryExponent) {
				currentBase = squared(currentBase);
				goto raise;
			}

		return (0 <= exponent) ? result : (1.0 / result);
	}


} // Math