#include "framework.hpp"
#include "Memory.hpp"
#include "OS Win32.hpp"
#include "OS Win32 File.hpp"

#if COMPILED_BY_VISUAL_CPP
	// Visual C++ intrinsics. On Clang and GCC, the __builtin_ counterparts are used, so no declaration is required.
	import {
		int C_CALL memcmp(void const *, void const *, size_t);
		void * C_CALL memmove(void * _Dst, void const * _Src, size_t _Size);
		void __movsb(void *, void const *, UCount);
		void __movsw(void *, void const *, UCount);
		void __movsd(void *, void const *, UCount);
#		if COMPILED_FOR_X86_64
			void __movsq(void *, void const *, UCount);
#		endif
		void __stosb(void *, Byte, UCount);
		void __stosw(void *, UInt2B, UCount);
		void __stosd(void *, UInt4B, UCount);
		void __stosq(void *, UInt8B, UCount);
	}
#endif



// The x86 REP MOVSB instruction (byte-wise copy).
void movsb(Byte * toDestination, Byte const * toSource, UCount count) {
#	if COMPILED_BY_VISUAL_CPP
		__movsb(toDestination, toSource, count);
#	elif COMPILED_BY_CLANG || COMPILED_BY_GCC
		__asm__("rep movsb" : : "D"(toDestination), "S"(toSource), "c"(count) : "%edi", "%esi", "%ecx");
#	else
#		error TODO
		auto const toEnd = toDestination + count;
		while(toDestination < toEnd) {
			*toDestination++ = *toSource++;
		}
#	endif
}
// The x86 REP MOVSW instruction (2-B-wise copy).
void movsw(UInt2B * toDestination, UInt2B const * toSource, UCount count) {
#	if COMPILED_BY_VISUAL_CPP
		__movsw(toDestination, toSource, count);
#	elif COMPILED_BY_CLANG || COMPILED_BY_GCC
		__asm__("rep movsw" : : "D"(toDestination), "S"(toSource), "c"(count) : "%edi", "%esi", "%ecx");
#	else
#		error TODO
		auto const toEnd = toDestination + count;
		while(toDestination < toEnd) {
			*toDestination++ = *toSource++;
		}
#	endif
}
// The x86 REP MOVSD instruction (4-B-wise copy).
void movsd(UInt4B * toDestination, UInt4B const * toSource, UCount count) {
#	if COMPILED_BY_VISUAL_CPP
		__movsd(toDestination, toSource, count);
#	elif COMPILED_BY_CLANG || COMPILED_BY_GCC
		__asm__("rep movsd" : : "D"(toDestination), "S"(toSource), "c"(count) : "%edi", "%esi", "%ecx");
#	else
#		error TODO
		auto const toEnd = toDestination + count;
		while(toDestination < toEnd) {
			*toDestination++ = *toSource++;
		}
#	endif
}
#if COMPILED_FOR_X86_64
	// The x86 REP MOVSQ instruction (8-B-wise copy).
	void movsq(UInt8B * toDestination, UInt8B const * toSource, UCount count) {
#		if COMPILED_BY_VISUAL_CPP
			__movsq(toDestination, toSource, count);
#		elif COMPILED_BY_CLANG || COMPILED_BY_GCC
			__asm__("rep movsq" : : "D"(toDestination), "S"(toSource), "c"(count) : "%edi", "%esi", "%ecx");
#		else
#			error TODO
			auto const toEnd = toDestination + count;
			while(toDestination < toEnd) {
				*toDestination++ = *toSource++;
			}
#		endif
	}
#endif

// The x86 REP STOSB instruction (byte-wise filling).
void stosb(Byte * toDestination, Byte value, UCount count) {
#	if COMPILED_BY_VISUAL_CPP
		__stosb(toDestination, value, count);
#	elif COMPILED_BY_CLANG || COMPILED_BY_GCC
		__asm__("rep stosb" : : "D"(toDestination), "a"(value), "c"(count) : "%edi", "%ecx");
#	else
#		error TODO
		auto const toEnd = toDestination + count;
		while(toDestination < toEnd) {
			*toDestination++ = value;
		}
#	endif
}
// The x86 REP STOSW instruction (2-B-wise filling).
void stosw(UInt2B * toDestination, UInt2B value, UCount count) {
#	if COMPILED_BY_VISUAL_CPP
		__stosw(toDestination, value, count);
#	elif COMPILED_BY_CLANG || COMPILED_BY_GCC
		__asm__("rep stosw" : : "D"(toDestination), "a"(value), "c"(count) : "%edi", "%ecx");
#	else
#		error TODO
		auto const toEnd = toDestination + count;
		while(toDestination < toEnd) {
			*toDestination++ = value;
		}
#	endif
}
// The x86 REP STOSD instruction (2-B-wise filling).
void stosd(UInt4B * toDestination, UInt4B value, UCount count) {
#	if COMPILED_BY_VISUAL_CPP
		__stosd(toDestination, value, count);
#	elif COMPILED_BY_CLANG || COMPILED_BY_GCC
		__asm__("rep stosd" : : "D"(toDestination), "a"(value), "c"(count) : "%edi", "%ecx");
#	else
#		error TODO
		auto const toEnd = toDestination + count;
		while(toDestination < toEnd) {
			*toDestination++ = value;
		}
#	endif
}
#if COMPILED_FOR_X86_64
// The x86 REP STOSQ instruction (2-B-wise filling).
	void stosq(UInt8B * toDestination, UInt8B value, UCount count) {
#		if COMPILED_BY_VISUAL_CPP
			__stosq(toDestination, value, count);
#		elif COMPILED_BY_CLANG || COMPILED_BY_GCC
			__asm__("rep stosq" : : "D"(toDestination), "a"(value), "c"(count) : "%edi", "%ecx");
#		else
#			error TODO
			auto const toEnd = toDestination + count;
			while(toDestination < toEnd) {
				*toDestination++ = value;
			}
#		endif
	}
#endif

// Compares the given memory for equality.
bool equals(
	void const * toA,
	void const * toB,
	USize        size
) {
#	if COMPILED_BY_VISUAL_CPP
		return 0 == memcmp(toA, toB, size);
#	elif COMPILED_BY_CLANG
		return 0 == __builtin_memcmp(toA, toB, size);
#	else
#		error TODO
#	endif
}

// Copies the given memory (byte-wise if overlapping).
void copy(
	Byte *       toDestination,
	Byte const * toSource,
	Byte const * toSourcesEnd
) {
	MUST(toSource <= toSourcesEnd);
	return copy(toDestination, toSource, USize(toSourcesEnd - toSource));
}
void copy(
	Byte *       toDestination,
	Byte const * toSource,
	USize        size
) {
#	if COMPILED_BY_VISUAL_CPP
		memmove(toDestination, toSource, size);
#	elif COMPILED_BY_CLANG
		__builtin_memmove(toDestination, toSource, size);
#	else
#		error TODO
#	endif
}

// Zeroes the given range of memory.
void zero(void * toRegion, void * toRegionsEnd) {
	MUST(toRegion <= toRegionsEnd);
	return zero(toRegion, byteDistance(toRegion, toRegionsEnd));
}
void zero(void * toRegion, USize size) {
#	if COMPILED_BY_VISUAL_CPP
		// Full hardware support for high performance (on latest Intel CPUs); minimal code size and register usage (everywhere).
		__stosb(toRegion, 0, size);
#	elif COMPILED_BY_CLANG
		__builtin_memset(toRegion, 0, size);
#	else
#		error TODO
#	endif
}

bool isZero(void const * memory, USize const size) {
	auto       toBeginning = reinterpret_cast<Byte const *>(memory);
	auto const toEnd = toBeginning + size;
	while(toBeginning < toEnd) {
		if(0 != *toBeginning++) {
			return no;
		}
	}
	return yes;
}

namespace Heap {


	// Reserves memory of the given size, in bytes.
	//   content is undefined
	//   must be release()d
	//   returns nullptr on failure
	OUTLINE_CODE Byte * tryToReserve(USize size) {
		return OS::Win32::tryToReserve(size);
	}

	// Reserves memory of the given size, in bytes.
	//   content is zeroed
	//   must be release()d
	//   returns nullptr on failure
	OUTLINE_CODE Byte * tryToReserveZeroed(USize size) {
		return OS::Win32::tryToReserveZeroed(size);
	}

	// Resizes the given memory block to the given size.
	//   input must not be nullptr
	//   new content is undefined
	//   must be release()d
	//   on failure, returns nullptr and leaves input untouched
	OUTLINE_CODE Byte * tryToResize(void const * baseAddress, USize size) {
		return OS::Win32::tryToReallocate(baseAddress, size);
	}

	OUTLINE_CODE void release(void const * baseAddress) {
		return OS::Win32::release(baseAddress);
	}


	Region & defaultPool() {
		return reinterpret_cast<Region &>(OS::Win32::processPool());
	}

	//  on failure, returns nullptr and reports the error
	//  otherwise, destroy() after use
	Region * tryToCreateRegion(Supervisor const * const supervisorOrNull) {
		auto const result = reinterpret_cast<Region *>(OS::Win32::tryToCreateRegion());
		if(nullptr == result) {
			report(supervisorOrNull, ::Error::outOfVirtualMemory);
		}
		return result;
	}

	// Reserves memory of the given size, in bytes.
	//   on failure
	//     reports the error to the given supervisor (pass nullptr to suppress)
	//     returns nullptr
	//   on success
	//     returns the memory with undefined content
	//   checks for array size overflow
	Byte * Region::tryToReserveRaw(
		USize const              elementSize,
		UCount const             numberOfElements,
		Supervisor const * const supervisorOrNull
	) {
		if(numberOfElements <= ~USize(0) / elementSize) { // No allocation size overflow?
			if(auto result = reinterpret_cast<OS::Win32::Region *>(this)->tryToReserve(elementSize * numberOfElements)) {
				return result;
			}
		}
		report(supervisorOrNull, Error::outOfVirtualMemory);
		return nullptr;
	}

	// Reserves memory of the given size, in bytes.
	//   content is zeroed
	//   returns nullptr on failure
	Byte * Region::tryToReserveZeroed(USize size) {
		return reinterpret_cast<OS::Win32::Region *>(this)->tryToReserveZeroed(size);
	}

	// Resizes the given memory block to the given size.
	//   input must not be nullptr
	//   on failure
	//     reports the error to the given supervisor (pass nullptr to suppress)
	//     does not touch the original block
	//     returns nullptr
	//   on success
	//     returns the memory with new content being undefined
	//   checks for array size overflow
	Byte * Region::tryToResizeRaw(
		void const * const       baseAddress,
		USize const              elementSize,
		UCount const             numberOfElements,
		Supervisor const * const supervisorOrNull
	) {
		if(numberOfElements <= ~USize(0) / elementSize) { // No allocation size overflow?
			if(auto result = reinterpret_cast<OS::Win32::Region *>(this)->tryToReallocate(baseAddress, elementSize * numberOfElements)) {
				return result;
			}
		}
		report(supervisorOrNull, Error::outOfVirtualMemory);
		return nullptr;
	}

	void Region::release(void const * const baseAddress) {
		reinterpret_cast<OS::Win32::Region *>(this)->release(baseAddress);
	}

	// Destroys the region and all memory in it.
	void Region::destroy() {
		reinterpret_cast<OS::Win32::Region *>(this)->destroy();
	}

	// Maps the file at the given path to read-only memory.
	//   on failure, returns no and reports the error
	//   unmap() after use
	//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
	bool tryToMapFileForReading(
		MappedFile_readOnly &        result,
		NativeCodeUnit const * const path,
		Supervisor const * const     supervisorOrNull
	) {
		auto returnValue = no;

		if(auto const file = OS::Win32::tryToOpenFile(
			path,
			OS::Win32::FileAccess::readOnly, OS::Win32::FileSharing::dontCare,
			OS::Win32::FileCaching::dontCare, // no effect on memory-mapped files
			supervisorOrNull
		)) {

			auto const size = file->size();
			if(0 == size) {

				// NT & Win32 cannot map empty files; instead return a dummy mapping:
				result.toBeginning = nullptr;
				result.size        = 0;
				returnValue        = yes;

#			if COMPILED_FOR_X86_32
			} else if(size > 0x7FFF'FFFF) {
				report(supervisorOrNull, ::Error::outOfVirtualMemory); // 32-bit systems cannot map files of 2 GiB or more
#			endif

			} else {

				// NT & Win32 do not track file size (only mapped size), so it must be provided separately:
				result.toBeginning = tryToMapForReading(*file, supervisorOrNull);
				result.size        = USize(size);
				returnValue        = nullptr != result.toBeginning;

			}

			file->close();
		}

		return returnValue;
	}

	void unmap(MappedFile_readOnly const & map) {
		if(nullptr != map.toBeginning) {
			OS::Win32::unmap(map.toBeginning);
		}
	}

	// Maps the file at the given path to memory.
	//   on failure, returns no and reports the error
	//   unmap() after use
	//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
	bool tryToMapFileForWriting(
		MappedFile_readWrite &       result,
		NativeCodeUnit const * const path,
		Supervisor const * const     supervisorOrNull
	) {
		auto returnValue = no;

		if(auto const file = OS::Win32::tryToOpenFile(
			path,
			OS::Win32::FileAccess::readWrite, OS::Win32::FileSharing::dontCare,
			OS::Win32::FileCaching::dontCare, // no effect on memory-mapped files
			supervisorOrNull
		)) {

			auto const size = file->size();
			if(0 == size) {

				// NT & Win32 cannot map empty files; instead return a dummy mapping:
				result.toBeginning = nullptr;
				result.size        = 0;
				returnValue = yes;

#			if COMPILED_FOR_X86_32
			} else if(size > 0x7FFF'FFFF) {
				report(supervisorOrNull, ::Error::outOfVirtualMemory); // 32-bit systems cannot map files of 2 GiB or more
#			endif

			} else {

				// NT & Win32 do not track file size (only mapped size), so it must be provided separately:
				result.toBeginning = tryToMapForWriting(*file, supervisorOrNull);
				result.size        = USize(size);
				returnValue        = nullptr != result.toBeginning;

			}

			file->close();
		}

		return returnValue;
	}

	void unmap(MappedFile_readWrite const & map) {
		if(nullptr != map.toBeginning) {
			OS::Win32::unmap(map.toBeginning);
		}
	}


} // Heap


void * C_CALL operator new(USize size) {
	if(auto const toNewMemory = OS::Win32::tryToReserve(size)) {
		return toNewMemory;
	} else {
		throw 0;
	}
}

void * C_CALL operator new[](USize size) {
	if(auto const toNewMemory = OS::Win32::tryToReserve(size)) {
		return toNewMemory;
	} else {
		throw 0;
	}
}

void C_CALL operator delete(void * const address) noexcept {
	if(nullptr != address) { // delete handles nullptr, but Heap::release() doesnt
		Heap::release(address);
	}
}

void C_CALL operator delete[](void * address) noexcept { // CAUTION: MUST NOT BE DECLARED const; VISUAL C++ 2010 WOULDNT ACCEPT IT THEN!
	if(nullptr != address) { // delete handles nullptr, but Heap::release() doesnt
		Heap::release(address);
	}
}

void operator delete(void * address, USize) noexcept { // sized delete since C++14
	if(nullptr != address) { // delete handles nullptr, but Heap::release() doesnt
		Heap::release(address);
	}
}

void operator delete [](void * address, USize) noexcept { // sized delete since C++14
	if(nullptr != address) { // delete handles nullptr, but Heap::release() doesnt
		Heap::release(address);
	}
}

// Visual C++ 2015 pulls in type_info for each and every new, even for the placement versions. Define its destructor here in
//  order to avoid error LNK2001: unresolved external symbol "const type_info::`vftable'" with CRT replacements.
extern "C" class type_info {

	virtual ~type_info() { }

};