// The x86 REP MOVSB instruction (byte-wise copy).
void movsb(Byte *, Byte const *, UCount);
// The x86 REP MOVSW instruction (2-B-wise copy).
void movsw(UInt2B *, UInt2B const *, UCount);
// The x86 REP MOVSD instruction (4-B-wise copy).
void movsd(UInt4B *, UInt4B const *, UCount);
#if COMPILED_FOR_X86_64
	// The x86 REP MOVSQ instruction (8-B-wise copy).
	void movsq(UInt8B *, UInt8B const *, UCount);
#endif

// The x86 REP STOSB instruction (byte-wise filling).
void stosb(Byte *, Byte const, UCount);
// The x86 REP STOSW instruction (2-B-wise filling).
void stosw(UInt2B *, UInt2B const, UCount);
// The x86 REP STOSD instruction (4-B-wise filling).
void stosd(UInt4B *, UInt4B const, UCount);
#if COMPILED_FOR_X86_64
	// The x86 REP STOSQ instruction (8-B-wise filling).
	void stosq(UInt8B *, UInt8B const, UCount);
#endif


// Compares the given memory for equality.
bool equals(
	void const *,
	void const *,
	USize
);

// Copies the given memory (byte-wise if overlapping).
void copy(
	Byte *       toDestination,
	Byte const * toSource,
	Byte const * toSourcesEnd
);
void copy(
	Byte *       toDestination,
	Byte const * toSource,
	USize
);

// Copies the given object.
template <typename T> void copy(
	T &       destination,
	T const & source
) {
	return ::copy(
		reinterpret_cast<Byte *>(&destination),
		reinterpret_cast<Byte const *>(&source), sizeof(T)
	);
}

// Copies the given array (byte-wise if overlapping).
template <typename T> T * copy(
	T *       toDestination,
	T const * toSource,
	T const * toSourcesEnd
) {
	copy(
		reinterpret_cast<Byte *>(toDestination),
		reinterpret_cast<Byte const *>(toSource), reinterpret_cast<Byte const *>(toSourcesEnd)
	);
	return byteOffset(toDestination, byteDistance(toSource, toSourcesEnd));
}

// Uses the x86 REP MOVS(B/W/D/Q) instruction.
//   element-wise if overlapping
//   returns a pointer to the end of the copy
template <typename T> T * movs(
	T *       toDestination,
	T const * toSource,
	UCount    length
) {
#	if COMPILED_FOR_X86_64
		if(0 == sizeof(T) % 8) {
			movsq((UInt8B *)toDestination, (UInt8B const *)toSource, sizeof(T) / 8 * length);
		} else
#	endif
	if(0 == sizeof(T) % 4) {
		movsd((UInt4B *)toDestination, (UInt4B const *)toSource, sizeof(T) / 4 * length);
	} else if(0 == sizeof(T) % 2) {
		movsw((UInt2B *)toDestination, (UInt2B const *)toSource, sizeof(T) / 2 * length);
	} else {
		movsb((Byte *)toDestination, (Byte const *)toSource, sizeof(T) * length);
	}
	return toDestination + length;
}

// Fills the given array with the given value.
//   supports elements of 1, 2, 4 B size (and also 8 B on x86-64)
//   returns a pointer to the end of the filled region
template <typename T> T * fill(
	T *       toDestination,
	T const & source,
	UCount    length
) {
#	if COMPILED_FOR_X86_64
		if(8 == sizeof(T)) {
			stosq((UInt8B *)toDestination, (UInt8B const &)source, length);
		} else
#	endif
	if(4 == sizeof(T)) {
		stosd((UInt4B *)toDestination, (UInt4B const &)source, length);
	} else if(2 == sizeof(T)) {
		stosw((UInt2B *)toDestination, (UInt2B const &)source, length);
	} else { MUST(1 == sizeof(T));
		stosb((Byte *)toDestination, (Byte const &)source, length);
	}
	return toDestination + length;
}

// Zeroes the given memory.
void zero(void * toRegion, void * toRegionsEnd);
void zero(void *, USize);

// Zeroes the given object.
template <typename T> void zero(T & x) {
	zero(&x, sizeof x);
}

bool isZero(void const *, USize);


namespace Heap {


	// Reserves memory of the given size, in bytes.
	//   content is undefined
	//   must be release()d (even with zero size)
	//   returns nullptr on failure
	Byte * tryToReserve(USize);

	// Reserves memory of the given size, in bytes.
	//   content is zeroed
	//   must be release()d (even with zero size)
	//   returns nullptr on failure
	Byte * tryToReserveZeroed(USize);

	// Resizes the given memory block to the given size.
	//   input must not be nullptr
	//   new content is undefined
	//   must be release()d (even with zero size)
	//   on failure, returns nullptr and leaves input untouched
	Byte * tryToResize(void const * baseAddress, USize);

	void release(void const * baseAddress);

	// Reserves an instance of the given type.
	//   content is undefined
	//   must be release()d
	//   returns nullptr on failure
	template <typename T> T * tryToReserve() {
		return reinterpret_cast<T *>(tryToReserve(sizeof(T)));
	}

	// Reserves an instance of the given type.
	//   content is zeroed
	//   must be release()d
	//   returns nullptr on failure
	template <typename T> T * tryToReserveZeroed() {
		return reinterpret_cast<T *>(tryToReserveZeroed(sizeof(T)));
	}

	// Reserves the given number of instances of the given type.
	//   content is undefined
	//   must be release()d (even with zero length)
	//   returns nullptr on failure
	template <typename T> T * tryToReserve(UCount length) {
		if(length > ~USize(0) / sizeof(T)) { // Allocation size overflow?
			return nullptr;
		}
		return reinterpret_cast<T *>(tryToReserve(length * sizeof(T)));
	}

	// Reserves the given number of instances of the given type.
	//   content is zeroed
	//   must be release()d (even with zero length)
	//   returns nullptr on failure
	template <typename T> T * tryToReserveZeroed(UCount length) {
		if(length > ~USize(0) / sizeof(T)) { // Allocation size overflow?
			return nullptr;
		}
		return reinterpret_cast<T *>(tryToReserveZeroed(length * sizeof(T)));
	}

	// Resizes the given array to the given length.
	//   input must not be nullptr
	//   new content is undefined
	//   must be release()d (even with zero size)
	//   on failure, returns no and leaves input untouched
	template <typename T> bool tryToResizeArray(T * & array, UCount newLength) {
		if(newLength > ~USize(0) / sizeof(T)) { // Allocation size overflow?
			return no;
		}
		if(auto toNew = tryToResize(array, newLength * sizeof(T))) {
			array = reinterpret_cast<T *>(toNew);
			return yes;
		}
		return no;
	}

	// Reserves an empty range with sufficient room for the given number of elements.
	//   content is undefined
	//   must be release()d (even with zero length)
	//   returns a nullptr beginning on failure
	template <typename T> Range<T> tryToReserveRange(UCount length) {
		auto result = tryToReserve<T>(length);
		return { result, result };
	}

	// Reserves a range with the given number of elements.
	//   content is undefined
	//   must be release()d (even with zero length)
	//   returns an empty range with nullptr beginning on failure
	template <typename T> Range<T> tryToReserveFullRange(UCount length) {
		if(length > ~USize(0) / sizeof(T)) { // Allocation size overflow?
			return { nullptr, nullptr };
		} else {
			auto size   = length * sizeof(T);
			auto result = reinterpret_cast<T *>(tryToReserve(size));
			return { result, byteOffset(result, size) };
		}
	}

	// Reserves a range with the given number of elements.
	//   content is zeroed
	//   must be release()d (even with zero length)
	//   returns an empty range with nullptr beginning on failure
	template <typename T> Range<T> const tryToReserveZeroedRange(UCount length) {
		if(length > ~USize(0) / sizeof(T)) { // Allocation size overflow?
			return { nullptr, nullptr };
		} else {
			auto size   = length * sizeof(T);
			auto result = reinterpret_cast<T *>(tryToReserveZeroed(size));
			return { result, byteOffset(result, size) };
		}
	}

	// Creates a copy of the given range on the default heap.
	//   returns an empty range with nullptr beginning on failure
	template <typename T> Range<T> copy(Range<T> const & r) {
		auto size   = sizeOf(r);
		auto result = reinterpret_cast<T *>(tryToReserve(size));
		if(nullptr == result) {
			return { nullptr, nullptr };
		} else {
			::copy((Byte*)result, (Byte*)r.toBeginning, size);
			return { result, byteOffset(result, size) };
		}
	}

	//  works even with altered end
	template <typename T> void release(Range<T> const & range) {
		release(range.toBeginning);
	}


	//  R E G I O N S :

	// A memory pool / arena / region for region-based memory management.
	// Regions combine the advantages of manual memory management with garbage collection: all memory from a region is released
	//  when the region is destroyed. This is highly efficient when allocations are grouped in regions by expected lifetime.
	struct Region : Opaque {

		// Reserves memory of the given size, in bytes.
		//   on failure
		//     reports the error to the given supervisor (pass nullptr to suppress)
		//     returns nullptr
		//   on success
		//     returns the memory with undefined content
		//   checks for array size overflow
		OUTLINE_CODE Byte * tryToReserveRaw(
			USize              elementSize,
			UCount             numberOfElements = 1,
			Supervisor const * supervisorOrNull = nullptr
		);

		// Reserves memory of the given size, in bytes.
		//   content is zeroed
		//   returns nullptr on failure
		Byte * tryToReserveZeroed(USize);

		// Resizes the given memory block to the given size.
		//   input must not be nullptr
		//   on failure
		//     reports the error to the given supervisor (pass nullptr to suppress)
		//     does not touch the original block
		//     returns nullptr
		//   on success
		//     returns the memory with new content being undefined
		//   checks for array size overflow
		OUTLINE_CODE Byte * tryToResizeRaw(
			void const *       baseAddress,
			USize              elementSize,
			UCount             numberOfElements,
			Supervisor const * supervisorOrNull = nullptr
		);

		void release(void const * baseAddress);

		// Reserves an instance of the given type.
		//   on failure
		//     reports the error to the given supervisor (pass nullptr to suppress)
		//     returns nullptr
		//   on success
		//     returns the memory with undefined content
		template <typename T> T * tryToReserve(
			Supervisor const * supervisorOrNull = nullptr
		) {
			return reinterpret_cast<T *>(tryToReserveRaw(sizeof(T), 1, supervisorOrNull));
		}

		// Reserves the given number of instances of the given type.
		//   on failure
		//     reports the error to the given supervisor (pass nullptr to suppress)
		//     returns nullptr
		//   on success
		//     returns the memory with undefined content
		//   checks for array size overflow
		template <typename T> T * tryToReserve(
			UCount const       length,
			Supervisor const * supervisorOrNull = nullptr
		) {
			return reinterpret_cast<T *>(tryToReserveRaw(sizeof(T), length, supervisorOrNull));
		}

		// Creates a copy of the given POD values.
		//   on failure
		//     reports the error to the given supervisor (pass nullptr to suppress)
		//     returns nullptr
		//   on success
		//     returns the memory with and copies the original content
		template <typename T> T * tryToCopy(
			T const *          originals,
			UCount const       length,
			Supervisor const * supervisorOrNull = nullptr
		) {
			if(auto const result = reinterpret_cast<T *>(tryToReserveRaw(sizeof(T), length, supervisorOrNull))) {
				movs(result, originals, length);
				return result;
			}
			return nullptr;
		}

		// Reserves the given number of instances of the given type.
		//   on failure
		//     reports the error to the given supervisor (pass nullptr to suppress)
		//     returns no
		//   on success
		//     returns the memory with undefined content
		//   checks for array size overflow
		template <typename T> bool tryToReserve(
			T * &              result,
			UCount const       length,
			Supervisor const * supervisorOrNull = nullptr
		) {
			result = reinterpret_cast<T *>(tryToReserveRaw(sizeof(T), length, supervisorOrNull));
			return nullptr != result;
		}

		// Reserves an empty range with sufficient room for the given number of elements.
		//   content is undefined
		//   returns a nullptr beginning on failure
		template <typename T> Range<T> tryToReserveRange(
			UCount const             length,
			Supervisor const * const supervisorOrNull = nullptr
		) {
			auto result = tryToReserve<T>(length, supervisorOrNull);
			return { result, result };
		}

		// Resizes the given memory block to the given size.
		//   input must not be nullptr
		//   on failure
		//     reports the error to the given supervisor (pass nullptr to suppress)
		//     does not touch the original block
		//     returns no
		//   on success
		//     returns the memory with new content being undefined
		//   checks for array size overflow
		template <typename T> bool tryToResize(
			T * &              result,
			UCount const       newLength,
			Supervisor const * supervisorOrNull = nullptr
		) {
			if(newLength > ~USize(0) / sizeof(T)) { // Allocation size overflow?
				return no;
			}
			if(auto const toNew = tryToResizeRaw(result, sizeof(T), newLength, supervisorOrNull)) {
				result = reinterpret_cast<T *>(toNew);
				return yes;
			}
			return no;
		}

		// Resizes the given array to the given new number of instances, or reserves them if nullptr.
		//   pass nullptr for initial allocation
		//   on failure
		//     reports the error to the given supervisor (pass nullptr to suppress)
		//     does not touch the original block
		//     returns no
		//   on success
		//     returns the memory with new content being undefined
		//   checks for array size overflow
		template <typename T> bool tryToResizeOrReserve(
			T * &              result,
			UCount const       length,
			Supervisor const * supervisorOrNull = nullptr
		) {
			if(nullptr == result) { // First allocation?
				return tryToReserve(result, length, supervisorOrNull);
			} else { // Subsequent allocation?
				return tryToResize(result, length, supervisorOrNull);
			}
		}

		//  works even with altered end
		template <typename T> void release(Range<T> const & range) {
			release(range.toBeginning);
		}

		// Destroys the region and all memory in it.
		void destroy();

	};

	Region & defaultPool();

	//  on failure, returns nullptr and reports the error
	//  otherwise, destroy() after use
	Region * tryToCreateRegion(Supervisor const *);

	// Reserves an empty range in the given region with sufficient room for the given number of elements.
	//   content is zeroed
	//   returns an a nullptr beginning on failure
	template <typename T> Range<T> const tryToReserveZeroedRange(Region & region, UCount length) {
		if(length > ~USize(0) / sizeof(T)) { // Allocation size overflow?
			return { nullptr, nullptr };
		} else {
			auto result = reinterpret_cast<T *>(region.tryToReserveZeroed(length * sizeof(T)));
			return { result, result };
		}
	}

	// Reserves a range in the given region with the given number of elements.
	//   content is undefined
	//   returns an empty range with nullptr beginning on failure
	template <typename T> Range<T> tryToReserveFullRange(Region & region, UCount length) {
		if(auto const base = region.tryToReserve<T>(length)) {
			return range(base, length);
		} {
			return { nullptr, nullptr };
		}
	}

	// Resizes the given array to the given length.
	//   input must not be nullptr
	//   new content is undefined
	//   must be release()d (even with zero size)
	//   on failure, returns no and leaves input untouched
	template <typename T> bool tryToResizeArray(Region & region, T * & array, UCount newLength) {
		if(newLength > ~USize(0) / sizeof(T)) { // Allocation size overflow?
			return no;
		}
		if(auto toNew = tryToResize(region, array, newLength * sizeof(T))) {
			array = reinterpret_cast<T *>(toNew);
			return yes;
		}
		return no;
	}

	//  works even with altered end
	template <typename T> void release(Region & region, Range<T> const & range) {
		region.release(range.toBeginning);
	}


	//  M E M O R Y   M A P P I N G :

	struct MappedFile_readOnly {
		Byte const * toBeginning;
		USize        size;
	};

	// Maps the file at the given path to read-only memory.
	//   on failure, returns no and reports the error
	//   unmap() after use
	//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
	bool tryToMapFileForReading(
		MappedFile_readOnly &  result,
		NativeCodeUnit const * path,
		Supervisor const *
	);

	void unmap(MappedFile_readOnly const &);


	struct MappedFile_readWrite {
		Byte * toBeginning;
		USize  size;
	};

	// Maps the file at the given path to memory.
	//   on failure, returns no and reports the error
	//   unmap() after use
	//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
	bool tryToMapFileForWriting(
		MappedFile_readWrite &  result,
		NativeCodeUnit const *  path,
		Supervisor const *
	);

	void unmap(MappedFile_readWrite const &);


} // Heap