// Namco's data compression method in Ace Combat 3, Ace Combat 5, and Time Crisis.



#include "framework.hpp"
#include "Memory.hpp"
#include "Namco ULZ.hpp"



namespace Namco {


	namespace ULZ {


		// Checks whether the given bytes mark the beginning of an ULZ stream.
		//   pass at least 20 bytes
		bool isULZ(Byte const data[], USize const dataSize) {
			return 20 < dataSize && reinterpret_cast<UInt4B_LE const *>(data)[0] == bigEndian(0x556C7A1Au);
		}

		// Returns the size of the extracted data. ULZ supports file sizes up to 24 MiB.
		UInt4B extractedSizeOf(Header const & header) {
			return littleEndian(header.typeAndSize) & 0xFF'FFFF;
		}

		// Extracts the given ULZ to the given address.
		//   returns no on failure
		//   requires sufficient space  use extractedSizeOf()
		bool extract(
			Byte *             toOutput,
			Byte const * const toInput,
			USize const        inputSize
		) {
			MUST(isULZ(toInput, inputSize));

			auto &     header     = *reinterpret_cast<Header const *>(toInput);
			auto const type       = littleEndian(header.typeAndSize) >> 24;
			auto const offsetBits = littleEndian(header.u_pos) >> 24u;
			auto const offsetMask = (1u << offsetBits) - 1;
			{
				// The sliding window cannot be larger than 2^15 B (because offset-length pairs are 16-bit integers):
				if(offsetBits > 15) {
					INVOKE_DEBUGGER; // TODO: error: bad window
					return no;
				}
				// All tables must be located within the file.
				// TODO
			}

			// There is one stream of new symbols:
			auto toUncompressedData = toInput + (littleEndian(header.u_pos) & 0x00FF'FFFF);
			// There is one stream of lookups (offset + length) into already-decompressed data:
			auto toCompressedData   = reinterpret_cast<UInt2B const *>(toInput + littleEndian(header.c_pos));
			// There is one stream of 32-bit flags, where each bit determines whether the next symbol is known (offset + length):
			auto toFlags            = header.flags;
			auto bytesLeft          = extractedSizeOf(header);
			if(0 == bytesLeft) {
				return yes; // nothing to do (never happens with real files)
			}

			// Both type 0 and type 2 consume bits (starting at the most significant position), but type 1 skips every 32nd bit:
			UInt4B flagMaskIfConsumed;
			switch(type) {
				case 0: flagMaskIfConsumed = 1; break;
				case 2: flagMaskIfConsumed = 0; break;
				default: INVOKE_DEBUGGER; return no; // TODO: error: bad type
			}
			UInt4B flags;
			UInt4B flagMask = flagMaskIfConsumed; // force refill on first iteration
			do {

				if(flagMaskIfConsumed == flagMask) { // Need to refill flags?
					flags    = littleEndian(*toFlags++);
					flagMask = 0x8000'0000u;
				}

				if(flags & flagMask) {

					// Copy an uncompressed byte:
					*toOutput++ = *toUncompressedData++;
					--bytesLeft;

				} else {

					// Duplicate length bytes starting backwards at offset in the result.
					auto const offsetAndLength = UInt4B(littleEndian(*toCompressedData++));
					auto const offset          = offsetAndLength & offsetMask;
					auto const length          = 3 + (offsetAndLength >> offsetBits);
					if(length > bytesLeft) {
						INVOKE_DEBUGGER; // TODO: error: overrun
						return no;
					}

					movsb(toOutput, toOutput - 1 - offset, length);
					toOutput += length;
					bytesLeft -= length;

				}

				flagMask >>= 1;
			} while(bytesLeft);

			return yes;
		}


	} // ULZ


} // Namco