// Helpers for Windows NT files & directories.


#include "framework.hpp"
#include "OS NT.hpp"
#include "OS NT files.hpp"
#include "Text.hpp" // for path processing
#include "errors/OS Win32.hpp" // TODO: make own error codes



import_from(ntdll) {

	struct IO_STATUS_BLOCK {
		OS::NT::Status Status;
		UInt4B         padding;
		USize          Information;
	};

	struct UNICODE_STRING_const {
		UInt2B            Length;
		UInt2B            MaximumLength;
		PAD_4B_ON_X86_64;
		UTF16Unit const * Buffer;
	};

	struct OBJECT_ATTRIBUTES {
		UInt4B                       Length;
		PAD_4B_ON_X86_64;
		OS::NT::Directory const *    RootDirectory;
		UNICODE_STRING_const const * ObjectName;
		UInt4B                       Attributes;
		PAD_4B_ON_X86_64;
		void *                       SecurityDescriptor;
		void *                       SecurityQualityOfService;
	};

	OS::NT::Status API_CALL NtCreateFile(
		OS::NT::File * *   FileHandle,
		UInt4B             DesiredAccess,
		OBJECT_ATTRIBUTES *,
		IO_STATUS_BLOCK *  ,
		UInt8B *           AllocationSize,
		UInt4B             FileAttributes,
		UInt4B             ShareAccess,
		UInt4B             CreateDisposition,
		UInt4B             CreateOptions,
		void *             EaBuffer,
		UInt4B             EaLength
	);

	struct FILE_STANDARD_INFORMATION {
		UInt8B AllocationSize;
		UInt8B EndOfFile;
		UInt4B NumberOfLinks;
		UInt1B DeletePending;
		UInt1B Directory;
		PAD_2B;
	};

	OS::NT::Status API_CALL NtQueryInformationFile(
		OS::NT::File const *,
		IO_STATUS_BLOCK *   ,
		void *              FileInformation,
		UInt4B              Length,
		UInt4B              FileInformationClass
	);

	OS::NT::Status API_CALL NtWriteFile(
		OS::NT::File *   ,
		void *           Event,
		void *           ApcRoutine,
		void *           ApcContext,
		IO_STATUS_BLOCK *,
		void const *     Buffer,
		UInt4B           Length,
		UInt8B *         ByteOffset,
		UInt4B *         Key
	);

	OS::NT::Status API_CALL NtClose(void const *);

	OS::NT::Status API_CALL NtCreateSection(
		OS::NT::Section * *  SectionHandle,
		UInt4B               DesiredAccess,
		OBJECT_ATTRIBUTES *  ,
		UInt8B *             MaximumSize,
		UInt4B               SectionPageProtection,
		UInt4B               AllocationAttributes,
		OS::NT::File const *
	);

	OS::NT::Status API_CALL NtMapViewOfSection(
		OS::NT::Section * ,
		OS::NT::Process * ,
		Byte * *          BaseAddress,
		USize             ZeroBits,
		USize             CommitSize,
		UInt8B *          SectionOffset,
		USize *           ViewSize,
		UInt4B            InheritDisposition,
		UInt4B            AllocationType,
		UInt4B            Win32Protect
	);

	OS::NT::Status API_CALL NtUnmapViewOfSection(
		OS::NT::Process *,
		void const *     BaseAddress
	);

	OS::NT::Status API_CALL NtClose(void const *);


} // ntdll.dll



namespace OS {


	namespace NT {


		struct Path_const {
			UNICODE_STRING_const string;
			OBJECT_ATTRIBUTES    attributes;
		};

		local void makeWin32Path(
			Path_const &    result,
			UTF16Unit const win32Path[]
		) {
			MUST(Text::beginsWith(win32Path, "\\??\\")); // Begins with \??\?
			MUST(Text::isAlphabetical(ASCIICharacter(win32Path[4])) && ':' == win32Path[5]); // Absolute?
			MUST(Text::lengthOf(win32Path) <= 32766);

			result.string.Buffer        = win32Path;
			result.string.Length        = UInt2B(2 * Text::lengthOf(win32Path)); // Bytes, not characters!
			result.string.MaximumLength = result.string.Length + 2; // the terminating zero

			result.attributes.Length                   = sizeof result.attributes;
			result.attributes.RootDirectory            = nullptr;
			result.attributes.ObjectName               = &result.string;
			result.attributes.Attributes               = 64; // OBJ_CASE_INSENSITIVE  Win32 ignores case
			result.attributes.SecurityDescriptor       = nullptr;
			result.attributes.SecurityQualityOfService = nullptr;
		}

		local void makeWin32Name(
			Path_const &      result,
			Directory const & parent,
			UTF16Unit const   win32Name[]
		) {
			MUST(Text::lengthOf(win32Name) <= 32766);

			result.string.Buffer        = win32Name;
			result.string.Length        = UInt2B(2 * Text::lengthOf(win32Name)); // Bytes, not characters!
			result.string.MaximumLength = result.string.Length + 2; // the terminating zero

			result.attributes.Length                   = sizeof result.attributes;
			result.attributes.RootDirectory            = &parent;
			result.attributes.ObjectName               = &result.string;
			result.attributes.Attributes               = 64; // OBJ_CASE_INSENSITIVE  Win32 ignores case
			result.attributes.SecurityDescriptor       = nullptr;
			result.attributes.SecurityQualityOfService = nullptr;
		}

		// Opens an existing sub-directory.
		//   on failure, returns nullptr and reports the error
		//   close() after use
		//   name:
		//     zero-terminated
		//     shorter than 32,766 characters
		Directory * Directory::tryToOpenDirectory(
			UTF16Unit const          win32Name[],
			ObjectSharing const      sharing,
			Supervisor const * const supervisorOrNull
		) {
			// Directories are opened just like files, but with different flags.

			Path_const      path;
			IO_STATUS_BLOCK io;
			File *          result;
			makeWin32Name(path, *this, win32Name);

			if(auto const error = NtCreateFile(
				&result, 32, // FILE_TRAVERSE  the directory can be part of a path (seems to be the only required right)
				&path.attributes, &io,
				nullptr,     // pre-allocated size  useless for directories
				0,           // attributes
				UInt4B(sharing),
				1,           // FILE_OPEN  open existing directory; fail otherwise
				1,           // FILE_DIRECTORY_FILE  fail if not a directory
				nullptr, 0   // no extended attributes
			)) {
				reportError(supervisorOrNull, error);
				return nullptr;
			}

			MUST(nullptr != result); // NT bug?
			return reinterpret_cast<Directory *>(result);
		}

		// Opens a sub-directory. If not existent, creates a new one.
		//   on failure, returns nullptr and reports the error
		//   close() after use
		//   name:
		//     zero-terminated
		//     shorter than 32,766 characters
		Directory * Directory::tryToOpenOrCreateDirectory(
			UTF16Unit const          win32Name[],
			ObjectSharing const      sharing,
			Supervisor const * const supervisorOrNull
		) {
			// Directories are opened just like files, but with different flags.

			Path_const      name;
			IO_STATUS_BLOCK io;
			File *          result;
			makeWin32Name(name, *this, win32Name);

			if(auto const error = NtCreateFile(
				&result, 32, // FILE_TRAVERSE  the directory can be part of a path (seems to be the only required right)
				&name.attributes, &io,
				nullptr,     // pre-allocated size  useless for directories
				0,           // attributes
				UInt4B(sharing),
				3,           // FILE_OPEN_IF  open existing directory; create otherwise
				1,           // FILE_DIRECTORY_FILE  fail if not a directory
				nullptr, 0   // no extended attributes
			)) {
				reportError(supervisorOrNull, error);
				return nullptr;
			}

			MUST(nullptr != result); // NT bug?
			return reinterpret_cast<Directory *>(result);
		}

		// Opens an existing file.
		//   on failure, returns nullptr and reports the error
		//   close() after use
		//   name:
		//     zero-terminated
		//     shorter than 32,766 characters
		File * Directory::tryToOpenFile(
			UTF16Unit const          win32Name[],
			FileAccess const         access,
			ObjectSharing const      sharing,
			FileCaching const        caching,
			Supervisor const * const supervisorOrNull
		) {
			Path_const      name;
			IO_STATUS_BLOCK io;
			File *          result;
			makeWin32Name(name, *this, win32Name);

			if(auto const error = NtCreateFile(
				&result,
				UInt4B(access),
				&name.attributes, &io,
				nullptr,     // pre-allocated size  useless
				0,           // attributes
				UInt4B(sharing),
				1,           // FILE_OPEN  open existing file; fail otherwise
				  64         // FILE_NON_DIRECTORY_FILE  fail if not a file
				| 512        // FILE_NO_EA_KNOWLEDGE  fail if the file has complicated metadata
				| 32         // FILE_SYNCHRONOUS_IO_NONALERT  file is synchronous
				| UInt4B(caching),
				nullptr, 0   // no extended attributes
			)) {
				reportError(supervisorOrNull, error);
				return nullptr;
			}

			MUST(nullptr != result); // NT bug?
			return result;
		}

		// Clears an existing file or creates a new one.
		//   on failure, returns nullptr and reports the error
		//   close() after use
		//   name:
		//     zero-terminated
		//     shorter than 32,766 characters
		File * Directory::tryToCreateEmptyFile(
			UTF16Unit const          win32Name[],
			FileAccess const         access,
			ObjectSharing const      sharing,
			FileCaching const        caching,
			Supervisor const * const supervisorOrNull
		) {
			Path_const      name;
			IO_STATUS_BLOCK io;
			File *          result;
			makeWin32Name(name, *this, win32Name);

			if(auto const error = NtCreateFile(
				&result,
				UInt4B(access),
				&name.attributes, &io,
				nullptr,     // pre-allocated size  useless
				0,           // attributes
				UInt4B(sharing),
				5,           // FILE_OVERWRITE_IF  open & clear existing file; create empty otherwise
				  64         // FILE_NON_DIRECTORY_FILE  fail if not a file
				| 512        // FILE_NO_EA_KNOWLEDGE  fail if the file has complicated metadata
				| 32         // FILE_SYNCHRONOUS_IO_NONALERT  file is synchronous
				| UInt4B(caching),
				nullptr, 0   // no extended attributes
			)) {
				reportError(supervisorOrNull, error);
				return nullptr;
			}

			MUST(nullptr != result); // NT bug?
			return result;
		}

		void Directory::close() {
			mustSucceed(NtClose(this));
		}

		// Opens an existing directory.
		//   on failure, returns nullptr and reports the error
		//   close() after use
		//   path:
		//     zero-terminated
		//     absolute
		//     begins with \\?\
		//     shorter than 32,766 characters
		Directory * tryToOpenDirectory(
			UTF16Unit const          win32Path[],
			ObjectSharing const      sharing,
			Supervisor const * const supervisorOrNull
		) {
			// Directories are opened just like files, but with different flags.

			Path_const      path;
			IO_STATUS_BLOCK io;
			File *          result;
			makeWin32Path(path, win32Path);

			if(auto const error = NtCreateFile(
				&result, 32, // FILE_TRAVERSE  the directory can be part of a path (seems to be the only required right)
				&path.attributes, &io,
				nullptr,     // pre-allocated size  useless for directories
				0,           // attributes
				UInt4B(sharing),
				1,           // FILE_OPEN  open existing directory; fail otherwise
				1,           // FILE_DIRECTORY_FILE  fail if not a directory
				nullptr, 0   // no extended attributes
			)) {
				reportError(supervisorOrNull, error);
				return nullptr;
			}

			MUST(nullptr != result); // NT bug?
			return reinterpret_cast<Directory *>(result);
		}

		// Opens a directory. If not existent, creates a new one.
		//   on failure, returns nullptr and reports the error
		//   close() after use
		//   path:
		//     zero-terminated
		//     absolute
		//     begins with \\?\
		//     shorter than 32,766 characters
		Directory * tryToOpenOrCreateDirectory(
			UTF16Unit const          win32Path[],
			ObjectSharing const      sharing,
			Supervisor const * const supervisorOrNull
		) {
			// Directories are opened just like files, but with different flags.

			Path_const      path;
			IO_STATUS_BLOCK io;
			File *          result;
			makeWin32Path(path, win32Path);

			if(auto const error = NtCreateFile(
				&result, 32, // FILE_TRAVERSE  the directory can be part of a path (seems to be the only required right)
				&path.attributes, &io,
				nullptr,     // pre-allocated size  useless for directories
				0,           // attributes
				UInt4B(sharing),
				3,           // FILE_OPEN_IF  open existing directory; create otherwise
				1,           // FILE_DIRECTORY_FILE  fail if not a directory
				nullptr, 0   // no extended attributes
			)) {
				reportError(supervisorOrNull, error);
				return nullptr;
			}

			MUST(nullptr != result); // NT bug?
			return reinterpret_cast<Directory *>(result);
		}


		// Obtains the files size.
		//   the file must have been opened with the readAttributes access right
		UInt8B File::size() const {
			IO_STATUS_BLOCK           io;
			FILE_STANDARD_INFORMATION info;
			mustSucceed(NtQueryInformationFile(this, &io, &info, sizeof info, 5)); // FileStandardInformation
			return info.EndOfFile;
		}

		// Writes the given number of bytes at the given offset to the file.
		//   on failure, returns no and reports the error
		//   size must be < 2 GiB
		bool File::tryToWrite(
			UInt8B                   offset,
			void const * const       data,
			USize const              size,
			Supervisor const * const supervisorOrNull
		) {
			MUST(size <= 0x7FFFFFFF);

			IO_STATUS_BLOCK io;
			if(auto const error = NtWriteFile(
				this,
				nullptr, nullptr, nullptr, // no event on completion
				&io, data, UInt4B(size), &offset,
				nullptr // no unlocking of locked regions
			)) {
				reportError(supervisorOrNull, error);
				return no;
			}

			return yes;
		}

		// Writes the given number of bytes to the file.
		//   on failure, returns no and reports the error
		//   size must be < 2 GiB
		bool File::tryToWrite(
			void const * const       data,
			USize const              size,
			Supervisor const * const supervisorOrNull
		) {
			MUST(size <= 0x7FFFFFFF);

			IO_STATUS_BLOCK io;
			if(auto const error = NtWriteFile(
				this,
				nullptr, nullptr, nullptr, // no event on completion
				&io, data, UInt4B(size), nullptr,
				nullptr // no unlocking of locked regions
			)) {
				reportError(supervisorOrNull, error);
				return no;
			}

			return yes;
		}

		void File::close() {
			mustSucceed(NtClose(this));
		}


		// Maps the given file to read-only memory.
		//   on failure, returns nullptr and reports the error
		//   unmap() after use
		//   the file must 
		//     not be empty
		//     be open with the read access right
		//   needs the files size because the Kernel doesnt track it
		Byte const * tryToMapForReading(
			File const &             file,
			Supervisor const * const supervisorOrNull
		) {
			Byte const * result = nullptr;

			// File mappings are internally realized as sections. Create a section for the file:
			Section * sectionHandle;
			auto const sectionStatus = NtCreateSection(
				&sectionHandle,
				4,           // SECTION_MAP_READ
				nullptr,     // unnamed
				nullptr,     // map the entire file
				2,           // PAGE_READONLY
				0x0800'0000, // SEC_COMMIT  reserved sections are for expanding shared memory
				&file
			);
			if(0 <= sectionStatus) {
				MUST(nullptr != sectionHandle); // NT bug?

				// Map the section to virtual memory:
				Byte * baseAddress = nullptr; // let the Kernel choose an appropriate address
				USize  size        = 0; // map the entire section
				auto const mapStatus = NtMapViewOfSection(
					sectionHandle, thisProcess,
					&baseAddress, 0,  // let the system choose proper alignment (will always be at least page-aligned)
					0,                // unused (affects only page file sections)
					nullptr, &size,   // map the entire section
					2,                // ViewUnmap
					0,                // MEM_COMMIT is implied, but passing it explicitly is illegal
					2                 // PAGE_READONLY
				);
				if(0 <= mapStatus) {
					MUST(nullptr != baseAddress); // NT bug?

					result = baseAddress;

				} else {
					reportError(supervisorOrNull, mapStatus);
				}

				// The section handle is useless now and can be released (the kernel is reference-counted):
				mustSucceed(NtClose(sectionHandle));
			} else {
				reportError(supervisorOrNull, sectionStatus);
			}

			return result;
		}

		//  works for any address within a mapped file
		//  for read-writeable maps, does not update the files modification date (due to lazy writing)!
		void unmap(void const * mappedFile) {
			mustSucceed(NtUnmapViewOfSection(thisProcess, mappedFile));
		}


	} // NT


} // OS