// Helpers for Windows NT files & directories.



namespace OS {


	namespace NT {


		enum class FileAccess : UInt4B {
			readOnly            = 0x120089, // FILE_GENERIC_READ
			writeOnly           = 0x120116, // FILE_GENERIC_WRITE
			appendOnly          = 4,        // can write at file end, but cannot overwrite existing data (FILE_APPEND_DATA)
			readWrite           = 0x12019F, // FILE_GENERIC_READ | FILE_GENERIC_WRITE
			readAttributesOnly  = 128,      // no cache pollution with actual file data (FILE_READ_ATTRIBUTES)
			writeAttributesOnly = 256,      // no cache pollution with actual file data (FILE_WRITE_ATTRIBUTES)
		};

		// Controls how other processes can access an open file, directory, device, 
		enum class ObjectSharing : UInt4B {
			none     = 0, // other processes cannot access the object
			readOnly = 5, // other processes can read from the object or delete it (FILE_SHARE_READ | FILE_SHARE_DELETE)
			all      = 7, // other processes have full access to the object (FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE)
			dontCare = readOnly
		};

		// Controls how a file is cached by the Memory Manager.
		//   does not affect memory-mapped files
		//   details: The Old New Thing, How do FILE_FLAG_SEQUENTIAL_SCAN and FILE_FLAG_RANDOM_ACCESS affect how the operating
		//    system treats my file?
		enum class FileCaching : UInt4B {
			sequential = 4,    // maximal prefetching; evict-behind (FILE_SEQUENTIAL_ONLY)
			adaptive   = 0,    // observe the access pattern and optimize dynamically
			random     = 2048, // no prefetching; evict-behind (FILE_RANDOM_ACCESS)
			dontCare   = adaptive
		};


		struct alignas(handleAlignment) Directory : Opaque {

			// Opens an existing sub-directory.
			//   on failure, returns nullptr and reports the error
			//   close() after use
			//   name:
			//     zero-terminated
			//     shorter than 32,766 characters
			Directory * tryToOpenDirectory(
				UTF16Unit const   win32Name[],
				ObjectSharing     ,
				Supervisor const *
			);

			// Opens a sub-directory. If not existent, creates a new one.
			//   on failure, returns nullptr and reports the error
			//   close() after use
			//   name:
			//     zero-terminated
			//     shorter than 32,766 characters
			Directory * tryToOpenOrCreateDirectory(
				UTF16Unit const   win32Name[],
				ObjectSharing     ,
				Supervisor const *
			);

			// Opens an existing file.
			//   on failure, returns nullptr and reports the error
			//   close() after use
			//   name:
			//     zero-terminated
			//     shorter than 32,766 characters
			File * tryToOpenFile(
				UTF16Unit const   win32Name[],
				FileAccess        ,
				ObjectSharing     ,
				FileCaching       ,
				Supervisor const *
			);

			// Clears an existing file or creates a new one.
			//   on failure, returns nullptr and reports the error
			//   close() after use
			//   name:
			//     zero-terminated
			//     shorter than 32,766 characters
			File * tryToCreateEmptyFile(
				UTF16Unit const   win32Name[],
				FileAccess        ,
				ObjectSharing     ,
				FileCaching       ,
				Supervisor const *
			);

			void close();

		};


		// Opens an existing directory.
		//   on failure, returns nullptr and reports the error
		//   close() after use
		//   path:
		//     zero-terminated
		//     absolute
		//     begins with \\?\
		//     shorter than 32,766 characters
		Directory * tryToOpenDirectory(
			UTF16Unit const   win32Path[],
			ObjectSharing     ,
			Supervisor const *
		);

		// Opens a directory. If not existent, creates a new one.
		//   on failure, returns nullptr and reports the error
		//   close() after use
		//   path:
		//     zero-terminated
		//     absolute
		//     begins with \\?\
		//     shorter than 32,766 characters
		Directory * tryToOpenOrCreateDirectory(
			UTF16Unit const   win32Path[],
			ObjectSharing     ,
			Supervisor const *
		);


		struct alignas(handleAlignment) File : Opaque {

			// Obtains the files size.
			//   the file must have been opened with the readAttributes access right
			UInt8B size() const;

			// Writes the given number of bytes at the given offset to the file.
			//   on failure, returns no and reports the error
			//   size must be < 2 GiB
			bool tryToWrite(
				UInt8B            offset,
				void const *      ,
				USize             size,
				Supervisor const *
			);

			// Writes the given number of bytes to the file.
			//   on failure, returns no and reports the error
			//   size must be < 2 GiB
			bool tryToWrite(
				void const *      ,
				USize             size,
				Supervisor const *
			);

			void close();

		};


		// Maps the given file to read-only memory.
		//   on failure, returns nullptr and reports the error
		//   unmap() after use
		//   the file must 
		//     not be empty
		//     be open with the read access right
		//   needs the files size because the Kernel doesnt track it
		Byte const * tryToMapForReading(
			File const &      ,
			Supervisor const *
		);

		//  works for any address within a mapped file
		//  for read-writeable maps, does not update the files modification date (due to lazy writing)!
		void unmap(void const *);


	} // NT


} // OS