// Helpers for Windows debugging (call stacks, symbol names, ).



#include "framework.hpp"
#include "OS Win32.hpp"
#include "OS Win32 DbgHelp.hpp"

import_from(Kernel32) {

	OS::Win32::Process * API_CALL GetCurrentProcess();

	UInt2B API_CALL RtlCaptureStackBackTrace(
		UInt4B         FramesToSkip,
		UInt4B         FramesToCapture,
		void const * * BackTrace,
		UInt4B *       BackTraceHash
	);

	void API_CALL OutputDebugStringA(ASCIICharacter const *);
	void API_CALL OutputDebugStringW(UTF16Unit const *);

} // Kernel32

import_from(Dbghelp) {

	struct SYMBOL_INFO {
		UInt4B         SizeOfStruct;
		UInt4B         TypeIndex;
		UInt8B         Reserved[2];
		UInt4B         Index;
		UInt4B         Size;
		UInt8B         ModBase;
		UInt4B         Flags;
		PAD_4B;
		UInt8B         Value;
		UInt8B         Address;
		UInt4B         Register;
		UInt4B         Scope;
		UInt4B         Tag;
		UInt4B         NameLen;
		UInt4B         MaxNameLen;
		ASCIICharacter Name[1];
		PAD_3B;
	};

	struct IMAGEHLP_LINE64 {
		UInt4B            SizeOfStruct;
		PAD_4B_ON_X86_64;
		void *            Key;
		UInt4B            LineNumber;
		PAD_4B_ON_X86_64;
		UTF16Unit const * FileName;
		UInt8B            Address;
	};

	int API_CALL SymInitialize(
		OS::Win32::Process *,
		UTF16Unit const * * UserSearchPath,
		SInt4B              fInvadeProcess
	);

	int API_CALL SymFromAddr(
		OS::Win32::Process *,
		UInt8B              Address,
		UInt8B *            Displacement,
		SYMBOL_INFO *
	);

	int API_CALL SymGetLineFromAddrW64( // also available on 32-bit
		OS::Win32::Process *,
		UInt8B              dwAddr,
		UInt4B *            pdwDisplacement,
		IMAGEHLP_LINE64 *
	);

} // Dbghelp



// DbgHelp constantly asks for the processs handle, so store it here. Fill it when loading debug information.
local OS::Win32::Process * thisProcess;

// Loads debug information for this process. This function must be called before any DbgHelp routine.
local void useDebugInformation() {
	static bool isLoaded = no;
	if(no == isLoaded) {
		thisProcess = GetCurrentProcess();

		SymInitialize(
			thisProcess,
			nullptr, // search the working directory, the executable directory, etc.
			1        // load for all loaded modules
		);

		isLoaded = yes;
	}
}


namespace OS {


	namespace Win32 {


		namespace DbgHelp {


			// Retrieves the current call stack. The upper frames can be skipped; this should be at least one (to ignore the
			//  currentCallStack() call).
			CallStack currentCallStack(UIntLeast1B howManyCalleesToSkip) {
				CallStack result = { }; // zero unused slots

				RtlCaptureStackBackTrace(howManyCalleesToSkip, UInt4B(capacityOf(result.callees)), result.callees, nullptr);

				return result;
			}

			// Returns the name of the symbol at the given address in the modules of the current process.
			SymbolName nameOfSymbolAt(void const * address) {
				useDebugInformation(); // requires debug symbols

				// SymFromAddr() appends the name to an internal structure. Place a character array there.
				struct {
					SYMBOL_INFO    data;
					ASCIICharacter name[256];
				} information;

				information.data.MaxNameLen   = 4 + 256; // use padding after data
				information.data.SizeOfStruct = sizeof information.data;
				if(1 == SymFromAddr(thisProcess, UInt8B(address), 0, &information.data)) {
					// SymFromAddr() does not append a terminating zero if the symbol name exceeds the buffer length.
					information.name[255] = 0;
				} else {
					information.data.Name[0] = 0;
				}

				return *reinterpret_cast<SymbolName const *>(information.data.Name);
			}

			// Returns the source file path and the line number of the symbol at the given address in the current process. If an error
			//  occured, the path is nullptr and the line is zero.
			SourceLink sourceOfSymbolAt(void const * address) {
				SourceLink result;

				useDebugInformation(); // requires debug symbols

				IMAGEHLP_LINE64 information;
				information.SizeOfStruct = sizeof information;
				UInt4B displacement = 0;
				if(1 == SymGetLineFromAddrW64(thisProcess, UInt8B(address), &displacement, &information)) {
					result.toZeroTerminatedPath = information.FileName;
					result.line                 = information.LineNumber;
				} else {
					result.toZeroTerminatedPath = nullptr;
					result.line                 = 0;
				}

				return result;
			}

			// Prints the given call stack as source file and line number to the debuggers output window. With Visual Studio, the
			//  output can be double-clicked to open the corresponding file.
			void print(CallStack const & callStack) {
				for(auto toAddress = toBeginningOf(callStack.callees); toAddress < toEndOf(callStack.callees); ++toAddress) {

					// Do not print below main().
					auto const & name = nameOfSymbolAt(*toAddress).zeroTerminated;
					if('c' == name[0] && 'r' == name[1] && 't' == name[2] && 'M' == name[3]) { // ain  sure enough
						break;
					}

					printSymbolAt(*toAddress);
				}
			}

			// Prints the source file path, line number, and name of the symbol at the given address.
			void printSymbolAt(void const * address) {
				useDebugInformation(); // requires debug symbols

				IMAGEHLP_LINE64 source;
				source.SizeOfStruct = sizeof source;
				UInt4B displacement = 0;
				if(1 == SymGetLineFromAddrW64(thisProcess, UInt8B(address), &displacement, &source)) {
					// See MSBuild Team Blog, MSBuild / Visual Studio aware error messages and message formats.

					// File path:
					OutputDebugStringA("    ");
					OutputDebugStringW(source.FileName);

					// Line number:
					ASCIICharacter chars[9] = { " (" };
					auto line = source.LineNumber;
					if(line > 100000) {
						chars[2] = '0' + line / 100000 % 10;
						chars[3] = '0' + line / 10000 % 10;
						chars[4] = '0' + line / 1000 % 10;
						chars[5] = '0' + line / 100 % 10;
						chars[6] = '0' + line / 10 % 10;
						chars[7] = '0' + line / 1 % 10;
					} else if(line > 10000) {
						chars[2] = '0' + line / 10000 % 10;
						chars[3] = '0' + line / 1000 % 10;
						chars[4] = '0' + line / 100 % 10;
						chars[5] = '0' + line / 10 % 10;
						chars[6] = '0' + line / 1 % 10;
					} else if(line > 1000) {
						chars[2] = '0' + line / 1000 % 10;
						chars[3] = '0' + line / 100 % 10;
						chars[4] = '0' + line / 10 % 10;
						chars[5] = '0' + line / 1 % 10;
					} else if(line > 100) {
						chars[2] = '0' + line / 100 % 10;
						chars[3] = '0' + line / 10 % 10;
						chars[4] = '0' + line / 1 % 10;
					} else if(line > 10) {
						chars[2] = '0' + line / 10 % 10;
						chars[3] = '0' + line / 1 % 10;
					} else {
						chars[2] = '0' + line / 1 % 10;
					}
					OutputDebugStringA(chars);
					OutputDebugStringA("): ");

				} else {
					OutputDebugStringA("    (unknown source): ");
				}

				auto const name = nameOfSymbolAt(address);
				if(0 == name.zeroTerminated[0]) {
					OutputDebugStringA("(unknown name)\n");
				} else {
					OutputDebugStringA(name.zeroTerminated);
					OutputDebugStringA("\n");
				}
			}


		} // DbgHelp


	} // Win32


} // OS