// Win32 file helpers.



#include "framework.hpp"
#include "Text.hpp"
#include "OS Win32.hpp"
#include "OS Win32 File.hpp"


import_from(Kernel32) {

	UInt4B API_CALL GetLastError();

	int API_CALL GetFileSizeEx(OS::Win32::File const *, UInt8B *);

	int API_CALL GetFileTime(
		OS::Win32::File const *,
		OS::Win32::FileDate *  lpCreationTime,
		OS::Win32::FileDate *  lpLastAccessTime,
		OS::Win32::FileDate *  lpLastWriteTime
	);

	int API_CALL SetFileTime(
		OS::Win32::File *           ,
		OS::Win32::FileDate const * lpCreationTime,
		OS::Win32::FileDate const * lpLastAccessTime,
		OS::Win32::FileDate const * lpLastWriteTime
	);

	void API_CALL GetSystemTimeAsFileTime(OS::Win32::FileDate *);

	struct SECURITY_ATTRIBUTES {
		UInt4B nLength;
		PAD_4B_ON_X86_64;
		void * lpSecurityDescriptor;
		int    bInheritHandle;
		PAD_4B_ON_X86_64;
	};

	OS::Win32::File * API_CALL CreateFileW(
		UTF16Unit const *           lpFileName,
		UInt4B                      dwDesiredAccess,
		UInt4B                      dwShareMode,
		SECURITY_ATTRIBUTES const * lpSecurityAttributes,
		UInt4B                      dwCreationDisposition,
		UInt4B                      dwFlagsAndAttributes,
		OS::Win32::File *           hTemplateFile
	);

	struct OVERLAPPED {
		USize  Internal;
		USize  InternalHigh;
		union {
			UInt8B Offset;
			void * Pointer;
		};
		void * hEvent;
	};

	int API_CALL ReadFile(
		OS::Win32::File const *,
		void const *           ,
		UInt4B                 nNumberOfBytesToRead,
		UInt4B *               lpNumberOfBytesRead,
		OVERLAPPED *
	);

	int API_CALL WriteFile(
		OS::Win32::File *,
		void const *     ,
		UInt4B           nNumberOfBytesToWrite,
		UInt4B *         lpNumberOfBytesWritten,
		OVERLAPPED *
	);

	int API_CALL CloseHandle(void *);

	OS::Win32::File * API_CALL CreateFileMappingW(
		OS::Win32::File const *,
		void *                 lpFileMappingAttributes,
		UInt4B                 flProtect,
		UInt4B                 dwMaximumSizeHigh,
		UInt4B                 dwMaximumSizeLow,
		UTF16Unit const *      lpName
	);

	Byte * API_CALL MapViewOfFile(
		OS::Win32::File * hFileMappingObject,
		UInt4B            dwDesiredAccess,
		UInt4B            dwFileOffsetHigh,
		UInt4B            dwFileOffsetLow,
		USize             dwNumberOfBytesToMap
	);

	int API_CALL UnmapViewOfFile(void const *);

	int API_CALL CreateDirectoryW(UTF16Unit const *, SECURITY_ATTRIBUTES *);

	int API_CALL CopyFileW(
		UTF16Unit const * lpExistingFileName,
		UTF16Unit const * lpNewFileName,
		int               bFailIfExists
	);

	int API_CALL MoveFileW(
		UTF16Unit const * lpExistingFileName,
		UTF16Unit const * lpNewFileName
	);

	int API_CALL DeleteFileW(UTF16Unit const *);

	int API_CALL RemoveDirectoryW(UTF16Unit const *);

	UInt4B API_CALL GetTempPathW(UInt4B nBufferLength, UTF16Unit *);

	UInt4B API_CALL GetTempFileNameW(
		UTF16Unit const * lpPathName,
		UTF16Unit const * lpPrefixString,
		UInt4B            uUnique,
		UTF16Unit *       lpTempFileName
	);

	UInt4B API_CALL GetFileAttributesW(UTF16Unit const *);

	int API_CALL SetFileAttributesW(UTF16Unit const *, UInt4B);

	UCount API_CALL FindFirstFileW(UTF16Unit const *, void *);

	int API_CALL FindNextFileW(UCount, void *);

	int API_CALL FindClose(UCount);

} // Kernel32

import_from(Shell32) {

	SInt4B API_CALL SHGetFolderPathW(
		void *      hwnd,
		int         csidl,
		void *      hToken,
		UInt4B      dwFlags,
		UTF16Unit *
	);

} // Shell32



namespace OS {


	namespace Win32 {


		auto const nullFile = reinterpret_cast<File *>(-1); // INVALID_HANDLE_VALUE


		FileDate now_asFileDate() {
			FileDate out;
			GetSystemTimeAsFileTime(&out);
			return out;
		}

		UInt8B File::size() const {
			UInt8B result;
			WIN32_SHOULD_SUCCEED(GetFileSizeEx(this, &result));
			return result;
		}

		// Copies bytes from the file to the given buffer.
		//   non-const because it uses the file pointer
		//   returns the requested size on success
		//   otherwise, reports errors to the given supervisor and returns the number of bytes actually read
		USize File::tryToRead(
			void *                   toDestination,
			USize                    remainingSize,
			Supervisor const * const supervisorOrNull
		) {
			USize totalBytesRead = 0;

			// Depending on the underlying storage, short reads might occur: ReadFile() returning fewer bytes than requested. This
			//  is not an error (it allows transparent block-based compression) and the application should just keep reading. See The
			//  Old New Thing, Changing the conditions under which ReadFile produces fewer bytes than requested.
			cycle {
				UInt4B bytesRead; // ReadFile sets this value to zero before doing any work or error checking.
				auto const result = ReadFile(
					this, toDestination,
					UInt4B(minimumOf(USize(0x7FFFFFFF), remainingSize)), // Win32 limit is 2 GiB
					&bytesRead, nullptr
				);
				totalBytesRead += bytesRead;
				remainingSize  -= bytesRead;
				if(0 == result) { // Error?
					reportLastError(supervisorOrNull);
					break;
				} else if(0 == bytesRead) { // End of file?
					break;
				} else if(0 == remainingSize) { // Complete?
					break;
				}
				toDestination = byteOffset(toDestination, bytesRead);
			}

			return totalBytesRead;
		}

		// Copies bytes from the file to the given buffer.
		//   starts at the given offset without modifying the file pointer (atomic seek-and-read)
		//   returns the requested size on success
		//   otherwise, reports errors to the given supervisor and returns the number of bytes actually read
		USize File::tryToRead(
			UInt8B                   offset,
			void *                   toDestination,
			USize                    remainingSize,
			Supervisor const * const supervisorOrNull
		) const {
			USize totalBytesRead = 0;

			// Overlapped I/O can be used to read bytes without touching the file pointer. See The Old New Thing, You can use an
			//  OVERLAPPED structure with synchronous I/O, too.
			OVERLAPPED command = { 0, 0, { offset }, nullptr };

			// Depending on the underlying storage, short reads might occur: ReadFile() returning fewer bytes than requested. This
			//  is not an error (it allows transparent block-based compression) and the application should just keep reading. See The
			//  Old New Thing, Changing the conditions under which ReadFile produces fewer bytes than requested.
			// This does not affect the OVERLAPPED structure:
			//
			//    The system updates the OVERLAPPED offset before ReadFile returns.
			//
			//   (MSDN, ReadFile function, Remarks, Synchronization and File Position)
			cycle {
				UInt4B bytesRead; // ReadFile sets this value to zero before doing any work or error checking.
				auto const result = ReadFile(
					this, toDestination,
					UInt4B(minimumOf(USize(0x7FFFFFFF), remainingSize)), // Win32 limit is 2 GiB
					&bytesRead, &command
				);
				totalBytesRead += bytesRead;
				offset         += bytesRead;
				remainingSize  -= bytesRead;
				if(0 == result) { // Error?
					reportLastError(supervisorOrNull);
					break;
				} else if(0 == bytesRead) { // End of file?
					break;
				} else if(0 == remainingSize) { // Complete?
					break;
				}
				toDestination = byteOffset(toDestination, bytesRead);
			}

			return totalBytesRead;
		}

		// Copies bytes to the given file.
		//   size must be < 2 GiB
		//   returns the requested size on success
		//   otherwise, reports errors to the given supervisor and returns the number of bytes actually written
		USize File::tryToWrite(
			void const *             toSource,
			USize const              size,
			Supervisor const * const supervisorOrNull
		) {
			MUST(size <= 0x7FFFFFFF);

			UInt4B bytesWritten; // WriteFile sets this value to zero before doing any work or error checking.
			if(0 == WriteFile(this, toSource, UInt4B(size), &bytesWritten, nullptr)) {
				reportLastError(supervisorOrNull);
			}

			return bytesWritten;
		}

		FileDate File::lastWriteDate() const {
			FileDate result;
			WIN32_SHOULD_SUCCEED(GetFileTime(this, nullptr, nullptr, &result));
			return result;
		}

		void File::lastWriteDateIs(FileDate const date) {
			WIN32_SHOULD_SUCCEED(SetFileTime(this, nullptr, nullptr, &date));
		}

		void File::lastWriteDateIsNow() {
			lastWriteDateIs(now_asFileDate());
		}

		// Overwrites a files time stamps.
		//   last access time is deliberately ignored because it is disabled since WindowsVista
		void File::fileDatesAre(
			FileDate const creationDate,
			FileDate const lastWriteDate
		) {
			WIN32_SHOULD_SUCCEED(SetFileTime(this, &creationDate, &lastWriteDate, &lastWriteDate));
		}

		void File::close() {
			WIN32_SHOULD_SUCCEED(CloseHandle(this));
		}

		// Opens an existing file.
		//   on failure, returns nullptr and reports the error
		//   close() after use
		//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		File * tryToOpenFile(
			UTF16Unit const          path[],
			FileAccess const         access,
			FileSharing const        sharing,
			FileCaching const        caching,
			Supervisor const * const supervisorOrNull
		) {
			auto const result = CreateFileW(path, UInt4B(access), UInt4B(sharing), nullptr, 3, UInt4B(caching), nullptr); // OPEN_EXISTING
			if(nullFile == result) {
				reportLastError(supervisorOrNull);
				return nullptr;
			}
			return result;
		}

		// Creates an empty file. If existent, clears the file.
		//   on failure, returns nullptr and reports the error
		//   close() after use
		//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		File * tryToCreateEmptyFile(
			UTF16Unit const          path[],
			FileAccess const         access,
			FileSharing const        sharing,
			FileCaching const        caching,
			Supervisor const * const supervisorOrNull
		) {
			auto const result = CreateFileW(path, UInt4B(access), UInt4B(sharing), nullptr, 2, UInt4B(caching), nullptr); // CREATE_ALWAYS
			if(nullFile == result) {
				reportLastError(supervisorOrNull);
				return nullptr;
			}
			return result;
		}

		// Opens a file. If not existent, creates a new one.
		//   on failure, returns nullptr and reports the error
		//   close() after use
		//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		File * tryToOpenOrCreateFile(
			UTF16Unit const          path[],
			FileAccess const         access,
			FileSharing const        sharing,
			FileCaching const        caching,
			Supervisor const * const supervisorOrNull
		) {
			auto const result = CreateFileW(path, UInt4B(access), UInt4B(sharing), nullptr, 4, UInt4B(caching), nullptr); // OPEN_ALWAYS
			if(nullFile == result) {
				reportLastError(supervisorOrNull);
				return nullptr;
			}
			return result;
		}

		// Creates a temporary file.
		//   returns nullptr on failure
		//   the file is created in the users TEMP directory with a semi-random name starting with c1c
		//   close() deletes the file
		//   the file is written to hard disk only on memory pressure (see Larry Osterman, Its only temporary)
		// Useful for parking rarely accessed data either in RAM or on HDD, depending on system load.
		File * tryToCreateTemporaryFile() {

			// Create a temporary file name. Consider this will also create a temporary file in order to reserve the name for
			//  concurrent calls.
			UTF16Unit path[261];
			if(no == proposeTemporaryPath(path)) {
				return nullptr;
			}

			return tryToCreateEmptyFile(
				path, FileAccess::readWrite,
#				if _DEBUG
					FileSharing::all, // allow analysis via 3rd party tool
#				else
					FileSharing::none, // keep the file private so it remains in memory
#				endif
				FileCaching::temporary, nullptr
			);
		}

		// In Win32, devices are files, too. Sending and receiving is realized as writing and reading.
		//   returns nullptr on failure
		File * tryToOpenDeviceForReading(UTF16Unit const path[]) {
			MUST('\\' == path[0] && '\\' == path[1] && '?' == path[2] && '\\' == path[3]); // Not a device?

			// Devices *must* be available to other processes, so use full sharing!
			return tryToOpenFile(path, FileAccess::readOnly, FileSharing::all, FileCaching::dontCare, nullptr);
		}


		// Handles file mapping (read-only as well as read-write).
		local Byte * tryToMap(
			File const &             file,
			UInt4B const             pageAccess,
			UInt4B const             mapAccess,
			Supervisor const * const supervisorOrNull
		) {
			// There is no need to carry all handles through the rest of the program:
			//
			//    To fully close a file mapping object, [] call the UnmapViewOfFile function and then [] the CloseHandle function.
			//     These functions can be called in any order.
			//        MSDN, CreateFileMappingNuma function (Windows)
			//
			// Therefore, once the file is mapped, the file mapping object is closed. (This also saves error handling code because in
			//  case of an error, these objects are closed anyway.)

			Byte * result = nullptr;

			// Create the file mapping object:
			if(auto const mapping = CreateFileMappingW(&file, nullptr, pageAccess, 0, 0, nullptr)) {

				// Map the file to memory:
				result = MapViewOfFile(mapping, mapAccess, 0, 0, 0);
				if(nullptr == result) {
					reportLastError(supervisorOrNull);
				}

				WIN32_SHOULD_SUCCEED(CloseHandle(mapping));
			} else {
				reportLastError(supervisorOrNull);
			}

			return result;
		}

		// Maps the given file to read-only memory.
		//   on failure, returns nullptr and reports the error
		//   unmap() after use
		//   the file must 
		//     not be empty
		//     be open with the read access right
		//   needs the files size because the Kernel doesnt track it
		Byte const * tryToMapForReading(
			File const &             file,
			Supervisor const * const supervisorOrNull
		) {
			return tryToMap(file, 2, 4, supervisorOrNull); // PAGE_READONLY, FILE_MAP_READ
		}

		// Maps the given file to memory.
		//   on failure, returns nullptr and reports the error
		//   unmap() after use
		//   the file must 
		//     not be empty
		//     be open with the readWrite access right
		//   needs the files size because the Kernel doesnt track it
		Byte * tryToMapForWriting(
			File const &             file,
			Supervisor const * const supervisorOrNull
		) {
			return tryToMap(file, 4, 2, supervisorOrNull); // PAGE_READWRITE, FILE_MAP_WRITE
		}

		//  works for any address within a mapped file
		//  for read-writeable maps, does not update the files modification date (due to lazy writing)!
		void unmap(void const * mappedFile) {
			WIN32_SHOULD_SUCCEED(UnmapViewOfFile(mappedFile));
		}


		UInt8B sizeOfFile(UTF16Unit const path[]) {
			// There is no file size by path function in Win32. Open a handle to the file, get its size, and close it.
			UInt8B result = 0;
			if(auto const file = tryToOpenFile(path, FileAccess::readAttributesOnly, FileSharing::all, FileCaching::dontCare, nullptr)) {
				WIN32_SHOULD_SUCCEED(GetFileSizeEx(file, &result));
				file->close();
			} else {
				WIN32_SHOULD_HAVE_SUCCEEDED;
			}
			return result;
		}

		// Ensures the directory at the given path exists.
		//   the path must be zero-terminated and shorter than 249 characters (32,767 with \\?\ prepended)
		//   returns yes iff the directory already exists or is newly created
		bool tryToOpenOrCreateDirectory(UTF16Unit const * path) {

			if(0 == CreateDirectoryW(path, nullptr) && 183 != GetLastError()) { // ERROR_ALREADY_EXISTS
				WIN32_SHOULD_HAVE_SUCCEEDED;
				return no;
			}

			return yes;
		}

		// Copies the file from the given source to the given destination.
		//   the paths must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		//   returns yes if the file was copied successfully
		//   returns no if the file already exists or if there was an error
		bool tryToCopyFile(UTF16Unit const * destination, UTF16Unit const * source) {
			return 0 != WIN32_SHOULD_SUCCEED(CopyFileW(source, destination, 1));
		}

		// Copies the file from the given source to the given destination. If it already exists, it is overwritten.
		//   the paths must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		//   returns yes iff the file was copied or overwritten successfully
		bool tryToCopyOrReplaceFile(UTF16Unit const * destination, UTF16Unit const * source) {
			return 0 != WIN32_SHOULD_SUCCEED(CopyFileW(source, destination, 0));
		}

		// Destroys the file at the given path (without moving it to the Recycle Bin).
		//   if there are handles to the file, it is deleted once the last handle is closed
		//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		//   returns yes iff the file was destroyed successfully
		bool tryToDestroyFileOnClose(UTF16Unit const * path) {
			return 0 != WIN32_SHOULD_SUCCEED(DeleteFileW(path));
		}

		// Moves the file or directory from the given source to the given destination. Can also rename.
		//   the paths must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		//   returns yes iff the file or directory was moved or renamed successfully
		bool tryToMoveOrRenameFileOrDirectory(UTF16Unit const * oldPath, UTF16Unit const * newPath) {
			return 0 != WIN32_SHOULD_SUCCEED(MoveFileW(oldPath, newPath));
		}

		local int destroyDirectoryRecursively(
			UTF16Unit * path,
			UTF16Unit * pathsEnd
		) {
			// Try to search for files and subdirectories. This requires \* at the paths end.
			pathsEnd[0] = '\\';
			pathsEnd[1] = '*';
			pathsEnd[2] = 0;
			FileSearch search;
			search.id = FindFirstFileW(path, &search.attributes);
			if(-1 == search.id) { // INVALID_HANDLE_VALUE
				auto status = GetLastError();
				if(2 == status || 3 == status) {
					//  no file or directory at that path (ERROR_FILE_NOT_FOUND)
					//  no such parent directory (ERROR_PATH_NOT_FOUND)
					return 1;
				}
				WIN32_SHOULD_HAVE_SUCCEEDED;
				return 0;
			}

			// Iterate the directory:
			int result = 1;
			cycle {
				if(0 != FindNextFileW(search.id, &search.attributes)) { // File or subdirectory found?
					auto newPathsEnd = Text::copyZeroTerminated(pathsEnd + 1, search.name);

					if(search.attributes & 0x10) { // FILE_ATTRIBUTE_DIRECTORY
						if(search.attributes & 0x400) { // FILE_ATTRIBUTE_REPARSE_POINT
							result &= WIN32_SHOULD_SUCCEED(RemoveDirectoryW(path)); // do not delete the content of symbolic directories
						} else {
							// Every directory contains . (itself) and .. (its parent). Skip those to avoid infinite loops. But do not skip
							//  ..., which unfortunately is a valid NTFS name.
							if('.' == search.name[0] && (0 == search.name[1] || ('.' == search.name[1] && 0 == search.name[2]))) {
								continue;
							}
							// Destroy the subdirectory and pass on the result.
							result &= destroyDirectoryRecursively(path, newPathsEnd);
						}
					} else {
						makeFileReadable(path);
						tryToDestroyFileOnClose(path);
					}

				} else { // Nothing more found?

					if(18 != GetLastError()) { // ERROR_NO_MORE_FILES
						result = 0;
					}
					abort(search);

					// Delete the empty directory:
					*pathsEnd = 0;
					result &= WIN32_SHOULD_SUCCEED(RemoveDirectoryW(path));
					break;
				}
			}
			return result;
		}

		// Destroys the directory at the given path and all (non-symbolic) files in it (without moving them to the Recycle Bin).
		//   the path must
		//     be absolute
		//     not end with a backslash
		//     be zero-terminated
		//     be shorter than 32,500 characters
		//   no files must be opened in the directory and any of its sub-directories
		//   returns yes if the entire directory was destroyed
		//   returns yes if no directory existed in the first place
		//   returns no on error
		bool tryToDestroyDirectory(UTF16Unit const * path) {
			// SHFileOperation() is useless because its limited to 260 characters. IFileOperation has huge COM and Shell overhead
			//  for such a simple operation. Therefore, use good ole FindFirstFile() recursively.

			// Dont let recusion eat up the stack by reserving 32 KiB for paths at each level; share one buffer for all enumerations!
			UTF16Unit   currentPath[pathLengthLimit];
			UTF16Unit * toPathsEnd;

			// Avoid problems by using extended-length paths exclusively:
			if('\\' != path[0]) {
				toPathsEnd = Text::copyUntilZero(currentPath, "\\\\?\\");
			} else {
				toPathsEnd = currentPath;
			}

			toPathsEnd = Text::copyZeroTerminated(toPathsEnd, path);
			return 0 != destroyDirectoryRecursively(currentPath, toPathsEnd);
		}

		// Writes the path to the users temporary files (%TEMP%).
		//   returns a pointer to the paths end
		//   absolute
		//   shorter than 262 characters
		//   always ends with a backslash
		UTF16Unit * pathOfTEMP(UTF16Unit result[261]) {
			return result + WIN32_SHOULD_SUCCEED(GetTempPathW(261, result));
		}

		// Writes the path to the Windows directory (%WINDIR%).
		//   returns a pointer to the paths end
		//   absolute
		//   shorter than 262 characters
		//   always ends with a backslash
		UTF16Unit * pathOfWINDIR(UTF16Unit result[262]) {

			// SHGetFolderPath() is reusable and XP-compatible, SHGetKnownFolderPath() is not
			if(0 != SHGetFolderPathW(nullptr, 36, nullptr, 0, result)) { // CSIDL_WINDOWS, SHGFP_TYPE_CURRENT
				return result;
			}
			// The returned path does not end with a trailing backslash; fix that:
			auto toEnd = Text::seekEndOf((UTF16Unit *)result);
			*toEnd++ = '\\';

			return toEnd;
		}

		// Writes the path to the users local application data (%USERPROFILE%\AppData\Local\).
		//   on failure, returns a pointer to the buffers beginning
		//   otherwise, returns a pointer to the paths end
		//   absolute
		//   shorter than 263 characters
		//   always ends with a backslash
		UTF16Unit * pathOfLocalAPPDATA(UTF16Unit result[262]) {

			// SHGetFolderPath() is reusable and XP-compatible, SHGetKnownFolderPath() is not
			if(0 != SHGetFolderPathW(nullptr, 28, nullptr, 0, result)) { // CSIDL_LOCAL_APPDATA, SHGFP_TYPE_CURRENT
				return result;
			}
			// The returned path does not end with a trailing backslash; fix that:
			auto toEnd = Text::seekEndOf((UTF16Unit *)result);
			*toEnd++ = '\\';

			return toEnd;
		}

		// Creates an empty file in the users temporary files and writes its path.
		//   returns no on failure
		bool proposeTemporaryPath(UTF16Unit (&result)[261]) {
			pathOfTEMP(result);
			return 0 != WIN32_SHOULD_SUCCEED(GetTempFileNameW(result, UTF16("c1c"), 0, result));
		}

		// Determines whether a file or directory exists at the given path.
		//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		bool isPathExistent(UTF16Unit const * path) {
			// PathFileExistsW() would be the function to use here, but unfortunately, its in Shlwapi.dll and I dont want to depend
			//  on 438 KiB of code just for a function which effectively redirects to GetFileAttributesW().
			return 0xFFFFFFFF != GetFileAttributesW(path);
		}

		// Returns a pointer to the extension of the given file path (where the dot is).
		//   the path must be zero-terminated
		//   returns the paths end if no extension is present
		UTF16Unit * findFileExtensionInPath(UTF16Unit * path) {
			UTF16Unit * toLastDot   = nullptr;
			auto        toCharacter = path;
			cycle {
				if('.' == *toCharacter) {
					toLastDot = toCharacter;
				} else if(0 == *toCharacter) {
					if(nullptr == toLastDot) {
						return toCharacter; // no file extension; return end of path
					}
					return toLastDot;
				}
				++toCharacter;
			}
		}
		UTF16Unit const * findFileExtensionInPath(UTF16Unit const * path) {
			return findFileExtensionInPath(const_cast<UTF16Unit *>(path));
		}

		// Returns a pointer to the filename of the given path.
		//   the path must be zero-terminated
		//   returns the paths end if no name is present
		UTF16Unit * findFilenameInPath(UTF16Unit * path) {
			UTF16Unit * toLastSlash = nullptr;
			auto        toCharacter = path;
			cycle {
				if('\\' == *toCharacter || '/' == *toCharacter) {
					toLastSlash = toCharacter + 1;
				} else if(0 == *toCharacter) {
					if(nullptr == toLastSlash) {
						return toCharacter; // no filename; return end of path
					}
					return toLastSlash;
				}
				++toCharacter;
			}
		}
		UTF16Unit const * findFilenameInPath(UTF16Unit const * path) {
			return findFilenameInPath(const_cast<UTF16Unit *>(path));
		}

		// Removes the read-only attribute from the given file. No effect if the file is already readable.
		void makeFileReadable(UTF16Unit const * path) {
			auto attributes = GetFileAttributesW(path);
			if(attributes & 1) { // FILE_ATTRIBUTE_READONLY
				WIN32_SHOULD_SUCCEED(SetFileAttributesW(path, attributes & ~1));
			}
		}

		// Checks whether the given file or directory is reserved for system use.
		//   the path must exist
		//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		bool isSystem(UTF16Unit const * path) {
			auto attributes = GetFileAttributesW(path);
			MUST(0xFFFFFFFF != attributes); // Non-existent file?
			return 0 != (attributes & 4); // FILE_ATTRIBUTE_SYSTEM
		}


		// Tries to enumerate files and directories which match the given path.
		//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		//   pass dir\\* to enumerate all files and directories in a directory
		//   pass dir\\*.x to enumerate all files and directories of extension .x in the directory
		//   returns no if nothing was found
		//   either call next() until finished or abort()
		bool tryToSearch(UTF16Unit const * path, FileSearch & search) {
			// Create a file search:
			search.id = FindFirstFileW(path, &search.attributes);
#			if _DEBUG
				if(-1 == search.id) { // INVALID_HANDLE_VALUE
					// Either no file matches the search string (ERROR_FILE_NOT_FOUND/ERROR_PATH_NOT_FOUND/ERROR_NO_MORE_FILES) or
					//  there was an error. Dont care in release builds.
					if(2 != GetLastError() && 3 != GetLastError() && 18 != GetLastError()) {
						WIN32_SHOULD_HAVE_SUCCEEDED;
					}
				}
#			endif
			return -1 != search.id;
		}

		// Returns yes and updates information if another file or directory was found. Returns no otherwise.
		bool next(FileSearch & search) {
			if(0 == FindNextFileW(search.id, &search.attributes)) {
				// Either the directory is empty or no matches were found or there was an error. Dont care in release builds.
#				if _DEBUG
					if(18 != GetLastError()) { // ERROR_NO_MORE_FILES
						WIN32_SHOULD_HAVE_SUCCEEDED;
					}
#				endif
				abort(search);
				return no;
			} else {
				return yes;
			}
		}

		// Checks whether the current result in the search is a directory. Its a file otherwise.
		bool isDirectory(FileSearch & search) {
			return 0 != (0x10 & search.attributes); // FILE_ATTRIBUTE_DIRECTORY
		}

		void abort(FileSearch & search) {
			WIN32_SHOULD_SUCCEED(FindClose(search.id));
		}


	} // Win32


} // OS