// Win32 file helpers.



namespace OS {


	namespace Win32 {


		enum class FileAccess : UInt4B {
			readOnly            = 0x80000000, // GENERIC_READ
			writeOnly           = 0x40000000, // GENERIC_WRITE
			appendOnly          = 4,          // can write at file end, but cannot overwrite existing data (FILE_APPEND_DATA)
			readWrite           = 0xC0000000, // GENERIC_READ | GENERIC_WRITE
			readAttributesOnly  = 128,        // no cache pollution with actual file data (FILE_READ_ATTRIBUTES)
			writeAttributesOnly = 256,        // no cache pollution with actual file data (FILE_WRITE_ATTRIBUTES)
		};

		// Controls how other processes can access an open file.
		enum class FileSharing : UInt4B {
			none     = 0, // other processes cannot access the file
			readOnly = 5, // other processes can read from the file or delete it (FILE_SHARE_READ | FILE_SHARE_DELETE)
			all      = 7, // other processes have full access to the file (FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE)
			dontCare = readOnly
		};

		// Controls how a file is cached by the Memory Manager.
		//   does not affect memory-mapped files
		//   details: The Old New Thing, How do FILE_FLAG_SEQUENTIAL_SCAN and FILE_FLAG_RANDOM_ACCESS affect how the operating
		//    system treats my file?
		enum class FileCaching : UInt4B {
			sequential = 0x08000000, // maximal prefetching; evict-behind (FILE_FLAG_SEQUENTIAL_SCAN)
			adaptive   = 0,          // observe the access pattern and optimize dynamically
			random     = 0x10000000, // no prefetching; evict-behind (FILE_FLAG_RANDOM_ACCESS)
			temporary  = 0x04000100, // kept in RAM as long as possible (FILE_FLAG_DELETE_ON_CLOSE | FILE_ATTRIBUTE_TEMPORARY)
			dontCare   = adaptive
		};

		// HANDLE in Win32s file and driver functions.
		//   can also be consoles or devices
		struct alignas(handleAlignment) File : Opaque {

			UInt8B size() const;

			// Copies bytes from the file to the given buffer.
			//   non-const because it uses the file pointer
			//   returns the size actually read, which may be lower than the requested size only on non-resumable errors like
			//     unexpected end of file
			//     hardware problems
			USize tryToRead(void *, USize, Supervisor const *);

			// Copies bytes from the file to the given buffer.
			//   starts at the given offset without modifying the file pointer (atomic seek-and-read)
			//   returns the requested size on success
			//   otherwise, reports errors to the given supervisor and returns the number of bytes actually read
			USize tryToRead(UInt8B offset, void * toDestination, USize, Supervisor const *) const;

			// Writes the given number of bytes to the given file.
			//   size must be < 2 GiB
			//   returns the size actually written
			USize tryToWrite(void const *, USize, Supervisor const *);

			FileDate lastWriteDate() const;

			void lastWriteDateIs(FileDate);

			void lastWriteDateIsNow();

			// Overwrites a files time stamps.
			//   last access time is deliberately ignored because it is disabled since WindowsVista
			void fileDatesAre(
				FileDate creationDate,
				FileDate lastWriteDate
			);

			void close();

		};

		// Opens an existing file.
		//   on failure, returns nullptr and reports the error
		//   close() after use
		//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		File * tryToOpenFile(
			UTF16Unit const    path[],
			FileAccess         ,
			FileSharing        ,
			FileCaching        ,
			Supervisor const *
		);

		// Creates an empty file. If existent, clears the file.
		//   on failure, returns nullptr and reports the error
		//   close() after use
		//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		File * tryToCreateEmptyFile(
			UTF16Unit const    path[],
			FileAccess         ,
			FileSharing        ,
			FileCaching        ,
			Supervisor const *
		);

		// Opens a file. If not existent, creates a new one.
		//   on failure, returns nullptr and reports the error
		//   close() after use
		//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		File * tryToOpenOrCreateFile(
			UTF16Unit const    path[],
			FileAccess         ,
			FileSharing        ,
			FileCaching        ,
			Supervisor const *
		);

		// Creates a temporary file.
		//   returns nullptr on failure
		//   the file is created in the users TEMP directory with a semi-random name starting with c1c
		//   close() deletes the file
		//   the file is written to hard disk only on memory pressure (see Larry Osterman, Its only temporary)
		// Useful for parking rarely accessed data either in RAM or on HDD, depending on system load.
		File * tryToCreateTemporaryFile();

		// In Win32, devices are files, too. Sending and receiving is realized as writing and reading.
		//   returns nullptr on failure
		File * tryToOpenDeviceForReading(UTF16Unit const path[]);

		FileDate now_asFileDate();


		// Maps the given file to read-only memory.
		//   on failure, returns nullptr and reports the error
		//   unmap() after use
		//   the file must 
		//     not be empty
		//     be open with the read access right
		//   needs the files size because the Kernel doesnt track it
		Byte const * tryToMapForReading(
			File const &      ,
			Supervisor const *
		);

		// Maps the given file to memory.
		//   on failure, returns nullptr and reports the error
		//   unmap() after use
		//   the file must 
		//     not be empty
		//     be open with the readWrite access right
		//   needs the files size because the Kernel doesnt track it
		Byte * tryToMapForWriting(
			File const &      ,
			Supervisor const *
		);

		//  works for any address within a mapped file
		//  for read-writeable maps, does not update the files modification date (due to lazy writing)!
		void unmap(void const *);



		UInt8B sizeOfFile(UTF16Unit const []);

		// Ensures the directory at the given path exists.
		//   the path must be zero-terminated and shorter than 249 characters (32,767 with \\?\ prepended)
		//   returns yes iff the directory already exists or is newly created
		bool tryToOpenOrCreateDirectory(UTF16Unit const *);

		// Copies the file from the given source to the given destination.
		//   the paths must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		//   returns yes if the file was copied successfully
		//   returns no if the file already exists or if there was an error
		bool tryToCopyFile(UTF16Unit const * destination, UTF16Unit const * source);

		// Copies the file from the given source to the given destination. If it already exists, it is overwritten.
		//   the paths must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		//   returns yes iff the file was copied or overwritten successfully
		bool tryToCopyOrReplaceFile(UTF16Unit const * destination, UTF16Unit const * source);

		// Destroys the file at the given path (without moving it to the Recycle Bin).
		//   if there are handles to the file, it is deleted once the last handle is closed
		//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		//   returns yes iff the file was destroyed successfully
		bool tryToDestroyFileOnClose(UTF16Unit const *);

		// Moves the file or directory from the given source to the given destination. Can also rename.
		//   the paths must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		//   returns yes iff the file or directory was moved or renamed successfully
		bool tryToMoveOrRenameFileOrDirectory(UTF16Unit const * oldPath, UTF16Unit const * newPath);

		// Destroys the directory at the given path and all (non-symbolic) files in it (without moving them to the Recycle Bin).
		//   the path must
		//     be absolute
		//     not end with a backslash
		//     be zero-terminated
		//     be shorter than 32,500 characters
		//   no files must be opened in the directory and any of its sub-directories
		//   returns yes if the entire directory was destroyed
		//   returns yes if no directory existed in the first place
		//   returns no on error
		bool tryToDestroyDirectory(UTF16Unit const *);

		// Writes the path to the users temporary files (%TEMP%).
		//   returns a pointer to the paths end
		//   absolute
		//   shorter than 262 characters
		//   always ends with a backslash
		UTF16Unit * pathOfTEMP(UTF16Unit[261]);

		// Writes the path to the Windows directory (%WINDIR%).
		//   returns a pointer to the paths end
		//   absolute
		//   shorter than 262 characters
		//   always ends with a backslash
		UTF16Unit * pathOfWINDIR(UTF16Unit[262]);

		// Writes the path to the users local application data (usually %USERPROFILE%\AppData\Local\).
		//   on failure, returns a pointer to the buffers beginning
		//   otherwise, returns a pointer to the paths end
		//   absolute
		//   shorter than 263 characters
		//   always ends with a backslash
		UTF16Unit * pathOfLocalAPPDATA(UTF16Unit[262]);

		// Creates an empty file in the users temporary files and writes its path.
		//   returns no on failure
		bool proposeTemporaryPath(UTF16Unit (&result)[261]);

		// Determines whether a file or directory exists at the given path.
		//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		bool isPathExistent(UTF16Unit const * toItsZeroTerminatedPath);

		// Returns a pointer to the extension of the given file path (where the dot is).
		//   the path must be zero-terminated
		//   returns the paths end if no extension is present
		UTF16Unit * findFileExtensionInPath(UTF16Unit *);
		UTF16Unit const * findFileExtensionInPath(UTF16Unit const *);

		// Returns a pointer to the filename of the given path.
		//   the path must be zero-terminated
		//   returns the paths end if no name is present
		UTF16Unit * findFilenameInPath(UTF16Unit *);
		UTF16Unit const * findFilenameInPath(UTF16Unit const *);

		// Removes the read-only attribute from the given file. No effect if the file is already readable.
		void makeFileReadable(UTF16Unit const *);

		// Checks whether the given file or directory is reserved for system use.
		//   the path must exist
		//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		bool isSystem(UTF16Unit const *);


	} // Win32


} // OS



// Functions for searching files and directories.



namespace OS {


	namespace Win32 {


#		pragma pack(push, 4)
			struct FileSearch {
				USize     id; // HANDLE of a file search
				// WIN32_FIND_DATA starts here:
				UInt4B    attributes;
				FileDate  creationDate;
				FileDate  lastAccessDate;
				FileDate  lastWriteDate;
				UInt4B    nFileSizeHigh;
				UInt4B    nFileSizeLow;
				UInt8B    reserved;
				UTF16Unit name[260]; // zero-terminated
				UTF16Unit name8_3[14]; // zero-terminated 8.3 format for legacy applications
			};
#		pragma pack(pop)

		// Tries to enumerate files and directories which match the given path.
		//   the path must be zero-terminated and shorter than 260 characters (32,767 with \\?\ prepended)
		//   pass dir\\* to enumerate all files and directories in a directory
		//   pass dir\\*.x to enumerate all files and directories of extension .x in the directory
		//   returns no if nothing was found
		//   either call next() until finished or abort()
		bool tryToSearch(UTF16Unit const *, FileSearch &);

		// Returns yes and updates information if another file or directory was found. Returns no otherwise.
		bool next(FileSearch &);

		// Checks whether the current result in the search is a directory. Its a file otherwise.
		bool isDirectory(FileSearch &);

		void abort(FileSearch &);


	} // Win32


} // OS