// Adds debugging tools to Windows heap API.
//   simple checks for writes before and past allocations, of course
//   memory leak information
//     call stacks for leaked allocations
//     call stacks for leaked heaps


// You can use this variable to cause a specific allocation to fail. Useful for testing out-of-memory sitatuations.
extern UInt4B volatile DEBUG_allocToFail;
       UInt4B volatile DEBUG_allocToFail = ~0u;

// Running index of every allocation.
extern UInt4B volatile DEBUG_allocCounter;
       UInt4B volatile DEBUG_allocCounter = 0;



#include "Parallel.hpp"

import_from(Kernel32) {

	struct PROCESS_HEAP_ENTRY {
		void * lpData;
		UInt4B cbData;
		UInt1B cbOverhead;
		UInt1B iRegionIndex;
		UInt2B wFlags;
		union {
			struct {
				void * hMem;
				UInt4B dwReserved[3];
				PAD_4B_ON_X86_64;
			} Block;
			struct {
				UInt4B dwCommittedSize;
				UInt4B dwUnCommittedSize;
				void * lpFirstBlock;
				void * lpLastBlock;
			} Region;
		};
	};

	int API_CALL HeapLock(OS::Win32::Region *);

	int API_CALL HeapWalk(OS::Win32::Region *, PROCESS_HEAP_ENTRY *);

	int API_CALL HeapUnlock(OS::Win32::Region *);

} // Kernel32



namespace OS {


	namespace Win32 {


	locals {


			// Header of a user-allocated heap.
			struct HeapInfo {
				OS::Win32::Region *           heap;
				OS::Win32::DbgHelp::CallStack callStack;
			};


			// Header of an allocated block.
			//   carries debug information (e.g. a call stack)
			//   placed before the actual payload
			struct AllocationInfo {
				OS::Win32::DbgHelp::CallStack callStack;
				// Size of the actual allocation, in B.
				USize payloadSize;
				// Always 0x13371337.
				//   can be used to detect writes before the buffer
				USize id;
				Byte  payload[];
			};

			auto constexpr allocationInfoPattern =
#			if COMPILED_FOR_X86_32
				0x1337'1337u;
#			elif COMPILED_FOR_X86_64
				0x1337'1337'1337'1337u;
#			endif

			// A list of all heaps which were created through the framework.
			//   nullptr entry if unused
			//   allows to enumerate leaking heaps
			HeapInfo userHeaps[256];
			// Protects the heap list.
			SpinLock heapListLock = SPIN_LOCK_INITIALIZER;



			// Constant for the run-time of the process, so store it globally to save kernel calls in allocation routines. The function
			//  resides in Kernel32.dll, so it's safe to be called upon DLL initialization.
#			pragma warning(disable : 4074) // blah blah yadda yadda
#			pragma init_seg(compiler)
			auto const processHeapsHandle = WIN32_SHOULD_SUCCEED(HeapCreate(0, 0, 0)); // do not interfere with CRT allocations (until the CRT is gone)


			Byte * DEBUG_HeapAlloc(
				OS::Win32::Region * heap,
				UInt4B const        flags,
				USize const         size
			) {
				if(DEBUG_allocToFail == atomicIncrement(DEBUG_allocCounter)) {
					return nullptr;
				}

				if(auto const toNewBlock = reinterpret_cast<AllocationInfo *>(HeapAlloc(heap, flags, sizeof(AllocationInfo) + size))) {
					toNewBlock->callStack   = DbgHelp::currentCallStack(2);
					toNewBlock->payloadSize = size;
					toNewBlock->id          = allocationInfoPattern;
					return toNewBlock->payload;
				} else {
					OutputDebugStringA("\nbad allocation request at\n");
					print(DbgHelp::currentCallStack(2));
					return nullptr;
				}
			}

			Byte * DEBUG_HeapReAlloc(
				OS::Win32::Region * heap,
				void const * const  oldAddress,
				USize const         newSize
			) {
				if(DEBUG_allocToFail == atomicIncrement(DEBUG_allocCounter)) {
					return nullptr;
				}

				auto const toOldBlock = reinterpret_cast<AllocationInfo const *>(reinterpret_cast<Byte const *>(oldAddress) - sizeof(AllocationInfo));
				MUST(allocationInfoPattern == toOldBlock->id); // Damaged?

				if(auto const toNewBlock = WIN32_SHOULD_SUCCEED(reinterpret_cast<AllocationInfo *>(HeapReAlloc(heap, 0, toOldBlock, newSize + sizeof(AllocationInfo))))) {
					toNewBlock->payloadSize = newSize;
					return toNewBlock->payload;
				} else {
					OutputDebugStringA("\nbad allocation request at\n");
					print(DbgHelp::currentCallStack(2));
					return nullptr;
				}
			}

			void DEBUG_HeapFree(
				OS::Win32::Region * heap,
				void const * const  address
			) {
				MUST(nullptr != address); // nullptr must not be freed

				auto const toBlock = reinterpret_cast<AllocationInfo const *>(reinterpret_cast<Byte const *>(address) - sizeof(AllocationInfo));
				MUST(allocationInfoPattern == toBlock->id); // Damaged?
				WIN32_SHOULD_SUCCEED(HeapFree(heap, 0, toBlock));
			}

			void verifyIntegrity(Region & region, bool const reportLeaks) {
				HeapLock(&region); // prevent other threads from modifying the heap during analysis

				// Enumerate all heap blocks:
				auto isFirstLeak = yes;
				PROCESS_HEAP_ENTRY currentEntry = { };
				while(1 == HeapWalk(&region, &currentEntry)) {

					if(0x0004 != (0x0004 & currentEntry.wFlags)) { // PROCESS_HEAP_ENTRY_BUSY
						continue; // only report committed blocks (no control structures)
					}

					auto const toBlock = reinterpret_cast<AllocationInfo const *>(currentEntry.lpData);
					MUST(sizeof *toBlock <= currentEntry.cbData && allocationInfoPattern == toBlock->id); // Damaged block?

					if(no == reportLeaks) {
						continue; // integrity has been checked; no printing leaks
					}

					if(isFirstLeak) {
						OutputDebugStringA("\nMemory leaks:\n\n");
						isFirstLeak = no;
					}

					currentEntry.cbData -= sizeof *toBlock; // count only the payload

#					define FORMAT_NUMBER_TEXT(NUMBER, POSTFIX) \
						ASCIICharacter text[] = "leaking      " POSTFIX "\n"; \
						if((NUMBER) >= 1000) { \
							text[8]  = '0' + (NUMBER) / 1000 % 10; \
							text[9]  = '0' + (NUMBER) / 100 % 10; \
							text[10] = '0' + (NUMBER) / 10 % 10; \
						} else if((NUMBER) >= 100) { \
							text[9]  = '0' + (NUMBER) / 100 % 10; \
							text[10] = '0' + (NUMBER) / 10 % 10; \
						} else if((NUMBER) >= 10) { \
							text[10] = '0' + (NUMBER) / 10 % 10; \
						} \
						text[11] = '0' + (NUMBER) % 10; \
						OutputDebugStringA(text);

					if(1024 * 1024 * 1024 <= currentEntry.cbData) {
						FORMAT_NUMBER_TEXT(currentEntry.cbData >> 30, "GiB");
					} if(1024 * 1024 <= currentEntry.cbData) {
						FORMAT_NUMBER_TEXT(currentEntry.cbData >> 20, "MiB");
					} else if(1024 <= currentEntry.cbData) {
						FORMAT_NUMBER_TEXT(currentEntry.cbData >> 10, "KiB");
					} else {
						FORMAT_NUMBER_TEXT(currentEntry.cbData, "B");
					}

#					undef FORMAT_NUMBER_TEXT

					DbgHelp::print(toBlock->callStack);
					OutputDebugStringA("\n");

				} // until heap blocks are enumerated

				HeapUnlock(&region); // other threads may proceed now
			}


		} // locals


		Region & processPool() {
			return *processHeapsHandle;
		}

		Byte * tryToReserve(USize const size) {
			return DEBUG_HeapAlloc(processHeapsHandle, 0, size);
		}

		Byte * tryToReserveZeroed(USize const size) {
			return DEBUG_HeapAlloc(processHeapsHandle, 8, size); // HEAP_ZERO_MEMORY
		}

		Byte * tryToReallocate(void const * oldAddress, USize const newSize) {
			return DEBUG_HeapReAlloc(processHeapsHandle, oldAddress, newSize);
		}

		void release(void const * const address) {
			DEBUG_HeapFree(processHeapsHandle, address);
		}

		//  must be destroy()d
		//  returns nullptr on failure
		Region * tryToCreateRegion() {
			if(atomicIncrement(DEBUG_allocCounter) == DEBUG_allocToFail) {
				return nullptr;
			}
			auto result = WIN32_SHOULD_SUCCEED(HeapCreate(0, 0, 0));
			lock(heapListLock);
			for(auto & h : userHeaps) {
				if(nullptr == h.heap) {
					h.heap = result;
					unlock(heapListLock);
					h.callStack = DbgHelp::currentCallStack(2);
					return result;
				}
			}
			unlock(heapListLock);
			INVOKE_DEBUGGER; // too many heaps
			return result;
		}

		Byte * Region::tryToReserve(USize const size) {
			return DEBUG_HeapAlloc(this, 0, size);
		}

		Byte * Region::tryToReserveZeroed(USize const size) {
			return DEBUG_HeapAlloc(this, 8, size); // HEAP_ZERO_MEMORY
		}

		Byte * Region::tryToReallocate(void const * const address, USize const size) {
			return DEBUG_HeapReAlloc(this, address, size);
		}

		void Region::release(void const * const address) {
			return DEBUG_HeapFree(this, address);
		}

		// Destroys the region and all memory in it.
		void Region::destroy() {
			lock(heapListLock);
			for(auto & h : userHeaps) {
				if(h.heap == this) {
					h.heap = nullptr;
					unlock(heapListLock);
					verifyIntegrity(*this, no);
					WIN32_SHOULD_SUCCEED(HeapDestroy(this));
					return;
				}
			}
			unlock(heapListLock);
			INVOKE_DEBUGGER; // not created with this API
		}

		void printMemoryLeaks() {
			for(auto & uh : userHeaps) {
				if(nullptr != uh.heap) {
					verifyIntegrity(*uh.heap, yes);
					INVOKE_DEBUGGER; // leaked a heap
				}
			}
			verifyIntegrity(*processHeapsHandle, yes);
		}

	} // Win32


} // OS