// Implements raw Windows heaps.



// The process heap is constant for the run-time of the process. Rather than calling into the Win32 API on each allocation,
//  store it in a global constant.
#if COMPILED_BY_VISUAL_CPP
	// However, the Visual C++ CRT will use the constant before it is initialized. Force initialization before CRT initializers run
	//  by placing an entry in the .CRT$XCAA section:
	//   the sections are sorted alphabetically
	//   .CRT$XCA is the lists beginning; .CRT$XCZ is the end
	//   using .CRT$XCA causes problems, probably because its shared with the pointer to the lists beginning
	//   .CRT$XCAA is the next free name (shared with the CRTs pre_cpp_initialization(), but it works)

	local OS::Win32::Region * processHeapsHandle;

	void C_CALL initializeProcessHeap() {
		processHeapsHandle = GetProcessHeap();
	}

	using StaticFunction = void C_CALL();
	#pragma section(".CRT$XCAA", read)
	__declspec(allocate(".CRT$XCAA")) StaticFunction * forceInitializeProcessHeap[] = { &initializeProcessHeap }; // CAUTION: Must not be const!
#else
	local auto const processHeapsHandle = GetProcessHeap();
#endif



namespace OS {


	namespace Win32 {


		Region & processPool() {
#			if COMPILED_BY_VISUAL_CPP
				forceInitializeProcessHeap; // DO NOT REMOVE: Forces heap initialization; see Visual C++ bug 3136901!
#			endif
			return *processHeapsHandle;
		}

		Byte * tryToReserve(USize const size) {
			return HeapAlloc(processHeapsHandle, 0, size);
		}

		Byte * tryToReserveZeroed(USize const size) {
			return HeapAlloc(processHeapsHandle, 8, size); // HEAP_ZERO_MEMORY
		}

		Byte * tryToReallocate(void const * const oldAddress, USize const newSize) {
			return HeapReAlloc(processHeapsHandle, 0, oldAddress, newSize);
		}

		void release(void const * const address) {
			WIN32_SHOULD_SUCCEED(HeapFree(processHeapsHandle, 0, address));
		}


		//  must be destroy()d
		//  returns nullptr on failure
		Region * tryToCreateRegion() {
			return WIN32_SHOULD_SUCCEED(HeapCreate(0, 0, 0));
		}

		Byte * Region::tryToReserve(USize const size) {
			return HeapAlloc(this, 0, size);
		}

		Byte * Region::tryToReserveZeroed(USize const size) {
			return HeapAlloc(this, 8, size); // HEAP_ZERO_MEMORY
		}

		Byte * Region::tryToReallocate(void const * const address, USize const size) {
			return HeapReAlloc(this, 0, address, size);
		}

		void Region::release(void const * const address) {
			WIN32_SHOULD_SUCCEED(HeapFree(this, 0, address));
		}

		// Destroys the region and all memory in it.
		void Region::destroy() {
			WIN32_SHOULD_SUCCEED(HeapDestroy(this));
		}


	} // Win32


} // OS