// Helpers for basic Win32 functionality (files, processes, threads, modules, ).



#if _DEBUG
	void breakOnWindowsError();

	// for To get extended error information, call GetLastError.
#	define WIN32_SHOULD_HAVE_SUCCEEDED breakOnWindowsError()

	// for If the function fails, the return value is zero. To get extended error information, call GetLastError.
	template <typename R> R WIN32_SHOULD_SUCCEED(R call) {
		if(0 == call) {
			breakOnWindowsError();
		}
		return call;
	}

#else
#	define WIN32_SHOULD_SUCCEED(CALL) (CALL)
#	define WIN32_SHOULD_HAVE_SUCCEEDED
#endif

uses namespace GFX {
	uses struct SBGRC_8888;
}



namespace IBM {

	// TODO: move to own header!
	// Escaped AT (set 2) scan codes of an IBM PC-compliant keyboard, translated to XT (set 1).
	//   default scan code on IBM PC-compatible keyboards
	//   fixed to mechanical keyboard layout, i.e. independent of input layout (useful for games)
	enum class ScanCode : UInt2B {
		none,
		escape, _1, _2, _3, _4, _5, _6, _7, _8, _9, _0, minus, equals, backspace,
		tab, q, w, e, r, t, y, u, i, o, p, leftBracket, rightBracket, return_,
		leftControl, a, s, d, f, g, h, j, k, l, semicolon, apostrophe, accentGrave,
		leftShift, backslash, z, x, c, v, b, n, m, comma, period, slash, rightShift,
		numpadMultiply,
		leftAlt, spaceBar, altGr,
		f1, f2, f3, f4, f5, f6, f7, f8, f9, f10, numLock, scrollLock,
		numpad7, numpad8, numpad9, numpadSubtract,
		numpad4, numpad5, numpad6, numpadAdd,
		numpad1, numpad2, numpad3,
		numpad0, numpadDecimal,
		f11 = 0x57, f12,
		// Escaped scan codes for extended keys:
		numpadEnter = 0xe01c, rightControl,
		numpadDivide = 0xe035,
		rightAlt = 0xe038,
		home = 0xe047, upArrow, pageUp,
		leftArrow = 0xe04b,
		rightArrow = 0xe04d,
		end = 0xe04f, downArrow, pageDown, insert, delete_,
		leftWindows = 0xe05b, rightWindows, apps,
		// Fake SHIFT keystrokes are inserted for backwards compatibility:
		fakeLeftShift = 0xe02a,
		fakeRightShift = 0xe036
	};

} // IBM


namespace OS {


	// Looks up the name of the key at the given scan code, in the language of the current keyboard layout.
	//   requires at least 20 characters
	//   returns a pointer to the new terminating zero (original pointer on failure)
	UTF16Unit * nameOf(IBM::ScanCode, UTF16Unit *);


	namespace Win32 {


		// Reports the last error code to the given supervisor.
		void OUTLINE_CODE reportLastError(Supervisor const *);

		// Prints the description of a Win32 error code.
		//   requires 65536 characters
		//   returns a pointer to the new terminating zero (original pointer on failure)
		//   uses current locale
		UTF16Unit * textFromErrorCode(
			UTF16Unit   text[65536],
			UIntLeast4B code
		);


		// Win32 handles are always a multiple of four (see The Old New Thing, Why are kernel HANDLEs always a multiple of four?).
		auto constexpr handleAlignment = 4;


		struct alignas(4096) Module;


		// HANDLE in Win32s process management functions.
		struct alignas(handleAlignment) Process;

		using ProcessID = UInt4B;


		// HANDLE in Win32s thread management functions.
		struct alignas(handleAlignment) Thread;

		using ThreadID = UInt4B;


		// HANDLE in Win32s file and driver functions.
		struct alignas(handleAlignment) File;

		UCount constexpr pathLengthLimit = 32767; // see MSDN, Naming Files, Paths, and Namespaces


		// FILETIME, the timestamp Win32 uses for file systems: a 8-B-integer representing the number of 100-nanosecond intervals
		//  since January 1, 1601 (UTC).
		using FileDate = UInt8B;


		// HMONITOR
		struct Monitor;


		// HICON/HCURSOR (since Win32s cursors are also icons).
		struct Icon;


		namespace GDI {


			// HDC, an abstract device and its state. Could be a window (left-to-right or right-to-left), a printer, or a metafile.
			struct DeviceContext;


			struct Bitmap;


		} // GDI


		// COORD
		struct PointOnConsole {
			UInt2B x;
			UInt2B y;
		};


		// A position in pixels relative to the top-left corner of the primary screen. POINT compatible.
		struct PointOnScreen {
			SInt4B x;
			SInt4B y;
		};


		// RECT
		struct Rectangle {
			SInt4B left;
			SInt4B top;
			SInt4B right;
			SInt4B bottom;
		};


		// The codes of all virtual keys (including the mouses).
		enum class VirtualKeyCode : UInt1B {
			leftMouseButton  = 0x01, // VK_LBUTTON
			rightMouseButton = 0x02, // VK_RBUTTON
			middleMouseButton= 0x04, // VK_MBUTTON
			x1MouseButton    = 0x05, // VK_XBUTTON1
			x2MouseButton    = 0x06, // VK_XBUTTON2
			backspace        = 0x08, // VK_BACK
			tab              = 0x09, // VK_TAB
			clear            = 0x0C, // VK_CLEAR
			enter            = 0x0D, // VK_RETURN
			shift            = 0x10, // VK_SHIFT
			control          = 0x11, // VK_CONTROL
			alt              = 0x12, // VK_MENU
			pause            = 0x13, // VK_PAUSE
			capsLock         = 0x14, // VK_CAPITAL
			escape           = 0x1B, // VK_ESCAPE
			spaceBar         = 0x20, // VK_SPACE
			pageUp           = 0x21, // VK_PRIOR
			pageDown         = 0x22, // VK_NEXT
			end              = 0x23, // VK_END
			home             = 0x24, // VK_HOME
			leftArrow        = 0x25, // VK_LEFT
			upArrow          = 0x26, // VK_UP
			rightArrow       = 0x27, // VK_RIGHT
			downArrow        = 0x28, // VK_DOWN
			select           = 0x29, // VK_SELECT
			print            = 0x2A, // VK_PRINT
			execute          = 0x2B, // VK_EXECUTE
			printScreen      = 0x2C, // VK_SNAPSHOT
			insert           = 0x2D, // VK_INSERT
			Delete           = 0x2E, // VK_DELETE
			help             = 0x2F, // VK_HELP
			_0 = '0', _1, _2, _3, _4, _5, _6, _7, _8, _9,
			a = 'A', b, c, d, e, f, g, h, i, j, k, l, m, n, o, p, q, r, s, t, u, v, w, x, y, z,
			leftWindows   = 0x5B, // VK_LWIN
			rightWindows  = 0x5C, // VK_RWIN
			applications  = 0x5D, // VK_APPS
			sleep         = 0x5F, // VK_SLEEP
			numpad0       = 0x60, // VK_NUMPAD0
			numpad1       = 0x61, // VK_NUMPAD1
			numpad2       = 0x62, // VK_NUMPAD2
			numpad3       = 0x63, // VK_NUMPAD3
			numpad4       = 0x64, // VK_NUMPAD4
			numpad5       = 0x65, // VK_NUMPAD5
			numpad6       = 0x66, // VK_NUMPAD6
			numpad7       = 0x67, // VK_NUMPAD7
			numpad8       = 0x68, // VK_NUMPAD8
			numpad9       = 0x69, // VK_NUMPAD9
			numpadMultiply= 0x6A, // VK_MULTIPLY
			numpadAdd     = 0x6B, // VK_ADD
			seperator     = 0x6C, // VK_SEPARATOR
			numpadSubtract= 0x6D, // VK_SUBTRACT
			numpadDecimal = 0x6E, // VK_DECIMAL
			numpadDivide  = 0x6F, // VK_DIVIDE
			f1            = 0x70, // VK_F1
			f2            = 0x71, // VK_F2
			f3            = 0x72, // VK_F3
			f4            = 0x73, // VK_F4
			f5            = 0x74, // VK_F5
			f6            = 0x75, // VK_F6
			f7            = 0x76, // VK_F7
			f8            = 0x77, // VK_F8
			f9            = 0x78, // VK_F9
			f10           = 0x79, // VK_F10
			f11           = 0x7A, // VK_F11
			f12           = 0x7B, // VK_F12
			numLock       = 0x90, // VK_NUMLOCK
			scrollLock    = 0x91, // VK_SCROLL
			numpadEquals  = 0x92, // VK_OEM_NEC_EQUAL
			oem1          = 0xBA, // VK_OEM_1
			plus          = 0xBB, // VK_OEM_PLUS
			comma         = 0xBC, // VK_OEM_COMMA
			minus         = 0xBD, // VK_OEM_MINUS
			period        = 0xBE  // VK_OEM_PERIOD
		};

		// Looks up the name of the given virtual key code, in the language of the current keyboard layout.
		//   requires at least 20 characters
		//   returns a pointer to the new terminating zero (original pointer on failure)
		UTF16Unit * nameOf(VirtualKeyCode, UTF16Unit *);


		//  M E M O R Y

		struct Region : Opaque {

			// Reserves memory of the given size, in bytes.
			//   content is undefined
			//   must be release()d (even with zero size)
			//   returns nullptr on failure
			Byte * tryToReserve(USize);

			// Reserves memory of the given size, in bytes.
			//   content is zeroed
			//   must be release()d (even with zero size)
			//   returns nullptr on failure
			Byte * tryToReserveZeroed(USize);

			// Resizes the given memory block to the given size.
			//   input must not be nullptr
			//   new content is undefined
			//   must be release()d (even with zero size)
			//   on failure, returns nullptr and leaves input untouched
			Byte * tryToReallocate(void const *, USize);

			void release(void const *);

			// Destroys the region and all memory in it.
			void destroy();

		};

		Region & processPool();

		//  must be destroy()d
		//  returns nullptr on failure
		Region * tryToCreateRegion();


		// Reserves memory of the given size, in bytes.
		//   content is undefined
		//   must be release()d (even with zero size)
		//   returns nullptr on failure
		Byte * tryToReserve(USize);

		// Reserves memory of the given size, in bytes.
		//   content is zeroed
		//   must be release()d (even with zero size)
		//   returns nullptr on failure
		Byte * tryToReserveZeroed(USize);

		// Resizes the given memory block to the given size.
		//   input must not be nullptr
		//   new content is undefined
		//   must be release()d (even with zero size)
		//   on failure, returns nullptr and leaves input untouched
		Byte * tryToReallocate(void const *, USize);

		void release(void const *);


		// Returns the amount of virtual memory that Windows has committed for the process.
		USize bytesCommittedForThisProcess();

		// Returns the amount of physical memory installed on the system.
		UInt8B bytesOfPhysicalMemory();

		// Removes as many pages as possible from the working set of this process.
		void minimizeWorkingSet();

		// Removes the given range of pages from the working set of this process.
		void pageOut(void *, USize);

		// Sets this processs stack reserve and stack commit size to Visual C++ default.
		// This is an ugly hack, but its required because most APIs and drivers create their threads with the default stack size of
		//  the process. If this is very large, hundreds of megabytes are wasted.
		void setDefaultStackOptions();


		//  M O D U L E S

		// A pseudo-variable which always points to the base address of the module the code is compiled to.
		extern Module * const toThisModule;


		//  P R O C E S S E S

		// Returns this processs command line as a zero-terminated string.
		UTF16Unit const * commandLine();

		// Returns the individual command line arguments of this process as an array of zero-terminated strings.
		//   must be called only once
		Range<UTF16Unit const *> parsedCommandLine();

		// Writes the path to the current directory of the process.
		//   returns a pointer to the terminating zero
		//   absolute
		//   does not always end with a backslash
		UTF16Unit * currentDirectory(UTF16Unit [260 + 1]);

		// Sets this processs current directory.
		bool currentDirectoryIs(UTF16Unit const *);

		// Exits the calling process normally, performing regular cleanup and detaching DLLs.
		WONT_RETURN exitProcess(UInt4B exitCode);

		// Terminates the calling process immediately without any cleanup or DLL detaching.
		WONT_RETURN terminateProcess(UInt4B exitCode);


		//  C O N S O L E

		// Returns the standard input. Symbolic; do not close!
		File & getStdin();

		// Returns the standard output. Symbolic; do not close!
		File & getStdout();

		// Returns the standard error output. Symbolic; do not close!
		File & getStderr();

		void resizeConsole(File & console, UInt2B, UInt2B);

		File * tryToCreateConsoleRedirection(UTF16Unit const * path);


		namespace stdin {

			// Waits for any input to stdin. Most commonly used for press any key to continue in console programs.
			void await();

			// Reads from stdin until a line break is hit.
			//   result is always zero-terminated (zero must be included in size)
			//   returns pointer to terminating zero
			//   excess input is discarded
			//   non-ASCII characters come out undefined
			//   on failure, returns the original pointer
			ASCIICharacter * readLine(ASCIICharacter *, UCount);

			// Reads from stdin until a line break is hit.
			//   result is always zero-terminated (zero must be included in size)
			//   returns pointer to terminating zero
			//   excess input is discarded
			//   on failure, returns the original pointer
			UTF16Unit * readLine(UTF16Unit *, UCount);

		}


		//  C U R S O R

		// System cursors are shared resources and must not be destroyed.
		Icon & systemCursor_arrow();
		Icon & systemCursor_cross();
		Icon & systemCursor_sizeLeftRight();
		Icon & systemCursor_sizeUpDown();
		Icon & systemCursor_sizeAll();
		Icon & systemCursor_busy();
		Icon & systemCursor_busyInBackground();

		void cursorPositionIs(PointOnScreen const &);

		void curserIconIs(Icon &);

		void curserIconIsBlank();


		//  C L I P B O A R D

		// Copies the text in the clipboard to the given zero-terminated UTF-16 string.
		//   returns no on failure or if the destination is too small
		bool copyFromClipboard(UTF16Unit * toResult, UTF16Unit * toEndOfResult);

		// Copies the given zero-terminated ASCII string to the clipboard.
		//   returns no on failure
		//   line breaks must be carriage return + line feed to work properly
		bool copyToClipboard(ASCIICharacter const *);

		// Copies the given zero-terminated UTF-16 string to the clipboard.
		//   returns no on failure
		//   line breaks must be carriage return + line feed to work properly
		bool copyToClipboard(UTF16Unit const *);

		// Copies the given bitmap to the clipboard.
		//   returns no on failure
		bool copyToClipboard_bottomUp(
			UIntLeast2B             widthInPixels,
			UIntLeast2B             heightInPixels,
			GFX::SBGRC_8888 const *
		);


		//  P O W E R   M A N A G E M E N T

		void enableStandbyForSystemAndScreen();
		void disableStandbyForSystemAndScreen();





		using Ticks = UInt8B;

		Ticks performanceCounterFrequencyInHz();

		// Returns the current value of the high-performance counter.
		Ticks performanceCounter();

		// Returns the CPUs internal time stamp counter. See http://en.wikipedia.org/wiki/Time_Stamp_Counter.
		UInt8B cpuCounter();

		// Prints the current date & time (Coordinated Universal Time) as YYYY-MM-DD hh-mm-ss. Useful for logs and unique file names.
		//   returns a pointer to the strings end
		UTF16Unit * writeCurrentDateAndTime(UTF16Unit[19]);


		namespace Settings {

			// Returns the factor for the mouse speed (which can be customized through the mouse Control Panel application).
			//   in the range of 1 (slowest) to 20 (fastest)
			//   on failure, returns the default of 10
			UInt4B tryToGetMouseSpeedMultiplier();

		} // Settings






		bool tryToCreatePipeForChildProcess(
			File * & out,
			File * & in
		);

		// Opens an existing pipe.
		//   the name must be zero-terminated and shorter than 249 characters
		//   returns no on failure
		//   must be close()d otherwise
		//   must be located on the local system
		//   opened for reading & writing
		File * tryToOpenLocalDuplexPipe(UTF16Unit const * name);


		// Captures the current call stack (excluding this call) as a zero-terminated string.
		//   writes an empty string on failure
		//   layout:
		//     our program.exe                  + 0x12345678
		//     <unknown module>
		//     Kernel32.dll                     + 0x12345678
		//     MaximalLengthIs32CharactersPerF + 0x12345678
		//     
		//   up to 16 entries
		void captureCallStack(UTF16Unit[16 * 48]);






		// A character in the current system codepage.
		using ANSICharacter = char;


		enum class CodePage : UIntLeast2B {
			unknown       = 0, // defaults to users setting in some cases
			cyrillic_1251 = 1251,
			latin_1252    = 1252,
			ascii         = 20127,
			utf8          = 65001
		};


		// Tries to detect the codepage of the given data.
		//   returns CodePage::unknown on failure
		CodePage detectCodePage(
			Byte const [],
			UCount     length
		);

		// Converts the given characters from the given code page to UTF-16.
		//   the UTF-16 buffer should be four times as long as the ANSI buffer (accounts for the worst case)
		//   returns the end of the result
		UTF16Unit * utf16FromCodePage(
			CodePage               codepage,
			UTF16Unit *            toResult,
			ASCIICharacter const * toANSI,
			UCount                 lengthOfANSI
		);
		UTF16Unit * utf16FromCodePage(
			CodePage               codepage,
			UTF16Unit *            toResult,
			ASCIICharacter const * toANSI,
			ASCIICharacter const * toEndOfANSI
		);


	} // Win32


} // OS