// An abstraction layer for Single Instruction, Multiple Data (SIMD). SIMD is available on all modern CPUs (SSE on x86, NEON
//  on ARM, AltiVec on PowerPC), but the syntax is highly compiler-dependent.



// Test Clang first (it may emulate GCC or Visual C++). Clang and GCC are mostly compatible (with very few exceptions).
#if __clang__ || __GNUC__

	using IMPL_FLOATX2   = float  __attribute__((vector_size( 8)));
	using IMPL_FLOATX4   = float  __attribute__((vector_size(16)));
	using IMPL_INTX4     = int    __attribute__((vector_size(16)));
	using IMPL_DOUBLEX2  = double __attribute__((vector_size(16)));
	using IMPL_UINT1BX4  = UInt1B __attribute__((vector_size(16))); // pad 12 lanes for SSEs native register size
	using IMPL_UINT1BX8  = UInt1B __attribute__((vector_size(16))); // pad eight lanes for SSEs native register size
	using IMPL_UINT1BX16 = UInt1B __attribute__((vector_size(16)));
	using IMPL_SINT1BX16 = SInt1B __attribute__((vector_size(16)));
	using IMPL_UINT2BX4  = UInt2B __attribute__((vector_size(16))); // pad four lanes for SSEs native register size
	using IMPL_UINT2BX8  = UInt2B __attribute__((vector_size(16)));
	using IMPL_UINT2BX16 = UInt2B __attribute__((vector_size(32)));
	using IMPL_UINT4BX4  = UInt4B __attribute__((vector_size(16)));
	using IMPL_SINT4BX4  = SInt4B __attribute__((vector_size(16)));
	using IMPL_UINT8BX2  = UInt8B __attribute__((vector_size(16)));

#elif _MSC_VER >= 1900 // Microsoft Visual C++ 2015 or later?

	// These are built-in types, but theyre only visible when defined.
	__if_not_exists(__m128) { // avoid collisions with Microsoft headers
		 union __declspec(intrin_type align(16)) __m128  {  float f32[4]; };
		 union __declspec(intrin_type align(16)) __m128i {    int i32[4]; };
		struct __declspec(intrin_type align(16)) __m128d { double f64[2]; };
	}
	using IMPL_UINT1BX4  = __m128i; // only first quarter in use
	using IMPL_UINT1BX8  = __m128i; // only first half in use
	using IMPL_UINT1BX16 = __m128i;
	using IMPL_SINT1BX16 = __m128i;
	using IMPL_UINT2BX4  = __m128i; // only first half in use
	using IMPL_UINT2BX8  = __m128i;
	struct IMPL_UINT2BX16 {
		__m128i a;
		__m128i b;
	};
	using IMPL_UINT4BX4 = __m128i;
#	define IMPL_FLOATX2  __m128 /* only first half in use */
#	define IMPL_FLOATX4  __m128
#	define IMPL_INTX4    __m128i
#	define IMPL_DOUBLEX2 __m128d

#else
#	error this code requires Visual C++ 2015, Clang, or GCC
#endif

#if TARGET_OS_IS_WIN32
#	if COMPILED_BY_VISUAL_CPP
#		define IMPL_SIMD_CALL __vectorcall
#	elif COMPILED_BY_CLANG || COMPILED_BY_GCC
#		define IMPL_SIMD_CALL __attribute__((vectorcall))
#	else
#		error TODO
#	endif
#elif TARGET_OS_IS_UNIX
	// Useless there is only one calling convention:
#	define IMPL_SIMD_CALL
#else
#	error TODO
#endif

// Usually __vectorcall, which improves function parameter handling with SIMD significantly.
#define SIMD_CALL IMPL_SIMD_CALL


// All modern CPU architectures offer SIMD: SSE on x86, NEON on ARM, AltiVec on PowerPC. Define basic types and operations.
//   use struct wrappers for vectors where possible:
//     built-in vector types on Clang & GCC:
//      > statically typed: int __attribute__((vector_size(16))) differs from char __attribute__((vector_size(8)))
//      > built-in operators (vector versions of scalar operators)
//      > overloading operators is not allowed
//     built-in vector types on Visual C++:
//      > dynamically-typed: 4דint is 16דchar is both __m128i
//      > no built-in operators
//      > overloading is allowed
//     struct wrappers are the only way to fit them both
//     no performance impact on Visual C++ 2015 or higher with global optimization

struct floatx2 { // two lanes of 32-bit floating-point numbers
	IMPL_FLOATX2 floats;
};
struct floatx4 { // four lanes of 32-bit floating-point numbers
	IMPL_FLOATX4 floats;
};
struct SInt4Bx4 { // four lanes of signed 32-bit integers
	IMPL_INTX4 ints;
};
using doublex2 = IMPL_DOUBLEX2; // two lanes of 64-bit floating-point numbers
struct UInt1Bx4 { // four lanes of 8-bit unsigned integers
	IMPL_UINT1BX4 uints;
};
struct UInt1Bx8 { // eight lanes of 8-bit unsigned integers
	IMPL_UINT1BX8 uints;
};
struct UInt1Bx16 { // 16 lanes of 8-bit unsigned integers
	IMPL_UINT1BX16 uints;
};
struct SInt1Bx16 { // 16 lanes of 8-bit signed integers
	IMPL_SINT1BX16 ints;
};
struct UInt2Bx4 { // four lanes of 16-bit unsigned integers
	IMPL_UINT2BX4 uints;
};
struct UInt2Bx8 { // eight lanes of 16-bit unsigned integers
	IMPL_UINT2BX8 uints;
};
struct UInt2Bx16 { // 16 lanes of 16-bit unsigned integers
	IMPL_UINT2BX16 uints;
};
struct UInt4Bx4 { // four lanes of 32-bit unsigned integers
	IMPL_UINT4BX4 uints;
};
// Four lanes of 8-bit integers.
//   saturation arithmetic with [0, 255] range
//   unsigned normalized:
//     0 is treated as 0
//     255 is treated as 1
//     64  128 == 32 (0.25  0.5 == 0.125)
struct UNorm1Bx4 {
	IMPL_UINT1BX4 unorms;
};




// Loads two elements from the given address.
//  no alignment required
//  memory order (lowest address becomes first element)
floatx2 SIMD_CALL loadx2(float const [2]);

// Creates a vector of two elements with the given value.
floatx2 SIMD_CALL loadx2(float);

// Extracts lane [0].
float SIMD_CALL a_of(floatx2);

// Extracts lanes [0] and [1].
floatx2 SIMD_CALL ab_of(floatx4);

// Broadcasts lane [0]s value.
floatx2 SIMD_CALL aa_of(floatx2);
floatx2 SIMD_CALL aa_of(floatx4);

// Broadcasts lane [1]s value.
floatx2 SIMD_CALL bb_of(floatx2);
floatx2 SIMD_CALL bb_of(floatx4);

// Extracts lanes [2] and [3].
floatx2 SIMD_CALL cd_of(floatx4);

floatx2 SIMD_CALL operator *(floatx2, floatx2);

floatx2 SIMD_CALL operator +(floatx2, floatx2);

floatx2 SIMD_CALL operator -(floatx2, floatx2);



// Loads a vector of four 32-bit floating-point numbers from the given address. The address may be unaligned.
floatx4 SIMD_CALL loadx4(float const *);
floatx4 SIMD_CALL xyz0From(float const *);
floatx4 SIMD_CALL xyzwFrom(float, float, float, float);
SInt4Bx4 SIMD_CALL xyzwFrom(SInt4B, SInt4B, SInt4B, SInt4B);
// Loads a vector of three 32-bit floating-point numbers from the given address. The address may be unaligned.
floatx4 SIMD_CALL xyzFrom(float const *);
// Loads a vector of three 32-bit signed integers from the given address. The address may be unaligned.
SInt4Bx4 SIMD_CALL xyzFrom(SInt4B const *);
// Loads a vector of two 32-bit floating-point numbers from the given address. The address may be unaligned.
floatx4 SIMD_CALL xy00From(float const *);
floatx4 SIMD_CALL xy__From(float, float);
floatx4 SIMD_CALL xy00From(float, float); // XY -> XY00
SInt4Bx4 SIMD_CALL xy00From(SInt4B, SInt4B);
floatx4 SIMD_CALL x0z0From(float, float); // XZ -> X0Z0
floatx4 SIMD_CALL xyxyFrom(floatx4, floatx4); // XYZW ABCD -> XYAB
SInt4Bx4 SIMD_CALL xyxyFrom(SInt4Bx4, SInt4Bx4); // XYZW ABCD -> XYAB
floatx4 SIMD_CALL yxyxFrom(floatx4, floatx4); // XYZW ABCD -> YXBA
floatx4 SIMD_CALL xyz_wFrom(floatx4, floatx4); // XYZW ABCD -> XYZD
floatx4 SIMD_CALL x___From(float); // X -> X???
floatx4 SIMD_CALL x000From(float); // X -> X000
SInt4Bx4 SIMD_CALL x000From(SInt4B);
// X -> XXXX
floatx4 SIMD_CALL loadx4(float);
SInt4Bx4 SIMD_CALL xxxxFrom(SInt4B);
// XY -> XY00
floatx4 SIMD_CALL xy00From(doublex2);
// Loads two 8-bit integers from the given address and casts them to 32-bit integers. The address may be unaligned.
SInt4Bx4 SIMD_CALL xy00From(SInt1B const *);
// Loads two 16-bit integers from the given address and casts them to 32-bit integers. The address may be unaligned.
SInt4Bx4 SIMD_CALL xy00From(UInt2B const *);
// Loads four 16-bit integers from the given address and casts them to 32-bit integers. The address may be unaligned.
SInt4Bx4 SIMD_CALL xyzwFrom(UInt2B const *);
// Loads four 8-bit integers from the given 4 B (in memory order) and casts them to 32-bit integers.
SInt4Bx4 SIMD_CALL xyzwFrom(UInt4B);
// Loads four 16-bit integers from the given address and casts them to 32-bit integers. The address may be unaligned.
SInt4Bx4 SIMD_CALL xyzwFrom(SInt2B const *);
SInt4Bx4 SIMD_CALL xyz0From(SInt2B const *);
SInt4Bx4 SIMD_CALL xy00From(SInt4B const *);
UInt4B SIMD_CALL roundedToSInt1Bx4(floatx4);
// X -> X0
doublex2 SIMD_CALL x0From(double);
// X -> XX
doublex2 SIMD_CALL xxFrom(double);
doublex2 SIMD_CALL xyFrom(double, double);
doublex2 SIMD_CALL xyFrom(double const *);
doublex2 SIMD_CALL xyFrom(SInt4B, SInt4B);
// Stores the X lane at the given address. The address may be unaligned.
void SIMD_CALL xTo(float *, floatx4 xyzw);
void SIMD_CALL xTo(double *, doublex2 xyzw);
// Stores the X and Y lanes at the given address. The address may be unaligned.
void SIMD_CALL xyTo(float *, floatx4 xyzw);
void SIMD_CALL xyTo(double *, doublex2 xyzw);
void SIMD_CALL xyTo(SInt2B *, SInt4Bx4 xyzw);
void SIMD_CALL xyTo(SInt4B *, SInt4Bx4 xyzw);
// Stores the Z and W lanes at the given address. The address may be unaligned.
void SIMD_CALL zwTo(float *, floatx4 xyzw);
// Stores the X, Y, and Z lanes at the given address. The address may be unaligned.
void SIMD_CALL xyzTo(float *, floatx4 xyzw);
void SIMD_CALL xyzTo(SInt2B *, SInt4Bx4 xyzw);
void SIMD_CALL xyzTo(SInt4B *, SInt4Bx4 xyzw);
// Stores the X, Y, Z, and W lanes at the given address. The address may be unaligned.
void SIMD_CALL xyzwTo(float *, floatx4 xyzw);
void SIMD_CALL xyzwTo(UInt2B *, SInt4Bx4 xyzw);
void SIMD_CALL xyzwTo(SInt2B *, SInt4Bx4 xyzw);

float SIMD_CALL xOf(floatx4);
float SIMD_CALL yOf(floatx4);
float SIMD_CALL zOf(floatx4);
float SIMD_CALL wOf(floatx4);
double SIMD_CALL xOf(doublex2);
SInt4B SIMD_CALL xOf(SInt4Bx4);
SInt4B SIMD_CALL yOf(SInt4Bx4);
SInt4B SIMD_CALL zOf(SInt4Bx4);
SInt4B SIMD_CALL wOf(SInt4Bx4);

floatx4 SIMD_CALL x0y0Of(floatx4);
floatx4 SIMD_CALL xxxxOf(floatx4);
floatx4 SIMD_CALL xxxyOf(floatx4);
floatx4 SIMD_CALL xxyyOf(floatx4);
floatx4 SIMD_CALL xy0_Of(floatx4);
floatx4 SIMD_CALL xyxxOf(floatx4);
floatx4 SIMD_CALL xyxyOf(floatx4);
floatx4 SIMD_CALL xyyxOf(floatx4);
floatx4 SIMD_CALL xyz0Of(floatx4);
floatx4 SIMD_CALL xyz1Of(floatx4);
floatx4 SIMD_CALL xyzxOf(floatx4);
floatx4 SIMD_CALL xz__Of(floatx4);
floatx4 SIMD_CALL xzywOf(floatx4);
floatx4 SIMD_CALL xzzzOf(floatx4);
floatx4 SIMD_CALL y___Of(floatx4);
floatx4 SIMD_CALL y_w_Of(floatx4);
floatx4 SIMD_CALL y000Of(floatx4);
floatx4 SIMD_CALL yx__Of(floatx4);
floatx4 SIMD_CALL yx00Of(floatx4);
floatx4 SIMD_CALL yxxyOf(floatx4);
floatx4 SIMD_CALL yxyxOf(floatx4);
floatx4 SIMD_CALL yyxxOf(floatx4);
floatx4 SIMD_CALL yyyyOf(floatx4);
floatx4 SIMD_CALL yz__Of(floatx4);
floatx4 SIMD_CALL yzxzOf(floatx4);
floatx4 SIMD_CALL yzxwOf(floatx4);
floatx4 SIMD_CALL yzzyOf(floatx4);
floatx4 SIMD_CALL yzwxOf(floatx4);
floatx4 SIMD_CALL z___Of(floatx4);
floatx4 SIMD_CALL zx__Of(floatx4);
floatx4 SIMD_CALL zxyzOf(floatx4);
floatx4 SIMD_CALL zxywOf(floatx4);
floatx4 SIMD_CALL zxwyOf(floatx4);
floatx4 SIMD_CALL zyxwOf(floatx4);
floatx4 SIMD_CALL zyzzOf(floatx4);
floatx4 SIMD_CALL zzxxOf(floatx4);
floatx4 SIMD_CALL zzyyOf(floatx4);
floatx4 SIMD_CALL zzzxOf(floatx4);
floatx4 SIMD_CALL zzzzOf(floatx4);
floatx4 SIMD_CALL zw__Of(floatx4);
floatx4 SIMD_CALL wz__Of(floatx4);
floatx4 SIMD_CALL wwyxOf(floatx4);
floatx4 SIMD_CALL wwwwOf(floatx4);
floatx4 SIMD_CALL negatedZ(floatx4);
SInt4Bx4 SIMD_CALL xzywOf(SInt4Bx4);
SInt4Bx4 SIMD_CALL yyyyOf(SInt4Bx4);
SInt4Bx4 SIMD_CALL zw__Of(SInt4Bx4);
SInt4Bx4 SIMD_CALL wwwwOf(SInt4Bx4);

floatx4 SIMD_CALL sumOfXY(floatx4); // X + Y (in first lane only!)
floatx4 SIMD_CALL sumOfXYZ(floatx4); // X + Y + Z (in first lane only!)

SInt4Bx4 SIMD_CALL dotOfXY(SInt4Bx4); // X + Y (in first lane only!)
floatx4 SIMD_CALL interleavedXY(floatx4, floatx4); // XYZW ABCD -> XAYB

// Converts the given 32-bit integers to 32-bit floating-point numbers.
floatx4 SIMD_CALL asFloats(SInt4Bx4);
// Converts the given 32-bit floating-point numbers in the X and Y lane to 64-bit floating-point numbers.
doublex2 SIMD_CALL asDoubles(floatx4);
// Converts the given 32-bit signed integers in the X and Y lane to 64-bit floating-point numbers.
doublex2 SIMD_CALL asDoubles(SInt4Bx4);

floatx4 SIMD_CALL floatx4Zero();
SInt4Bx4 SIMD_CALL SInt4Bx4_zero();
floatx4 SIMD_CALL floatx4Infinity();
// A mask for the sign bits of all four floating-point numbers.
floatx4 SIMD_CALL floatx4SignMask_XYZW();
// A mask for the sign bits of the floating-point numbers in the Y and W lane.
floatx4 SIMD_CALL floatx4SignMask_YW();
// A mask for the sign bits of the floating-point numbers in the X and Z lane.
floatx4 SIMD_CALL floatx4SignMask_XZ();
// A mask for the sign bit of the floating-point number in the X lane.
floatx4 SIMD_CALL floatx4SignMask_X();
// A mask for the sign bit of the floating-point number in the Y lane.
floatx4 SIMD_CALL floatx4SignMask_Y();
// A mask for the sign bit of the floating-point number in the Z lane.
floatx4 SIMD_CALL floatx4SignMask_Z();
// A mask for the sign bit of the floating-point number in the W lane.
floatx4 SIMD_CALL floatx4SignMask_W();
// Generates a mask for the sign bits of two IEEE754 floating-point numbers.
doublex2 SIMD_CALL doublex2SignMask();

floatx4 SIMD_CALL operator +(floatx4, floatx4);
floatx4 & SIMD_CALL operator +=(floatx4 &, floatx4);
floatx4 SIMD_CALL operator -(floatx4, floatx4);
floatx4 & SIMD_CALL operator -=(floatx4 &, floatx4);
floatx4 SIMD_CALL operator *(floatx4, floatx4);
floatx4 & SIMD_CALL operator *=(floatx4 &, floatx4);
floatx4 SIMD_CALL operator /(floatx4, floatx4);
floatx4 & SIMD_CALL operator /=(floatx4 &, floatx4);
floatx4 SIMD_CALL operator &(floatx4, floatx4);
floatx4 SIMD_CALL operator |(floatx4, floatx4);
floatx4 SIMD_CALL operator ^(floatx4, floatx4);
#if COMPILED_BY_VISUAL_CPP // TODO: built-in on GCC/Windows
	doublex2 SIMD_CALL operator +(doublex2, doublex2);
	doublex2 SIMD_CALL operator *(doublex2, doublex2);
	doublex2 SIMD_CALL operator &(doublex2, doublex2);
	SInt4Bx4 SIMD_CALL operator +(SInt4Bx4, SInt4Bx4);
	SInt4Bx4 SIMD_CALL operator -(SInt4Bx4, SInt4Bx4);
	SInt4Bx4 & SIMD_CALL operator -=(SInt4Bx4 &, SInt4Bx4);
	SInt4Bx4 SIMD_CALL operator ^(SInt4Bx4, SInt4Bx4);
	SInt4Bx4 SIMD_CALL operator &(SInt4Bx4, SInt4Bx4);
	SInt4Bx4 SIMD_CALL operator <<(SInt4Bx4, UIntLeast1B);
#endif
floatx4 SIMD_CALL andnot(floatx4, floatx4); // a & ~b
SInt4Bx4 SIMD_CALL andnot(SInt4Bx4, SInt4Bx4); // a & ~b
floatx4 SIMD_CALL minimumOf(floatx4, floatx4);
floatx4 SIMD_CALL minimumOf(floatx4, floatx4, floatx4);
SInt4Bx4 SIMD_CALL minimumOf(SInt4Bx4, SInt4Bx4);
doublex2 SIMD_CALL minimumOf(doublex2, doublex2);
floatx4 SIMD_CALL maximumOf(floatx4, floatx4);
floatx4 SIMD_CALL maximumOf(floatx4, floatx4, floatx4);
SInt4Bx4 SIMD_CALL maximumOf(SInt4Bx4, SInt4Bx4);
doublex2 SIMD_CALL maximumOf(doublex2, doublex2);

// Broadcasts the greatest value in the vector.
//   returns all NaNs if any lane is NaN
floatx4 SIMD_CALL maximumOf(floatx4);

// Broadcasts the sum of all lanes.
floatx4 SIMD_CALL sumOf(floatx4);

// Merges a and b according to the given bitmask.
//   picks bits from a if the according mask bit is set, and from b otherwise
floatx4 SIMD_CALL selectedByBitmask(floatx4 a, floatx4 b, floatx4 mask);

// Shifts each lanes value to the right.
//   shifts in zeroes
//   shift must not exceed 31 bits
SInt4Bx4 SIMD_CALL shiftRightLogical(SInt4Bx4, UIntLeast1B);

// Shifts each lanes value to the right.
//   replicates the sign bit
//   shift must not exceed 31 bits
SInt4Bx4 SIMD_CALL shiftRightArithmetic(SInt4Bx4, UIntLeast1B);

// Casts all elements to 1-B unsigned integers.
//   all lanes must be in [0, 255] range
UInt1Bx4 SIMD_CALL asUInt1B(SInt4Bx4);


//  no for negative zero or NaN
bool SIMD_CALL areAllLanesPositive(floatx4);

//  no for NaN
bool SIMD_CALL isAnyLaneGreater(floatx4, floatx4);
bool SIMD_CALL isAnyLaneGreater(floatx4, float);

//  no for NaN
bool SIMD_CALL areAllLanesLower(floatx4, floatx4);
bool SIMD_CALL areAllLanesLower(floatx4, float);

//  no for NaN
bool SIMD_CALL areAllLanesLowerOrEqual(floatx4, floatx4);
bool SIMD_CALL areAllLanesLowerOrEqual(floatx4, float);

//  no for NaN
bool SIMD_CALL areAllLanesFinite(floatx4);

//  no for NaN
bool SIMD_CALL areAllLanesWithin(float lowerLimit, floatx4, float upperLimit);

bool SIMD_CALL equalXY(floatx4, floatx4);
bool SIMD_CALL equalXY(SInt4Bx4, SInt4Bx4);
bool SIMD_CALL isNonZero_xOrYOrZ(floatx4);
bool SIMD_CALL equalXYZ(floatx4, floatx4);
bool SIMD_CALL equalXYZ(SInt4Bx4, SInt4Bx4);
bool SIMD_CALL xyBelow(floatx4, floatx4);
bool SIMD_CALL xyBelow(SInt4Bx4, SInt4Bx4);
bool SIMD_CALL xyzBelow(floatx4, floatx4);
bool SIMD_CALL xyzBelow(SInt4Bx4, SInt4Bx4);
bool SIMD_CALL xyzLowerOrEqual(floatx4, floatx4);
bool SIMD_CALL xyzLowerOrEqual(SInt4Bx4, SInt4Bx4);
bool SIMD_CALL xyzWithin(SInt4B lowerLimit, SInt4Bx4, SInt4B upperLimit);

// Returns a bit mask per lane:
//   1 where left < right
//   0 where left >= right
//   0 where either side is NaN
floatx4 SIMD_CALL mask_isLower(floatx4 l, floatx4 r);

// Returns a bit mask per lane:
//   1 where left <= right
//   0 where left > right
//   0 where either side is NaN
floatx4 SIMD_CALL mask_isLowerOrEqual(floatx4 l, floatx4 r);

// Returns a bit mask per lane:
//   1 where left == right
//   0 where left != right
//   0 where either side is NaN
floatx4 SIMD_CALL mask_isEqual(floatx4 l, floatx4 r);

// Returns a bit mask per lane:
//   1 where left != right
//   0 where left == right
//   0 where either side is NaN
floatx4 SIMD_CALL mask_isNotEqual(floatx4 l, floatx4 r);

// Returns a bit mask per lane:
//   1 where left > right
//   0 where left <= right
//   0 where either side is NaN
floatx4 SIMD_CALL mask_isGreater(floatx4 l, floatx4 r);

// Returns a bit mask per lane:
//   1 where finite
//   0 where NaN/infinite
floatx4 SIMD_CALL mask_isFinite(floatx4 abcd);


bool SIMD_CALL isWithin(
	SInt4Bx4 lowerLimit,
	SInt4Bx4 value,
	SInt4Bx4 upperLimit
);

floatx4 SIMD_CALL clamped(
	floatx4 lowerLimit,
	floatx4 value,
	floatx4 upperLimit
);

floatx4 SIMD_CALL absoluteOf(floatx4);
SInt4Bx4 SIMD_CALL absoluteOf(SInt4Bx4);

// Checks if X, Y, or Z contains a subnormal, non-zero floating-point number.
bool SIMD_CALL isSubnormal_xOrYOrZ(floatx4);

floatx4 SIMD_CALL doubled(floatx4);

// Converts to integer by truncating all digits after the decimal point.
//   all lanes must be in [-2 147 483 648, +2 147 483 520] range
//     deviation from signed 4-B integer maximum is due to floating-point accuracy
//   other interpretations:
//     floored() for positive numbers; ceiled() for negative numbers
//     rounding towards zero
SInt4Bx4 SIMD_CALL truncated(floatx4);

// Rounds to integer.
//   all lanes must be in [-2 147 483 648, +2 147 483 520] range
//     deviation from signed 4-B integer minimum/maximum is due to floating-point accuracy
SInt4Bx4 SIMD_CALL rounded(floatx4);

// Converts to integer by rounding towards negative infinity.
//   all lanes must be in [-2 147 483 648, +2 147 483 520] range
//     deviation from signed 4-B integer maximum is due to floating-point accuracy
SInt4Bx4 SIMD_CALL floored(floatx4);

// Rounds four floating-point numbers in the range of [01] to four integers in the range of [0255].
//   input below 0.0 clamps to 0; input above 1.0 clamps to 255
//   useful for color handling
UInt4B SIMD_CALL UNorm1Bx4From(floatx4);

// Casts all elements to 2-B unsigned integers.
//   all lanes must be in [0, 65 535] range
UInt2Bx4 SIMD_CALL asUInt2B(SInt4Bx4);



// Loads four elements from the given address.
//  no alignment required
//  memory order (lowest address becomes first element)
UInt1Bx4 SIMD_CALL loadx4(UInt1B const [4]);

// Creates a vector of four elements with the given value.
UInt1Bx4 SIMD_CALL loadx4(UInt1B);

// Writes all four elements to the given address.
//  no alignment required
//  memory order (lowest address gets first element)
void SIMD_CALL store(UInt1B[4], UInt1Bx4);

// Expands all elements to 4-B signed integers.
SInt4Bx4 SIMD_CALL asSInt4B(UInt1Bx4);

// Expands all elements to 2-B unsigned integers.
UInt2Bx4 SIMD_CALL asUInt2B(UInt1Bx4);




// Creates a vector of four elements with the given value.
UInt2Bx4 SIMD_CALL loadx4(UInt2B);

// Broadcasts lane [0]s value.
UInt2Bx4 SIMD_CALL aaaa_of(UInt2Bx4);

// Broadcasts lane [1]s value.
UInt2Bx4 SIMD_CALL bbbb_of(UInt2Bx4);

// Broadcasts lane [2]s value.
UInt2Bx4 SIMD_CALL cccc_of(UInt2Bx4);

// Broadcasts lane [3]s value.
UInt2Bx4 SIMD_CALL dddd_of(UInt2Bx4);

UInt2Bx4 SIMD_CALL operator *(UInt2Bx4, UInt2Bx4);

UInt2Bx4 SIMD_CALL operator +(UInt2Bx4, UInt2Bx4);

UInt2Bx4 SIMD_CALL operator -(UInt2Bx4, UInt2Bx4);





// Shifts each lanes value to the right.
//   shifts in zeroes
//   implementation-defined if shift exceeds lane bit length (much like standard C++ shifts)
UInt4Bx4 SIMD_CALL shiftRightLogical(UInt4Bx4, UIntLeast1B);

UInt4Bx4 SIMD_CALL operator +(UInt4Bx4, UInt4Bx4);

// Merges a and b according to the given bitmask.
//   for each lane, picks a if the according mask bit is zero, and b otherwise
UInt4Bx4 SIMD_CALL selectedByBitmask(UInt4Bx4 a, UInt4Bx4 b, UInt4Bx4 mask);

// Converts all elements to 1-B unsigned integers.
//   over- and underflows are undefined
UInt1Bx4 SIMD_CALL asUInt1B(UInt4Bx4);




// Loads four elements from the given address.
//  no alignment required
//  memory order (lowest address becomes first element)
UNorm1Bx4 SIMD_CALL loadUNormx4(UInt1B const [4]);

// Creates a vector of four lanes with the given value.
UNorm1Bx4 SIMD_CALL loadUNorm1Bx4(UInt1B);

// Creates a vector of four lanes with the given values.
UNorm1Bx4 SIMD_CALL loadUNorm1Bx4(UInt1B, UInt1B, UInt1B, UInt1B);

// Converts the given floating-point numbers in [0, 255] range to unsigned normalized numbers.
//   round-to-nearest
//   over- and underruns are saturated
UNorm1Bx4 SIMD_CALL asUNorm1B_0_255(floatx4);

// Converts the given floating-point numbers in [0, 1] range to unsigned normalized numbers.
//   round-to-nearest
//   over- and underruns are saturated
UNorm1Bx4 SIMD_CALL asUNorm1B(floatx4);

UNorm1Bx4 SIMD_CALL operator +(UNorm1Bx4, UNorm1Bx4);

UNorm1Bx4 SIMD_CALL operator -(UNorm1Bx4, UNorm1Bx4);

UNorm1Bx4 SIMD_CALL operator *(UNorm1Bx4, UNorm1Bx4);

UNorm1Bx4 SIMD_CALL operator |(UNorm1Bx4, UNorm1Bx4);

// For each lane, computes lx  rx + ly  ry.
//   useful for image processing:
//     alpha blending: foreground * alpha + background * (1 - alpha)
//     lighting: ambient * light + diffuse * light
UNorm1Bx4 SIMD_CALL dotProductOf(
	UNorm1Bx4 lx,
	UNorm1Bx4 rx,
	UNorm1Bx4 ly,
	UNorm1Bx4 ry
);

// Broadcasts lane [3]s value.
UNorm1Bx4 SIMD_CALL dddd_of(UNorm1Bx4);

// Writes all four elements to the given address.
//  no alignment required
//  memory order (lowest address gets first element)
void SIMD_CALL store(UInt1B[4], UNorm1Bx4);







// Converts the given number to a zero-padded hexadecimal string. Returns a register with eight ASCII characters in memory
//  order.
UInt1Bx8 SIMD_CALL toHexadecimalZeroPaddedString_reg(UInt4B);

// Expands all elements to 2-B unsigned integers.
UInt2Bx8 SIMD_CALL asUInt2B(UInt1Bx8);




// Loads 16 elements from the given address.
//  no alignment required
//  memory order (lowest address becomes first element)
SInt1Bx16 SIMD_CALL loadx16(ASCIICharacter const [16]);

// Creates a vector of 16 elements with the given value.
SInt1Bx16 SIMD_CALL loadx16(ASCIICharacter);

// Returns a bit mask per lane:
//   1 where left < right
//   0 where left >= right
//   0 where either side is NaN
SInt1Bx16 SIMD_CALL mask_isLower(SInt1Bx16, SInt1Bx16);

// Returns a bit mask per lane:
//   1 where left > right
//   0 where left <= right
//   0 where either side is NaN
SInt1Bx16 SIMD_CALL mask_isGreater(SInt1Bx16, SInt1Bx16);

// Returns a bit mask per lane:
//   1 where left == right
//   0 where left != right
//   0 where either side is NaN
SInt1Bx16 SIMD_CALL mask_isEqual(SInt1Bx16, SInt1Bx16);


SInt1Bx16 SIMD_CALL operator &(SInt1Bx16, SInt1Bx16);
SInt1Bx16 SIMD_CALL operator |(SInt1Bx16, SInt1Bx16);

//  memory order (0 is first lane from memory)
//  masked means all bits set
//  returns 16 if no lane is found
UIntLeast1B SIMD_CALL indexOfFirstMaskedLaneIn(SInt1Bx16);




// Loads 16 elements from the given address.
//  no alignment required
//  memory order (lowest address becomes first element)
UInt1Bx16 SIMD_CALL loadx16(UInt1B const [16]);

// Creates a vector of 16 elements with the given value.
UInt1Bx16 SIMD_CALL loadx16(UInt1B);

// Loads eight elements from the first address and appends eight elements from the second address.
//  no alignment required
//  memory order (lowest address becomes first element)
//  useful for matrix transposing
UInt1Bx16 SIMD_CALL loadx8x2(UInt1B const *, UInt1B const *);

// Converts the given number to a zero-padded hexadecimal string. Returns a register with 16 ASCII characters in memory order.
UInt1Bx16 SIMD_CALL toHexadecimalZeroPaddedString_reg(UInt8B);

// Expands all elements to 2-B unsigned integers.
UInt2Bx16 SIMD_CALL asUInt2B(UInt1Bx16);

// Writes all 16 elements to the given address.
//  no alignment required
//  memory order (lowest address gets first element)
void SIMD_CALL store(UInt1B[16], UInt1Bx16);

// Writes the first eight elements to the first address and the last eight to the other.
//  no alignment required
//  memory order (lowest address gets first element)
//  useful for matrix transposing
void SIMD_CALL store8x2(UInt1B *, UInt1B *, UInt1Bx16);

bool SIMD_CALL areAllLanesZero(UInt1Bx16);

UInt1Bx16 SIMD_CALL operator &(UInt1Bx16, UInt1Bx16);
UInt1Bx16 SIMD_CALL operator |(UInt1Bx16, UInt1Bx16);
UInt1Bx16 SIMD_CALL operator ^(UInt1Bx16, UInt1Bx16);
UInt1Bx16 SIMD_CALL operator <<(UInt1Bx16, UIntLeast1B); // componentwise left shift

// Merges a and b according to the given bitmask.
//   for each lane, picks a if the according mask bit is zero, and b otherwise
UInt1Bx16 SIMD_CALL selectedByBitmask(UInt1Bx16 a, UInt1Bx16 b, UInt1Bx16 mask);

UInt1Bx16 & SIMD_CALL operator |=(UInt1Bx16 &, UInt1Bx16);



// Writes all eight elements to the given address.
//  no alignment required
//  memory order (lowest address gets first element)
void SIMD_CALL store(UInt2B[8], UInt2Bx8);




// Writes all 16 elements to the given address.
//  no alignment required
//  memory order (lowest address gets first element)
void SIMD_CALL store(UInt2B[16], UInt2Bx16);

UInt2Bx8 SIMD_CALL lanes0To7Of(UInt2Bx16);