// PlayStation 1 Image (*.TIM) loader. Compatible with the generic image loader in Formats ImageFile.hpp.
//
// References:
//   best: Sony Computer Entertainment America, File Formats (March 2000 version at
//    http://www.psxdev.net/forum/viewtopic.php?f=49&t=146)
//   overly complicated and wrong in general file layout:
//     Klarth, TIM Graphic Formats (PSX 2D Graphics)
//     QhimmWiki, PSX/TIM format



#include "framework.hpp"
#include "SIMD.hpp"
#include "GFX.hpp"
#include "Formats.hpp"
#include "Sony PSX TIM.hpp"
#include "errors/Sony PSX TIM.hpp"



namespace Sony {


	namespace PSX {


		namespace TIM {


			struct Header {
				// File signature.
				//   always 16
				//   Sonys definition is overly complicated:
				//     bits 815: version (always zero)
				//     bits 1631: reserved (always zero)
				UInt4B_LE signature;
				// bit    meaning
				//  0 2  pixel mode according to Flags enumeration
				//  3     Palette present?
				//  431  reserved; must be zero
				UInt4B_LE flags;
			};


			// A rectangle for direct transfer to the frame buffer.
			//   TIM uses this for palettes and for pixels
			struct FrameBufferRectangle {
				// Size of this structure plus the payload.
				UInt4B_LE sizeInBytes;
				// Origin of the data in the PSX frame buffer.
				//   in pixels relative to the top-left corner
				//   for palettes, X must be a multiple of 16
				UInt2B_LE originX;
				UInt2B_LE originY;
				// Width of the rectangle, in 2-B words.
				//   for palettes, this equals the width in pixels
				//   for pixels, the color depth must be taken into account to compute the width in pixels
				UInt2B_LE widthInWords;
				// Height of the rectangle, in pixels.
				UInt2B_LE heightInPixels;
			};


			namespace Flags { enum V {
				i4         = 0, // 4-bit indexed colors
				i8         = 1, // 8-bit indexed colors
				r5g5b5c1   = 2,
				r8g8b8     = 3,
				mixed      = 4,
				hasPalette = 8
			}; }


			// Default palette for 4-bit TIMs which dont bring their own palette.
			local GFX::SBGRC_5551_LE const default4BitPalette[16] = {
				{ 0x8000 }, { 0x8842 }, { 0x9084 }, { 0x98c6 }, // grayscale
				{ 0xa108 }, { 0xa94a }, { 0xb18c }, { 0xb9ce },
				{ 0xc631 }, { 0xce73 }, { 0xd6b5 }, { 0xdef7 },
				{ 0xe739 }, { 0xef7b }, { 0xf7bd }, { 0xffff }
			};

			// Default palette for 8-bit TIMs which dont bring their own palette.
			local GFX::SBGRC_5551_LE const default8BitPalette[256] = {
				{ 0x8000 }, { 0x8000 }, { 0x8000 }, { 0x8000 }, { 0x8000 }, { 0x8000 }, { 0x8000 }, { 0x8000 }, // grayscale (wastes bits, but whatever)
				{ 0x8421 }, { 0x8421 }, { 0x8421 }, { 0x8421 }, { 0x8421 }, { 0x8421 }, { 0x8421 }, { 0x8421 },
				{ 0x8842 }, { 0x8842 }, { 0x8842 }, { 0x8842 }, { 0x8842 }, { 0x8842 }, { 0x8842 }, { 0x8842 },
				{ 0x8c63 }, { 0x8c63 }, { 0x8c63 }, { 0x8c63 }, { 0x8c63 }, { 0x8c63 }, { 0x8c63 }, { 0x8c63 },
				{ 0x9084 }, { 0x9084 }, { 0x9084 }, { 0x9084 }, { 0x9084 }, { 0x9084 }, { 0x9084 }, { 0x9084 },
				{ 0x94a5 }, { 0x94a5 }, { 0x94a5 }, { 0x94a5 }, { 0x94a5 }, { 0x94a5 }, { 0x94a5 }, { 0x94a5 },
				{ 0x98c6 }, { 0x98c6 }, { 0x98c6 }, { 0x98c6 }, { 0x98c6 }, { 0x98c6 }, { 0x98c6 }, { 0x98c6 },
				{ 0x9ce7 }, { 0x9ce7 }, { 0x9ce7 }, { 0x9ce7 }, { 0x9ce7 }, { 0x9ce7 }, { 0x9ce7 }, { 0x9ce7 },
				{ 0xa108 }, { 0xa108 }, { 0xa108 }, { 0xa108 }, { 0xa108 }, { 0xa108 }, { 0xa108 }, { 0xa108 },
				{ 0xa529 }, { 0xa529 }, { 0xa529 }, { 0xa529 }, { 0xa529 }, { 0xa529 }, { 0xa529 }, { 0xa529 },
				{ 0xa94a }, { 0xa94a }, { 0xa94a }, { 0xa94a }, { 0xa94a }, { 0xa94a }, { 0xa94a }, { 0xa94a },
				{ 0xad6b }, { 0xad6b }, { 0xad6b }, { 0xad6b }, { 0xad6b }, { 0xad6b }, { 0xad6b }, { 0xad6b },
				{ 0xb18c }, { 0xb18c }, { 0xb18c }, { 0xb18c }, { 0xb18c }, { 0xb18c }, { 0xb18c }, { 0xb18c },
				{ 0xb5ad }, { 0xb5ad }, { 0xb5ad }, { 0xb5ad }, { 0xb5ad }, { 0xb5ad }, { 0xb5ad }, { 0xb5ad },
				{ 0xb9ce }, { 0xb9ce }, { 0xb9ce }, { 0xb9ce }, { 0xb9ce }, { 0xb9ce }, { 0xb9ce }, { 0xb9ce },
				{ 0xbdef }, { 0xbdef }, { 0xbdef }, { 0xbdef }, { 0xbdef }, { 0xbdef }, { 0xbdef }, { 0xbdef },
				{ 0xc210 }, { 0xc210 }, { 0xc210 }, { 0xc210 }, { 0xc210 }, { 0xc210 }, { 0xc210 }, { 0xc210 },
				{ 0xc631 }, { 0xc631 }, { 0xc631 }, { 0xc631 }, { 0xc631 }, { 0xc631 }, { 0xc631 }, { 0xc631 },
				{ 0xca52 }, { 0xca52 }, { 0xca52 }, { 0xca52 }, { 0xca52 }, { 0xca52 }, { 0xca52 }, { 0xca52 },
				{ 0xce73 }, { 0xce73 }, { 0xce73 }, { 0xce73 }, { 0xce73 }, { 0xce73 }, { 0xce73 }, { 0xce73 },
				{ 0xd294 }, { 0xd294 }, { 0xd294 }, { 0xd294 }, { 0xd294 }, { 0xd294 }, { 0xd294 }, { 0xd294 },
				{ 0xd6b5 }, { 0xd6b5 }, { 0xd6b5 }, { 0xd6b5 }, { 0xd6b5 }, { 0xd6b5 }, { 0xd6b5 }, { 0xd6b5 },
				{ 0xdad6 }, { 0xdad6 }, { 0xdad6 }, { 0xdad6 }, { 0xdad6 }, { 0xdad6 }, { 0xdad6 }, { 0xdad6 },
				{ 0xdef7 }, { 0xdef7 }, { 0xdef7 }, { 0xdef7 }, { 0xdef7 }, { 0xdef7 }, { 0xdef7 }, { 0xdef7 },
				{ 0xe318 }, { 0xe318 }, { 0xe318 }, { 0xe318 }, { 0xe318 }, { 0xe318 }, { 0xe318 }, { 0xe318 },
				{ 0xe739 }, { 0xe739 }, { 0xe739 }, { 0xe739 }, { 0xe739 }, { 0xe739 }, { 0xe739 }, { 0xe739 },
				{ 0xeb5a }, { 0xeb5a }, { 0xeb5a }, { 0xeb5a }, { 0xeb5a }, { 0xeb5a }, { 0xeb5a }, { 0xeb5a },
				{ 0xef7b }, { 0xef7b }, { 0xef7b }, { 0xef7b }, { 0xef7b }, { 0xef7b }, { 0xef7b }, { 0xef7b },
				{ 0xf39c }, { 0xf39c }, { 0xf39c }, { 0xf39c }, { 0xf39c }, { 0xf39c }, { 0xf39c }, { 0xf39c },
				{ 0xf7bd }, { 0xf7bd }, { 0xf7bd }, { 0xf7bd }, { 0xf7bd }, { 0xf7bd }, { 0xf7bd }, { 0xf7bd },
				{ 0xfbde }, { 0xfbde }, { 0xfbde }, { 0xfbde }, { 0xfbde }, { 0xfbde }, { 0xfbde }, { 0xfbde },
				{ 0xffff }, { 0xffff }, { 0xffff }, { 0xffff }, { 0xffff }, { 0xffff }, { 0xffff }, { 0xffff }
			};


			// Checks whether the given 12 B mark the beginning of a TIM with some certainty.
			bool isTIM(
				Byte const * const toFirst12B,
				USize const        size
			) {
				MUST(size > 12);

				// Check the header:
				//   the signature must be 16
				//   the 28 reserved flags must be zero
				//   the pixel format (in the three least significant bits) must be one of the five known values
				auto & header = *reinterpret_cast<Header const *>(toFirst12B);
				if(littleEndian(16u) != header.signature) {
					return no;
				}
				auto const flags = littleEndian(header.flags);
				if(flags > 15) {
					return no;
				}
				if((flags & 7) > 4) { // five known formats
					return no;
				}

				// A pixels block or a palettes block follows. We dont care; they both start with a size.
				//   the size must not exceed the file size
				//   the size should not be below the minimum size either (this catches zeroed files)
				auto const pixelsOrPalettesSize = littleEndian(reinterpret_cast<FrameBufferRectangle const *>(toFirst12B + 8)->sizeInBytes);
				if(pixelsOrPalettesSize < 12 || pixelsOrPalettesSize > size - 8) {
					return no;
				}

				return yes; // pretty sure here
			}


			// Parses the given TIM file.
			//   must have been checked successfully via isTIM()
			//   returns no on failure
			//   otherwise, must be destroy()d after use
			//   keeps references to the given bytes
			bool tryToLoad(
				File &                   result,
				Byte const * const       toBeginning,
				Byte const * const       toEnd,
				Supervisor const * const supervisorOrNull
			) {
				MUST(toEnd >= toBeginning);
				MUST(isTIM(toBeginning, USize(toEnd - toBeginning)));
				auto toBytes = toBeginning;

				// Read & verify the header:
				//   there must be sufficient room
				//   the signature must be 16
				//   the 28 reserved flags must be zero
				//   the color depth (in the three least significant bits) must be one of the five known values
				auto const & header = *reinterpret_cast<Header const *>(toBytes);
				auto const   flags = littleEndian(header.flags);
				if(flags > 15) {
					report(supervisorOrNull, Error::badFlags);
					return no;
				}
				auto const formatFlag = Flags::V(flags & 7);
				if(formatFlag == Flags::mixed) {
					report(supervisorOrNull, Error::unsupportedFormat);
					return no;
				} else if(formatFlag > Flags::mixed) {
					report(supervisorOrNull, Error::badPixelFormat);
					return no;
				}
				toBytes += sizeof header;

				// If a palette is implied, read & verify it:
				if(flags & Flags::hasPalette) {
					//   there must be sufficient room
					//   the size must be valid
					if(sizeof(FrameBufferRectangle) >= toEnd - toBytes) {
						report(supervisorOrNull, Error::eofInPalettesHeader);
						return no;
					}
					auto const & palettesHeader = *reinterpret_cast<FrameBufferRectangle const *>(toBytes);
					auto const   palettesSize   = littleEndian(palettesHeader.sizeInBytes);
					auto const   originX        = littleEndian(palettesHeader.originX);
					auto const   originY        = littleEndian(palettesHeader.originY);
					auto const   palettesWidth  = littleEndian(palettesHeader.widthInWords);
					auto const   palettesHeight = littleEndian(palettesHeader.heightInPixels);

					// Prevent integer overflows:
					if(palettesWidth > 32768) {
						report(supervisorOrNull, Error::imageTooWide);
						return no;
					}
					if(palettesHeight > 32768) {
						report(supervisorOrNull, Error::imageTooWide);
						return no;
					}

					// Check for sufficient room:
					//   dimensions are limited to 32768, so 2 is unreachable
					if(sizeof palettesHeader + palettesWidth * palettesHeight * 2 > palettesSize || palettesSize > USize(toEnd - toBytes)) {
						report(supervisorOrNull, Error::eofInPalettes);
						return no;
					}

					//  the PSX frame buffer is limited to 1024512 pixels
					//  palettes must be aligned to multiples of 16 on X
					//  we can still load other layouts, but a PSX cant
					if(originX + palettesWidth > 1024 || originY + palettesHeight > 512) {
						if(0 == report(supervisorOrNull, Warning::palettesOff)) {
							return no;
						}
					}
					if(0 != originX % 16) {
						if(0 == report(supervisorOrNull, Warning::palettesNotAligned)) {
							return no;
						}
					}

					result.palettesOriginX = originX;
					result.palettesOriginY = originY;
					if(Flags::i4 == formatFlag) {
						result.colorsPerPalette = 16;
					} else if(Flags::i8 == formatFlag) {
						result.colorsPerPalette = 256;
					} else {
						if(0 == report(supervisorOrNull, Warning::needlessPalettes)) {
							return no;
						}
					}

					// Some games place their palettes side-by-side (Ace Combat 2) and other stack them vertically (Ace Combat 3). Try
					//  to account for all combinations.
					result.palettesWidthInPixels  = palettesWidth;
					result.palettesHeightInPixels = palettesHeight;
					result.numberOfPalettes       = palettesWidth * palettesHeight / result.colorsPerPalette;
					result.toPalettes             = reinterpret_cast<GFX::SBGRC_5551_LE const *>(toBytes + sizeof palettesHeader);
					toBytes += palettesSize;

				} else { // No palettes block present (e.g. in Japanese Ace Combat 3)?
					result.palettesWidthInPixels  = 0;
					result.palettesHeightInPixels = 0;
					result.colorsPerPalette       = 0;
					result.numberOfPalettes       = 0;
					result.palettesOriginX        = 0;
					result.palettesOriginY        = 0;
				}

				// Read & verify the pixels:
				if(sizeof(FrameBufferRectangle) > toEnd - toBytes) {
					report(supervisorOrNull, Error::eofInPixelsHeader);
					return no;
				}
				auto const & pixelsHeader   = *reinterpret_cast<FrameBufferRectangle const *>(toBytes);
				auto         pixelsSize     = littleEndian(pixelsHeader.sizeInBytes);
				auto const   pixelsX        = littleEndian(pixelsHeader.originX);
				auto const   pixelsY        = littleEndian(pixelsHeader.originY);
				auto const   widthInWords   = littleEndian(pixelsHeader.widthInWords);
				auto const   heightInPixels = littleEndian(pixelsHeader.heightInPixels);

				// Prevent integer overflows:
				if(widthInWords > 32768) {
					report(supervisorOrNull, Error::imageTooWide);
					return no;
				}
				if(heightInPixels > 32768) {
					report(supervisorOrNull, Error::imageTooWide);
					return no;
				}

				// Check for sufficient room:
				//   dimensions are limited to 32768, so 2 is unreachable
				//   there are TIMs with invalid size information (e.g. in Japanese Ace Combat 3)
				if(sizeof pixelsHeader + widthInWords * heightInPixels * 2 > pixelsSize) {
					if(no == report(supervisorOrNull, Warning::badPixelsSize)) {
						return no;
					}
					pixelsSize = sizeof pixelsHeader + widthInWords * heightInPixels * 2;
				}
				if(pixelsSize > USize(toEnd - toBytes)) {
					report(supervisorOrNull, Error::eofInPixels);
					return no;
				}

				//  the PSX frame buffer is limited to 1024512 pixels
				//  we can still load other layouts, but a PSX cant
				if(pixelsX + widthInWords > 1024 || pixelsY + heightInPixels > 512) {
					if(0 == report(supervisorOrNull, Warning::pixelsOff)) {
						return no;
					}
				}

				if(0 == widthInWords || 0 == heightInPixels) {
					if(no == report(supervisorOrNull, Warning::noPixels)) {
						return no;
					}
				}

				// The width is given in 2-B steps, and the actual width (in pixels) must be computed from it depending on the format:
				if(Flags::i4 == formatFlag) {
					result.format        = Formats::PixelFormat::indexed_4bit;
					result.widthInPixels = 4 * widthInWords;
				} else if(Flags::i8 == formatFlag) {
					result.format        = Formats::PixelFormat::indexed_8bit;
					result.widthInPixels = 2 * widthInWords;
				} else if(Flags::r5g5b5c1 == formatFlag) {
					result.format        = Formats::PixelFormat::bgrc_5551_le;
					result.widthInPixels = widthInWords;
				} else { MUST(Flags::r8g8b8 == formatFlag);
					result.format        = Formats::PixelFormat::bgr_888;
					result.widthInPixels = widthInWords * 2 / 3;
				}

				result.pixelsOriginX = pixelsX;
				result.pixelsOriginY = pixelsY;
				result.widthInWords   = widthInWords;
				result.heightInPixels = heightInPixels;
				result.toPixels       = toBytes + sizeof pixelsHeader;
				toBytes += pixelsSize;

				// Warn if there are extra data after the end of the TIM. The user might miss something.
				if(toEnd > toBytes) {
					report(supervisorOrNull, Info::extraData);
				}
				result.sizeInBytes = UInt4B(toBytes - toBeginning); // dimensions are limited to 32768, so 2 is unreachable

				return yes;
			}

			// Returns the number of palettes in the TIM.
			//   zero under the following circumstances:
			//     the TIM is a true-color image
			//     the TIM does not bring its own palettes (happens, for example, in some Ace Combat 3 files)
			UIntLeast2B numberOfPalettesIn(File const & tim) {
				return tim.numberOfPalettes;
			}

			// Returns the TIMs properties.
			//   width and height may be zero (happens, for example, in some Ace Combat 3 files)
			Formats::ImageProperties propertiesOf(File const & tim) {
				return { tim.widthInPixels, tim.heightInPixels, Formats::ImageType::palette, tim.format, tim.colorsPerPalette };
			}

			// Returns the size of the TIM, in bytes.
			USize sizeOf(File const & tim) {
				return tim.sizeInBytes;
			}

			// Extracts the image using the given palette.
			//   returns no on failure
			//   writes the pixels to the given array, starting with the first line
			//   specify the distance between two consecutive lines (negative for bottom-up)
			bool extractImage(
				File const &             tim,
				UIntLeast2B const        paletteIndex,
				GFX::SBGRC_8888 * const  toResult,
				SSize const              linePadding,
				Supervisor const * const supervisorOrNull
			) {
				MUST(nullptr != toResult);

				if(0 == tim.widthInPixels || 0 == tim.heightInPixels) {
					return yes; // nothing to do
				}

				switch(tim.format) {

					case Formats::PixelFormat::indexed_4bit: {

						GFX::SBGRC_5551_LE const * toPalette;
						if(tim.colorsPerPalette >= 16) {
							MUST(paletteIndex < tim.numberOfPalettes);
							toPalette = tim.toPalettes + paletteIndex * tim.colorsPerPalette;
						} else {
							if(0 == report(supervisorOrNull, Warning::noPalettes)) {
								return no;
							}
							toPalette = default4BitPalette;
						}

						// Instead of doing a color conversion on every pixel, convert the palette.
						GFX::SBGRC_8888 palette[16];
						Formats::bgrc_to_bgrc(palette, 0, toPalette, 0, CAPACITY_OF(palette), 1);

						// TIMs are top-down, whereas the result is bottom-up. Mirror vertically.
						Formats::i4l_to_bgrc_flipY(
							toResult, linePadding,
							tim.toPixels, palette, 0, tim.widthInPixels, tim.heightInPixels
						);

					} break;

					case Formats::PixelFormat::indexed_8bit: {

						GFX::SBGRC_5551_LE const * toPalette;
						if(tim.colorsPerPalette >= 256) {
							MUST(paletteIndex < tim.numberOfPalettes);
							toPalette = tim.toPalettes + paletteIndex * tim.colorsPerPalette;
						} else {
							if(0 == report(supervisorOrNull, Warning::noPalettes)) {
								return no;
							}
							toPalette = default8BitPalette;
						}

						// Instead of doing a color conversion on every pixel, convert the palette.
						GFX::SBGRC_8888 palette[256];
						Formats::bgrc_to_bgrc(palette, 0, toPalette, 0, CAPACITY_OF(palette), 1);

						// TIMs are top-down, whereas the result is bottom-up. Mirror vertically.
						Formats::i8_to_bgrc_flipY(
							toResult, linePadding,
							tim.toPixels, palette, 0, tim.widthInPixels, tim.heightInPixels
						);

					} break;

					case Formats::PixelFormat::bgrc_5551_le:
						MUST(0 == paletteIndex);

						// TIMs are top-down, whereas the result is bottom-up. Mirror vertically.
						Formats::bgrc_to_bgrc_flipY(
							toResult, linePadding,
							reinterpret_cast<GFX::SBGRC_5551_LE const *>(tim.toPixels), 0,
							tim.widthInPixels, tim.heightInPixels
						);

						break;

					case Formats::PixelFormat::bgr_888:
						MUST(0 == paletteIndex);

						// TIMs are top-down, whereas the result is bottom-up. Mirror vertically.
						Formats::bgr_to_bgrc_flipY(
							toResult, linePadding,
							reinterpret_cast<GFX::SBGR_888 const *>(tim.toPixels), 0,
							tim.widthInPixels, tim.heightInPixels
						);

						break;

					default:
						UNREACHABLE;
				}

				return yes;
			}

			// Extracts the palette of the given index from the TIM.
			//   returns no on failure
			//   palette length can be queried via propertiesOf()
			//   writes the colors in original order to the given array
			bool extractPalette(
				File const &      tim,
				UIntLeast2B const paletteIndex,
				GFX::SBGRC_8888 * toResult
			) {
				MUST(paletteIndex < tim.numberOfPalettes);
				MUST(0 < tim.colorsPerPalette);

				auto const toEndOfResult = toResult + tim.colorsPerPalette;
				auto       toSource      = tim.toPalettes + paletteIndex * tim.colorsPerPalette;
				do {
					store(*toResult, sBGRC_8888_from(load(*toSource)));
				} while(++toSource, ++toResult < toEndOfResult);

				return yes;
			}

			void destroy(File &) { }


		} // TIM


	} // PSX


} // Sony