// PlayStation 1 Image (*.TIM) loader. Compatible with the generic image loader in Formats ImageFile.hpp.
//
//  can be found in almost every PlayStation game
//  part of the official SDK
//  designed for the PlayStations GPU  even defines coordinates in the frame buffer to look up on drawing
//  true-color TIMs contain exactly one image
//  paletted TIMs contain just one set of color indices, but different palettes for these indices
//    allows to encode more than one image
//    allows to encode different regions with individual palettes
//    no region-to-palette mapping; can only be found in the according draw instructions to the PlayStation GPU (texture window
//     & palette coordinates)
//
// Done:
//   4-bit indexed
//   8-bit indexed
//   16-bit High Color
//   24-bit True Color
//   excessive palettes
//   no palettes
//
// To do:
//   decode concatenated files (Resident Evil)



namespace Sony {


	namespace PSX {


		// Default image encoding on PlayStation.
		namespace TIM {


			// Checks whether the given 12 B mark the beginning of a TIM with some certainty.
			bool isTIM(
				Byte const * toFirst12B,
				USize        size
			);


			// A parsed TIM file, loaded into memory.
			struct File {
				// The files color depth.
				Formats::PixelFormat format;
				// The number of colors per individual palette. 0 if the image is true-color.
				UInt2B               colorsPerPalette;
				UInt2B               palettesWidthInPixels;
				UInt2B               palettesHeightInPixels;
				// The number of palettes in the file.
				//   0 if the image is true-color
				//   0 if the image does not bring its own palette
				UInt2B               numberOfPalettes;
				UInt2B               widthInWords;
				UInt2B               widthInPixels;
				UInt2B               heightInPixels;
				// Total file size (including headers etc.).
				UInt4B               sizeInBytes;
				// Points to the data of all palettes in a consecutive memory block.
				//   undefined if number of palettes is zero
				GFX::SBGRC_5551_LE const * toPalettes;
				// Points to all pixels in a consecutive memory block.
				Byte const *         toPixels;
				// Origin of the pixels in the frame buffer, in pixels.
				UInt2B               pixelsOriginX;
				UInt2B               pixelsOriginY;
				// Origin of the palettes in the frame buffer, in pixels.
				UInt2B               palettesOriginX;
				UInt2B               palettesOriginY;
			};


			// Parses the given TIM file.
			//   returns no on failure
			//   otherwise, must be destroy()d after use
			//   keeps references to the given bytes
			bool tryToLoad(
				File &            ,
				Byte const *      toBeginning,
				Byte const *      toEnd,
				Supervisor const *
			);

			// Returns the number of palettes in the TIM.
			//   zero under the following circumstances:
			//     the TIM is a true-color image
			//     the TIM does not bring its own palettes (happens, for example, in some Ace Combat 3 files)
			UIntLeast2B numberOfPalettesIn(File const &);

			// Returns the TIMs properties.
			//   width and height may be zero (happens, for example, in some Ace Combat 3 files)
			Formats::ImageProperties propertiesOf(File const &);

			// Returns the size of the TIM, in bytes.
			USize sizeOf(File const &);

			// Extracts the image using the given palette.
			//   returns no on failure
			//   writes the pixels to the given array, starting with the first line
			//   specify the distance between two consecutive lines (negative for bottom-up)
			bool extractImage(
				File const &      ,
				UIntLeast2B       paletteIndex,
				GFX::SBGRC_8888 * ,
				SSize             linePadding,
				Supervisor const *
			);

			// Extracts the palette of the given index from the TIM.
			//   returns no on failure
			//   palette length can be queried via propertiesOf()
			//   writes the colors in original order to the given array
			bool extractPalette(
				File const &     ,
				UIntLeast2B      paletteIndex,
				GFX::SBGRC_8888 *
			);

			void destroy(File &);


		} // TIM


	} // PSX


} // Sony


#define SONY_PSX_TIM 1 // use this format in any loader that is #included later