#include "framework.hpp"
#include "SIMD.hpp"
#include "Text.hpp"
#include "Memory.hpp"
#include "Parallel.hpp" // dtoa.c requires synchronization
#if COMPILED_BY_VISUAL_CPP
#	include <vadefs.h> // va_arg()
#	define va_start __crt_va_start
#	define va_arg   __crt_va_arg
#	define va_end   __crt_va_end
#endif

#if COMPILED_BY_VISUAL_CPP
	// Visual C++ intrinsics. On Clang and GCC, the __builtin_ counterparts are used, so no declaration is required.
	import {
		int memcmp(void const *, void const *, size_t);
		UCount strlen(char const *);
	}
#endif

// There does not seem to be a ftoa() or (floating-point compatible) sprintf() in Win32, so David M. Gays dtoa.c is used.
// It requires four external functions: two for dynamic memory allocation, and two for multithreading. See dtoa.c for details.

	local void * dtoa_alloc(USize const size) {
		return Heap::tryToReserve(size);
	}

	local void dtoa_free(void * const memory) {
		return Heap::release(memory);
	}

	local SpinLock dtoa_locks[2] = { SPIN_LOCK_INITIALIZER, SPIN_LOCK_INITIALIZER };

	local void dtoa_lock(int const index) {
		MUST(isWithin(0, index, 1));

		lock(dtoa_locks[index]);
	}

	local void dtoa_unlock(int const index) {
		MUST(isWithin(0, index, 1));

		unlock(dtoa_locks[index]);
	}


#	if _DEBUG
#		define DEBUG
#	endif

	// use as little STL as possible
#	define Bad_float_h
#	define NO_ERRNO

#	if COMPILED_FOR_X86_32 || COMPILED_FOR_X86_64
#		define IEEE_8087 // IEEE 754 Little Endian
#	else
#		error TODO
#	endif

	// Do not use malloc()/free(), but supplements instead:
#	define Omit_Private_Memory
#	define MALLOC dtoa_alloc
#	define FREE dtoa_free

	// Make dtoa() thread-safe
#	define MULTIPLE_THREADS
#	define ACQUIRE_DTOA_LOCK dtoa_lock
#	define FREE_DTOA_LOCK dtoa_unlock

	// Omit recognition of hexadecimal floating-point values:
#	define NO_HEX_FP
#	define No_Hex_NaN

#	if COMPILED_BY_VISUAL_CPP
#		pragma warning(push)
		// warning C4018: '<=': signed/unsigned mismatch
		// warning C4244: '=': conversion from '__int64' to 'int', possible loss of data
		// warning C4273: 'strtod': inconsistent dll linkage
		// warning C4505: unreferenced local function has been removed
		// warning C4701: potentially uninitialized local variable 'bb' used
		// warning C4703: potentially uninitialized local pointer variable 'bb' used
		// warning C4706: assignment within conditional expression
#		pragma warning(disable: 4018 4244 4273 4505 4701 4703 4706)
#	endif

#	include "external/dtoa.inl"

#	if COMPILED_BY_VISUAL_CPP
#		pragma warning(pop)
#	endif

import_from(ole32) {

	int API_CALL StringFromGUID2(
		UUID const &,
		UTF16Unit * ,
		int
	);

} // ole32



namespace Text {


	// Returns the number of characters in the given, zero-terminated string.
	UCount lengthOf(ASCIICharacter const * toZeroTerminatedString) {
#		if COMPILED_BY_VISUAL_CPP
			return strlen(toZeroTerminatedString);
#		elif COMPILED_BY_CLANG
			return __builtin_strlen(toZeroTerminatedString);
#		else
#			error TODO
#		endif
	}
	UCount lengthOf(UTF16Unit const * toZeroTerminatedString) {

		auto toEnd = toZeroTerminatedString;
		while(0 != *toEnd) {
			++toEnd;
		}

		return UCount(toEnd - toZeroTerminatedString);
	}

	template <typename Character> Character * seekEndOf_template(Character * toText) {
		while(0 != *toText) {
			++toText;
		}
		return toText;
	}

	// Returns a pointer to the terminating zero in the given string.
	ASCIICharacter const * seekEndOf(ASCIICharacter const * toText) {
		return seekEndOf_template(toText);
	}
	ASCIICharacter * seekEndOf(ASCIICharacter * toText) {
		return seekEndOf_template(toText);
	}
	UTF16Unit const * seekEndOf(UTF16Unit const * toText) {
		return seekEndOf_template(toText);
	}
	UTF16Unit * seekEndOf(UTF16Unit * toText) {
		return seekEndOf_template(toText);
	}

	// Creates a range from the given zero-terminated string.
	Range<ASCIICharacter const> range(ASCIICharacter const * toBeginning) {
		return { toBeginning, toBeginning +
#			if COMPILED_BY_VISUAL_CPP
				strlen(toBeginning)
#			elif COMPILED_BY_CLANG
				__builtin_strlen(toBeginning)
#			else
#				error TODO
#			endif
		};
	}
	Range<ASCIICharacter> range(ASCIICharacter * toBeginning) {
		return { toBeginning, toBeginning +
#			if COMPILED_BY_VISUAL_CPP
				strlen(toBeginning)
#			elif COMPILED_BY_CLANG
				__builtin_strlen(toBeginning)
#			else
#				error TODO
#			endif
		};
	}
	Range<UTF16Unit const> range(UTF16Unit const * toBeginning) {
		return { toBeginning, toBeginning + lengthOf(toBeginning) };
	}

	template <typename S, typename D> D * copyZeroTerminated_template(D * toDestination, S const * toSource) {
		cycle {
			*toDestination = D(*toSource);
			if(0 == *toSource) {
				return toDestination;
			}
			++toDestination;
			++toSource;
		}
	}
	ASCIICharacter * copyZeroTerminated(ASCIICharacter * toD, ASCIICharacter const * toS) {
		return copyZeroTerminated_template(toD, toS);
	}
	ASCIICharacter * copyZeroTerminated(ASCIICharacter * toDestination, UTF16Unit const * toSource) { // replaces non-ASCII characters with question marks
		cycle {
			if(*toSource < 128) {
				*toDestination = ASCIICharacter(*toSource);
			} else {
				*toDestination = '?';
			}
			if(0 == *toSource) {
				return toDestination;
			}
			++toDestination;
			++toSource;
		}
	}
	UTF16Unit * copyZeroTerminated(UTF16Unit * toD,ASCIICharacter const * toS) {
		return copyZeroTerminated_template(toD, toS);
	}
	UTF16Unit * copyZeroTerminated(UTF16Unit * toD, UTF8Unit const * toS) {
		return copyZeroTerminated_template(toD, toS);
	}
	UTF16Unit * copyZeroTerminated(UTF16Unit * toD, UTF16Unit const * toS) {
		return copyZeroTerminated_template(toD, toS);
	}
	UTF8Unit * copyZeroTerminated(UTF8Unit * toD, ASCIICharacter const * toS) {
		return copyZeroTerminated_template(toD, toS);
	}
	UTF8Unit * copyZeroTerminated(UTF8Unit * toD, UTF8Unit const * toS) {
		return copyZeroTerminated_template(toD, toS);
	}

	ASCIICharacter * copy(
		ASCIICharacter *                    toDestination,
		Range<ASCIICharacter const> const & source
	) {
		return copy(toDestination, source.toBeginning, source.toEnd);
	}
	ASCIICharacter * copy(
		ASCIICharacter *       toDestination,
		ASCIICharacter const * toSource,
		ASCIICharacter const * toSourceEnd
	) {
		return movs(toDestination, toSource, UCount(toSourceEnd - toSource));
	}
	UTF16Unit * copy(
		UTF16Unit *       toDestination,
		UTF16Unit const * toSource,
		UCount            length
	) {
		return movs(toDestination, toSource, length);
	}
	ASCIICharacter * copy(
		ASCIICharacter *             toDestination,
		ASCIICharacter const * const toSource,
		UCount const                 length
	) {
		return movs(toDestination, toSource, length);
	}
	ASCIICharacter * copyAndAppendZero(
		ASCIICharacter *                    toDestination,
		Range<ASCIICharacter const> const & source
	) {
		return copyAndAppendZero(toDestination, source.toBeginning, source.toEnd);
	}
	ASCIICharacter * copyAndAppendZero(
		ASCIICharacter *             toDestination,
		ASCIICharacter const * const toSource,
		ASCIICharacter const * const toSourceEnd
	) {
		auto const length = UCount(toSourceEnd - toSource);
		toDestination = movs(toDestination, toSource, length);
		*toDestination = 0;
		return toDestination;
	}
	UTF16Unit * copyAndAppendZero(
		UTF16Unit *                    toDestination,
		Range<UTF16Unit const> const & source
	) {
		return copyAndAppendZero(toDestination, source.toBeginning, source.toEnd);
	}
	UTF16Unit * copyAndAppendZero(
		UTF16Unit *             toDestination,
		UTF16Unit const * const toSource,
		UTF16Unit const * const toSourceEnd
	) {
		auto const length = UCount(toSourceEnd - toSource);
		toDestination = movs(toDestination, toSource, length);
		*toDestination = 0;
		return toDestination;
	}
	ASCIICharacter * copyAndAppendZero(
		ASCIICharacter *             toDestination,
		ASCIICharacter const * const toSource,
		UCount const                 length
	) {
		toDestination = movs(toDestination, toSource, length);
		*toDestination = 0;
		return toDestination;
	}
	UTF16Unit * copyAndAppendZero(
		UTF16Unit *            toDestination,
		ASCIICharacter const * toSource,
		UCount                 length
	) {
		while(0 < length) {
			*toDestination = *toSource;
			++toDestination;
			++toSource;
			--length;
		}
		*toDestination = 0;
		return toDestination;
	}


	template <typename Source, typename Destination> Destination * copyUntilZero_template(
		Destination *  toDestination,
		Source const * toSource
	) {
		while(0 != *toSource) {
			*toDestination = *toSource;
			++toDestination;
			++toSource;
		}
		return toDestination;
	}
	ASCIICharacter * copyUntilZero(
		ASCIICharacter *       toDestination,
		ASCIICharacter const * toSource
	) {
		return copyUntilZero_template(toDestination, toSource);
	}
	UTF16Unit * copyUntilZero(
		UTF16Unit *            toDestination,
		ASCIICharacter const * toSource
	) {
		return copyUntilZero_template(toDestination, toSource);
	}
	UTF16Unit * copyUntilZero(
		UTF16Unit *       toDestination,
		UTF16Unit const * toSource
	) {
		return copyUntilZero_template(toDestination, toSource);
	}

	bool isLowerCaseAlphabetical(ASCIICharacter character) {
		return isWithin('a', character, 'z');
	}

	bool isUpperCaseAlphabetical(ASCIICharacter character) {
		return isWithin('A', character, 'Z');
	}

	bool isAlphabetical(ASCIICharacter character) {
		return isLowerCaseAlphabetical(character) || isUpperCaseAlphabetical(character);
	}

	bool isDecimalDigit(ASCIICharacter character) {
		return isWithin('0', character, '9');
	}
	bool isDecimalDigit(UTF16Unit character) {
		return isWithin(UTF16('0'), character, UTF16('9'));
	}
	bool isLowerCaseHexadecimalDigit(ASCIICharacter character) {
		return isWithin('0', character, '9') || isWithin('a', character, 'f');
	}
	bool isUpperCaseHexadecimalDigit(ASCIICharacter character) {
		return isWithin('0', character, '9') || isWithin('A', character, 'F');
	}
	bool isOctalDigit(ASCIICharacter character) {
		return isWithin('0', character, '7');
	}

	// Checks whether the given ASCII character is a control character (zero terminator, bell, line feed, ).
	bool isControlCharacter(ASCIICharacter const c) {
		return UInt1B(c) < 32;
	}

	// Checks whether the given UTF-16 code unit is a ASCII control character (zero terminator, bell, line feed, ) but NOT a
	//  Unicode control character (e.g. right-to-left control characters).
	bool isASCIIControlCharacter(UTF16Unit const c) {
		return c < 32;
	}

	ASCIICharacter toDecimalDigit(UIntLeast1B number) {
		MUST(isWithin(0u, number, 9u));
		return ASCIICharacter('0' + number);
	}

	UIntLeast1B decimalDigitToNumber(ASCIICharacter character) {
		MUST(isWithin('0', character, '9'));
		return character - '0';
	}
	UIntLeast1B decimalDigitToNumber(UTF16Unit character) {
		MUST(isWithin(UTF16('0'), character, UTF16('9')));
		return character - '0';
	}

	// Returns zero for the ASCII character 0, one for 1, etc.
	//   returns a number larger than 9 if the character was no decimal digit
	UIntLeast1B numberFromDecimalDigit_unchecked(ASCIICharacter character) {
		return character - '0';
	}
	UIntLeast2B numberFromDecimalDigit_unchecked(UTF16Unit character) {
		return character - '0';
	}

	// Returns zero for the ASCII character 0, one for 1, etc.
	//   returns a number larger than 7 if the character was no octal digit
	UIntLeast1B numberFromOctalDigit_unchecked(ASCIICharacter character) {
		return character - '0';
	}
	UIntLeast2B numberFromOctalDigit_unchecked(UTF16Unit character) {
		return character - '0';
	}

	UIntLeast1B numberFromUpperCaseHexadecimalDigit(ASCIICharacter character) {
		if(isWithin('A', character, 'F')) {
			return character + (10 - 'A');
		} else {
			return decimalDigitToNumber(character);
		}
	}

	UIntLeast1B numberFromOctalDigit(ASCIICharacter character) {
		MUST(isWithin('0', character, '7'));
		return character - '0';
	}

	ASCIICharacter toHexadecimalDigit(UIntLeast1B number) {
		MUST(isWithin(0u, number, 15u));
		return "0123456789ABCDEF"[number];
	}

	bool isWhitespace(ASCIICharacter const character) {
		// The Unicode standard defines ASCII characters that conforming applications should recognize as whitespace:
		//   9: tab
		//  10: line feed
		//  11: vertical tab (next vertical item)
		//  12: form feed (next page of document)
		//  13: carriage return (revert to beginning of line)
		//  32: space
		return isWithin(9u, UIntLeast1B(character), 13u) || ' ' == character;
	}
	bool isWhitespace(UTF16Unit const character) {
		// U+0009: character tabulation
		// U+000A: line feed
		// U+000B: line tabulation
		// U+000C: form feed
		// U+000D: carriage return
		return isWithin(0x0009, character, 0x000D)
		// U+0020: space
		    || 0x0020 == character
		// U+0085: next line
		    || 0x0085 == character
		// U+00A0: no-break space
		    || 0x00A0 == character
		// U+1680: ogham space mark
		    || 0x1680 == character
		// U+180E: mongolian vowel separator
		    || 0x180E == character
		// U+2000: en quad
		// U+2001: em quad
		// U+2002: en space
		// U+2003: em space
		// U+2004: three-per-em space
		// U+2005: four-per-em space
		// U+2006: six-per-em space
		// U+2007: figure space
		// U+2008: punctuation space
		// U+2009: thin space
		// U+200A: hair space
		// U+200B: zero width space
		// U+200C: zero width non-joiner
		// U+200D: zero width joiner
		    || isWithin(0x2000, character, 0x200D)
		// U+2028: line separator
		// U+2029: paragraph separator
		    || isWithin(0x2028, character, 0x2029)
		// U+202F: narrow no-break space
		    || 0x202F == character
		// U+205F: medium mathematical space
		// U+2060: word joiner
		    || isWithin(0x205F, character, 0x2060)
		// U+3000: ideographic space
		    || 0x3000 == character
		// U+FEFF: zero width non-breaking space
		    || 0xFEFF == character;
	}

	bool isLineBreak(ASCIICharacter const character) {
		// The Unicode standard defines ASCII characters that conforming applications should recognize as line terminators:
		//  10: line feed
		//  11: vertical tab (next vertical item)
		//  12: form feed (next page of document)
		//  13: carriage return (revert to beginning of line)
		return isWithin(10u, UIntLeast1B(character), 13u);
	}
	bool isLineBreak(UTF16Unit const character) {
		// The Unicode standard defines ASCII characters that conforming applications should recognize as line terminators:
		//  10:     line feed
		//  11:     vertical tab (next vertical item)
		//  12:     form feed (next page of document)
		//  13:     carriage return (revert to beginning of line)
		//  U+0085: NEL (next line)
		//  U+2028: line separator
		//  U+2029: paragraph separator
		return isWithin(10u, UIntLeast2B(character), 13u) || 0x85 == character || isWithin(0x2028u, UIntLeast2B(character), 0x2029u);
	}

	bool isWhitespaceButNotLineBreak(ASCIICharacter const character) {
		// See implementations of isWhitespace() and isLineBreak() for details:
		return '\t' == character || ' ' == character;
	}

	// Returns the uppercase version of the given character.
	ASCIICharacter upperCase(ASCIICharacter character) {
		return isWithin('a', character, 'z') ? (character - 32) : character;
	}
	UTF16Unit upperCase(UTF16Unit character) {
		return isWithin(L'a', character, L'z') ? (character - 32) : character;
	}

	// Returns the lowercase version of the given character.
	ASCIICharacter lowerCase(ASCIICharacter character) {
		return isWithin('A', character, 'Z') ? (character + 32) : character;
	}
	UTF16Unit lowerCase(UTF16Unit character) {
		return isWithin(L'A', character, L'Z') ? (character + 32) : character;
	}

	// Lowers the case of the given string.
	ASCIICharacter * lowerCase(
		ASCIICharacter *       toDestination,
		ASCIICharacter const * toSource,
		ASCIICharacter const * toSourcesEnd
	) {
		while(toSource < toSourcesEnd) {
			*toDestination++ = lowerCase(*toSource++);
		}
		return toDestination;
	}

	// Determines whether the given strings equal.
	bool equal(
		Range<ASCIICharacter const> const & a,
		Range<ASCIICharacter const> const & b
	) {
		auto       toA     = a.toBeginning;
		auto       toB     = b.toBeginning;
		auto const toAsEnd = a.toEnd;
		// Different lengths? Not equal!
		if(b.toEnd - toB != toAsEnd - toA) {
			return no;
		}
		// Zero length? Equal!
		if(toA == toAsEnd) {
			return yes;
		}
		// Otherwise, compare unit by unit
		do {
			if(*toA != *toB) {
				return no;
			}
		} while(++toB, ++toA < toAsEnd);
		return yes;
	}

	bool equal(
		Range<ASCIICharacter const> a,
		ASCIICharacter const *      toZeroTerminatedB
	) {
		// Skip the equal part.
		cycle {

			if(a.toBeginning == a.toEnd) {
				// a has ended, so they are equal only if b ends, too.
				return 0 == *toZeroTerminatedB;
			}

			if(*a.toBeginning != *toZeroTerminatedB) {
				// Also reached if b is shorter than a.
				return no;
			}

			++a.toBeginning;
			++toZeroTerminatedB;
		}
	}

	bool equal(ASCIICharacter const * toZeroTerminatedA, ASCIICharacter const * toZeroTerminatedB) {
		cycle {
			if(*toZeroTerminatedA != *toZeroTerminatedB) {
				return no;
			}
			if(0 == *toZeroTerminatedA) {
				return yes;
			}
			++toZeroTerminatedA, ++toZeroTerminatedB;
		}
	}

	// Compares two zero-terminated strings for sorting, just like strcmp().
	//   return value is 
	//     zero if both strings are equal
	//     negative if the left string comes first
	//     positive if the right string comes first
	SIntLeast1B compare(ASCIICharacter const * toZeroTerminatedA, ASCIICharacter const * toZeroTerminatedB) {
		cycle {
			if(*toZeroTerminatedA != *toZeroTerminatedB) {
				return int(*toZeroTerminatedA - *toZeroTerminatedB);
			}
			if(0 == *toZeroTerminatedA) {
				return yes;
			}
			++toZeroTerminatedA, ++toZeroTerminatedB;
		}
	}
	SIntLeast1B compare(ASCIICharacter const * toA, UCount aLength, ASCIICharacter const * toB, UCount bLength) {
		cycle {
			if(0 == (aLength | bLength)) { // Both zero? (saves one branch)
				return 0; // same length
			}
			if(0 == aLength) {
				return -1; // a comes first
			}
			if(0 == bLength) {
				return 1; // b comes first
			}
			if(*toA != *toB) {
				return int(*toA - *toB);
			}
			++toA, ++toB;
			--aLength, --bLength;
		}
	}

	bool equal(UTF16Unit const * toZeroTerminatedA, UTF16Unit const * toZeroTerminatedB) {
		cycle {
			if(*toZeroTerminatedA != *toZeroTerminatedB) {
				return no;
			}
			if(0 == *toZeroTerminatedA) {
				return yes;
			}
			++toZeroTerminatedA, ++toZeroTerminatedB;
		}
	}

	bool beginsWith(ASCIICharacter const * toZeroTerminatedA, ASCIICharacter const * toZeroTerminatedB) {
		cycle {
			if(*toZeroTerminatedA != *toZeroTerminatedB) {
				return 0 == *toZeroTerminatedB; // End of substring?
			}
			if(0 == *toZeroTerminatedB) {
				return yes;
			}
			++toZeroTerminatedA, ++toZeroTerminatedB;
		}
	}

	bool beginsWith(UTF16Unit const * toZeroTerminatedA, ASCIICharacter const * toZeroTerminatedB) {
		cycle {
			if(*toZeroTerminatedA != *toZeroTerminatedB) {
				return 0 == *toZeroTerminatedB; // End of substring?
			}
			if(0 == *toZeroTerminatedB) {
				return yes;
			}
			++toZeroTerminatedA, ++toZeroTerminatedB;
		}
	}

	// Determines whether the given strings equal.
	// (complexity: linear in length of shorter string)
	bool equal(Range<UTF16Unit const> const & a, Range<UTF16Unit const> const & b) {
		auto       toA     = a.toBeginning;
		auto       toB     = b.toBeginning;
		auto const toAsEnd = a.toEnd;
		// Different lengths? Not equal!
		if(b.toEnd - toB != toAsEnd - toA) {
			return no;
		}
		// Zero length? Equal!
		if(toA == toAsEnd) {
			return yes;
		}
		// Otherwise, compare unit by unit
		do {
			if(*toA != *toB) {
				return no;
			}
		} while(++toB, ++toA < toAsEnd);
		return yes;
	}

	// Determines whether the given strings equal without case sensitivity.
	// (complexity: linear in length of shorter string)
	bool equalCaseInsensitive(
		Range<ASCIICharacter const> const & a,
		Range<ASCIICharacter const> const & b
	) {
		if(sizeOf(a) != sizeOf(b)) {
			return no;
		}

		for(auto toA = a.toBeginning, toB = b.toBeginning; toA < a.toEnd; ++toA, ++toB) {
			if(lowerCase(*toA) != lowerCase(*toB)) {
				return no;
			}
		}
		return yes;
	}

	// Determines whether the given string equals the given zero-terminated lowercase literal.
	// Used to detect keywords quickly.
	// (complexity: linear in length of literal)
	bool equalsCaseInsensitiveLiteral(
		ASCIICharacter const * toString,
		ASCIICharacter const * toStringsEnd,
		ASCIICharacter const * toLiteral
	) {
		MUST(0 != toLiteral[0]); // literals must not be empty

		cycle {
			MUST(lowerCase(*toLiteral) == *toLiteral); // literals must be all lowercase
			if(lowerCase(*toString) != *toLiteral) {
				return no;
			}
			++toLiteral;
			++toString;
			if(toString == toStringsEnd) {
				return 0 == *toLiteral; // Same length or end of literal?
			}
		}
	}
	bool equalsCaseInsensitiveLiteral(
		Range<ASCIICharacter const> const & string,
		ASCIICharacter const *              toLiteral
	) {
		MUST(0 != toLiteral[0]); // literals must not be empty

		auto toCharacter = string.toBeginning;
		cycle {
			MUST(lowerCase(*toLiteral) == *toLiteral); // literals must be all lowercase
			if(lowerCase(*toCharacter) != *toLiteral) {
				return no;
			}
			++toLiteral;
			++toCharacter;
			if(toCharacter == string.toEnd) {
				return 0 == *toLiteral; // Same length?
			}
		}
	}
	bool equalsCaseInsensitiveLiteral(
		UTF16Unit const *      toString,
		ASCIICharacter const * toLiteral
	) {
		MUST(0 != toLiteral[0]); // literals must not be empty

		cycle {
			MUST(lowerCase(*toLiteral) == *toLiteral); // literals must be all lowercase
			if(lowerCase(*toString) != *toLiteral) {
				return no;
			}
			++toLiteral;
			++toString;
			if(0 == *toString) {
				return 0 == *toLiteral; // Same length?
			}
		}
	}

	// Detects the given keyword at the beginning of the given text.
	//   requires
	//     the entire literal being found at the beginning of the text (case sensitive)
	//     whitespace following or the text ending
	//   on match, returns yes and advances toText to just after the match
	//   otherwise, returns no and leaves toText untouched
	//   the literal must be zero-terminated
	bool parse(
		ASCIICharacter const * &     toText,
		ASCIICharacter const * const toEnd,
		ASCIICharacter const *       toLiteral
	) {
		MUST(0 != toLiteral[0]); // literals must not be empty

		auto toNextCharacter = toText;
		cycle {

			if(toNextCharacter == toEnd) { // End of text?
				// The text was matched so far. If the literal ends now as well, its a match.
				if(0 == *toLiteral) {
					goto match;
				}
				return no;
			}
			if(*toNextCharacter != *toLiteral) {
				if(0 == *toLiteral && isWhitespace(*toNextCharacter)) {
					goto match;
				}
				return no;
			}

			++toNextCharacter, ++toLiteral;
		}

	match:
		toText = toNextCharacter;
		return yes;
	}

	// Detects the given keyword at the beginning of the given text.
	//   requires
	//     the entire literal being found at the beginning of the text (case insensitive)
	//     whitespace following or the text ending
	//   on match, returns yes and advances toText to just after the match
	//   otherwise, returns no and leaves toText untouched
	//   the literal must
	//     contain only lowercase characters
	//     be zero-terminated
	bool parse_caseInsensitive(
		ASCIICharacter const * &     toText,
		ASCIICharacter const * const toEnd,
		ASCIICharacter const *       toLiteral
	) {
		MUST(0 != toLiteral[0]); // literals must not be empty

		auto toNextCharacter = toText;
		cycle {
			MUST(lowerCase(*toLiteral) == *toLiteral); // literals must be all lowercase

			if(toNextCharacter == toEnd) { // End of text?
				// The text was matched so far. If the literal ends now as well, its a match.
				if(0 == *toLiteral) {
					goto match;
				}
				return no;
			}
			if(lowerCase(*toNextCharacter) != *toLiteral) {
				if(0 == *toLiteral && isWhitespace(*toNextCharacter)) {
					goto match;
				}
				return no;
			}

			++toNextCharacter, ++toLiteral;
		}

	match:
		toText = toNextCharacter;
		return yes;
	}

	// Moves the given ranges beginning to the first non-whitespace character, or to the end. Any white space, control characters
	//  or line breaks are skipped.
	void skipWhitespace(Range<ASCIICharacter const> & text) {
		while(text.toBeginning < text.toEnd && isWhitespace(*text.toBeginning)) {
			++text.toBeginning;
		}
	}
	void skipWhitespace(ASCIICharacter const * & toText, ASCIICharacter const * toEnd) {
		while(toText < toEnd && isWhitespace(*toText)) {
			++toText;
		}
	}

	// Moves the given ranges beginning to the first non-whitespace character, line break, or to the end. Any white space or
	//  control characters are skipped.
	void skipWhitespaceButNotLineBreaks(Range<ASCIICharacter const> & text) {
		while(text.toBeginning < text.toEnd && isWhitespaceButNotLineBreak(*text.toBeginning)) {
			++text.toBeginning;
		}
	}
	void skipWhitespaceButNotLineBreaks(ASCIICharacter const * & toText, ASCIICharacter const * toEnd) {
		while(toText < toEnd && isWhitespaceButNotLineBreak(*toText)) {
			++toText;
		}
	}

	// Moves the given ranges beginning to behind the current line.
	// Useful for parsers to skip single-line comments.
	void skipLine(Range<ASCIICharacter const> & text) {
		while(text.toBeginning < text.toEnd && no == isLineBreak(*text.toBeginning)) {
			++text.toBeginning;
		}
	}
	void skipLine(ASCIICharacter const * & toText, ASCIICharacter const * toEnd) {
		cycle {
			if(toText == toEnd) {
				return;
			}
			if(isLineBreak(*toText)) {
				++toText; // skip the line break
				return;
			}
			++toText;
		}
	}

	// Tries to read text enclosed in double quotes.
	//   Returns yes on success, moves toText to behind the quotation, and assigns the quotations beginning and end (without
	//    the double quotes).
	//   Returns no on failure.
	bool tryToReadQuotation(
		ASCIICharacter const * & toText,
		ASCIICharacter const *   toEndOfText,
		ASCIICharacter const * & toQuotation,
		ASCIICharacter const * & toQuotationsEnd
	) {
		if(toEndOfText - toText < 2 || '"' != *toText) {
			return no; // too short or no double quotes
		}
		++toText; // skip the double quote

		toQuotation = toText;
		while('"' != *toText) {
			if(toText == toEndOfText) {
				return no; // unexpected end of quotation
			}
			++toText;
		}
		toQuotationsEnd = toText;
		++toText; // skip the double quote

		return yes;
	}

	// Tries to read text enclosed in square brackets (common in INI files).
	//   Returns yes on success, moves toText to behind the name, and assigns the names beginning and end (without the
	//    square brackets).
	//   Returns no on failure.
	bool tryToReadINISectionName(
		ASCIICharacter const * & toText,
		ASCIICharacter const *   toEndOfText,
		ASCIICharacter const * & toName,
		ASCIICharacter const * & toNamesEnd
	) {
		if(toEndOfText - toText < 2 || '[' != *toText) {
			return no; // too short or no square brackets
		}
		++toText; // skip the opening bracket

		toName = toText;
		while(']' != *toText) {
			if(toText == toEndOfText) {
				return no; // unexpected end of quotation
			}
			++toText;
		}
		toNamesEnd = toText;
		++toText; // skip the double quote

		return yes;
	}

	// Tries to read a property from an INI file.
	//   Returns yes on success. toText is moved to the end of the property. The name is assigned. If a value exists, it is
	//    assigned; otherwise, it is empty. The value is not filtered for whitespace.
	//   Returns no on failure. toText is moved to the point of failure; the other iterators are undefined.
	bool tryToReadINIProperty(
		ASCIICharacter const * & toText,
		ASCIICharacter const *   toTextsEnd,
		ASCIICharacter const * & toName,
		ASCIICharacter const * & toNamesEnd,
		ASCIICharacter const * & toValue,
		ASCIICharacter const * & toValuesEnd
	) {
		if(toText == toTextsEnd) {
			return no; // empty
		}

		// Read the property name:
		toName = toText;
		{
			auto nextCharacter = *toText;
			if(isAlphabetical(nextCharacter)) { // An identifier?
				do {
					++toText;
				} while(toText < toTextsEnd && '=' != *toText && no == isWhitespace(*toText));
			} else if(isDecimalDigit(nextCharacter)) { // A number?
				do {
					++toText;
				} while(toText < toTextsEnd && isDecimalDigit(*toText));
			} else {
				return no;
			}
			toNamesEnd = toText;
		}

		// Check if a value is assigned (equals sign after optional whitespace):
		while(toText < toTextsEnd && isWhitespace(*toText)) {
			++toText;
		}
		if(toText < toTextsEnd && '=' == *toText) {
			++toText; // skip =
			toValue = toText;

			// The value is terminated either by file end or by line breaks.
			while(toText < toTextsEnd) {
				auto nextCharacter = *toText;
				if(isLineBreak(nextCharacter)) {
					break;
				}
				++toText;
			}

		} else {
			toValue = toText;
		}
		toValuesEnd = toText;

		return yes;
	}

	// Counts the lines of text in the given range.
	// Useful for parsers: error messages typically require a line where the error occured. Instead of complicating the parser by
	//  tracking the current line, it will just track the current text position. Only if an error occurs, the number of lines from
	//  the beginning of the text to the current position is counted to print an error message.
	// (complexity: linear in length of text)
	UCount countLinesIn(Range<ASCIICharacter const> const & text) {
		if(isEmpty(text)) {
			return 0; // empty text does not occupy any lines
		}

		// The number of lines is often requested in two cases:
		//  1) An error occurred while parsing and the parser needs to count the number of lines until the current position.
		//  2) A script file (e.g. INI) is pre-parsed to allocate resources according to the number of lines in it.
		// Case 1) has the text at the current position cached, because it has just been parsed. In case 2), it is useful to keep
		//  the beginning of the text in cache because it is interpreted right after this function returns.
		// Therefore, some cache misses can be saved by actually parsing the text back-to-front.
		UCount result = 1u; // non-empty text occupies at least one line
		auto toCharacter = toEndOf(text) - 1u;
		do {
			if('\n' == *toCharacter) {
				++result;
			}
		} while(--toCharacter != toBeginningOf(text));

		return result;
	}

	// Converts the given ASCII string to a UTF-16 string.
	// (complexity: linear in length of string)
	void toUTF16(
		Range<ASCIICharacter const> const & source,
		Range<UTF16Unit> const &            destination
	) {
		MUST(lengthOf(destination) >= lengthOf(source)); // Insufficient space for ASCII to UTF-16 conversion?

		// ASCII to UTF16 conversion is very easy: just zero-extend the character value to 16 bits. It is, however, important that
		//  there are only ASCII characters (i.e. 7-bit characters) in the source string.
		auto toSource = source.toBeginning;
		auto toDestination = destination.toBeginning;
		while(toSource < source.toEnd) {
			MUST(128u > UInt1B(*toSource)); // Non-ASCII character in ASCII to UTF-16 conversion?
			*(toDestination++) = UTF16Unit(*(toSource++));
		}

	}

	// Converts the given unsigned integer number to a string. The string must be of sufficient room; a pointer to the end of the
	//  written number is returned.
	// highestPowerOfTen is the largest power of ten that has influence on the result, e.g. 100 for numbers up to 255.
	template <
		typename        UnsignedInteger,
		typename        Character,
		UnsignedInteger highestPowerOfTen
	> local Character * formatUnsignedInteger(
		UnsignedInteger value,
		Character *     toDestination
	) {
		MUST(nullptr != toDestination);

		// Handle the case that nothing would be written if the number was zero.
		if(0 == value) {
			*toDestination = '0';
			return toDestination + 1;
		}

		auto * toCharacter = toDestination;

		// Iterate through all relevant powers of ten.
		for(auto currentBase = highestPowerOfTen; currentBase > 0; currentBase /= 10) {
			auto const currentDigit = value / currentBase;

			// If at least one digit has already been written or if the current digit is non-zero, write it.
			if((toCharacter > toDestination) || (0 < currentDigit)) {
				*(toCharacter++) = Character('0' + currentDigit);
				value %= currentBase; // (result is recycled from the division a few lines above)
			} // if number must be printed

		} // for each base

		return toCharacter;
	}


	// Converts the given integer number to a string of decimal digits.
	//   the destination must be large enough to store the result
	//   no terminating zero
	//   returns the a pointer to the end of the written string
	ASCIICharacter * toDecimalString(
		ASCIICharacter result[sizeof "255"],
		UInt1B         number
	) {
		// Perform the conversion using the native integer type, which is the fastest.
		return formatUnsignedInteger<UIntLeast1B, ASCIICharacter, 100>(number, result);
	}

	ASCIICharacter * toDecimalString(
		ASCIICharacter result[sizeof "65536"],
		UInt2B         number
	) {
		// Perform the conversion using the native integer type, which is the fastest.
		return formatUnsignedInteger<UIntLeast2B, ASCIICharacter, 10000>(number, result);
	}

	ASCIICharacter * toDecimalString(
		ASCIICharacter result[sizeof "4294967295"],
		UInt4B         number
	) {
		// Perform the conversion using the native integer type, which is the fastest.
		return formatUnsignedInteger<UIntLeast4B, ASCIICharacter, 1000000000>(number, result);
	}

	ASCIICharacter * toDecimalString(
		ASCIICharacter result[sizeof "18446744073709551615"],
		UInt8B         number
	) {
		return formatUnsignedInteger<UInt8B, ASCIICharacter, 10000000000000000000ull>(number, result);
	}

	ASCIICharacter * toDecimalString(
		ASCIICharacter result[sizeof "-128"],
		SInt1B         number
	) {
		auto const isNegative = 0 > number;
		if(isNegative) {
			*result++ = '-';
		}
		// Perform the conversion using the native integer type, which is the fastest.
		return formatUnsignedInteger<UIntLeast1B, ASCIICharacter, 100>(isNegative ? 0 - UIntLeast1B(number) : UIntLeast1B(number), result);
	}

	ASCIICharacter * toDecimalString(
		ASCIICharacter result[sizeof "-32768"],
		SInt2B         number
	) {
		auto const isNegative = 0 > number;
		if(isNegative) {
			*result++ = '-';
		}
		// Perform the conversion using the native integer type, which is the fastest.
		return formatUnsignedInteger<UIntLeast2B, ASCIICharacter, 10000>(isNegative ? 0u - UIntLeast2B(number) : UIntLeast2B(number), result);
	}

	ASCIICharacter * toDecimalString(
		ASCIICharacter result[sizeof "-2147483648"],
		SInt4B         number
	) {
		auto const isNegative = 0 > number;
		if(isNegative) {
			*result++ = '-';
		}

		// Perform the conversion using the native integer type, which is the fastest.
		return formatUnsignedInteger<UIntLeast4B, ASCIICharacter, 1000000000u>(
			isNegative ? 0u - UIntLeast4B(number) : UIntLeast4B(number),
			result
		);
	}

	ASCIICharacter * toDecimalString(
		ASCIICharacter result[sizeof "-9223372036854775808"],
		SInt8B         number
	) {
		auto const isNegative = 0 > number;
		if(isNegative) {
			*result++ = '-';
		}

		// Perform the conversion using the native integer type, which is the fastest.
		return formatUnsignedInteger<UInt8B, ASCIICharacter, 1000000000000000000ull>(
			isNegative ? 0ull - UInt8B(number) : UInt8B(number),
			result
		);
	}

	UTF16Unit * toDecimalString(
		UTF16Unit result[sizeof "255"],
		UInt1B    number
	) {
		// Perform the conversion using the native integer type, which is the fastest.
		return formatUnsignedInteger<UIntLeast1B, UTF16Unit, 100>(number, result);
	}

	UTF16Unit * toDecimalString(
		UTF16Unit result[sizeof "65536"],
		UInt2B    number
	) {
		// Perform the conversion using the native integer type, which is the fastest.
		return formatUnsignedInteger<UIntLeast2B, UTF16Unit, 10000>(number, result);
	}

	UTF16Unit * toDecimalString(
		UTF16Unit result[sizeof "4294967295"],
		UInt4B    number
	) {
		// Perform the conversion using the native integer type, which is the fastest.
		return formatUnsignedInteger<UIntLeast4B, UTF16Unit, 1000000000>(number, result);
	}

	UTF16Unit * toDecimalString(
		UTF16Unit result[sizeof "18446744073709551615"],
		UInt8B    number
	) {
		return formatUnsignedInteger<UInt8B, UTF16Unit, 10000000000000000000ull>(number, result);
	}

	UTF16Unit * toDecimalString(
		UTF16Unit result[sizeof "-128"],
		SInt1B    number
	) {
		auto const isNegative = 0 > number;
		if(isNegative) {
			*result++ = '-';
		}
		// Perform the conversion using the native integer type, which is the fastest.
		return formatUnsignedInteger<UIntLeast1B, UTF16Unit, 100>(isNegative ? 0 - UIntLeast1B(number) : UIntLeast1B(number), result);
	}

	UTF16Unit * toDecimalString(
		UTF16Unit result[sizeof "-32768"],
		SInt2B    number
	) {
		auto const isNegative = 0 > number;
		if(isNegative) {
			*result++ = '-';
		}
		// Perform the conversion using the native integer type, which is the fastest.
		return formatUnsignedInteger<UIntLeast2B, UTF16Unit, 10000>(isNegative ? 0u - UIntLeast2B(number) : UIntLeast2B(number), result);
	}

	UTF16Unit * toDecimalString(
		UTF16Unit result[sizeof "-2147483648"],
		SInt4B    number
	) {
		auto const isNegative = 0 > number;
		if(isNegative) {
			*result++ = '-';
		}

		// Perform the conversion using the native integer type, which is the fastest.
		return formatUnsignedInteger<UIntLeast4B, UTF16Unit, 1000000000u>(
			isNegative ? 0u - UIntLeast4B(number) : UIntLeast4B(number),
			result
		);
	}

	UTF16Unit * toDecimalString(
		UTF16Unit result[sizeof "-9223372036854775808"],
		SInt8B    number
	) {
		auto const isNegative = 0 > number;
		if(isNegative) {
			*result++ = '-';
		}

		// Perform the conversion using the native integer type, which is the fastest.
		return formatUnsignedInteger<UInt8B, UTF16Unit, 1000000000000000000ull>(
			isNegative ? 0ull - UInt8B(number) : UInt8B(number),
			result
		);
	}

	// Convertes the given integer to a string of hexadecimal digits with zero padding. The destination provided must be large
	//	enough to store the result. There is no terminating zero; a pointer to the end of the written number is returned.
	ASCIICharacter * toHexadecimalZeroPaddedString(
		UInt2B         number,
		ASCIICharacter result[4]
	) {
		result[0] = toHexadecimalDigit((number >> 12) & 0xF);
		result[1] = toHexadecimalDigit((number >>  8) & 0xF);
		result[2] = toHexadecimalDigit((number >>  4) & 0xF);
		result[3] = toHexadecimalDigit((number >>  0) & 0xF);
		return result + 4;
	}
	UTF16Unit * toHexadecimalZeroPaddedString(
		UInt2B    number,
		UTF16Unit result[4]
	) {
		result[0] = toHexadecimalDigit((number >> 12) & 0xF);
		result[1] = toHexadecimalDigit((number >>  8) & 0xF);
		result[2] = toHexadecimalDigit((number >>  4) & 0xF);
		result[3] = toHexadecimalDigit((number >>  0) & 0xF);
		return result + 4;
	}
	UTF16Unit * toHexadecimalZeroPaddedString(
		UTF16Unit result[8],
		UInt4B    number
	) {
		// Use the SIMD-optimized ASCII version, then convert to UTF-16 via zero extension to 16 bits.
		store(reinterpret_cast<UInt2B *>(result), asUInt2B(toHexadecimalZeroPaddedString_reg(number)));
		return result + 8;
	}
	ASCIICharacter * toHexadecimalZeroPaddedString(
		ASCIICharacter result[16],
		UInt8B const   number
	) {
		store(reinterpret_cast<UInt1B *>(result), toHexadecimalZeroPaddedString_reg(number));
		return result + 16;
	}
	UTF16Unit * toHexadecimalZeroPaddedString(
		UTF16Unit    result[16],
		UInt8B const number
	) {
		// Use the SIMD-optimized ASCII version, then convert to UTF-16 via zero extension to 16 bits.
		store(reinterpret_cast<UInt2B *>(result), asUInt2B(toHexadecimalZeroPaddedString_reg(number)));
		return result + 16;
	}

	// Converts the given floating-point number to a decimal string.
	//   inserts the given symbol as decimal point
	//   enforcePlus forces a + for positive number (no prefix otherwise)
	ASCIICharacter * toDecimalString(
		double const         value,
		UIntLeast1B const    precision,
		ASCIICharacter const decimalPoint,
		bool const           enforcePlus,
		ASCIICharacter       result[349]
	) {
		auto toEnd = result;

		// dtoa() generates a string of digits and outputs other properties, like the index of the decimal point.
		//   the result is allocated by dtoa(); release it later
		//   it may be nullptr in out-of-memory conditions
		//   use mode 2: max(1, ndigits) significant digits. This gives a return value similar to that of ecvt, except that
		//    trailing zeros are suppressed.
		//   reserve two characters in the result (sign & decimal point)
		int              indexOfDecimalPoint;
		int              isNegative;
		ASCIICharacter * toEndOfDigits;
		if(auto digits = dtoa(value, 2, precision, &indexOfDecimalPoint, &isNegative, &toEndOfDigits)) {
			auto const numberOfDigits = UCount(toEndOfDigits - digits); // differs from precision in rare cases

			// Write the sign, if necessary.
			if(isNegative) {
				*toEnd++ = '-';
			} else if(enforcePlus) {
				*toEnd++ = '+';
			}

			if(0 < indexOfDecimalPoint) {

				// All digits before the decimal point?
				if(indexOfDecimalPoint >= int(numberOfDigits)) {

					toEnd = movs(toEnd, digits, numberOfDigits); // copy all digits

					toEnd = fill(toEnd, '0', indexOfDecimalPoint - int(numberOfDigits)); // pad zeroes

					*toEnd++ = decimalPoint; // write the decimal point
					*toEnd++ = '0'; // guarantee at least one trailing zero

				} else { // Point splits the digits?

					auto toDigits = digits;
					toEnd = movs(toEnd, toDigits, indexOfDecimalPoint); // copy the digits before the decimal point
					toDigits += indexOfDecimalPoint;

					*toEnd++ = decimalPoint; // write the decimal point

					toEnd = movs(toEnd, toDigits, int(numberOfDigits) - indexOfDecimalPoint); // copy the remaining digits

				}

			} else { // Zero padding before digits?

				*toEnd++ = '0'; // guarantee at least one zero
				*toEnd++ = decimalPoint; // write the decimal point

				toEnd = fill(toEnd, '0', 0 - indexOfDecimalPoint); // pad zeroes

				toEnd = movs(toEnd, digits, numberOfDigits); // copy the digits
			}

			freedtoa(digits);
		}

		return toEnd;
	}

	// Converts a decimal string at toText to a 8-B unsigned integer and advances toText to the first character that was no
	//  decimal digit. Returns zero if no number was found at toText. Does not skip whitespace.
	UInt8B uInt8BFromDecimal(
		ASCIICharacter const * & toText,
		ASCIICharacter const *   toEnd
	) {
		UInt8B result = 0;

		while(toText < toEnd) {
			auto const character = *toText;
			if(no == isDecimalDigit(character)) {
				break;
			}
			result = 10 * result + decimalDigitToNumber(character);
			++toText;
		}

		return result;
	}
	UInt4B uInt4BFromDecimal(
		ASCIICharacter const * & toText,
		ASCIICharacter const *   toEnd
	) {
		return UInt4B(uInt8BFromDecimal(toText, toEnd) & 0xFFFFFFFF);
	}
	UInt2B uInt2BFromDecimal(
		ASCIICharacter const * & toText,
		ASCIICharacter const *   toEnd
	) {
		return UInt2B(uInt8BFromDecimal(toText, toEnd) & 0xFFFF);
	}
	UCount numberFromDecimal(
		ASCIICharacter const * & toText,
		ASCIICharacter const *   toEnd
	) {
#		if COMPILED_FOR_X86_64
			return uInt8BFromDecimal(toText, toEnd);
#		elif COMPILED_FOR_X86_32
			return uInt4BFromDecimal(toText, toEnd);
#		else
#			error TODO
#		endif
	}

	UInt8B uInt8BFromDecimal(
		UTF16Unit const * & toText,
		UTF16Unit const *   toEnd
	) {
		UInt8B result = 0;

		while(toText < toEnd) {
			auto const character = *toText;
			if(no == isDecimalDigit(character)) {
				break;
			}
			result = 10 * result + decimalDigitToNumber(character);
			++toText;
		}

		return result;
	}
	UInt4B uInt4BFromDecimal(
		UTF16Unit const * & toText,
		UTF16Unit const *   toEnd
	) {
		return UInt4B(uInt8BFromDecimal(toText, toEnd) & 0xFFFFFFFF);
	}
	UInt2B uInt2BFromDecimal(
		UTF16Unit const * & toText,
		UTF16Unit const *   toEnd
	) {
		return UInt2B(uInt8BFromDecimal(toText, toEnd) & 0xFFFF);
	}
	UCount numberFromDecimal(
		UTF16Unit const * & toText,
		UTF16Unit const *   toEnd
	) {
#		if COMPILED_FOR_X86_64
			return uInt8BFromDecimal(toText, toEnd);
#		elif COMPILED_FOR_X86_32
			return uInt4BFromDecimal(toText, toEnd);
#		else
#			error TODO
#		endif
	}

	// Computes 10 raised to the given power. Optimized for integer powers.
	// (complexity: logarithmic in magnitude of exponent)
	local double tenRaisedToPower(SIntLeast2B exponent) {
		auto result = 1.0;
		auto currentBase = 10.0;
		auto currentBinaryExponent = exponent >= 0 ? exponent : -exponent;

		raise:

			if(0u < (currentBinaryExponent & 1u)) {
				result = result * currentBase;
			}
			currentBinaryExponent = currentBinaryExponent >> 1u;
			if(0u < currentBinaryExponent) {
				currentBase = currentBase * currentBase;
				goto raise;
			}

		return (0 <= exponent) ? result : (1.0 / result);
	}

	template <typename Character> double float8BFromDecimal_template(Character const * & toBeginning, Character const * toEnd) {
		MUST(toBeginning <= toEnd);

		if(toBeginning == toEnd) {
			return NAN_FLOAT8B; // empty strings arent valid numbers
		}

		// The floating-point number is read as a tuple:
		//   One bool determines the sign of the result.
		bool isPositive;
		//   One signed integer serves as decimal exponent. Since IEEE 754 binary64s have an exponent of 11 bits, up to 308
		//    decimal digits are representable. One additional digit is required to round to zero or to infinity. Therefore, the
		//    exponent has a range of [-309, 309].
		SIntLeast2B exponent = 0;
		//   One unsigned integer serves as mantissa. Since IEEE 754 binary64s have a mantissa of 52+1 bits, up to 16 decimal
		//    digits must be represetable and one additional digit is required for correct rounding; i.e. the mantissa is a 64-bit
		//    integer in the range of [0, 99999999999999999]. Any additional digits are ignored in the mantissa but considered in
		//    the exponent.
		UInt8B mantissa = 0;

		// Determine the sign and skip the according character, if present.
		if('-' == *toBeginning) {
			isPositive = no;
			++toBeginning;
		} else if('+' == *toBeginning) {
			isPositive = yes;
			++toBeginning;
		} else {
			isPositive = yes; // no sign is implicitly positive
		}
		if(toBeginning == toEnd) { // Just a sign with no value?
			return NAN_FLOAT8B;
		}

		// Read the numbers integer part. Consider that digits before the decimal point are optional  .1 is valid.
		auto hasDigitsBeforeDecimalPoint = yes;
		{
			auto digit = numberFromDecimalDigit_unchecked(*toBeginning);
			if(digit <= 9) {
				do {
					if(mantissa < 9999999999999999) {
						mantissa = 10 * mantissa + digit;
					} else {
						++exponent;
					}
					++toBeginning;
					if(toBeginning == toEnd) {
						break;
					}
					digit = numberFromDecimalDigit_unchecked(*toBeginning);
				} while(digit <= 9);
			} else {
				// If there are no digits in the integer part, then there must at least be a decimal point.
				if('.' != *toBeginning) {
					return NAN_FLOAT8B;
				}
				hasDigitsBeforeDecimalPoint = no;
			}
		}

		// Read the numbers optional fraction.
		if(toBeginning < toEnd && '.' == *toBeginning) {
			++toBeginning; // skip the decimal point

			// If no digits appear before the decimal point, at least one must appear after the decimal point.
			if(no == hasDigitsBeforeDecimalPoint) {
				if(toBeginning >= toEnd || !isDecimalDigit(*toBeginning)) {
					return NAN_FLOAT8B;
				}
			}
			cycle {
				if(toBeginning == toEnd) {
					break;
				}
				auto const digit = numberFromDecimalDigit_unchecked(*toBeginning);
				if(digit <= 9) {
					if(mantissa < 9999999999999999) {
						mantissa = 10 * mantissa + digit;
						--exponent;
					}
				} else {
					break;
				}
				++toBeginning;
			}
		}

		// Read the numbers optional exponent.
		if(toBeginning < toEnd && ('e' == *toBeginning || 'E' == *toBeginning || 'd' == *toBeginning || 'D' == *toBeginning)) {
			++toBeginning; // skip the d, D, e, or E

			// The exponent must not be empty.
			if(toBeginning == toEnd) {
				return NAN_FLOAT8B;
			}

			// Determine the sign and skip the according character, if present.
			bool isExponentPositive;
			if('-' == *toBeginning) {
				isExponentPositive = no;
				++toBeginning;
			} else if('+' == *toBeginning) {
				isExponentPositive = yes;
				++toBeginning;
			} else {
				isExponentPositive = yes; // no sign is implicitly positive
			}

			// Consider: The exponent must not run over or under. Since the binary exponent of a binary64 is eleven bits, the decimal
			//  exponent is in the range of [-308, +308]. Therefore, any value outside of this range is zero or infinity anyway, and
			//  the value of the trailing digits is irrelevant and can be skipped.
			SIntLeast2B newExponent = 0;

			// There must be at least one digit in the exponent.
			if(toBeginning == toEnd) {
				return NAN_FLOAT8B;
			}
			auto digit = numberFromDecimalDigit_unchecked(*toBeginning);
			if(9 < digit) {
				return NAN_FLOAT8B;
			}

			do {
				if(isWithin(-308, newExponent, 308)) {
					newExponent = 10 * newExponent + SIntLeast1B(digit);
				}
				++toBeginning;
				if(toBeginning == toEnd) {
					break;
				}
				digit = numberFromDecimalDigit_unchecked(*toBeginning);
			} while(digit <= 9);

			if(no == isExponentPositive) {
				newExponent = -newExponent;
			}

			exponent = exponent + newExponent;
		}

		// TODO: Find an arbitrary-precision integer math solution here
		auto const absolute = double(mantissa) * tenRaisedToPower(exponent);
		return isPositive ? absolute : -absolute;
	}

	// Converts the given string range to an IEEE 754 binary64 floating-point number, or returns NaN if not successful. After the
	//  function returns, the given ranges beginning either equals its end the first character that was not recognized as part of
	//  the number.
	// The lexical format of a floating-point number is
	//    [sign] [digits] [.digits] [ {d|D|e|E}[sign]digits ]
	//  where
	//    sign is {+|-}; and
	//    digits is a string of decimal digits.
	double float8BFromDecimal(ASCIICharacter const * & toBeginning, ASCIICharacter const * toEnd) {
		return float8BFromDecimal_template(toBeginning, toEnd);
	}
	double float8BFromDecimal(UTF16Unit const * & toBeginning, UTF16Unit const * toEnd) {
		return float8BFromDecimal_template(toBeginning, toEnd);
	}
	double float8BFromDecimal(Range<ASCIICharacter const> & string) {
		return float8BFromDecimal_template(string.toBeginning, string.toEnd);
	}

} // Text



locals {


	void append(
		ASCIICharacter     character,
		ASCIICharacter * & text_toBeginning,
		ASCIICharacter *   text_toEnd
	) {
		MUST(text_toBeginning < text_toEnd);

		*text_toBeginning = character;
		++text_toBeginning;
	}

	void appendRepeated(
		ASCIICharacter     character,
		UCount             count,
		ASCIICharacter * & text_toBeginning,
		ASCIICharacter *   text_toEnd
	) {
		MUST(text_toBeginning < text_toEnd && UCount(text_toEnd - text_toBeginning) > count);
		stosb(reinterpret_cast<Byte *>(text_toBeginning), character, count);
		text_toBeginning += count;
	}

	void append(
		ASCIICharacter const * toStringsBeginning,
		UCount                 stringsLength,
		ASCIICharacter * &     text_toBeginning,
		ASCIICharacter *       text_toEnd
	) {
		MUST(text_toBeginning < text_toEnd && UCount(text_toEnd - text_toBeginning) > stringsLength);
		movsb(reinterpret_cast<Byte *>(text_toBeginning), reinterpret_cast<Byte const *>(toStringsBeginning), stringsLength);
		text_toBeginning += stringsLength;
	}


	// flag definitions
	#define FL_SIGN       0x0001 // put plus or minus in front
	#define FL_SIGNSP     0x0002 // put space or minus in front
	#define FL_LEFT       0x0004 // left justify
	#define FL_LEADZERO   0x0008 // pad with leading zeros
	#define FL_LONG       0x0010 // long value given
	#define FL_SHORT      0x0020 // short value given
	#define FL_SIGNED     0x0040 // signed data given
	#define FL_ALTERNATE  0x0080 // alternate form requested
	#define FL_NEGATIVE   0x0100 // value is negative
	#define FL_FORCEOCTAL 0x0200 // force leading 0 for octals
	#define FL_I64        0x0400 // __int64 value given

	// state definitions
	enum ST {
		ST_NORMAL, // normal state; outputting literal chars
		ST_PERCENT,// just read %
		ST_FLAG,   // just read flag character
		ST_WIDTH,  // just read width specifier
		ST_DOT,    // just read .
		ST_PRECIS, // just read precision specifier
		ST_SIZE,   // just read size specifier
		ST_TYPE,   // just read type specifier
		NUMSTATES
	};

	enum CharacterType {
		CH_OTHER,  // character with no special meaning
		CH_PERCENT,// %
		CH_DOT,    // .
		CH_STAR,   // *
		CH_ZERO,   // 0
		CH_DIGIT,  // 19
		CH_FLAG,   //  , +, -, #
		CH_SIZE,   // h, l, L, N, F, w
		CH_TYPE    // type specifying character
	};

	// Actual implementation of Text::format(). No overhead for retrieving the argument list is desired, so this is an inline
	//  function which is used by both variations.
	INLINE_CODE Range<ASCIICharacter> formatImplementation(
		Range<ASCIICharacter> const & destination,
		ASCIICharacter const *        toZeroTerminatedFormat,
		va_list &                     parameters
	) {
		MUST(destination.toBeginning <= destination.toEnd);
		MUST(nullptr != parameters);

		ASCIICharacter * toEndOfOutput = destination.toBeginning;
		int flags     = 0;         // flag word  see #defines for values
		ST  state     = ST_NORMAL; // current state
		int fldwidth  = 0;         // selected field width  0 means default
		int precision = 0;         // selected precision  -1 means default
		int prefixlen = 0;         // length of prefix  0 means no prefix

		// The state table.  This table is actually two tables combined into one.
		// The lower nybble of each byte gives the character class of any character; while the uper nybble of the byte gives the next
		//  state to enter. See the macros below the table for details.
		// The table is generated by maketabc.c  use this program to make changes.
		static UInt1B const lookupTable[] = {
			(ST_NORMAL << 4) | CH_FLAG,   //  
			(ST_NORMAL << 4) | CH_OTHER,  // !
			(ST_NORMAL << 4) | CH_OTHER,  // "
			(ST_NORMAL << 4) | CH_FLAG,   // #
			(ST_NORMAL << 4) | CH_OTHER,  // $
			(ST_NORMAL << 4) | CH_PERCENT,// %
			(ST_NORMAL << 4) | CH_OTHER,  // &
			(ST_NORMAL << 4) | CH_OTHER,  // '
			(ST_PERCENT << 4)| CH_OTHER,  // (
			(ST_NORMAL << 4) | CH_OTHER,  // )
			(ST_NORMAL << 4) | CH_STAR,   // *
			(ST_NORMAL << 4) | CH_FLAG,   // +
			(ST_NORMAL << 4) | CH_OTHER,  // ,
			(ST_NORMAL << 4) | CH_FLAG,   // -
			(ST_NORMAL << 4) | CH_DOT,    // .
			(ST_PERCENT << 4)| CH_OTHER,  // /
			(ST_NORMAL << 4) | CH_ZERO,   // 0
			(ST_DOT << 4)    | CH_DIGIT,  // 1
			(ST_DOT << 4)    | CH_DIGIT,  // 2
			(ST_DOT << 4)    | CH_DIGIT,  // 3
			(ST_NORMAL << 4) | CH_DIGIT,  // 4
			(ST_NORMAL << 4) | CH_DIGIT,  // 5
			(ST_NORMAL << 4) | CH_DIGIT,  // 6
			(ST_NORMAL << 4) | CH_DIGIT,  // 7
			(ST_NORMAL << 4) | CH_DIGIT,  // 8
			(ST_WIDTH << 4)  | CH_DIGIT,  // 9
			(ST_WIDTH << 4)  | CH_OTHER,  // :
			(ST_NORMAL << 4) | CH_OTHER,  // ;
			(ST_PRECIS << 4) | CH_OTHER,  // <
			(ST_NORMAL << 4) | CH_OTHER,  // =
			(ST_NORMAL << 4) | CH_OTHER,  // >
			(ST_NORMAL << 4) | CH_OTHER,  // ?
			(ST_NORMAL << 4) | CH_OTHER,  // @
			(ST_FLAG << 4)   | CH_OTHER,  // A (disable %A)
			(ST_FLAG << 4)   | CH_OTHER,  // B
			(ST_WIDTH << 4)  | CH_TYPE,   // C
			(ST_PRECIS << 4) | CH_OTHER,  // D
			(ST_PRECIS << 4) | CH_TYPE,   // E
			(ST_NORMAL << 4) | CH_SIZE,   // F
			(ST_NORMAL << 4) | CH_TYPE,   // G
			(ST_NORMAL << 4) | CH_OTHER,  // H
			(ST_WIDTH << 4)  | CH_SIZE,   // I
			(ST_WIDTH << 4)  | CH_OTHER,  // J
			(ST_WIDTH << 4)  | CH_OTHER,  // K
			(ST_PRECIS << 4) | CH_SIZE,   // L
			(ST_PRECIS << 4) | CH_OTHER,  // M
			(ST_NORMAL << 4) | CH_SIZE,   // N
			(ST_NORMAL << 4) | CH_OTHER,  // O
			(ST_NORMAL << 4) | CH_OTHER,  // P
			(ST_FLAG << 4)   | CH_OTHER,  // Q
			(ST_FLAG << 4)   | CH_OTHER,  // R
			(ST_NORMAL << 4) | CH_TYPE,   // S
			(ST_NORMAL << 4) | CH_OTHER,  // T
			(ST_NORMAL << 4) | CH_OTHER,  // U
			(ST_NORMAL << 4) | CH_OTHER,  // V
			(ST_NORMAL << 4) | CH_OTHER,  // W
			(ST_NORMAL << 4) | CH_TYPE,   // X
			(ST_SIZE << 4)   | CH_OTHER,  // Y
			(ST_SIZE << 4)   | CH_TYPE,   // Z
			(ST_SIZE << 4)   | CH_OTHER,  // [
			(ST_SIZE << 4)   | CH_OTHER,  // backslash
			(ST_SIZE << 4)   | CH_OTHER,  // ]
			(ST_SIZE << 4)   | CH_OTHER,  // ^
			(ST_NORMAL << 4) | CH_OTHER,  // _
			(ST_NORMAL << 4) | CH_OTHER,  // `
			(ST_TYPE << 4)   | CH_OTHER,  // a (disable %a)
			(ST_TYPE << 4)   | CH_OTHER,  // b
			(ST_TYPE << 4)   | CH_TYPE,   // c
			(ST_TYPE << 4)   | CH_TYPE,   // d
			(ST_TYPE << 4)   | CH_TYPE,   // e
			(ST_TYPE << 4)   | CH_TYPE,   // f
			(ST_NORMAL << 4) | CH_TYPE,   // g
			(ST_NORMAL << 4) | CH_SIZE,   // h
			(ST_NORMAL << 4) | CH_TYPE,   // i
			(ST_NORMAL << 4) | CH_OTHER,  // j
			(ST_NORMAL << 4) | CH_OTHER,  // k
			(ST_NORMAL << 4) | CH_SIZE,   // l
			(ST_NORMAL << 4) | CH_OTHER,  // m
			(ST_NORMAL << 4) | CH_OTHER,  // n (disable %n)
			(ST_NORMAL << 4) | CH_TYPE,   // o
			(ST_NORMAL << 4) | CH_TYPE,   // p
			(ST_NORMAL << 4) | CH_OTHER,  // q
			(ST_NORMAL << 4) | CH_OTHER,  // r
			(ST_NORMAL << 4) | CH_TYPE,   // s
			(ST_NORMAL << 4) | CH_OTHER,  // t
			(ST_NORMAL << 4) | CH_TYPE,   // u
			(ST_NORMAL << 4) | CH_OTHER,  // v
			(ST_NORMAL << 4) | CH_SIZE,   // w
			(ST_NORMAL << 4) | CH_TYPE    // x
		};

		// main loop  loop while format character exist and no I/O errors
		ASCIICharacter ch; // character just read
		while((ch = *toZeroTerminatedFormat++) != 0) {
			auto const chclass = CharacterType(isWithin(' ', ch, 'x') ? (lookupTable[ch - ' '] & 0x0F) : CH_OTHER);
			state = ST(lookupTable[chclass * NUMSTATES + state] >> 4);
			switch(state) {

				case ST_NORMAL:
				NORMAL_STATE:
					append(ch, toEndOfOutput, destination.toEnd);
					break;

				case ST_PERCENT:
					// set default value of conversion parameters
					prefixlen = 0;
					fldwidth  = 0;
					flags     = 0;
					precision = -1;
					break;

				case ST_FLAG:
					switch(ch) { // set flag based on which flag character
						case '-': flags |= FL_LEFT;      break; // - => left justify
						case '+': flags |= FL_SIGN;      break; // + => force sign indicator
						case ' ': flags |= FL_SIGNSP;    break; //   => force sign or space
						case '#': flags |= FL_ALTERNATE; break; // # => alternate form
						case '0': flags |= FL_LEADZERO;  break; // 0 => pad with leading zeros
						default: UNREACHABLE;
					}
					break;

				case ST_WIDTH:
					// update width value
					if('*' == ch) {
						fldwidth = va_arg(parameters, SInt4B);
						if(0 > fldwidth) {
							// ANSI says negative fld width means - flag and pos width
							flags |= FL_LEFT;
							fldwidth = -fldwidth;
						}
					} else {
						fldwidth = fldwidth * 10 + (ch - '0'); // add digit to current field width
					}
					break;

				case ST_DOT:
					// zero the precision, since dot with no number means 0 not default, according to ANSI
					precision = 0;
					break;

				case ST_PRECIS: // update precison value
					if('*' == ch) {
						precision = va_arg(parameters, SInt4B);
						if(0 > precision) {
							precision = -1; // neg precision means default
						}
					} else {
						precision = precision * 10 + (ch - '0'); // add digit to current precision
					}
					break;

				case ST_SIZE: {
					switch(ch) { // just read a size specifier, set the flags based on it

						case 'l':
							// In order to handle the ll case, we depart from the simple deterministic state machine.
							if(*toZeroTerminatedFormat == 'l') {
								++toZeroTerminatedFormat;
								flags |= FL_I64; // ll => long long
							} else {
								flags |= FL_LONG; // l => long int or wchar_t
							}
							break;

						case 'I':
							// In order to handle the I, I32, and I64 size modifiers, we depart from the simple deterministic state machine.
							//  The code below scans for characters following the I, and defaults to 64 bit on Win64 and 32 bit on Win32.
#							if COMPILED_FOR_X86_64
								flags |= FL_I64; // I => __int64 on Win64 systems
#							elif COMPILED_FOR_X86_32
								// i32 is default
#							else
#								error TODO
#							endif
							if(toZeroTerminatedFormat[0] == '6' && toZeroTerminatedFormat[1] == '4') {
								toZeroTerminatedFormat += 2;
								flags |= FL_I64; // I64 => __int64
							} else if(toZeroTerminatedFormat[0] == '3' && toZeroTerminatedFormat[1] == '2') {
								toZeroTerminatedFormat += 2;
								flags &= ~FL_I64; // I32 => __int32
							} else if(('d' == *toZeroTerminatedFormat) ||
								        ('i' == *toZeroTerminatedFormat) ||
								        ('o' == *toZeroTerminatedFormat) ||
								        ('u' == *toZeroTerminatedFormat) ||
								        ('x' == *toZeroTerminatedFormat) ||
								        ('X' == *toZeroTerminatedFormat)
							) {
								// Nothing further needed. %Id (et al) is handled just like %d, except that it defaults to 64 bits on x64. Fall
								//  through to the next iteration.
							} else {
								state = ST_NORMAL;
								goto NORMAL_STATE;
							}
							break;

						case 'h':
							flags |= FL_SHORT; // h => short int or char
							break;

						case 'w': // ISO wide character
							UNREACHABLE; // %Iw format specifier disabled

						default:
							UNREACHABLE;

					} break; // switch size specifier

				} break; // ST_SIZE == state

				case ST_TYPE: {
					// we have finally read the actual type character, so we now format and print the output. We use a big switch state-
					//  -ment that sets text to point to the text that should be printed, and textlen to the length of this text.
					//  Common code later on takes care of justifying it and other miscellaneous chores. Note that cases share code, in
					//  particular, all integer formatting is done in one place. Look at those funky goto statements!
					ASCIICharacter * toNewText;
					int              radix;              // current conversion radix
					int              hexadd = 0;         // offset to add to number to get af
					ASCIICharacter   prefix[2];          // numeric prefix  up to two characters
					ASCIICharacter   buffer[309 + 40 + 6];
					int              newTextsLength = 0; // length of the text to be printed
					switch(ch) {

						case 'C': // ISO wide character
							UNREACHABLE; // %C format specifier disabled

						case 'c': { // format multibyte character  this is an extension of ANSI
							auto const character = va_arg(parameters, SInt4B);
							MUST(127 >= character); // wide characters not supported by Text::format(ASCII)
							buffer[0] = ASCIICharacter(character);
							toNewText = buffer;
							newTextsLength = 1;
						} break;

						case 'z': // size_t in late Visual C++ versions; counted string before
						case 'Z': //{ // print a counted string
						//	struct CountedString {
						//		UInt2B           length;
						//		UInt2B           capacity;
						//		ASCIICharacter * toBuffer;
						//	} * const toString = va_arg(parameters, CountedString *);
						//	toNewText      = toString->toBuffer;
						//	newTextsLength = toString->length;
						//} break;
							UNREACHABLE;

						case 'S': // ISO wide character string
							UNREACHABLE; // %S format specifier disabled

						case 's': { // print a string
							// ANSI rules on how much of string to print:
							//   all if precision is default;
							//   min(precision, length) if precision given; or
							//   (null) if a null string is passed.

							// At this point it is tempting to use strlen(), but if a precision is specified, were not allowed to scan past
							//  there, because there might be no null at all. Thus, we must do our own scan.
							int i = precision; // -1 for default
							toNewText = va_arg(parameters, ASCIICharacter *);
							// scan for null upto i characters
							MUST(nullptr != toNewText);
							ASCIICharacter * p = toNewText;
							while(i-- && *p) {
								++p;
							}
							newTextsLength = int(p - toNewText); // length of the string
						} break;

						case 'n':
							UNREACHABLE; // %n disabled

						case 'E':
						case 'G':
						case 'A':
							ch += 'a' - 'A'; // lowercase
							// drop through
						case 'e':
						case 'f':
						case 'g':
						case 'a': { // floating-point conversion  use _gcvt_s() for that
							flags |= FL_SIGNED; // always signed
							toNewText = buffer; // put result in buffer

							if(0 > precision) {
								precision = 6u; // default
							} else if(precision == 0 && ch == 'g') {
								precision = 1u; // specified by ANSI
							} else if(precision > int(capacityOf(buffer))) {
								precision = int(capacityOf(buffer));
							}

							auto const value = va_arg(parameters, double);
							if(isFinite(value)) {
								Text::toDecimalString(value, precision, '.', no, buffer)[0] = 0; // zero-terminate result
							} else if(isNumber(value)) {
								toNewText[0] = 'i';
								toNewText[1] = 'n';
								toNewText[2] = 'f';
								toNewText[3] = 0;
							} else {
								toNewText[0] = 'n';
								toNewText[1] = 'a';
								toNewText[2] = 'n';
								toNewText[3] = 0;
							}

							// check if result was negative, save - for later and point to positive part (for 0 padding)
							if('-' == *toNewText) {
								flags |= FL_NEGATIVE;
								++toNewText;
							}

#							if COMPILED_BY_VISUAL_CPP
								newTextsLength = int(strlen(toNewText));
#							elif COMPILED_BY_CLANG
								newTextsLength = int(__builtin_strlen(toNewText));
#							else
#								error TODO
#							endif
						} break;

						case 'd': // signed decimal output
						case 'i':
							flags |= FL_SIGNED;
							radix = 10;
							goto COMMON_INT;

						case 'u':
							radix = 10;
							goto COMMON_INT;

						case 'p':
							// write a pointer  this is like an integer or long except we force precision to pad with zeros and output in big
							//  hex.
							precision = 2 * sizeof(void *); // number of hex digits needed
#							if COMPILED_FOR_X86_64
								flags |= FL_I64; // pretend were converting an int64
#							elif COMPILED_FOR_X86_32
								// i32 is default
#							else
#								error TODO
#							endif
							// drop through to hex formatting
						case 'X': // unsigned upper hex output
							hexadd = 'A' - '9' - 1;
							goto COMMON_HEX;

						case 'x': // unsigned lower hex output
							hexadd = 'a' - '9' - 1;
							// drop through to COMMON_HEX
						COMMON_HEX:
							radix = 16;
							if(flags & FL_ALTERNATE) { // means 0x prefix
								prefix[0] = '0';
								prefix[1] = ASCIICharacter('x' - 'a' + '9' + 1 + hexadd); // x or X
								prefixlen = 2;
							}
							goto COMMON_INT;

						case 'o': // unsigned octal output
							radix = 8;
							if(flags & FL_ALTERNATE) {
								flags |= FL_FORCEOCTAL; // force a leading zero
							}
							// drop through to COMMON_INT
						COMMON_INT: {
							// This is the general integer formatting routine. Basically, we get an argument, make it positive if necessary,
							//  and convert it according to the correct radix, setting toNewText and newTextsLength appropriately.
							int    digit;  // ASCII value of digit
							SInt8B l;      // temp long value
							UInt8B number; // absolute value to print

							// 1. read argument into l, sign extend as needed
							if(flags & FL_I64) {
								l = va_arg(parameters, SInt8B);
							} else { // even 2-B integers are promoted to int and unsigned
								if(flags & FL_SIGNED) {
									l = va_arg(parameters, int);
								} else {
									l = va_arg(parameters, unsigned);
								}
							}

							// 2. check for negative; copy into number
							if((flags & FL_SIGNED) && l < 0) {
								number = -l;
								flags |= FL_NEGATIVE; // remember negative sign
							} else {
								number = l;
							}

							if(0 == (flags & FL_I64)) {
								// Unless printing a full 64-bit value, ensure values here are not in canonical longword format to prevent the
								//  the sign extended upper 32-bits from being printed.
								number &= 0xFFFFFFFF;
							}

							// 3. check precision value for default; non-default turns off 0 flag according to ANSI
							if(precision < 0) {
								precision = 1; // default
							} else {
								flags &= ~FL_LEADZERO;
								precision = minimumOf(precision, int(capacityOf(buffer)));
							}

							// 4. check if data is 0; if so, turn off hex prefix
							if(0 == number) {
								prefixlen = 0;
							}

							// 5. Convert data to ASCII  note if precision is zero and number is zero, we get no digits at all.
							toNewText = toEndOf(buffer) - 1; // last digit at end of buffer
							while(0 < precision-- || 0 != number) {
								digit = int(number % radix) + '0';
								number /= radix; // reduce
								if('9' < digit) {
									digit += hexadd; // a hex digit, make it a letter
								}
								*toNewText-- = ASCIICharacter(digit);
							}

							newTextsLength = int(toEndOf(buffer) - 1 - toNewText);
							++toNewText; // text points to first digit now

							// 6. force a leading zero if FORCEOCTAL flag set
							if((flags & FL_FORCEOCTAL) && (newTextsLength == 0 || toNewText[0] != '0')) {
								*--toNewText = '0';
								++newTextsLength;
							}
						} break; // COMMON_INT

						default:
							UNREACHABLE;

					} // switch format character

					// At this point, we have done the specific conversion, and toNewText points to text to print; newTextsLength is
					//  length. Now we justify it, put on prefixes, leading zeros, and then print it.
					{
						if(flags & FL_SIGNED) {
							if(flags & FL_NEGATIVE) {
								prefix[0] = '-';
								prefixlen = 1;
							} else if(flags & FL_SIGN) {
								prefix[0] = '+';
								prefixlen = 1;
							} else if(flags & FL_SIGNSP) {
								prefix[0] = ' ';
								prefixlen = 1;
							}
						}

						// calculate amount of padding  might be negative, but this will just mean zero
						int padding = maximumOf(0, fldwidth - newTextsLength - prefixlen);

						// put out the padding, prefix, and text, in the correct order

						if(0 == (flags & (FL_LEFT + FL_LEADZERO))) {
							appendRepeated(' ', padding, toEndOfOutput, destination.toEnd); // pad on left with blanks
						}

						append(prefix, prefixlen, toEndOfOutput, destination.toEnd);

						if((flags & FL_LEADZERO) && 0 == (flags & FL_LEFT)) {
							appendRepeated('0', padding, toEndOfOutput, destination.toEnd); // write leading zeros
						}

						append(toNewText, newTextsLength, toEndOfOutput, destination.toEnd);

						if(flags & FL_LEFT) {
							appendRepeated(' ', padding, toEndOfOutput, destination.toEnd); // pad on right with blanks
						}

						// Done!
					}

				} break; // ST_TYPE == state

				case NUMSTATES:
				default:
					UNREACHABLE; // State transition table broken?
			} // switch state

		} // until zero character
		*toEndOfOutput = 0;

		return range(destination.toBeginning, toEndOfOutput);
	}


} // locals


namespace Text {


	// Formats the given string in the same way sprintf() would. Returns the range of characters that has been formatted.
	Range<ASCIICharacter> format(
		Range<ASCIICharacter> const & destination,
		ASCIICharacter const *        toZeroTerminatedFormat,
		...
	) {
		MUST(destination.toBeginning <= destination.toEnd);

		va_list parameters;
		va_start(parameters, toZeroTerminatedFormat);
		auto result = formatImplementation(destination, toZeroTerminatedFormat, parameters);
		va_end(parameters);

		return result;
	}
	Range<ASCIICharacter> format(
		Range<ASCIICharacter> const &     destination,
		ASCIICharacter const *            toZeroTerminatedFormat,
		struct VariableParameters const * toParameters
	) {
		MUST(destination.toBeginning <= destination.toEnd);

		auto parameters = va_list(toParameters);
		auto result = formatImplementation(destination, toZeroTerminatedFormat, parameters);
		va_end(parameters);

		return result;
	}


} // Text


// Writes the given UUID to the given string.
//   requires 38 characters + one terminating zero
//   format is {XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX}
//   returns a pointer to the terminating zero
UTF16Unit * asString(UUID const & uid, UTF16Unit result[39]) {
	if(39 != StringFromGUID2(uid, result, 39)) {
		UNREACHABLE;
	}
	return result + 38;
}