// Contains Visual C++ run-time replacement code which is EXE-specific, e.g. the entry point. Code that is common to EXEs and
//  DLLs (e.g. static initialization) is defined in 2015.cpp.



SInt4B C_CALL main(); // the program

#if _DEBUG
	namespace OS {
		namespace Win32 {
			void printMemoryLeaks();
		}
	}
#endif

#if COMPILED_FOR_X86_32
	import_from(User32) SInt4B API_CALL MessageBoxA(
		void *                 hWnd,
		ASCIICharacter const * lpText,
		ASCIICharacter const * lpCaption,
		UInt4B                 uType
	);
#endif



int C_CALL atexit(StaticFunction * toNewDestructor) {
	MUST(nullptr != toNewDestructor);

#	if _DEBUG
		if(toEndOfDestructorsToBeExecuted >= toEndOf(destructorsToBeExecuted)) {
			INVOKE_DEBUGGER;
		}
		*toEndOfDestructorsToBeExecuted++ = toNewDestructor;
#	else
		// deployed executables dont reach destructors  they are terminated for quickest shutdown
#	endif
	return 0;
}

// The entry point. Prepares the environment for the program to run (e.g. by static initialization), calls main(), and
//  cleans up. See MSDN, /ENTRY (Entry-Point Symbol).
#pragma comment(linker, "/ENTRY:crtMain")
void API_CALL crtMain() {

#	if COMPILED_FOR_X86_32
		// SSE2 is crucial for the framework. The operating system must support it as well, so dont check via CPUID.
		// All Intel CPUs after 2001 (2003 for AMD) support SSE2, but in case we find one without, showing an error dialog is
		//  certainly better than crashing in seemingly random places.
		if(0 == IsProcessorFeaturePresent(10)) { // PF_XMMI64_INSTRUCTIONS_AVAILABLE
			MessageBoxA(
				nullptr,
				"SSE2-compliant CPU required", "CPU too old",
				0x00010010 // MB_OK + MB_ICONERROR + MB_SETFOREGROUND
			);
			TerminateProcess(GetCurrentProcess(), 0x8000055E); // 055E == no SSE
		}
#	elif COMPILED_FOR_X86_64
		// SSE2 is part of the specification; no need to check
#	else
#		error TODO
#	endif

	// Initialize all C/C++ objects with static storage duration.
	// Consider: The very first elements of toCInitSectionsBegin and toCppInitSectionsBegin are defined to nullptr at
	//  declaration, so they can be skipped safely.
	executeConstructorTable(toCInitSectionsBegin + 1, toCInitSectionsEnd);
	executeConstructorTable(toCppInitSectionsBegin + 1, toCppInitSectionsEnd);

	auto result = main();

#	if _DEBUG
		// Execute static destructors in reverse order before terminating the process. Deployed executables dont reach
		//  destructors  they are terminated for quickest shutdown.
		executeDestructorTable(destructorsToBeExecuted, toEndOfDestructorsToBeExecuted);

		OS::Win32::printMemoryLeaks();
#	endif

	ExitProcess(result);
}

// Visual C++ 2015 by default does not link the CRT if the entry point is overriden. Force linking. Macros according to
//  MSDN, Predefined Macros. Library names according to Visual C++ Team Blog, Introducing the Universal CRT.
#if _DEBUG
#	define VCRT_D "d"
#else
#	define VCRT_D
#endif
#if _DLL
#	define VCRT_LIB
#else
#	define VCRT_LIB "lib"
#endif
#pragma comment(lib, VCRT_LIB "vcruntime" VCRT_D)
#pragma comment(lib, VCRT_LIB "ucrt" VCRT_D)
#undef VCRT_D
#undef VCRT_LIB