// Visual C++ 2015 C-Runtime replacements which are common to EXEs and DLLs.
//   compatible to Visual Studio 2017 as well
//   also requires
//     either 2015 x86-32.asm or 2015 x86-64.asm
//     2015 intrinsics.cpp
//   2015 EXE.inl contains EXE-specific replacements
//   2015 DLL.cpp contains DLL-specific replacements



// Required Win32 functions:
namespace OS {
	namespace Win32 {
		struct Process;
	}
}
import_from(Kernel32) {

	OS::Win32::Process * API_CALL GetCurrentProcess();

	WONT_RETURN API_CALL ExitProcess(UInt4B);

	SInt4B API_CALL TerminateProcess(
		OS::Win32::Process *,
		UInt4B              uExitCode
	);

#	if COMPILED_FOR_X86_32
		int API_CALL IsProcessorFeaturePresent(UInt4B); // to check the users CPU for SSE2 support
#	endif

#	if _DEBUG
		void API_CALL OutputDebugStringA(ASCIICharacter const *);
#	endif

} // Kernel32



locals {


	// A parameter-less C function. Those are generated by Visual C++ to construct and destroy static C/C++ objects.
	using StaticFunction = void C_CALL();

	// The CRT uses certain data sections of the name XI* to store pointers to construction functions for static objects.
	//  Declare two sections  ending with the letters A and Z  manually and store a pointer in each of them. The linker sorts
	//  all sections by their letters, so eventually the pointers will point to before the first and behind the last
	//  initialization function pointer in the table.

	// C:
#	pragma section(".CRT$XIA", read)
	__declspec(allocate(".CRT$XIA")) StaticFunction * toCInitSectionsBegin[] = { nullptr }; // CAUTION: Must not be const!
#	pragma section(".CRT$XIZ", read)
	__declspec(allocate(".CRT$XIZ")) StaticFunction * toCInitSectionsEnd[]   = { nullptr }; // CAUTION: Must not be const!
	// C++:
#	pragma section(".CRT$XCA", read)
	__declspec(allocate(".CRT$XCA")) StaticFunction * toCppInitSectionsBegin[] = { nullptr }; // CAUTION: Must not be const!
#	pragma section(".CRT$XCZ", read)
	__declspec(allocate(".CRT$XCZ")) StaticFunction * toCppInitSectionsEnd[]   = { nullptr }; // CAUTION: Must not be const!

#	pragma comment(linker, "/merge:.CRT=.rdata") // merge all CRT data into the read-only data segment

#	if _DEBUG // release builds wont reach destructors  terminated earlier for quickest shutdown
		// A list of all destructors to be executed at program exit (in reverse order). The standard does not require more than 32
		//  of those being saved.
		StaticFunction *   destructorsToBeExecuted[32];
		StaticFunction * * toEndOfDestructorsToBeExecuted = destructorsToBeExecuted;
#	endif

	// Executes the given constructor function table. Used to call constructors of static objects at startup.
	void executeConstructorTable(
		StaticFunction * const * const toBegin,
		StaticFunction * const * const toEnd
	) {
#		if _DEBUG
			bool hasAnEntry = no;
#		endif
		for(auto toCurrent = toBegin; toCurrent < toEnd; ++toCurrent) {
			if(nullptr != *toCurrent) { // the compiler inserts many nullptrs into the tables for unknown reasons
#				if _DEBUG
					if(no == hasAnEntry) {
						OutputDebugStringA("\nInitializing globals (check if you can replace these by static data):\n");
						hasAnEntry = yes;
					}
					OS::Win32::DbgHelp::printSymbolAt(*toCurrent);
#				endif
				(*toCurrent)();
			}
		}
#		if _DEBUG
			if(hasAnEntry) {
				OutputDebugStringA("\n");
			}
#		endif
	}

	// Executes the given destructor function table in reverse order. Used to call destructors of static objects at
	//  shutdown.
	void executeDestructorTable(
		StaticFunction * const * const toBegin,
		StaticFunction * const * const toEnd
	) {
		for(auto toCurrent = toEnd; toCurrent > toBegin; --toCurrent) {
			(*(toCurrent - 1))(); // the list has been written by our atexit(), we know no null pointers occur
		}
	}


} // locals


#if !_DEBUG // debug builds are too complicated to replace
	extern "C" {


		// Called on pure virtual method errors.
		// Even if a program never calls pure virtual methods, Visual C++ references this function once pure virtual methods occur
		//  anywhere in the program (probably to initialize virtual function tables).
		int C_CALL _purecall() {
			INVOKE_DEBUGGER;
			UNREACHABLE;
		}

		// An old trick in the CRT sources to include floating-point code only when its used. Nowadays useless but still linked.
		int _fltused = 1;


	} // extern "C"
#endif